#!/usr/bin/env python

import sys, string, re, operator
import vs, textedit, vsprogram
import copy
import pickle


# ---------------------------------------------------------------------------
# Abstract event types - These are constructed as they need to be handled by
#                        emacs. similar to those defined in textedit.py,
#                        but with a few minor emacs related changes.

CLIPBOARD = None

# Abstract base class
class Abstract:
	def __init__(self):
		self.type = 'abstract'
	
	def __cmp__(self, newevent):
		return cmp(self.__dict__, newevent.__dict__)	
		
	def __str__(self):
		return '%s()' % self.type
	__repr__ = __str__

# Insert class.
class Insert(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)
		self.type = 'insert'
		# self.text is either a string, or None (bottom)
		self.text = outstate.contents
		self.instate = instate
		self.outstate = outstate

	def __str__(self):
		return '%s(text=%s)' % (self.type, `self.text`)
	__repr__ = __str__

# Select class.
class Select(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)
		self.type = 'select'
		self.instate = instate
		self.outstate = outstate

	def __str__(self):
		return '%s(%s, %s)' % (self.type, self.instate.pos,
						   self.outstate.pos)
	__repr__ = __str__

# Delete class.
class Delete(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)
		self.type = 'delete'
		self.instate = instate
		self.outstate = outstate
	
	def __str__(self):
		return '%s(%s, %s)' % (self.type, self.instate.pos,
						   self.outstate.pos)
		return '%s(Location)' % self.type
	__repr__ = __str__

# Move class.
class Move(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)

		self.type = 'move'
		self.instate = instate
		self.outstate = outstate
	
	def __str__(self):
		return '%s(%s, %s)' % (self.type, self.instate.pos,
						   self.outstate.pos)
	__repr__ = __str__

# Copy class.
class Copy(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)
		self.type = 'copy'
		self.instate = instate
		self.outstate = outstate

# Paste class.
class Paste(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)
		self.type = 'paste'
		self.instate = instate
		self.outstate = outstate

# Cut class.
class Cut(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)
		self.type = 'cut'
		self.instate = instate
		self.outstate = outstate
	
# ----------------------------------------------------------------------
#
# BufferState class
#      Stores the state of the emacs buffer.
#
# ----------------------------------------------------------------------

class BufferState:
	def __init__(self, cursorpos, text, selection, clipboard ):

		self.set_defaults()
		self.contents = text
		self.pos = string.atoi(cursorpos)
		self.clipboard = clipboard
		p = re.compile(',')
		(a,b) = re.split(p, selection)
		if (string.atoi(a) == string.atoi(b)):
			self.selection_extent = ()
		else:
			(x,y) = textedit.rowColTuple( text, string.atoi(a) )
			(v,w) = textedit.rowColTuple( text, string.atoi(b) )
			if (x,y) < (v,w):
				self.selection_extent = ('%i.%i' % (x,y),
							 '%i.%i' % (v,w))
			else:
				self.selection_extent = ('%i.%i' % (v,w),
							 '%i.%i' % (x,y))
	def set_defaults(self):
		self.contents = ''
		self.pos = '1.0'
		self.clipboard = ''
		self.selection_extent = ()
		self.iteration_number = -1

	def __setstate__(self, dict):
		self.set_defaults()
		for key in dict.keys():
			self.__dict__[key] = dict[key]

	def __repr__(self):
		return 'Pos: %s Clipboard: <%s> Selection: <%s> Iteration %i END' %\
		       ('self.pos', 'self.clipboard', \
			'self.selection_extent', self.iteration_number)

	def __cmp__(self, other):
		if self is other:
			return 0

		cmp_contents = cmp(self.contents, other.contents)
		if cmp_contents != 0:
			return cmp_contents
		else:
			return cmp(self.pos, other.pos)

#------------------------------------------------------------------------#
#
# EmacsInterface
#    This is the interface that bridges emacs with vsprogram
#
#------------------------------------------------------------------------#
class EmacsInterface:
	def __init__(self):
		self.num_actions = 0
		self.oldEvent = ''
		self.oldState = None
		self.newState = None
		self.result_set = None
		self.vs = vsprogram.VSProgram()
		self.current_action = None
		self.recording = 0
		self.alive = 1

	# Fixup removes the \ character escaping : and \
	def fixup(self, str):
		return string.replace(string.replace(str, "\:", ":"),
				      "\\\\", "\\")
	# fancyup adds an escape character ('\') to : and \ characters.
	def fancyup(self, str):
		return string.replace(string.replace(str, "\\", "\\\\"),
				      ":", "\:")

	# getInput retrieves a message from stdin.
	def getInput(self):
		
		str = sys.stdin.read()

		# we split on ':', as long as they're not escaped.
		p = re.compile('[^\\\]:')
		sorted = re.split(p, str)

		# signals that the current recording has ended.
		if( sorted[0] == "END_RECORDING" ):
			self.end_recording( sorted )

		# signals that we've reset the recording, starting all over.
		elif( sorted[0] == "RESET" ):
			self.reset_recording()

		# we want the next guess.
		elif( sorted[0] == "NEXT_GUESS" ):
			self.guess_again()

		# Accept the current step and move on to the next.
		elif( sorted[0] == "STEP" ):
			self.step_through( sorted )

		# restore the previous buffer state.
		elif( sorted[0] == "RESTORE" ):
			self.current_action = None
			self.oldEvent = ''

		# Time to stut it all down.
		elif( sorted[0] == "DIE_DIE_DIE"):
			self.alive = 0
			
		# Handle overwriting? Here, or in the elisp code? !!!
		elif( sorted[0] == "SAVE_PROGRAM"):
			self.write_program(sorted[1])

		# Handle badly formed program files? !!!
		elif( sorted[0] == "LOAD_PROGRAM"):
			self.read_program(sorted[1])
			self.vs.reset()
		
		else:
			#If nothing else, we'll assume it's a new event.
			self.add_event( sorted )

	# Classifies event_type given in the argument, returns a new event
	# of from the event classes given in this file.
	def classify(self, in_state, out_state, event_type):
		in_state = copy.copy(in_state)
		out_state = copy.copy(out_state)

		# This is a check to make sure the ordering is correct on
		# the selection_extent boundaries.
		if in_state.selection_extent:
			sel_ext = (map(int,
				       string.split(in_state.selection_extent[0], '.')),
				   map(int,
				       string.split(in_state.selection_extent[1], '.')))
			if sel_ext[0] > sel_ext[1]:
				in_state.selection_extent = \
				(in_state.selection_extent[1],
				 in_state.selection_extent[0])
		
		if out_state.selection_extent:
			sel_ext=(map(int,
				     string.split(out_state.selection_extent[0],
						  '.')),
				 map(int,
				     string.split(out_state.selection_extent[1],
						  '.')))
			if sel_ext[0] > sel_ext[1]:
				out_state.selection_extent = \
				(out_state.selection_extent[1],
				 out_state.selection_extent[0])

		# move is straight forward.
		if( event_type == 'move' ):
			return(Move(in_state, out_state)) 

		elif( event_type == 'select' ):
			# select doesn't like having an existing selection
			# on instate.
			in_state.selection_extent = ()
			return(Select(in_state, out_state))

		# paste events are easy.
		elif( event_type == 'paste' ):
			return(Paste(in_state, out_state))

		# both cut and cut selection are essentially the same.
		elif( event_type == 'cut' or
		      event_type == 'cutSelection'):
			return(Cut(in_state, out_state))

		# copy is straight forward.
		elif( event_type == 'copy' ):
			return(Copy(in_state, out_state))

		# delete is straight forward.
		elif( event_type == 'delete' ):
			return(Delete(in_state, out_state))

		# insert is straight forward.
		elif( event_type == 'insert' ):
			return(Insert(in_state, out_state))

       		else:
			# othewise we don't have a clue.
			sys.stdout.write('\n:2:unknown action\n')
			return None

	# Add event takes an array of elements that will comprise an
	# event and parses it out appropriately.
	def add_event(self, sorted):

		self.recording = 1

	        # If it's a cut or a copy, we need to make a selection first.
		# We add the select, then move on to add the copy/cut.
		if( self.oldEvent == 'copy' or
		    self.oldEvent == 'cutSelection'):
			self.oldState.pos = self.newState.pos
			e = self.classify( self.oldState,
					   self.newState,
					   'select' )
			if e:
				try:
					self.vs.feed_nth(e)
				except vs.VSError, e:
					sys.stdout.write('\n:2:feed nth\n')
					
			self.num_actions = self.num_actions + 1

		# For a cut w/o select, we assume it's bounded by the cursor,
		# and we generate a forward or backward selection based on the
		# length of the text in the clipboard.
		if self.oldEvent == 'cut':

			tempstate = BufferState( sorted[1],
						 self.fixup( sorted[4] ),
						 sorted[2],
						 self.fixup(sorted[3]))
			

			#We've done a backward cut
			if(tempstate.pos < self.newState.pos):
				a = self.newState.pos-len(tempstate.clipboard)
			#otherwise it's forward.
			else:
				a = self.newState.pos+len(tempstate.clipboard)
				
			(x,y)=textedit.rowColTuple( self.newState.contents, a)
			(v,w)=textedit.rowColTuple( self.newState.contents,
						  tempstate.pos)
			#make sure we add them in the right order.
			if (x,y) < (v,w):
				self.newState.selection_extent = \
				('%i.%i' % (x,y), '%i.%i' % (v,w))
			else:
				self.newState.selection_extent = \
				('%i.%i' % (v,w), '%i.%i' % (x,y))
				
			self.oldState.pos = self.newState.pos
			e = self.classify( self.oldState,
					   self.newState,
					   'select' )
			if e:
				try:
					self.vs.feed_nth(e)
				except vs.VSError, e:
					sys.stdout.write('\n:2:feed nth\n')
					
			self.num_actions = self.num_actions + 1

		# Now we move on with a normal addition.  Note that if the
		# event recieved was a cut or copy, the select has already
		# been added and we're now moving on to add the cut/copy
		# event itself.
		self.oldState = self.newState
		self.newState = BufferState( sorted[1],
					     self.fixup( sorted[4] ),
					     sorted[2], self.fixup(sorted[3]))

		if( self.oldEvent != '' ):
			sys.stdout.write( self.fancyup(self.oldEvent) )
			sys.stdout.write( ' event added\n' )

			# get the actual event
			e = self.classify( self.oldState,
					   self.newState, self.oldEvent )

			if e:
				try:
					# feed it into the vsprogram
					self.vs.feed_nth(e)
				except vs.VSError, e:
					sys.stdout.write('\n:2:feed nth\n')
					
				self.num_actions = self.num_actions + 1
		#store this off.
		self.oldEvent = sorted[0]


	# end_recording signals a stop to the current recording, this
	# means we need to add the last event and feed_end.
	def end_recording(self, sorted):
		if not self.recording:
			return
		self.add_event( sorted )
		self.vs.feed_end()
		self.recording = 0
		self.oldEvent = ''
		sys.stdout.write("end recording\n\n")

	# resets all the defaults if a reset is called.
	def reset_recording(self):
		self.oldEvent = ''
		self.oldState = None
		self.newState = None
		self.num_actions = 0
		self.vs = vs.VSProgram()
		self.result_set = None
		self.current_action = None
		self.recording = 0

	# if we get a guess again, we need to perform the next action in
	# the hypothesis list.  (perform means message it to stdout here)
	def guess_again(self):
		if self.result_set:
			(hyplist, totalprob) = self.result_set.pick_next_best()
			self.perform_action(hyplist, totalprob )

	# This is where we message the hypothesis out to emacs.
	def perform_action( self, hyplist, totalprob ):

		#setup the hypothesis
		hyp = hyplist[0][0].hyp
		(command, arg, callback) = hyp
		self.current_action = (command, arg, callback)

		# We need to make sure the probability is formatted correctly.
		# Also, we don't want to return 100%, or 0% probability unless
		# that is REALLY the case, so we account for some round off.
		probability = totalprob*100
		if probability != 100:
			if probability > 99:
				probability = 99
		if probability < 1:
			probability = 1

		#format the probability as a string.
		probstr = '[%2.f%% prob.]' % probability 

		# Then we need to message the arguments back to stdout.
		# Note that each event has a different format, but there is
		# regularity.  Each message starts with a '\n:', if it is
		# a hypothesis the first char will be a 1, 2 if it's an error
		# message.  The next field is the type of event, signaled by
		# a character 0-7.  The rest depend on the event.
		if vs.hyp_inapplicable( arg ):
			# Send back the gloss, and a probability.
			sys.stdout.write( '\n:1:0:' + 
					  self.fancyup('Terminate due to "' +
						       hyplist[0][0].reason +
						       '"') + 
					  ':' + probstr + ':\n' )

		elif command == 'insert':
			# Send back the text to be inserted, gloss,
			# and probability.
			sys.stdout.write( '\n:1:1:' + repr(arg) + ':' +
					  self.fancyup(hyplist[0][0].reason) +
					  ':' + probstr + ':\n')
			
		elif command == 'move':
			# Send the new position, gloss and probability.
			position = textedit.positionFromTuple(
				self.newState.contents, arg )
			sys.stdout.write( '\n:1:2:' + repr(position) + ':' +
					  self.fancyup(hyplist[0][0].reason) +
					  ':' + probstr + ':\n')

		elif command == 'delete':
			# Send the boundaries for the delete, a gloss, and the
			# probability.
			(a,b) = arg
			a = textedit.positionFromTuple(
				self.newState.contents, a)
			b = textedit.positionFromTuple(
				self.newState.contents, b)
			sys.stdout.write( '\n:1:3:' +
					  repr(a) + ':' +
					  repr(b) + ':' +
					  self.fancyup(hyplist[0][0].reason) +
					  ':' + probstr + ':\n' )

		elif (command == 'cutSelection' or command == 'cut'):
			# Send back the gloss and a probability
			sys.stdout.write( '\n:1:4:' +
					  self.fancyup(hyplist[0][0].reason) +
					  ':' + probstr + ':\n' )

		elif command == 'copy':
			# Send the gloss and a probability.
			sys.stdout.write( '\n:1:5:' +
					  self.fancyup(hyplist[0][0].reason) +
					  ':' + probstr + ':\n' )

		elif command == 'paste':
			# Send the gloss and a probability.
			sys.stdout.write( '\n:1:6:' +
					  self.fancyup(hyplist[0][0].reason) +
					  ':' + probstr + ':\n' )

		elif command == 'select':
			# Select is special because the i believe the
			# inapplicable hypothesis comes in to here. (maybe)
			# anyway.  It sends a tuple of the new selection
			# boundaries, gloss, and probability.
			if arg:
				(a,b) = arg
				a=textedit.positionFromTuple(
					self.newState.contents, a )
			else:
				a = none
			sys.stdout.write( '\n:1:7:' + repr(a) + ':' +
					  self.fancyup(hyplist[0][0].reason) +
					  ':' + probstr + ':\n' )

	# Might be better named accept hypothesis, but i'm borring convention
	# from the collect.py code.  This sets up a new hypothesis array.
	def step_through( self, sorted ):

		# if we're recording (should never happen in SmEdit)
		# stop and continue.
		if self.recording:
			self.add_event(sorted)
			self.vs.feed_end()
			self.recording = 0

		# If we don't have a macro we need to bail.
		if not self.num_actions:
			sys.stdout.write('\n:2:No macro learned!\n')
			return

		# If we have some hypothesis we need to accept it.
		# This should be the case for every instance except the first
		# time we call this function.
		if self.current_action:
			self.finish_step( sorted )


		# get a new result set.
		self.result_set =\
				self.vs.execute_nth( \
			self.newState,
			self.vs.get_action_number())

		#if this fails, we're done.  can't get any more predictions.
		if not self.result_set:
			sys.stdout.write('\n:2:Error: no more preditions\n')
			return

		self.result_set.initialize()


	# This is called for steps that the user has authorized smartedit to
	# make.
	def finish_step( self, sorted ):
		if self.current_action:
			try:
				# we need a post_execute, slightly different
				# from a normal add event.
				self.post_execute(self.current_action, sorted)
			finally:
				self.current_action = None
		else:
			#never ever seen this error, don't expect to.
			sys.stdout.write( '\n:2:FINISH CALLED W/o EVENT\n' )

	# We do a post execute smartedit makes a correct guess as to the
	# action.
	def post_execute( self, action, sorted ):

		(command, arg, callback) = action

		if vs.hyp_inapplicable(arg):
			sys.stdout.write(
				'\n:2:prediction of inappliable executed\n')
		
		self.oldState = self.newState
		self.newState = BufferState( sorted[1],
					     self.fixup( sorted[4] ),
					     sorted[2], self.fixup(sorted[3]))

		if( command != '' ):

			e = self.classify( self.oldState, self.newState,
					   command )

			if e:
				try:
					self.vs.feed_nth(e)
				except vs.VSError, e:
					sys.stdout.write(
						'\n:2:feeding post_execute\n')
				self.num_actions = self.num_actions + 1

# This is the main loop, once the file is loaded this loops until
# someone signals it to die, getting input.

	def write_program(self, filename):
		sys.stdout.write( filename )
		f = open(filename, 'w')
		pickle.dump(self.vs, f)

	def read_program(self, filename):
		sys.stdout.write( filename )
		f = open(filename, 'r')
		self.vs = pickle.load(f)

		
interface = EmacsInterface()
while( interface.alive ):

	interface.getInput()
