#!/usr/bin/env python

from Tkinter import *
import tkMessageBox
try:
	import Pmw
except:
	import PmwLocal as Pmw
import os, string, re, anydbm, gzip, cPickle, cStringIO, glob
import textedit
import HistoryCombo

DEBUG = 0

DBDIR = os.environ.get('SMARTEDITSCENARIODIR', 'scenarios')

def get_scenarios():
	global DBDIR
	if not os.path.exists(DBDIR):
		print 'Scenario directory %s does not exist; using "."' % DBDIR
		DBDIR = '.'

	files = glob.glob(os.path.join(DBDIR, '*.pck.gz'))
	# Strip absolute path
	files = map(lambda x: os.path.basename(x), files)
	# Strip trailing .pck and .gz
	files = map(lambda x: os.path.splitext(x)[0], files)
	files = map(lambda x: os.path.splitext(x)[0], files)
	files.sort()
	return files
	
class Example:
	"""An example in a scenario: initial program state, sequence of
	low-level events, final program state."""
	def __init__(self, before, events, after):
		self.beforestate = before
		self.events = events
		self.afterstate = after
	
def load_scenario(name):
	if DEBUG: print 'Loading scenario', name
	fname = os.path.join(DBDIR, '%s.pck.gz' % name)
	if not os.path.exists(fname):
		raise IOError, 'scenario %s not found' % name
	
	s = Scenario(name)
	# Workaround gzip/cPickle bug
	try:
		file = cStringIO.StringIO(gzip.open(fname, 'rb').read())
	except ValueError, e:
		print 'Error uncompressing scenario:', e
		dict = {'description':'', 'examples':[]}
	else:
		try:
			dict = cPickle.load(file)
			file.close()
		except cPickle.UnpicklingError, e:
			print 'Error unpickling scenario:', e
			dict = {'description':'', 'examples':[]}
	s.description = dict['description']
	s.examples = dict['examples']
	return s

class Scenario:
	def __init__(self, name):
		self.name = name
		self.description = ''
		# List of Example instances for this scenario
		self.examples = []

	def save(self):
		if not self.name:
			print 'Unable to save unnamed scenario!'
			return

		fname = os.path.join(DBDIR, '%s.pck.gz' % self.name)
		file = gzip.open(fname, 'wb')
		dict = {'description': self.description,
			'examples': self.examples}
		cPickle.dump(dict, file, 1)
		file.close()
		m = tkMessageBox.showinfo(title='Save successful',
			message=('Scenario %s saved' % self.name))

	def add_iteration(self, beforestate, events, afterstate, exnum=None):
		"""Add a single iteration as a tuple of (state, events, state).
		If exnum is None, add to the end, otherwise replace exnum."""
		example = Example(beforestate, events, afterstate)
		if exnum is None:
			self.examples.append(example)
		else:
			self.examples[exnum] = example
	
	def __len__(self):
		return len(self.examples)
	
	def __getitem__(self, index):
		return self.examples[index]

	def __delitem__(self, index):
		del self.examples[index]
	
	def __getslice__(self, i, j):
		return self.examples[i:j]
	
	def delete_all(self):
		self.examples = []
	
class ScenarioPopup(Toplevel):
	def __init__(self, master, scenario = None):
		Toplevel.__init__(self)
		self.title('Scenario editor')
		self.master = master
		scenario_list = get_scenarios()
		if scenario is None:
			if scenario_list:
				self.scenario = load_scenario(scenario_list[0])
			else:
				self.scenario = None
		else:
			self.scenario = scenario

		# Create widgets
		self.name = HistoryCombo.HistoryCombo(self,
			label_text='Scenario name:',
			labelpos='w', selectioncommand=self.choose_scenario,
			scrolledlist_items = scenario_list)
		if scenario_list:
			self.name.selectitem(0)
		self.name.pack(side=TOP, expand=1, fill=X)

		l = Label(self, text='Description:', anchor='w')
		l.pack(side=TOP, expand=0, fill=X)
		self.description = Pmw.ScrolledText(self, text_height=10,
			text_width=70, text_wrap='word')
		self.description.pack(side=TOP, expand=1, fill=BOTH)

		# How many labelled examples this scenario has
		f = Frame(self)
		f.pack(side=TOP, expand=1, fill=X)
		l = Label(f, text='Labelled examples:')
		l.pack(side=LEFT)
		if self.scenario is None:
			numiters = 0
		else:
			numiters = len(self.scenario)
		self.numiters = Label(f, text=`numiters`)
		self.numiters.pack(side=LEFT)

		# Iteration manipulation
		g = Pmw.Group(self, tag_text='Create new examples')
		g.pack(side=TOP, expand=1, fill=BOTH)
		self.iterationbuttons = []
		b = Button(g.interior(), text='Start recording examples',
			command=self.start_recording)
		b.pack(side=TOP, expand=0, fill=NONE)
		self.iterationbuttons.append(b)
		self.start_recording_button = b
		b = Button(g.interior(), text='Next example',
			command=self.record_iteration)
		b.pack(side=TOP, expand=0, fill=NONE)
		self.iterationbuttons.append(b)
		b = Button(g.interior(), text='Finished recording',
			command=self.done_recording)
		b.pack(side=TOP, expand=0, fill=NONE)
		self.iterationbuttons.append(b)
		self.finish_button = b

		# Manipulate existing examples
		g = Pmw.Group(self, tag_text='Manipulate recorded examples')
		g.pack(side=TOP, expand=1, fill=BOTH)
		self.example_list = Pmw.ScrolledListBox(g.interior(),
			listbox_selectmode='extended')
		self.example_list.pack(side=LEFT, expand=1, fill=BOTH)
		h = Frame(g.interior())
		h.pack(side=RIGHT, expand=1, fill=BOTH)

		b = Button(h, text='Delete example(s)',
			command=self.delete)
		b.pack(side=TOP, expand=0, fill=NONE)
		self.iterationbuttons.append(b)
		b = Button(h, text='Example initstate',
			command=self.example_initstate)
		b.pack(side=TOP, expand=0, fill=NONE)
		self.iterationbuttons.append(b)
		b = Button(h, text='Playback example(s)',
			command=self.playback_example)
		b.pack(side=TOP, expand=0, fill=NONE)
		self.iterationbuttons.append(b)
		b = Button(h, text='Rerecord this example',
			command=self.rerecord_example)
		b.pack(side=TOP, expand=0, fill=NONE)
		self.iterationbuttons.append(b)
		b = Button(h, text='Test this scenario',
			command=self.test_scenario)
		b.pack(side=TOP, expand=0, fill=NONE)
		self.iterationbuttons.append(b)
#		b = Button(h, text='Cancel playback',
#			command=self.cancel_playback)
#		b.pack(side=TOP, expand=0, fill=NONE)
#		self.iterationbuttons.append(b)

		# Button box
		buttonbox = Pmw.ButtonBox(self)
		buttonbox.pack(side=BOTTOM, expand=0, fill=X)
		buttonbox.add('OK', command=self.ok)
		buttonbox.add('Save', command=self.save)
		buttonbox.add('Cancel', command=self.destroy)
		buttonbox.setdefault('OK')
		buttonbox.alignbuttons()

		self.update_description()
		self.update_examples()
	
	def save_description(self):
		self.scenario.description = self.description.get('1.0', 'end-1c')

	def choose_scenario(self, name):
		scenario_list = get_scenarios()
		if not name:
			self.scenario = None
		elif name in scenario_list:
			self.scenario = load_scenario(name)
		else:
			self.scenario = Scenario(name)
		self.update_description()
		self.update_examples()
		self.master.reset()
		if len(self.scenario) >= 1:
			self.master.set_text_state(self.scenario[0].beforestate)

	def save(self):
		name = self.name.get()
		if not name:
			tkMessageBox.showerror('No scenario',
				'Please name the scenario first')
			return 0
		self.scenario.description = self.description.get('1.0', 'end-1c')
		self.scenario.save()
		return 1
	
	def ok(self):
		if self.save():
			self.destroy()
	
	def change_view(self, view):
		if view == 'Hypothesis':
			hyp = self.scenario.str_hypotheses()
			if hyp is None:
				self.set_hypothesis('')
			else:
				self.set_hypothesis(hyp)
		else:
			match = re.match('Example #(\d+)', view)
			if match is None:
				return
			num = int(match.group(1))
			actions = self.scenario.get_example(num-1)
			if actions is None:
				self.set_hypothesis("Example #%i doesn't exist!" % i)
			else:
				self.set_hypothesis(string.join(map(str, actions),
					'\n'))

	def start_recording(self):
		self.start_recording_button.configure(state='disabled')
		self.master.record_scenario()
	
	def done_recording(self):
		self.start_recording_button.configure(state='normal')
		self.master.done_scenario(self.scenario)
		self.update_examples()

	def record_iteration(self):
		self.master.iterate_scenario(self.scenario)

	def delete(self):
		assert(self.scenario)

		# Grab the selection
		selected = self.example_list.curselection()
		if not selected: return

		# Transform the selection to a list of indexes
		items = map(int, selected)

		# Reverse sort the indexes so that their deletions won't
		# interfere with each other.
		items.sort()
		items.reverse()
		
		assert(items[0] <= len(self.scenario))

		# Delete all the items
		for i in items:
			del self.scenario[i]

		# Reflect the changes in the example list on the screen.
		self.update_examples()
	
	def delete_all(self):
		self.scenario.delete_all()
		self.update_examples()

	def example_initstate(self):
		assert(self.scenario)
		selected = self.example_list.curselection()
		if not selected: return
		item = int(selected[0])
		self.master.set_text_state(self.scenario[item].beforestate)
	
	def playback_example(self):
		assert(self.scenario)
		selected = self.example_list.curselection()
		if not selected: return
		items = map(int, selected)
		assert(max(items) <= len(self.scenario))
		self.master.replay_scenario_examples(self.scenario, items)
	
	def rerecord_example(self):
		assert(self.scenario)
		selected = self.example_list.curselection()
		if not selected: return
		if len(selected) != 1:
			tkMessageBox.showerror('Wrong number of examples',
				'Please select only one example for re-recording.')
			return
		exnum = int(selected[0])
		# Disable all buttons
		for b in self.iterationbuttons:
			b.configure(state='disabled')
		self.finish_button.configure(state='normal')
		self.master.record_single_iteration(self.scenario[exnum])
		self.finish_button.configure(command=lambda s=self, e=exnum:
			s.finish_single_example(e))
	
	def finish_single_example(self, exnum):
		self.master.done_single_iteration(self.scenario, exnum)
		self.update_examples()
		for b in self.iterationbuttons:
			b.configure(state='normal')
		self.finish_button.configure(command=self.done_recording)
	
	def cancel_playback(self):
		self.master.cancel_run()
	
	def test_scenario(self):
		assert(self.scenario)
		self.master.test_scenario(self.scenario)

	def update_description(self):
		self.description.delete('1.0', 'end')
		if self.scenario is not None:
			self.description.insert('end', self.scenario.description)

	def update_examples(self):
		"Update example list from (changed) scenario"
		self.example_list.delete(0, 'end')
		if self.scenario is None:
			self.numiters.configure(text='0')
			for b in self.iterationbuttons:
				b.configure(state='disabled')
		else:
			self.numiters.configure(text=`len(self.scenario)`)
			for b in self.iterationbuttons:
				b.configure(state='normal')
			for i in range(len(self.scenario)):
				self.example_list.insert('end', 'Example %i' % (i + 1))

def main():
	Popup = ScenarioPopup(None)
	Popup.mainloop()

if __name__ == '__main__':
	main()

