#!/usr/bin/env python

import copy, operator, string
import vs, vslocation, vsinsert, textedit

class VSActionUnion(vs.VersionSpace):
	def __init__(self, act):
		self.type = act.type
		if act.type == 'move':
			self.action = VSMove(act.instate, act.outstate)
		elif act.type == 'insert':
			self.action = VSInsert(act.instate, act.outstate)
		elif act.type == 'delete':
			self.action = VSDelete(act.instate, act.outstate)
		elif act.type == 'select':
			self.action = VSSelect(act.instate, act.outstate)
		elif act.type == 'cut':
			self.action = VSCut(act.instate, act.outstate)
		elif act.type == 'copy':
			self.action = VSCopy(act.instate, act.outstate)
		elif act.type == 'paste':
			self.action = VSPaste(act.instate, act.outstate)
		elif act.type == 'deleteselection':
			self.action = VSDeleteSelection(act.instate, act.outstate)
		else:
			print 'Unknown action type:', act.type
	
	def __repr__(self):
		return 'VS:' + self.type

	# Classif is 0 or 1; instate/outstate are domain-dependent
	def feed(self, act):
		if act.type != self.type:
			#DEBUG!!
			#print 'Collapsing due to clashing action types', \
			#	 act.type, 'and', self.type
			self.action = None
		elif self.action:
			self.action.feed(act.instate, act.outstate)
	
	# Return a ResultSet instance
	def execute(self, instate):
		if not self.action:
			return vs.ResultSet()
		else:
			return self.action.execute(instate)
	
	def collapsed(self):
		if self.action:
			return self.action.collapsed()
		else:
			return 1

	def num_hyps(self):
		if self.action:
			return self.action.num_hyps()
		else:
			return 0

# ----------------------------------------------------------------------
# Action-specific version spaces
# ----------------------------------------------------------------------

class VSAction(vs.VersionSpace):
	def __init__(self, instate, outstate):
		pass
	def execute(self, instate):
		pass
	def feed(self, instate, outstate):
		pass
class VSMove(VSAction):
	def __init__(self, instate, outstate):
		self.loc = vslocation.VSLocation(instate, outstate)

	def feed(self, instate, outstate):
		self.loc.feed(instate, outstate)

	def execute(self, instate):
		rs = vs.ResultSet()
		locresults = self.loc.execute(instate)
		for (hyp, prob) in locresults:
			new_hyp = vs.Hypothesis(('move', hyp.hyp),
				'Move to %s' % hyp.reason)
			rs.add(new_hyp, prob)
		return rs

	def collapsed(self):
		return self.loc.collapsed()

	def num_hyps(self):
		return self.loc.num_hyps()

class VSDelete(vs.VersionSpaceUnion):
	def __init__(self, instate, outstate):
		uni = VSUnidirectionalDelete(instate, outstate)
		bi = VSBidirectionalDelete(instate, outstate)
		self.vses = [
			(uni, 0.5, "BidirectionalDelete"),
			(bi, 0.5, "UnidirectionalDelete")
			]
		# All probabilities must sum to one
		assert(reduce(operator.add, map(lambda x: x[1], self.vses))==1.0)
	
class VSUnidirectionalDelete(VSAction):
	"A version space representing a unidirectional delete action."
	
	def __init__(self, instate, outstate):
		self.has_collapsed = 0
		
		forward = outstate.pos
		backward = outstate.pos + len(instate.contents) - \
				 len(outstate.contents)
		outstate = copy.copy(instate)
		if backward == instate.pos:
			# The whole delete extent is forward
			outstate.pos = forward
		elif forward == instate.pos:
			outstate.pos = backward
		else:
			#DEBUG!!
			print 'VSUniDelete collapsing because nothing stayed still.'
			self.has_collapsed = 1
			
		# Create location instance in which staying still is
		# allowed for string matching.
		self.loc = vslocation.VSLocation(instate, outstate, 0)

	def feed(self, instate, outstate):
		if self.has_collapsed:
			return
		
		forward = outstate.pos
		backward = outstate.pos + len(instate.contents) - \
				 len(outstate.contents)
		outstate = copy.copy(instate)
		if backward == instate.pos:
			# The whole delete extent is forward
			outstate.pos = forward
		elif forward == instate.pos:
			outstate.pos = backward
		else:
			#DEBUG!!
			print 'VSUniDelete collapsing because nothing stayed still.'
			self.has_collapsed = 1
			
		# Create location instance in which staying still is
		# allowed for string matching.
		self.loc.feed(instate, outstate)

	def execute(self, instate):
		rs = vs.ResultSet()

		if self.has_collapsed:
			return rs
		
		results = self.loc.execute(instate)
		here = textedit.rowColTuple(instate.contents, instate.pos)
		for (hyp, prob) in results:
			if vs.hyp_inapplicable(hyp.hyp):
				new_hyp = vs.Hypothesis(('delete', hyp.hyp),
								 'Delete to %s' % hyp.reason)
			else:
				if here < hyp.hyp:
					left = here
					right = hyp.hyp
					left_reason = 'here'
					right_reason = hyp.reason
				else:
					left = hyp.hyp
					right = here
					left_reason = hyp.reason
					right_reason = 'here'
				new_hyp = vs.Hypothesis(('delete', (left, right)),
								 'Delete from %s to %s' %
								 (left_reason, right_reason))

			rs.add(new_hyp, prob)

		return rs
	
	def collapsed(self):
		return self.has_collapsed or self.loc.collapsed()

	def num_hyps(self):
		return self.loc.num_hyps()

class VSBidirectionalDelete(VSAction):
	"""A version space representing a bidirectional delete action.


	Incorporates one VS representing the backwards extent of the
	deleted region and another representing the forwards extent so
	that it can delete around a mark.
	
	Does a join of two Location version spaces to determine the extents of
	the deletion region."""
	
	def __init__(self, instate, outstate):
		outstate_back = copy.copy(instate)
		outstate_back.pos = outstate.pos
		outstate_fore = copy.copy(instate)
		outstate_fore.pos = outstate.pos + len(instate.contents) - \
						len(outstate.contents)
		# Create location instances in which staying still is
		# allowed for string matching.
		self.loc_back = vslocation.VSLocation(instate, outstate_back, 0)
		self.loc_fore = vslocation.VSLocation(instate, outstate_fore, 0)

	def feed(self, instate, outstate):
		outstate_back = copy.copy(instate)
		outstate_back.pos = outstate.pos
		outstate_fore = copy.copy(instate)
		outstate_fore.pos = outstate.pos + len(instate.contents) - \
						len(outstate.contents)
		self.loc_back.feed(instate, outstate_back)
		self.loc_fore.feed(instate, outstate_fore)

	def execute(self, instate):
		rs = vs.ResultSet()
		loc_backresults = self.loc_back.execute(instate)
		loc_foreresults = self.loc_fore.execute(instate)
		
		loc = textedit.rowColTuple(instate.contents, instate.pos)

		# Weed out hypotheses that are not joinable with anything.
		# Also, spit out inapplicable hypotheses right away.

		back_total = 0.0
		new_backresults = []
		for (back_hyp, back_prob) in loc_backresults:
			if back_hyp.isinapplicable() or back_hyp.hyp > loc:
				hyp = vs.Hypothesis(('delete',
							   vs.hyp_make_inapplicable()),
							  'Delete with backward bound %s' %
							  back_hyp.reason)
				rs.add(hyp, back_prob)
			else:
				new_backresults.append((back_hyp, back_prob))
				back_total = back_total + back_prob
		loc_backresults = new_backresults
			
		new_foreresults = []
		for (fore_hyp, fore_prob) in loc_foreresults:
			if fore_hyp.isinapplicable() or fore_hyp.hyp < loc:
				hyp = vs.Hypothesis(('delete',
							   vs.hyp_make_inapplicable()),
							  'Delete with forward bound %s' %
							  fore_hyp.reason)
				rs.add(hyp, fore_prob * back_total)
			else:
				new_foreresults.append((fore_hyp, fore_prob))
		loc_foreresults = new_foreresults

		for (back_hyp, back_prob) in loc_backresults:
			for (fore_hyp, fore_prob) in loc_foreresults:
				## Should it be possible to predict deleting nothing? 
				# Reacting to this would create an instance of a
				# non-independent join
				#if back_hyp.hyp == fore_hyp.hyp:
					#DEBUG!!
					#print 'Allowing prediction of delete nothing.'
					#pass

				if back_hyp.hyp == loc:
					back_reason = 'here'
				else:
					back_reason = back_hyp.reason
				if fore_hyp.hyp == loc:
					fore_reason = 'here'
				else:
					fore_reason = fore_hyp.reason
				hyp = vs.Hypothesis(('delete', (back_hyp.hyp,
									    fore_hyp.hyp)),
							  'Delete from %s to %s' %
							  (back_reason, fore_reason))
				
				rs.add(hyp, back_prob*fore_prob)

		# Should be normalized
		#rs.normalize()
		return rs
	
	def collapsed(self):
		return self.loc_back.collapsed() or self.loc_fore.collapsed()
	
	def num_hyps(self):
		return self.loc_back.num_hyps() * self.loc_fore.num_hyps()

# Hypotheses are of the form:
# ('select', INAPPLICABLE) or
# ('select', ((tuple for selection extent), (tuple where point ends up)))
class VSSelect(VSAction):
	def __init__(self, instate, outstate):
		self.hypothesis = ['right', 'left', 'opposite', 'same']
		self.value = {'right' : 0.3,
				    'opposite' : 0.6,
				    'left' : 0.05,
				    'same' : 0.05}
		if vs.FORCE_SAME:
			self.hypothesis = ['same']
		
		clean_outstate = self.check(instate, outstate)
		self.loc = vslocation.VSLocation(instate, clean_outstate)

	def feed(self, instate, outstate):
		clean_outstate = self.check(instate, outstate)
		self.loc.feed(instate, clean_outstate)

	def execute(self, instate):
		rs = vs.ResultSet()
		locresults = self.loc.execute(instate)
		for (hyp, prob) in locresults:
			#HACK!!! I'm *not* checking whether a selection
			# exists so that I can explicitly allow making a
			# selection when one is already there. This is
			# largely for smemacs.
			if not self.hypothesis or \
			   vs.hyp_inapplicable(hyp.hyp):
				arg = vs.hyp_make_inapplicable()
				hyp = vs.Hypothesis(('select', arg), 'Select up to %s' %
							  hyp.reason)
				rs.add(hyp, prob)
			else:
				in_pos = textedit.rowColTuple(instate.contents,
										instate.pos)
				if hyp.hyp > in_pos:
					right = hyp.hyp
					left = in_pos
				else:
					right = in_pos
					left = hyp.hyp
				for side in self.hypothesis:
					if side == 'left':
						arg = (hyp.hyp, left)
					elif side == 'right':
						arg = (hyp.hyp, right)
					elif side == 'opposite':
						arg = (hyp.hyp, hyp.hyp)
					elif side == 'same':
						arg = (hyp.hyp, in_pos)
					else:
						print 'UNKNOWN HYPOTHESIS!'
					new_hyp = vs.Hypothesis(('select', arg), 'Select up to %s and end on the %s side' %
								  (hyp.reason, side))
					rs.add(new_hyp, prob * self.value[side])
					
		if vs.ALWAYS_NORMALIZE:
			rs.normalize()
		return rs

	def collapsed(self):
		return not self.hypothesis or self.loc.collapsed()

	def num_hyps(self):
		if not self.hypothesis:
			return 0
		else:
			return self.loc.num_hyps()

	def check(self, instate, outstate):
		"""Constructs an output state which simulates a move
		to the far side of the selection. Also maintains the hypotheses
		explaining the side the cursor ends up on."""

		if len(outstate.selection_extent) == 0:
			self.hypothesis = []
			return outstate

		sel_range = (tuple(textedit.mark_split(outstate.selection_extent[0])),
			     tuple(textedit.mark_split(outstate.selection_extent[1])))
		sel_end = max(sel_range)
		sel_begin = min(sel_range)
		in_pos = textedit.rowColTuple(instate.contents, instate.pos)
		out_pos = textedit.rowColTuple(outstate.contents, outstate.pos)

		# Update based on the states
		if out_pos == sel_end and 'left' in self.hypothesis:
			# Disproves left
			self.hypothesis.remove('left')

		if out_pos == sel_begin and 'right' in self.hypothesis:
			# Disproves right
			self.hypothesis.remove('right')

		if out_pos != in_pos and 'same' in self.hypothesis:
			# Disproves same
			self.hypothesis.remove('same')

		if out_pos == in_pos and 'opposite' in self.hypothesis:
			# Disproves opposite
			self.hypothesis.remove('opposite')

		# Build the artificial output state
		newstate = copy.copy(outstate)
		if out_pos == in_pos:
			if out_pos == sel_end:
				new_pos = textedit.positionFromTuple(outstate.contents,
											  sel_begin)
			else:
				new_pos = textedit.positionFromTuple(outstate.contents,
											  sel_end)
			newstate.pos = new_pos

		return newstate

class VSInsert(vs.VersionSpaceUnion,VSAction):
	"""The Insert VS is a union of: a constant string VS, a
	number-and-string VS, and an indented-string VS."""

	def __init__(self, instate, outstate):
		"instate and outstate are both program states"

		strhyp = vs.VSConstString()
		strhyp.outputfunc = 'inserted_string'
		strhyp.outputfunc_up = 'toinsert'
		strhyp.reasonfunc = 'insert_reason'

		inthyp = vsinsert.VSStringNumberString()
		inthyp.outputfunc = 'inserted_string'
		inthyp.outputfunc_up = 'toinsert'
		inthyp.reasonfunc = 'insert_reason'

		pyindent = vsinsert.VSPythonIndent()
		pyindent.outputfunc = 'inserted_string'
		pyindent.outputfunc_up = 'toinsert'
		pyindent.inputfunc = 'get_curline'
		pyindent.reasonfunc = 'insert_reason'

		self.vses = [
			(strhyp, 0.25, 'Constant string'),
			(inthyp, 0.40, 'Number string'),
			(pyindent, 0.35, 'Python indentation')
		]
		assert(reduce(operator.add, map(lambda x: x[1], self.vses))==1.0)
		self.feed(instate, outstate)

	def inserted_string(self, instate, outstate):
		start = instate.pos
		end = outstate.pos
		return outstate.contents[start:end]

	def toinsert(self, hyp, instate):
		return ('insert', hyp.hyp)
	
	def insert_reason(self, hyp, instate):
		return 'Insert %s' % hyp.reason

	def get_curline(self, instate):
		thislinebegins = string.rfind(instate.contents, '\n', 0,
			instate.pos)

		# TODO: check for quoted strings, out of bounds errors, etc.
		while 1:
			while thislinebegins >= 0 and \
				 instate.contents[thislinebegins-1] == '\\':
				thislinebegins = string.rfind(instate.contents,
					'\n', 0, thislinebegins)
			if thislinebegins < 0:
				thislinebegins = 0
			bracketcount = 0
			parencount = 0
			thisline = instate.contents[thislinebegins:instate.pos]
			for i in range(len(thisline)-1, -1, -1):
				if thisline[i] == ')':
					parencount = parencount + 1
				elif thisline[i] == ']':
					bracketcount = bracketcount + 1
				elif thisline[i] == '(':
					parencount = parencount - 1
				elif thisline[i] == '[':
					bracketcount = bracketcount - 1
			if bracketcount > 0 or parencount > 0:
				thislinebegins = string.rfind(instate.contents,
					'\n', 0, thislinebegins)
			else:
				break
		return thisline

class VSCut(VSAction):
	def __init__(self, instate, outstate):
		pass

	def feed(self, instate, outstate):
		pass

	def execute(self, instate):
		rs = vs.ResultSet()
		if instate.selection_extent:
			arg = None
		else:
			arg = vs.hyp_make_inapplicable()
			
		rs.add(vs.Hypothesis(('cut', arg), 'Cut the selection'), 1.0)
		return rs
	
	def num_hyps(self):
		return 1
	
	def collapsed(self):
		return 0

class VSPaste(VSAction):
	def __init__(self, instate, outstate):
		pass

	def feed(self, instate, outstate):
		pass

	def execute(self, instate):
		rs = vs.ResultSet()
		rs.add(vs.Hypothesis(('paste', None), 'Paste the selection'), 1.0)
		return rs

	def num_hyps(self):
		return 1
	
	def collapsed(self):
		return 0

class VSCopy(VSAction):
	def __init__(self, instate, outstate):
		pass

	def feed(self, instate, outstate):
		pass

	def execute(self, instate):
		rs = vs.ResultSet()
		if instate.selection_extent:
			arg = None
		else:
			arg = vs.hyp_make_inapplicable()
			
		rs.add(vs.Hypothesis(('copy', arg), 'Copy the selection'), 1.0)

		return rs

	def num_hyps(self):
		return 1

	def collapsed(self):
		return 0

class VSDeleteSelection(VSAction):
	def __init__(self, instate, outstate):
		pass

	def feed(self, instate, outstate):
		pass

	def execute(self, instate):
		rs = vs.ResultSet()
		if instate.selection_extent:
			arg = None
		else:
			arg = vs.hyp_make_inapplicable()
			
		rs.add(vs.Hypothesis(('deleteselection', arg),
					   'Delete the selection'), 1.0)
		return rs

	def num_hyps(self):
		return 1

	def collapsed(self):
		return 0


