#!/usr/bin/env python

# Version spaces dealing with locations in a text buffer

DEBUG = 0
import string, math, copy, sys, operator, re, os
from types import *
import textedit, vs, char_classes

# A version space for (instate, outstate) cursor location states
class VSLocation(vs.VersionSpaceUnion):
	def __init__(self, instate, outstate, stay_allowed = 0):
		DISJUNCT_ALLOWED = 0 # change to 1 to allow disjunctive strings
		
		# DEBUG!!
		#print 'Input state:', instate
		#print 'Output state:', outstate

		# Initialize version spaces
		rowcol = VSRowCol(instate, outstate)
		offset = VSCharOffset(instate.pos, outstate.pos)
		self.shift  = VSWordShift(instate, outstate)
		charclass = VSCharClass(instate, outstate)
		findfix = VSFindFix(instate, outstate, stay_allowed)

		self.shift.outputfunc_up = 'shift_outputfunc_up'

		offset.inputfunc = 'char_offset_inputfunc'
		offset.outputfunc = 'char_offset_outputfunc'
		offset.reasonfunc = 'char_offset_reasonfunc'
		offset.outputfunc_up = 'char_offset_outputfunc_up'

		if DISJUNCT_ALLOWED:
			disjunct = VSStringDisjunct(instate, outstate)

			self.vses = [
				(charclass, 0.05, 'CharClass'),
				(self.shift, 0.05, 'WordShift'),
				(disjunct, 0.10, 'StringDisjunct'),
				(offset, 0.10, 'CharOffset'),
				(rowcol, 0.20, 'RowCol'),
				(findfix, 0.50, 'FindFix')
				]
		else:
			self.vses = [
				# VS, prior probability, name, reason_function(hyp,instate)
				(self.shift, 0.10, 'WordShift'),
				(offset, 0.10, 'CharOffset'),
				(charclass, 0.10, 'CharClass'),
				(rowcol, 0.20, 'RowCol'),
				(findfix, 0.50, 'FindFix')
				]

		# All probabilities must sum to one
		assert(reduce(operator.add, map(lambda x: x[1], self.vses))==1.0)

	def shift_outputfunc_up(self, hyp, instate):
		if type(hyp.hyp) is IntType:
			newpos = self.shift.getnewpos(instate)
			return textedit.rowColTuple(instate.contents, newpos)
		return vs.hyp_make_inapplicable()

	def char_offset_inputfunc(self, instate):
		return instate.pos

	def char_offset_outputfunc(self, instate, outstate):
		return outstate.pos

	def char_offset_outputfunc_up(self, hyp, instate):
		if hyp.hyp > len(instate.contents) or hyp.hyp < 0:
			return vs.hyp_make_inapplicable()
		return textedit.rowColTuple(instate.contents, hyp.hyp)

	def char_offset_reasonfunc(self, hyp, instate):
		offset = hyp.hyp - instate.pos
		if offset < 0:
			offset = offset * -1
			return "%i characters left" % offset
		return "%i characters right" % offset		

import trie
class VSStringDisjunct(vs.VersionSpace):
	def __init__(self, instate, outstate):
##		print '** VSHeadache.__init__(self, instate, outstate) **'
		
		self.limit = 20 # max size of any hypothesis
		# list of (text, position) tuples that is the history of moves
		self.history = []
		self.examples = trie.Trie()
		self.hypothesis = 'start'
		
		self.feed(instate, outstate)

##		print '** END VSHeadache.__init__(self, instate, outstate) **'
	
	# Classif is 0 or 1; instate/outstate are domain-dependent
	def feed(self, instate, outstate):
##		print '** VSHeadache.feed(self, instate, outstate) **'

		self.history.append((instate.contents[instate.pos:outstate.pos + self.limit], \
				    outstate.pos - instate.pos))

		self.add_pos_examples(self.history[-1])
		self.add_neg_examples(self.history[-1])

##		print '** END VSHeadache.feed(self, instate, outstate) **'

	def add_pos_examples(self, (text, position)):
##		print '  ** VSHeadache.add_pos_examples(self, text, position) **'

		i = position + 1
		while i <= position + self.limit:
			self.examples.insert(text[position:i], 1)
			i = i + 1

##		print '  ** END VSHeadache.add_pos_examples(self, text, position) **'

	def add_neg_examples(self, (text, position)):
##		print '  ** VSHeadache.add_neg_examples(self, text, position) **'
		
		#!!! start the count from 1? (from position 0, it's a postitive)
		#i = 0
		i = 1
		while i < position:
			j = i + 1
			while j <= i + self.limit:
				self.examples.insert(text[i:j], 0)
				j = j + 1
			i = i + 1

##		print '  ** END VSHeadache.add_neg_examples(self, text, position) **'


	def execute(self, instate):
##		print '** VSHeadache.execute(self, instate) **'
		
		rs = vs.ResultSet()
		if self.collapsed():
			return rs

		predict = {}

##		print '  saving trie...'
		self.examples.save()
##		print '  DONE saving trie'
		
##		print '  counting examples...'

		i = instate.pos + 1
		while i < len(instate.contents):
			predict[i] = 0.0

			if len(instate.contents) < i + self.limit:
				max = len(instate.contents)
			else:
				max = i + self.limit

			# temporarily add the negative examples
			j = instate.pos + 1
			if i > j + self.limit:
				j = i - self.limit
			while j < i:
				self.examples.insert(instate.contents[j:i], 0)
				j = j + 1

			j = i + 1
			chars = instate.contents[i:j]
			while self.examples.search(chars) == 0 and \
			      j <= max:
				j = j + 1
				chars = instate.contents[i:j]

			k = j
			chars = instate.contents[i:k]
			while self.examples.search(chars) >= 1 and \
			      k <= max:
				k = k + 1
				chars = instate.contents[i:k]
				predict[i] = predict[i] + 1

			i = i + 1

##		print '  converting map to list...'
		predict_list = [0] * len(predict)
		i = 0
		for x in predict.keys():
			predict_list[i] = (predict[x], x)
			i = i + 1

		del predict

##		print '  acquiring total...'
		i = 0
		total = 0.0
		while i < len(predict_list):
			value = predict_list[i][0]
			
			total = total + value
			i = i + 1

##		print '  number of hypotheses: ', len(predict_list)
##		print '  loading hypothesis into ResultSet...'
		i = 0
		while i < len(predict_list):
			value = predict_list[i][0]
			position = predict_list[i][1]
			
			position = textedit.rowColTuple(instate.contents, position)
			hyp = vs.Hypothesis(position, 'disjunctive string class')
			if value != 0:
				rs.add(hyp, value / total)
			i = i + 1

##		print '  restoring trie...'
		self.examples.restore()
##		print '  DONE restoring trie'

##		print '** END VSHeadache.execute(self, instate) **'
		return rs		

	def collapsed(self):
		"""Returns true iff the version space has collapsed and is empty."""

		return self.hypothesis == None

	def num_hyps(self):
		return 0

#!!! known bug:
# the tokenizer cannot convert '\n' into the whitespace token.  problem
#   is from VSFindPrefix--it uses rows and cols to make predictions,
#   but changing all '\n's into the whitespace token makes everything
#   one row.  A work-around has yet to present itself (see
#   token_outputfunc_up)
class VSFindFix(vs.VersionSpaceUnion):
	def __init__(self, instate, outstate, stay_allowed=0):
		self.classes = self.get_classes()

		tok_instate = self.tokenize_state(instate)
		tok_outstate = self.tokenize_state(outstate)
		
		prefix = VSFindPrefix(instate, outstate, stay_allowed)
		tok_prefix = VSFindPrefix(tok_instate, tok_outstate, stay_allowed)
		suffix = VSFindSuffix(instate, outstate, stay_allowed)
		tok_suffix = VSFindSuffix(tok_instate, tok_outstate, stay_allowed)

		prefix.outputfunc_up = 'token_outputfunc_up'

		tok_prefix.inputfunc = 'token_inputfunc'
		tok_prefix.outputfunc = 'token_outputfunc'
		tok_prefix.outputfunc_up = 'token_outputfunc_up'
		tok_prefix.reasonfunc = 'token_reasonfunc'

		suffix.outputfunc_up = 'token_outputfunc_up'

		tok_suffix.inputfunc = 'token_inputfunc'
		tok_suffix.outputfunc = 'token_outputfunc'
		tok_suffix.outputfunc_up = 'token_outputfunc_up'
		tok_suffix.reasonfunc = 'token_reasonfunc'

		# Removed until I handle inapplicable hypotheses
#		prefsuff = VSConjPrefSuff(instate, outstate,
#			stay_allowed)

		self.vses = [
			(tok_suffix, 0.125, 'TokenFindSuffix'),
			(tok_prefix, 0.125, 'TokenFindPrefix'),
			(suffix, 0.375, 'FindSuffix'),
			(prefix, 0.375, 'FindPrefix')
			]
#		self.vses = [
#			(tok_suffix, 0.1, 'TokenFindSuffix'),
#			(tok_prefix, 0.1, 'TokenFindPrefix'),
#			(suffix, 0.25, 'FindSuffix'),
#			(prefix, 0.25, 'FindPrefix'),
#			(prefsuff, 0.3, 'PrefixSuffix')
#			]

		# All probabilities must sum to one
		assert(reduce(operator.add, map(lambda x: x[1], self.vses))==1.0)

 	def token_inputfunc(self, instate):
 		return self.tokenize_state(instate)

 	def token_outputfunc(self, instate, outstate):
 		return self.tokenize_state(outstate)

	def token_outputfunc_up(self, hyp, instate):
		return hyp.hyp

 	def token_reasonfunc(self, hyp, instate):
		return hyp.reason + ' (tokenized)'

	def tokenize_state(self, state):
		# make an explicit copy of the text and tokenize it
		newstate = copy.deepcopy(state)
		newstate.contents = self.tokenize(newstate.contents)

		return newstate

	def tokenize(self, text):
		return string.translate(text, self.classes)

	def get_classes(self):
		s_from = []
		s_to = []
		whitespace = string.whitespace
		uppercase = string.uppercase
		lowercase = string.lowercase
		digits = string.digits
		punctuation = '`~!@#$%^&*()-_=+[{]}\\|;:\'",<.>/?'

		# mark all characters as 'no class'
		i = 0
		while i < 256:
			s_from.append(chr(i))
			s_to.append(chr(i))
			i = i + 1

		# mark each of the above classes
		for x in whitespace:
##			s_to[s_from.index(x)] = ' '
 			if x != '\n':
 				s_to[s_from.index(x)] = ' '
		for x in uppercase:
			s_to[s_from.index(x)] = 'A'
		for x in lowercase:
			s_to[s_from.index(x)] = 'a'
		for x in digits:
			s_to[s_from.index(x)] = '1'
		for x in punctuation:
			s_to[s_from.index(x)] = '!'

		s_from = reduce(lambda x,y: x+y, s_from)
		s_to = reduce(lambda x,y: x+y, s_to)

		return string.maketrans(s_from, s_to)

# arbitrary character class VS
class VSCharClass(vs.VersionSpace):
	def __init__(self, instate, outstate):
		self.bias = 0.501 # bias for first positive character
		self.NUM_LETTERS = 256
		self.examples = {}
		self.broke = 0

		# mark all characters as unseen
		i = 0
		while i < self.NUM_LETTERS:
			self.examples[chr(i)] = 'nei'
			i = i + 1
		self.classes = self.get_classes()

		self.direction = cmp(outstate.pos, instate.pos)

		self.feed(instate, outstate)

	def get_classes(self):
		# start with the full path to the smartedit script
		basepath = sys.argv[0]
		# follow the symbolic links until we get to a real file
		if sys.platform != 'win32':
			# Windows doesn't have symbolic links
			while os.path.islink(basepath):
				basepath = os.path.normpath(
					os.path.join(
						# directory where basepath lives
						os.path.split(basepath)[0],
						# contents of link
						os.readlink(basepath)))
		# Pull out the directory where the actual smartedit script lives
		path = os.path.split(basepath)[0]

		classes = {}
		for x in self.examples.keys():
			classes[x] = []
		
		for (name, number, chars) in char_classes.char_classes:
			self.examples[name] = 'nei'
			for char in chars:
				classes[char].append(name)
			classes[name] = number

		# add the all class
		self.examples['all'] = 'pos'
		for x in classes.keys():
			if len(x) < 2:
				classes[x].append('all')
		classes['all'] = 1.0
		
		return classes

	# Classif is 0 or 1; instate/outstate are domain-dependent
	def feed(self, instate, outstate):
		direction = cmp(outstate.pos, instate.pos)

		if direction != self.direction:
			# collapse
##			print 'VSCharClass collapsed due to direction change!'
			self.broke = 1
			return

		if outstate.pos >= len(outstate.contents) or \
		   instate.pos >= len(instate.contents):
			# collapse
## 			print 'VSCharClass collapsed - reached EOF!'
			self.broke = 1
			return
		
		# assume we're seeding the VS with the character the cursor is on
		#   and move forward from now on
		if direction == 0:
			self.direction = direction = 1

		i = instate.pos
		# if we've seen the first character
		if self.examples[instate.contents[instate.pos]] != 'nei':
			i = i + direction # skip it
		while (direction > 0 and i < outstate.pos) or \
		      (direction < 0 and i > outstate.pos):
			if self.examples[instate.contents[i]] == 'pos':
				# collapse
##				print 'VSCharClass collapsed due to pos and neg example found!'
				self.broke = 1
				return
			self.examples[instate.contents[i]] = 'neg'
			
			# mark the examples this character belongs to as 'neg'
			for x in self.classes[instate.contents[i]]:
				if x != 'all':
					self.examples[x] = 'neg'
				
			i = i + direction
		# if the character we've moved to is already negative
		if self.examples[instate.contents[outstate.pos]] == 'neg':
			# collapse
##			print 'VSCharClass collapsed due to pos and neg example found!'
			self.broke = 1
			return
		self.examples[instate.contents[outstate.pos]] = 'pos'

		# mark the examples this character belongs to as 'pos'
		for x in self.classes[instate.contents[i]]:
			if self.examples[x] == 'nei':
				self.examples[x] = 'pos'

	def execute(self, instate):
		rs = vs.ResultSet()
		if self.collapsed():
			return rs

		positive = None
		neither = []
		seen = {}

		i = instate.pos + self.direction

		while i < len(instate.contents) and i >= 0:
			char = instate.contents[i]
			if self.examples[char] == 'pos':
				reason = self.getreasonstr(char, seen.keys())
				positive = (char, i, reason)
				break
			elif self.examples[char] == 'nei':
				if not seen.has_key(char):
					reason = self.getreasonstr(char, seen.keys())
					neither.append((char, i, reason))
					seen[char] = 'neg'
			i = i + self.direction

		if positive == None:
			num = 1.0
		else:
			position = textedit.rowColTuple(instate.contents, positive[1])
			reason = positive[2]
			hyp = vs.Hypothesis(position,reason)
			if len(neither) == 0:
				rs.add(hyp,1.0)
				return rs
			rs.add(hyp,self.bias)
			num = 1.0 - self.bias

		total = 0.0
		alt_neither = []
		for (char, pos, reason) in neither:
			number = 0.0
			for x in self.classes[char]:
				if self.examples[x] == 'pos':
					number = number + self.classes[x]
					total = total + self.classes[x]
			alt_neither.append((char,pos,reason,number))
		neither = alt_neither
		for (char, pos, reason, number) in neither:
			position = textedit.rowColTuple(instate.contents, pos)
			hyp = vs.Hypothesis(position,reason)
			rs.add(hyp, num * (number / total))

		return rs

	def getreasonstr(self, char='', neg_list=[]):
		if char == '':
			char_type = 'null'
		else:
			char_type = self.examples[char]

		pos_list = []

		# for every ASCII character indexed 0 to NUM_LETTERS 
		for key in map(chr, range(self.NUM_LETTERS)):
			type = self.examples[key]
			if type == 'pos':
				pos_list.append(key)
			elif type == 'neg':
				neg_list.append(key)

		for i in range(len(pos_list)):
			key = pos_list[i]
			if key == '\n':
				pos_list[i] = '\\n'
			elif key == '\t':
				pos_list[i] = '\\t'

		for i in range(len(neg_list)):
			key = neg_list[i]
			if key == '\n':
				neg_list[i] = '\\n'
			elif key == '\t':
				neg_list[i] = '\\t'

		# to ensure that the lists are non-empty and the reduce
		#   statements won't bomb
		pos_list.append('')
		neg_list.append('')

 		pos_str = '[' + reduce(lambda x, y: x+y, pos_list) + ']'
 		neg_str = '[' + reduce(lambda x, y: x+y, neg_list) + ']'

		if char_type == 'nei':
			if char == '\n':
				char = "+'\\n'"
			elif char == '\t':
				char = "+'\\t'"
			else:
				char = "+'%s'" % char
			pos_str = pos_str + char

		ret_str = 'char class containing ' + pos_str
		if len(ret_str) + len(neg_str) < 50:
			ret_str = ret_str + ' and ^' + neg_str

		return ret_str
                	
	def collapsed(self):
		"""Returns true iff the version space has collapsed and is empty."""
		return self.broke

	def num_hyps(self):
		values = self.examples.values()
		neither = len(filter(lambda x: x == 'nei', values))
		is_pos = (filter(lambda x: x == 'pos', values) != [])
		return neither + is_pos

# the hypotheses of this version space are the number of words shifted
#   in the text.  the position is always moved to the next or previous word,
#   so it'll have at least 1 (or -1) has a hypothesis
# whitespace before EOF does not count as a word
class VSWordShift(vs.VersionSpace):
	# self.hypothesis == None means collapsed
	# self.hypothesis == non-zero int means # words shifted
	
	def __init__(self, instate, outstate):
		ipos = instate.pos
		opos = outstate.pos
		text = outstate.contents

		# Possible BUG!!! Should this check against len(text) + 1?
		if opos >= len(text) or opos == ipos:
			self.hypothesis = None
			return
		if ipos < opos:
			self.direction = 'forward'
		else:
			self.direction = 'backward'

		# only output locations that aren't whitespace
		#   immediately following whitespace are considered
		if (opos == 0 or self.iswhite(text[opos-1])) and \
		   not self.iswhite(text[opos]):
			if self.direction == 'forward':
				offset_text = text[ipos:opos]
			else:
				offset_text = text[opos:ipos]
			num_words = 0
			words = self.getwords(offset_text)
 			for (type,num) in words:
 				if type == 'non-white':
 					num_words = num_words + 1
			# special case (from middle of whitespace going
			# forward) fix
			if words[0][0] == 'white' and \
			   self.direction == 'forward':
				num_words = num_words + 1
			if self.direction == 'forward':
				self.hypothesis = num_words
			else:
				self.hypothesis = -num_words
		else:
			self.hypothesis = None

	def feed(self, instate, outstate):
		newpos = self.getnewpos(instate)
		if newpos == -1:
			self.hypothesis = None
		if newpos != outstate.pos:
			self.hypothesis = None
	
	# Return a ResultSet instance. If ALWAYS_NORMALIZE is true, the
	# ResultSet should be normalized before being returned.
	def execute(self, instate):
		rs = vs.ResultSet()
		if type(self.hypothesis) is IntType:
			hyp = vs.Hypothesis(self.hypothesis,
					 self.getreasonstr())
			rs.add(hyp, 1)
		return rs
	
	def getreasonstr(self):
		if self.hypothesis == 1:
			return '1 word forward'
		if self.hypothesis == -1:
			return '1 word backward'
		if self.hypothesis < 0:
			return '%i words backward' % -self.hypothesis
		return '%i words forward' % self.hypothesis

	def collapsed(self):
		return self.hypothesis == None

	def outputfunc_up_o(self, hyp, instate):
		newpos = self.getnewpos(instate)
		if newpos == -1:
			return vs.hyp_make_inapplicable()
		return textedit.rowColTuple(instate.contents, newpos)

	def getnewpos(self, instate):
		if self.hypothesis == None:
			return -1
		words = self.getwords(instate.contents,\
				      instate.pos,\
				      self.hypothesis)
		newpos = instate.pos
		num_words = 0

		if self.direction == 'backward':
			for (type,num) in words:
				if type != 'white':
					num_words = num_words + 1
				newpos = newpos - num
				if num_words == -self.hypothesis:
					break
			else:
				return -1
		else:
			for (type,num) in words:
				if num_words == self.hypothesis:
					break
				if type == 'white':
					num_words = num_words + 1
       				newpos = newpos + num
			else:
				return -1
		if newpos > len(instate.contents) or newpos < 0:
			return -1
		return newpos

	# word-manipulating helper methods

	# returns a list of (type,num_chars), delimited by whitespace
	# ex:  "hello  all you silly   people" returns
	#   [('non-white',5),('white',2),('non-white',3),('white',1)...
	# the start position is the index in the text where the count should
	#   begin, and hyp is the number of words it should count through
	#   before stopping (negative numbers are fine)
	def getwords(self, text, start=0, hyp=0):
		# Possible BUG!!! Should this check against len(text) + 1?
		if start < 0 or start >= len(text):
			start = 0
		if hyp == 0:
			# Possible BUG!!! Should this check against len(text) + 1?
			hyp = len(text)
		ret_list=[]
		# Possible BUG!!! Should this check against len(text) + 1?
		if len(text) == 0:
			return ret_list
		if hyp < 0:
			if start == 0: return ret_list
			start = start - 1 # fix for reverse scanning
		if self.iswhite(text[start]):
			type = 'white'
		else:
			type = 'non-white'
		num = 0
		i = start
		abshyp = abs(hyp)
		num_words = 0
		# Possible BUG!!! Should this check against len(text) + 1?
		while num_words < abshyp and i >= 0 and i < len(text):
			x = text[i]
			if self.iswhite(x) and type == 'white':
				num = num + 1
			elif self.iswhite(x) and type == 'non-white':
				ret_list.append((type,num))
				type = 'white'
				num = 1
				if hyp < 0:
					num_words = num_words + 1
			elif not self.iswhite(x) and type == 'white':
				ret_list.append((type,num))
				type = 'non-white'
				num = 1
				if hyp > -1:
					num_words = num_words + 1
			elif not self.iswhite(x) and type == 'non-white':
				num = num + 1
			if hyp < 0:
				i = i - 1
			else:
				i = i + 1
		ret_list.append((type,num))
		return ret_list

	# returns 1 if char is whitespace and 0 if it isn't
	def iswhite(self, char):
		if char == '\011' or \
		   char == '\012' or \
		   char == '\013' or \
		   char == '\014' or \
		   char == '\015' or \
		   char == ' ':
			return 1
		return 0

	def num_hyps(self):
		if type(self.hypothesis) is IntType:
			return 1
		else:
			return 0

# Join of a row and a column
class VSRowCol(vs.VersionSpaceJoin):
	def __init__(self, instate, outstate):
		row = VSRow(self.rowinput(instate),
				  self.rowoutput(instate, outstate))
		row.inputfunc = 'rowinput'
		row.outputfunc = 'rowoutput'

		col = VSAllCol(instate, outstate)

		self.vses = [
			(row, 'Row'),
			(col, 'Col'),
			]
	
	# Both joinfunc and joinreasonfunc are called with the untransformed
	# state; they can transform it if necessary by calling the transform
	# functions.
	# These are special member functions that are used by execute.
	def joinfunc(self, hyps, instate):
		"Mark nonexistent locations as inapplicable."
		assert(len(hyps) == 2)

		row = hyps[0].hyp
		col = hyps[1].hyp
		
		# We've received two hypotheses, one from the VSRow object
		# and another from the VSAllCol object.  The VSAllCol hyp.
		# may be a negative number (reverse column number) and we'll
		# need to convert it to a normal column number.
		if col < 0:
			# Convert to the number of colunms back from the end:
			col = col + 1
			try:
				posofline = textedit.positionFromTuple(instate.contents,
											    (row, 0))
			except ValueError:
				return vs.hyp_make_inapplicable()
			# Get the position of the end of the this line.
			if textedit.hasRow(instate.contents, row+1):
				#DEBUG!!
				#print 'Has row', row+1
				posofhyp = textedit.positionFromTuple(instate.contents,
											   (row + 1, 0))
				posofhyp = posofhyp - 1
			else:
				#DEBUG!!
				#print 'NO  row', row+1
				posofhyp = len(instate.contents)
			posofhyp = posofhyp + col
			if posofhyp < posofline:
				return vs.hyp_make_inapplicable()
			col = posofhyp	- posofline
		hyp = (row, col)
		try:
			pos = textedit.positionFromTuple(instate.contents, hyp)
		except ValueError:
			hyp = vs.hyp_make_inapplicable()
		return hyp

	# Both joinfunc and joinreasonfunc are called with the untransformed
	# state; they can transform it if necessary by calling the transform
	# functions.
	# These are special member functions that are used by execute.
	def joinreasonfunc(self, hyps, instate):
		"Provide gloss for joined reason."
		assert(len(hyps) == 2)
		hyp = (hyps[0].hyp, hyps[1].hyp)
		inxy = textedit.rowColTuple(instate.contents, instate.pos)
		if tuple(map(lambda x: x.hyp, hyps)) == inxy:
			return 'here'
		else:
			return '%s and %s' % (hyps[0].reason, hyps[1].reason)
	
	def rowinput(self, instate):
		inxy = textedit.rowColTuple(instate.contents, instate.pos)
		return inxy[0]

	def rowoutput(self, instate, outstate):
		outxy = textedit.rowColTuple(outstate.contents, outstate.pos)
		return outxy[0]


	def colinput(self, instate):
		inxy = textedit.rowColTuple(instate.contents, instate.pos)
		return inxy[1]

	def coloutput(self, instate, outstate):
		outxy = textedit.rowColTuple(outstate.contents, outstate.pos)
		return outxy[1]

# Version space that contains two VSCol objects (one of which is the column from
# the beginning of the line, the other is the column from the end of the line).  
class VSAllCol(vs.VersionSpaceUnion):
	def __init__(self, instate, outstate):
		col = VSCol(self.colinput(instate),
				  self.coloutput(instate, outstate))
		revcol = VSCol(self.revcolinput(instate),
					self.revcoloutput(instate, outstate))
		col.inputfunc = 'colinput'
		col.outputfunc = 'coloutput'
		revcol.inputfunc = 'revcolinput'
		revcol.outputfunc = 'revcoloutput'
		revcol.outputfunc_up = 'revcoloutputfunc_up'
		revcol.reasonfunc = 'revcolreasonfunc'
		self.vses = [
			(col, .75, 'Col'),
			(revcol, .25, 'RevCol')
			]

	def colinput(self, instate):
		inxy = textedit.rowColTuple(instate.contents, instate.pos)
		return inxy[1]

	def coloutput(self, instate, outstate):
		outxy = textedit.rowColTuple(outstate.contents, outstate.pos)
		return outxy[1]

	#BUG!!! ?? Not handling end of file properly?
  	def revcolinput(self, instate):
  		inendpos = string.find(instate.contents, '\n', instate.pos)
  		if inendpos == -1:
  			inendpos = len(instate.contents) 
  		return inendpos - instate.pos

  	def revcoloutput(self, instate, outstate):
  		outendpos = string.find(outstate.contents, '\n', outstate.pos)
  		if outendpos == -1:
  			outendpos = len(outstate.contents) 
  		return outendpos - outstate.pos

	def revcoloutputfunc_up(self, hyp, instate):
		if vs.hyp_inapplicable(hyp.hyp):
			return hyp.hyp
		else:
			return (hyp.hyp * -1) - 1

	def revcolreasonfunc(self, hyp, instate):
		return hyp.reason + " (if you turn your head upside down)"
		

# Union of AbsRow and RelRow
class VSRow(vs.VersionSpaceUnion):
	def __init__(self, inrow, outrow):
		absrow = vs.VSConstInt(inrow, outrow)
		absrow.reasonfunc = 'absreason'
		relrow = vs.VSLinearInt(inrow, outrow)
		relrow.reasonfunc = 'relreason'
		self.vses = [
			(absrow, 0.25, "AbsRow"),
			(relrow, 0.75, "RelRow")
			]
		# All probabilities must sum to one
		assert(reduce(operator.add, map(lambda x: x[1], self.vses))==1.0)
	
	def absreason(self, hyp, instate):
		return 'row %s' % hyp.reason
	
	def relreason(self, hyp, instate):
		delta = hyp.hyp - instate
		if delta == 0:
			return 'this row'
		elif delta < 0:
			return '%i rows up' % -delta
		else:
			return '%i rows down' % delta

class VSCol(vs.VersionSpaceUnion):
	def __init__(self, incol, outcol):
		abscol = vs.VSConstInt(incol, outcol)
		abscol.reasonfunc = 'absreason'
		relcol = vs.VSLinearInt(incol, outcol)
		relcol.reasonfunc = 'relreason'
		abscol.outputfunc_up = 'check_neg'
		relcol.outputfunc_up = 'check_neg'
		
		self.vses = [
			(abscol, 0.25, "AbsCol"),
			(relcol, 0.75, "RelCol")
			]
		# All probabilities must sum to one
		assert(reduce(operator.add, map(lambda x: x[1], self.vses))==1.0)

	def check_neg(self, hyp, instate):
		"""Check the hypothesis to ensure that it isn't negative."""

		if hyp.hyp >= 0:
			return hyp.hyp
		else:
			#DEBUG!!
			#print 'DANGER: negative column hypothesis!'
			return vs.hyp_make_inapplicable()
	
	def absreason(self, hyp, instate):
		return 'column %i' % hyp.hyp
	
	def relreason(self, hyp, instate):
		delta = hyp.hyp - instate
		if delta == 0:
			return 'this column'
		elif delta < 0:
			return '%i columns left' % -delta
		else:
			return '%i columns right' % delta

# This version space governs locataions based on a character offset.
# It is essentially a VSLinearInt with the appropriate transformation
# functions added.  
class VSCharOffset(vs.VSLinearInt):
	def __init__(self, inval=None, outval=None):
		# call the parent class constructor
		vs.VSLinearInt.__init__(self, inval, outval)

class VSStringFromInt(vs.VSString):
	"""Contains one possible function: f(i) = str(i)."""

	def feed(self, intval, sval):
		if self.is_bottom():
			return
		elif self.is_everything():
			self.hypothesis = sval
		else:
			# given that our function is consistent
			if `intval` != sval:
				# if the new example isn't, set the hyp to bottom
				self.hypothesis = vs.StringBottom()
	
	def execute(self, instate):
		row = textedit.rowColTuple(instate.contents, instate.pos)[0]
		rs = vs.ResultSet()
		if self.is_everything():
			return rs
		elif self.is_bottom():
			return rs
		else:
			rs.add(vs.Hypothesis(`row`, 'the row number %i' % row),
				prob=1.0)
		return rs
	
	def num_hyps(self):
		if self.is_everything():
			raise VSError, 'infinite VSStringFromInt space'
		elif self.is_bottom():
			return 0
		else:
			return 1

# Currently I don't bother separating FindPrefix from its underlying
# SuffStr atomic version space, and lump them together into one.  Ditto for
# FindSuffix/PrefStr.
class VSFindSuffix(vs.VSString):

	def __init__(self, instate, outstate, stay_allowed = 0):
		# The direction of search either 1 for forward or 0 for
		# no move or -1 for back.
		self.direction = cmp(outstate.pos, instate.pos)

		# Decides whether searching at the current point is allowed.
		self.stay_allowed = stay_allowed
		
		vs.VSString.__init__(self)

		self.feed(instate, outstate)

	def find_next(self, text, full_string, start, end, min_len, direction):
		"""Finds the next matching location in text for full_string of length
		at least min_len. Returns -1 if no such match exists. Otherwise,
		returns a tuple of the location where the match is and the
		length of the longest string that matches at that location."""
		
		if min_len > len(full_string):
			return -1
		
		if direction == 1:
			search = string.find
		else:
			search = string.rfind

		loc = search(text, full_string[:min_len], start, end + min_len)
		
		if loc < 0:
			return -1
		
		j = min_len
		try:
			while full_string[j] == text[loc + j]:
				j = j + 1
		except IndexError:
			pass
		
		return (loc, j)
		
	def feed(self, instate, outstate):
		# Positive example
		direction = cmp(outstate.pos, instate.pos)

		if direction == 0 and not self.stay_allowed:
			#DEBUG!!
			#print 'No movement in VSFindSuffix when ' \
			#	 'staying is disallowed.'
			self.hypothesis = vs.StringBottom()
			self.lower_bound = ''
			return
			
		start = instate.pos
		if not self.stay_allowed:
			start = start + direction
		stop = outstate.pos
		suffix = instate.contents[stop:stop + vs.WINDOWSIZE]

		# If we're going backward, switch the bounds
		if direction == -1:
			temp = start
			start = stop
			stop = temp

		# If the given direction of search is opposite the current
		# direction, then we're fresh out of hypotheses
		if (self.direction == 1 and direction == -1) or \
		   (self.direction == -1 and direction == 1):
			print 'VS %s collapsed due to reversal of direction.' %\
				`self`
			self.hypothesis = vs.StringBottom()
			self.lower_bound = ''
			return
		elif self.direction == 0:
			self.direction = direction
		
		if self.is_everything():
			# Specialize from everything to this example
			# Still prune hypotheses that are too short (below).
			self.hypothesis = suffix
			self.lower_bound = suffix[:1]

		if self.is_bottom():
			return


		i = len(self.lower_bound)
		while 1:
			suffixloc = self.find_next(instate.contents,
								  self.hypothesis,
								  start, stop, i, direction)

			if suffixloc == -1:
				# DEBUG!!
				#print 'No matching hypothesis.'
				self.hypothesis = vs.StringBottom()
				self.lower_bound = ''
				break

			(loc, length) = suffixloc

			# Check if this location is the one that matches the
			# given output state.
			if loc == outstate.pos:
				# DEBUG!!
				#print 'Found match from %s to %s.' % \
				#	 (self.hypothesis[:i], self.hypothesis[:length])
				self.lower_bound = self.hypothesis[:i]
				self.hypothesis = self.hypothesis[:length]
				break

			# Otherwise, loc is short of the target, which
			# means we need to keep searching for longer strings.
			i = length + 1
			if direction == 1:
				start = loc + 1
			else:
				stop = loc - 1

	def execute(self, instate):
		rs = vs.ResultSet()

		if self.direction <= 0:
			self.execute_directed(instate, rs, -1)

		if self.direction >= 0:
			self.execute_directed(instate, rs, 1)

		return rs

	def execute_directed(self, instate, rs, direction):
		"""Execute in the given direction, accumulating votes."""

		assert(direction == -1 or direction == 1)

		if self.is_everything():
			raise VSError, 'executing complete FindSuffix space'
		elif self.is_bottom():
			return


		if self.stay_allowed:
			bump = 0
		else:
			bump = 1
		
		if direction == 1:
			start = instate.pos + bump
			end = len(instate.contents)
		else:
			start = 0
			end = instate.pos - bump

		i = len(self.lower_bound)
		while i <= len(self.hypothesis):
			#substr = self.hypothesis[:i]
			#if not substr:
			#	print 'Yikes, suffix substring is empty!'
			#	print 'LB:', `self.lower_bound`
			#	print 'UB:', `self.hypothesis`
			#	continue
			suffixloc = self.find_next(instate.contents, self.hypothesis,
								  start, end, i, direction)
			if suffixloc == -1:
				# Can't find this hypothesis. How could we possibly
				# find a longer one? Enter all remaining hypotheses
				# as predicting inapplicable.
				for j in range(i, len(self.hypothesis)):
					substr = self.hypothesis[:j]
					weight = self.weight(substr)
					if not weight:
						continue
					hyp = vs.Hypothesis(hyp=vs.hyp_make_inapplicable(),
								  reason='the start of %s' %
								  `substr`)
					rs.add(hyp, weight)
				break

			(loc, length) = suffixloc

			if direction == 1:
				start = loc + 1
			else:
				end = loc - 1

			# Add a hypothesis for each string that matches at this
			# location.
			row, col = textedit.rowColTuple(instate.contents, loc)
			for j in range(i, length + 1):
				substr = self.hypothesis[:j]
				# DEBUG!!
				#print 'Found suffix match for %s at (%s, %s)' % \
				#	 (`substr`, row, col)
				weight = self.weight(substr)
				if not weight:
					continue
				hyp = vs.Hypothesis(hyp=(row, col),
							  reason='the start of %s' %
							  `substr`)
				rs.add(hyp, weight)

			i = length + 1

		if vs.ALWAYS_NORMALIZE:
			rs.normalize()

	def num_hyps(self):
		if self.is_bottom():
			return 0
		else:
			assert (len(self.hypothesis) - len(self.lower_bound)) >= 0
			return len(self.hypothesis) - len(self.lower_bound) + 1


# Fixed (probably)
class VSFindPrefix(vs.VSString):

	def __init__(self, instate, outstate, stay_allowed = 0):
		# The direction of search either 1 for forward or 0 for
		# no move or -1 for back. self.direction takes on 0 when the
		# direction is unknown.
		self.direction = cmp(outstate.pos, instate.pos)

		# Decides whether searching at the current point is allowed.
		self.stay_allowed = stay_allowed

		vs.VSString.__init__(self)

		self.feed(instate, outstate)
		
	def find_next(self, text, full_string, start, end, min_len, direction):
		"""Finds the next matching location in text for full_string
		of length min_len. Returns -1 if no such match
		exists. Otherwise, returns a tuple of the location where the
		match is and the length of the longest string that matches
		at that location."""

		if min_len > len(full_string):
			return -1
		
		if direction == 1:
			search = string.find
		else:
			search = string.rfind

		loc = search(text, full_string[-min_len:], max(0, start - min_len),
				   end)

		if loc < 0:
			return -1
		else:
			# Adjust to the actual location of the match.
			loc = loc + min_len

		# Search for longer matches
		j = min_len + 1
		try:
			while full_string[-j] == text[loc - j]:
				j = j + 1
		except IndexError:
			pass

		return (loc, j - 1)
		
	def feed(self, instate, outstate):
		# Positive example
		direction = cmp(outstate.pos, instate.pos)

		if not self.stay_allowed and direction == 0:
			#DEBUG!!
			#print 'No movement in VSFindPrefix when ' \
			#	 'staying is disallowed.'
			self.hypothesis = vs.StringBottom()
			self.lower_bound = ''
			return
			
		start = instate.pos
		if not self.stay_allowed:
			start = start + direction
		stop = outstate.pos
		prefix = instate.contents[max(0, stop - vs.WINDOWSIZE):stop]

		# If we're going backward, switch the bounds
		if direction == -1:
			temp = start
			start = stop
			stop = temp

		# If the given direction of search is opposite the current
		# direction, then we're fresh out of hypotheses
		if (self.direction == 1 and direction == -1) or \
		   (self.direction == -1 and direction == 1):
			print 'VS collapsed due to reversal of direction.'
			self.hypothesis = vs.StringBottom()
			self.lower_bound = ''
			return
		elif self.direction == 0:
			self.direction = direction
		
		if self.is_everything():
			# Specialize from everything to this example
			# Still prune hypotheses that are too short (below).
			self.hypothesis = prefix
			self.lower_bound = prefix[-1:]

		if self.is_bottom():
			return

		i = len(self.lower_bound)
		while 1:
			prefixloc = self.find_next(instate.contents,
								  self.hypothesis,
								  start, stop, i, direction)

			# Check if there was even a match
			if prefixloc == -1:
				# DEBUG!!
				#print 'No matching hypothesis.'
				self.hypothesis = vs.StringBottom()
				self.lower_bound = ''
				break

			(loc, length) = prefixloc

			# Check if this location is the one that matches the
			# given output state.
			if loc == outstate.pos:
				# DEBUG!!
				#print 'Found match from %s to %s.' % \
				#	 (self.hypothesis[-i:],
				#	  self.hypothesis[-length:])
				self.lower_bound = self.hypothesis[-i:]
				self.hypothesis = self.hypothesis[-length:]
				break

			# Otherwise, loc is short of the target, which
			# means we need to keep searching for longer strings.
			i = length + 1
			if direction == 1:
				start = loc + 1
			else:
				stop = loc - 1
	
	def execute(self, instate):
		rs = vs.ResultSet()
		# Do the stupid/slow thing for prefix/suffix search

		if self.direction <= 0:
			self.execute_directed(instate, rs, -1)

		if self.direction >= 0:
			self.execute_directed(instate, rs, 1)

		return rs

	def execute_directed(self, instate, rs, direction):
		"""Execute in the given direction, accumulating votes."""

		assert(direction == 1 or direction == -1)

		# Do the stupid/slow thing for prefix/suffix search
		if self.is_everything():
			raise VSError, 'executing complete FindPrefix space'
		elif self.is_bottom():
			return

		if self.stay_allowed:
			bump = 0
		else:
			bump = 1

		if direction == 1:
			start = instate.pos + bump
			end = len(instate.contents)
		else:
			start = 0
			end = instate.pos - bump

		i = len(self.lower_bound)
		while 1:
			#substr = self.hypothesis[-i:]
			prefixloc = self.find_next(instate.contents, self.hypothesis,
								  start, end, i, direction)
			if prefixloc == -1:
				# Can't find this hypothesis. How could we possibly
				# find a longer one? Enter all remaining hypotheses
				# as predicting inapplicable.
				for j in range(i, len(self.hypothesis)):
					substr = self.hypothesis[-j:]
					weight = self.weight(substr)
					if not weight:
						continue
					hyp = vs.Hypothesis(hyp=vs.hyp_make_inapplicable(),
								  reason='the end of %s' %
								  `substr`)
					rs.add(hyp, weight)
				break

			(loc, length) = prefixloc

			if direction == 1:
				start = loc + 1
			else:
				end = loc - 1
			
			# Add a hypothesis for each string that matches at this
			# location.
			row, col = textedit.rowColTuple(instate.contents, loc)
			for j in range(i, length + 1):
				substr = self.hypothesis[-j:]
				# DEBUG!!
				#print 'Found prefix match for %s at (%s, %s)' % \
				#	 (`substr`, row, col)
				weight = self.weight(substr)
				if not weight:
					continue
				hyp = vs.Hypothesis(hyp=(row, col),
							  reason='the end of %s' %
							  `substr`)
				rs.add(hyp, weight)

			i = length + 1

		if vs.ALWAYS_NORMALIZE:
			rs.normalize()

	def num_hyps(self):
		if self.is_bottom():
			return 0
		else:
			assert (len(self.hypothesis) - len(self.lower_bound)) >= 0
			return len(self.hypothesis) - len(self.lower_bound) + 1

class VSConjPrefSuff(vs.VersionSpace):
	def __init__(self, instate = None, outstate = None, stay_allowed=0):
		# self.prefix_upper is the upper bound on the prefix
		#   it contains the longest consistent prefix string
		self.prefix_upper = vs.StringTop()
		# self.suffix_upper is the upper bound on the suffix
		#   it contains the longest consistent suffix string
		self.suffix_upper = vs.StringTop()

		# To maintain the lower bound, we maintain an array of length
		# len(self.prefix_upper).  Each element in this array indexes
		# into the prefix string and indicates the shortest consistent
		# suffix in conjunction with this particular prefix.
		self.lower_bound = []

		self.stay_allowed = 0
		self.direction = 0
		if instate and outstate:
			self.feed(instate, outstate)

	def is_everything(self, hyp):
		return isinstance(hyp, vs.StringTop)
	
	def is_bottom(self, hyp):
		return isinstance(hyp, vs.StringBottom)
	
	def feed(self, instate, outstate):
		self.feed_prefix(instate, outstate)
		self.feed_suffix(instate, outstate)
		self.update_lower_bound(instate, outstate)
		if DEBUG:
			print '<<< VS contains:'
			print self.enumerate()
		
	def feed_prefix(self, instate, outstate):
		# Update only the upper boundary of the prefix space
		direction = cmp(outstate.pos, instate.pos)

		if not self.stay_allowed and direction == 0:
			#DEBUG!!
			#print 'No movement in VSFindPrefix when ' \
			#	 'staying is disallowed.'
			self.prefix_upper = vs.StringBottom()
			return

		start = instate.pos
		if not self.stay_allowed:
			start = start + direction
		stop = outstate.pos
		prefix = instate.contents[max(0, stop - vs.WINDOWSIZE):stop]

		# If we're going backward, switch the bounds
		if direction == -1:
			start, stop = stop, start

		# If the given direction of search is opposite the current
		# direction, then we're fresh out of hypotheses
		if (self.direction == 1 and direction == -1) or \
		   (self.direction == -1 and direction == 1):
			if DEBUG:
				print 'VSConjPrefSuff collapsed; reversal of direction'
			self.prefix_upper = vs.StringBottom()
			return
		elif self.direction == 0:
			self.direction = direction
		
		if self.is_everything(self.prefix_upper):
			# Specialize from everything to this example
			if DEBUG: print 'Spec pref from TOP to', `prefix`
			self.prefix_upper = prefix
		elif self.is_bottom(self.prefix_upper):
			return
		else:
			# Get the longest common suffix of self.prefix_upper and
			# prefix
			s = self.longest_suffix(self.prefix_upper, prefix)
			if not s:
				self.prefix_upper = vs.StringBottom()
				if DEBUG: print 'Prefix collapsed; no match'
			else:
				# Specialize prefix to this new string
				self.prefix_upper = s
				if DEBUG: print 'Spec pref to', `self.prefix_upper`

	def feed_suffix(self, instate, outstate):
		# Update only the upper boundary of the suffix space
		direction = cmp(outstate.pos, instate.pos)

		if direction == 0 and not self.stay_allowed:
			#DEBUG!!
			#print 'No movement in VSFindSuffix when ' \
			#	 'staying is disallowed.'
			self.suffix_upper = vs.StringBottom()
			return
			
		start = instate.pos
		if not self.stay_allowed:
			start = start + direction
		stop = outstate.pos
		suffix = instate.contents[stop:stop + vs.WINDOWSIZE]
		if not suffix:
			self.suffix_upper = vs.StringBottom()
			return

		# If we're going backward, switch the bounds
		if direction == -1:
			start, stop = stop, start

		# If the given direction of search is opposite the current
		# direction, then we're fresh out of hypotheses
		if (self.direction == 1 and direction == -1) or \
		   (self.direction == -1 and direction == 1):
			print 'VSConjPrefSuff collapsed; reversal of direction'
			self.suffix_upper = vs.StringBottom()
			return
		elif self.direction == 0:
			self.direction = direction
		
		if self.is_everything(self.suffix_upper):
			# Specialize from everything to this example
			# Still prune hypotheses that are too short (below).
			self.suffix_upper = suffix
			if DEBUG: print 'Spec suff from TOP to', `suffix`
		elif self.is_bottom(self.suffix_upper):
			return
		else:
			# Find the longest common prefix of self.suffix_upper and
			# suffix
			s = self.longest_prefix(self.suffix_upper, suffix)
			if not s:
				self.suffix_upper = vs.StringBottom()
			else:
				self.suffix_upper = s
				if DEBUG: print 'Spec suff to',\
					`self.suffix_upper`

	def update_lower_bound(self, instate, outstate):
		# Search for the two letters nearest the prefix/suffix split
		if self.is_bottom(self.prefix_upper) or\
			self.is_bottom(self.suffix_upper):
			return

		# The self.lower_bound variable is a list of integers, one for
		# each suffix of the prefix upper bound.  Each integer indicates
		# the shortest prefix of the suffix upper bound consistent with
		# the data.  The 0th integer in the list corresponds to the
		# unit-length prefix.
		if not self.lower_bound:
			self.lower_bound = [0] * len(self.prefix_upper)
		elif len(self.lower_bound) > len(self.prefix_upper):
			self.lower_bound = self.lower_bound[:len(self.prefix_upper)]
		
		# Sanity check
		if not self.prefix_upper or not self.suffix_upper:
			print 'prefix upper:', `self.prefix_upper`
			print 'suffix upper:', `self.suffix_upper`
			raise RuntimeError, 'upper bound is empty!'

		base = self.prefix_upper[-1] +\
			self.suffix_upper[0]
		if DEBUG:
			print 'prefix_upper:', `self.prefix_upper`
			print 'Updating with base', `base`
#			print 'Looking for base after position', start
#			print 'Ending at position', end
#			print 'Text in that range:', `instate.contents[start:end]`
			print 'instate.pos =', instate.pos
			print 'outstate.pos =', outstate.pos

		if self.direction == 1 or self.direction == 0:
			find = string.find
			start = instate.pos
			end = outstate.pos + 1
		elif self.direction == -1:
			find = string.rfind
			start = outstate.pos - 1
			end = instate.pos
		else:
			raise RuntimeError, 'unknown direction', self.direction

		while 1:
			i = find(instate.contents, base, start, end)
			if i == -1:
				# error!
				print 'instate.pos:', instate.pos
				print 'outstate.pos:', outstate.pos
				print 'Contents:', `instate.contents[start:end]`
				print 'start and end:', start, end
				print 'direction:', self.direction
				print 'Base:', `base`
				raise RuntimeError, 'base not found in text'
			i = i + 1
			if i == outstate.pos:
				# The first occurrence is at the example
				# No negative examples
				if DEBUG: print 'No need to update lower bounds'
				break
			else:
				# Found negative example
				# Move to middle of the base string
				if DEBUG:
					print 'Base split is at position', i
					print 'text after position:', `instate.contents[i:]`
				if self.direction == -1:
					end = i
				else:
					start = i
				# Find the longest matching prefix and suffix
				prefmatch = self.longest_suffix(instate.contents[:i],
					self.prefix_upper)
				suffmatch = self.longest_prefix(instate.contents[i:],
					self.suffix_upper)
				if prefmatch + suffmatch == self.prefix_upper +\
					self.suffix_upper:
					# collapse!
					self.prefix_upper = vs.StringBottom()
					self.suffix_upper = vs.StringBottom()
					return
				# The length of these two matched strings carve out a
				# box in the lower bound
				preflth = len(prefmatch)
				sufflth = len(suffmatch)
				if DEBUG: print 'preflth, sufflth:', preflth, sufflth
				for idx in range(preflth):
					if self.lower_bound[idx] < sufflth:
						self.lower_bound[idx] = sufflth
		if DEBUG:
			print 'new lower bound:', self.lower_bound
			print 'prefix bound:', `self.prefix_upper`
			print 'suffix bound:', `self.suffix_upper`

	def longest_prefix(self, s1, s2):
		"""Return the longest common prefix between strings s1 and s2."""
		i = 0
		while len(s1) > i and len(s2) > i and\
			s1[i] == s2[i]:
			i = i + 1
		if i == 0:
			return ''
		else:
			return s1[:i]

	def longest_suffix(self, s1, s2):
		"""Return the longest common suffix between strings s1 and s2."""
		i = -1
		while len(s1) > -i-1 and len(s2) > -i-1 and\
			s1[i] == s2[i]:
			i = i - 1
		if i == -1:
			return ''
		else:
			return s1[i+1:]

	def enumerate(self):
		"""Return a list of (str, len) pairs such that str is a
		consistent conjunctive hypothesis and the join point is
		len chars into the string str."""
		out = []
		if DEBUG:
			print '-----------'
			print 'Enumerating:'
			print 'Suffix_upper:', `self.suffix_upper`
		if self.is_bottom(self.prefix_upper) or\
			self.is_bottom(self.suffix_upper):
			return out

		for i in range(len(self.prefix_upper)):
			pref = self.prefix_upper[-i-1:]
			if DEBUG:
				print 'Prefix:', `pref`
				print 'i:', i
				print 'lower_bound[i]:', self.lower_bound[i]
			if self.lower_bound[i] == len(self.suffix_upper):
				if DEBUG:
					print 'Skipping all suffixes for prefix', prefix
			else:
				for j in range(self.lower_bound[i],
					len(self.suffix_upper)):
					suff = self.suffix_upper[:j+1]
					if DEBUG: print 'Appending:', `pref+suff`
					out.append((pref + suff, len(pref)))
		return out

	def execute(self, instate):
		rs = vs.ResultSet()
		hyps = self.enumerate()
		# Do it the easy way, as a first cut
		for (strhyp, preflen) in hyps:
			sufflen = len(strhyp) - preflen
			i = string.find(instate.contents, strhyp, instate.pos)
			if i != -1:
				i = i + preflen
				wt = self.weight(strhyp)
				if wt:
					pos = textedit.rowColTuple(instate.contents, i)
					hyp = vs.Hypothesis(pos, 'between %s and %s' %
						(`strhyp[:preflen]`, `strhyp[preflen:]`))
					rs.add(hyp, wt)
		return rs

	def weight(self, str):
		## Some constants to define the prior probability of a
		## given hypothesis. They are used according to the formula:
		##
		## P = e^-(SCALE*((len(hypothesis) + BASE)^2))
		GAUSS_SCALE = 0.5
		GAUSS_BASE = -5.0
		try:
			wt = math.exp(-(GAUSS_SCALE * (len(str) + GAUSS_BASE))**2)
		except OverflowError:
			return 0
		else:
			return wt

class TestState: pass

def main():
	v = VSConjPrefSuff()
	state1 = TestState()
	state1.contents  = 'A replay of a dreplane was replayed, but replay.'
	state1.pos = 0
	state2 = TestState()
	state2.contents  = 'A replay of a dreplane was replayed, but replay.'
	state2.pos = 4
	state3 = TestState()
	state3.contents  = 'A replay of a dreplane was replayed, but replay.'
	state3.pos = 29
	print '---'
	v.feed(state1, state2)
	print '---'
	print v.execute(state2)
	print '---'
	v.feed(state2, state3)
	print '---'
	print v.execute(state3)

if __name__ == '__main__':
	main()

