;;; dogfood.el --- Utilities for running SMARTedit from emacs

;; Copyright (C) 2000 Steve Wolfman

;; Author: 2000 Steve Wolfman <wolf@cs.washington.edu>
;; Created:    Jul 2000
;; Version:    1.0
;; Keywords:   python SMARTedit dogfood

;; This file is under the GNU public license.

;; This is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This software is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; This is a package of utilities for running the SMARTedit editing system
;; from within emacs. Practically, it could be used to run any editor or
;; other file manipulation system. We intend to use it to "dogfood" our
;; research editor. Running the editor to perform repetitive tasks we notice
;; (the target of SMARTedit) while editing files in emacs. The concept is
;; based on Tessa Lau's (identical) idea for vi.

;; Users should set the string 'dogfood-editor' to the program name they'll
;; use. Also, we use the syntax -l lineno -c colno filename for the args to
;; the editor. You'll need to change that if your program does something 
;; different. The line and column numbers are so we pop up in the same place
;; in the file in our dogfood as in emacs.

;; The main user-accessible commands are dogfood-current-buffer (which
;; starts SMARTedit on the current buffer at the current position) and
;; dogfood-note-selection/dogfood-note-file both of which log some text
;; with a message from you telling us what your repetitive task is.

;;; Code:

;; Everyone should provide dogfood to their dogs.
;; This file provides the dogfood interface for running SMARTedit from emacs.
(provide 'dogfood)

;; The name of the dogfed editor. If the name is blank ("" or nil), defaults
;; to our system's dogfood location.
(defvar dogfood-editor "" "Editor to run when dogfeeding.")
(if (or (null dogfood-editor) (string= dogfood-editor ""))
    (setq dogfood-editor "/projects/ai4/smartedit/smartedit"))
(defvar dogfood-note-archive-directory 
  "/projects/ai4/smartedit/scenarios/dogfood/" 
  "Directory in which to store dogfood notes.")
(defvar dogfood-suppress-confirm nil 
  "*If non-nil, treat user as expert; suppress yes-no prompts on some things.")
(defvar dogfood-note-header 
  "This text demonstrates a SMARTedit task. 
Instructions on what the task is follow.
----------------------------------------------------------------------------
----------------------------------------------------------------------------
BEGIN COMMENT

" 
  "Text to put at the top of a dogfood note.")
(defvar dogfood-note-footer 
  "

END COMMENT
----------------------------------------------------------------------------
----------------------------------------------------------------------------
BEGIN EXAMPLE

" 
  "Text to put at the bottom of a dogfood note, just before the file text.")

(defvar dogfood-note-mode nil)

;; Create a keymap to finish a dogfood recording window.
;; Initialization; should be run only once at load time.
(if dogfood-note-mode
    nil
  (progn
    (setq dogfood-note-mode (make-sparse-keymap))
    (define-key dogfood-note-mode "\C-c\C-c" 'dogfood-finish-note)))

(defun dogfood-helper-type-checker (type-fun arg)
  "Checks if argument is of the type listed. Raises an error if not."
  ; Sadly, this assumes that type-fun is a function.

  (or (funcall type-fun arg)
      (error "Wrong type argument: %s, %s"
	     (symbol-name type-fun) arg)))

(defun dogfood-file (file row col)
  "Opens SMARTedit on the given file. 
Positions the cursor at the given row and column."
  (interactive "fFile to load: \nnStarting row: \nnStarting col: ")
  (dogfood-helper-type-checker 'stringp file)
  (dogfood-helper-type-checker 'integerp row)
  (dogfood-helper-type-checker 'integerp col)
  (call-process dogfood-editor nil t nil "-l" (format "%d" (+ row 1)) 
			 "-c" (format "%d" col) (file-truename file))
  )

(defun dogfood-current-line ()
  "Returns the current line counting from 0."
  (+ (count-lines 1 (point))
     (if (= (current-column) 0) 1 0)
     -1))

;; Modified from vc.el function vc-buffer-sync
(defun dogfood-helper-buffer-sync (&optional not-urgent)
  ;; Make sure the current buffer and its working file are in sync
  ;; NOT-URGENT means it is ok to continue if the user says not to save.
  (if (buffer-modified-p)
      (if (or dogfood-suppress-confirm
	      (y-or-n-p (format "Buffer %s modified; save it? " (buffer-name))))
	  (save-buffer)
	(if not-urgent
	    nil
	  (error "Aborted")))))

(defun dogfood-current-buffer (&optional confirm)
  "Opens SMARTedit on the current file at the current line and column.
When you're done in SMARTedit, reloads the current file in the buffer.

If confirm is non-nil, asks the user before loading the dogfed file
back into the emacs buffer."
  ;; I like dogs :\)

  (interactive "P")
  (let ((file   (buffer-file-name))  
	   (row    (dogfood-current-line))
	   (col    (current-column)))
    (if file
	   (progn
		;; Save buffer before switching over. 
		(dogfood-helper-buffer-sync)
		;; Let 'er rip.
		(dogfood-file file row col)
		;; Reload the file into the current buffer to reflect changes.
		(revert-buffer nil (not confirm)))
	 (error "No visited file in current buffer."))))

(defun dogfood-helper-unique-file (directory file &optional num)
  ;; Directory is the directory in which to put the file.
  ;; File should be a format string with a single %d in it describing
  ;; the file format.

  ;; Default num to 1
  (if (null num)
	 (setq num 1))

  (dogfood-helper-type-checker 'stringp directory)
  (dogfood-helper-type-checker 'stringp file)
  (dogfood-helper-type-checker 'integerp num)
  (if (file-directory-p directory)
	 (setq directory (file-name-as-directory directory))
    (error "Directory '%s' does not exist." directory))
  (if (file-exists-p (format (concat directory file) num))
	 (dogfood-helper-unique-file directory file (+ num 1))
    (format file num)))

(defun dogfood-finish-note ()
  "Save the note aside."
  (interactive)
  
  ;; Get a filename for the file
  (let ((default-file (concat (user-login-name) "-%03d.ex.txt")))
    (setq default-file (dogfood-helper-unique-file 
				    dogfood-note-archive-directory
				    default-file))
    (let ((filename (read-file-name "Enter a file name for the example: "
							 dogfood-note-archive-directory
							 default-file
							 nil
							 default-file)))
	 
	 (write-file (concat (file-name-as-directory 
					  dogfood-note-archive-directory) filename))
	 (kill-buffer (current-buffer)))))

(defun dogfood-note-current-buffer ()
  "Save the contents of the current buffer as a dogfood note."
  (interactive)
  (dogfood-note-text (buffer-string)))

(defun dogfood-note-selection (start end)
  "Make a note of the current selection as a dogfood example."
  (interactive "r")
  (dogfood-note-text (buffer-substring start end)))

(defun dogfood-note-text (text)
  "Creates a buffer for a repetitive task which is then saved to a note file."
  (dogfood-helper-type-checker 'stringp text)

  ;; Generate a buffer with 'text' in it.
  (let ((new-buffer (generate-new-buffer "*dogfood*")))
    (switch-to-buffer new-buffer)
    (insert dogfood-note-header)
    (insert dogfood-note-footer)
    (insert text)
    (goto-char (+ (length dogfood-note-header) 1)))
    
  
  ;; On the key combo, save the buffer to a filename in the dogfood-directory
  ;; that's constructed to be meaningful but let the user modify the name.
  (use-local-map dogfood-note-mode)

  ;; Give a message to the user that she can use some key combo 
  ;; to finish editing
  (message "Type C-c C-c to save and exit."))



;;; dogfood.el ends here
