#!/usr/bin/env python

from interaction import *
import sys
import random
import operator
import getopt
import math

def vec_dot_product(u, v):
    return reduce(operator.add, map(operator.mul, u, v), 0)

def vec_length(u):
    return math.sqrt(vec_dot_product(u, u))

class WeightedChooserTester:
    def __init__(self, chooser_style, num_attributes, num_choices,
                 dump_level = 0, output = None):
        self.regen = 1
        self.chooser_style = chooser_style
        self.num_attributes = num_attributes
        self.num_choices = num_choices
        self.dump_level = dump_level
        if output is None:
            self.output = sys.stdout
        else:
            self.output = output

    def _dump(self, string, level):
        if self.dump_level >= level:
            self.output.write(('[%d]  ' % level) + string)
            self.output.write('\n')

    def _dump_header(self, steps, target, init):
        self._dump("INITIATING NEW RUN", 0)
        self._dump("STEPS: %s" % steps, 0)
        self._dump("TARGET: %s" % target, 1)
        self._dump("START: %s" % init, 2)
        self._dump("-----------------------------------------------", 0)

    def _dump_footer(self):
        self._dump("-----------------------------------------------", 0)

    def _score(self, vec1, vec2):
        return vec_dot_product(vec1, vec2)

    def test(self, steps, dump_level = None, target = None,
             seed = None, init = None):
        """Runs a test of a given number of steps.

        Outputs a header including the target, initial weights, and
        number of steps. Then, outputs the attribute set, ranking of
        choices, and actual ranking at each step.

        If init is the string 'random', creates a random initial state.
        """

        return_error = []
        return_similarity = []

        if init is None:
            init = [0.0] * self.num_attributes

        if init == 'random':
            init = []
            for i in range(self.num_attributes):
                init.append(random.random() * 2 - 1)

        if target is None:
            random.seed(seed)
            target = map(apply, [lambda : random.random() * 2 - 1] * \
                         self.num_attributes, [()] * self.num_attributes)

        if dump_level is not None:
            self.dump_level = dump_level

        self._dump("DEBUG: target: %s init %s" % (target, init), 10)
        
        self._dump_header(steps, target, init)

        chooser = self.chooser_style(self.num_attributes,
                                     self.num_choices, init)

        # Calculate scaling factor to normalize error (scales target
        # vector to 1).
        scaling_factor = 1.0 / vec_length(target)

        for i in range(steps):
            self._dump('STEP #: %s' % (i + 1), 5)
            
            # Construct the new attribute vectors
            rand = lambda : random.random*2 - 1
            att_vectors = []
            for i in range(self.num_choices):
                att_vectors.append([])
                for j in range(self.num_attributes):
                    att_vectors[i].append(random.random() * 2 - 1)

            # Dump the attribute vectors.
            for vector in att_vectors:
                self._dump('VECTOR: %s' % vector, 1)
                    
            # Calculate the correct and chooser's ranking/scores
            correct_scores = map(lambda vec, s=self, t=target:
                                 s._score(t, vec), att_vectors)
            chooser.rescore(att_vectors)
            chooser_scores = chooser.score_list()

            # Calculate rebukes
            best_score = max(chooser_scores)
            best_index = chooser_scores.index(best_score)

            correct_score = max(correct_scores)
            correct_index = correct_scores.index(correct_score)
            chosen_score = correct_scores[best_index]

            est_correct_score = chooser_scores[correct_index]

            error = correct_score - chosen_score

            self._dump('CHOICES: CORRECT_INDEX: %s CORRECT_SCORE: %s CHOSEN_INDEX: %s CHOSEN_SCORE: %s' % (correct_index, correct_score, best_index, chosen_score), 1)
            self._dump('ERROR: %s' % (error * scaling_factor), 0)
            return_error.append(error * scaling_factor)
            if vec_length(chooser.weights) == 0 or vec_length(target) == 0:
                concept_angle = 0.0
            else:
                concept_angle = vec_dot_product(chooser.weights, target) / \
                                vec_length(chooser.weights) / \
                                vec_length(target)
            self._dump('CONCEPT_ANGLE: %s' % concept_angle, 0)
            return_similarity.append(concept_angle)

            spank_update = 0
            best_update = 0
            if error > 0:
                # Calculate whether to spank
                if chosen_score < 0:
                    spank_update = InteractionInterface.REBUKE_UPDATE

                # Calculate the update for the correct choice
                best_update = InteractionInterface.SILENT_UPDATE + \
                              (best_score - est_correct_score)
                

            self._dump('UPDATES: SPANK: %s LIFT: %s' % (spank_update, best_update), 2)
            
            # Update the chooser
            chooser.feedback(best_index, -spank_update)
            chooser.feedback(correct_index, best_update)
            chooser.complete_step()

            # Print new weights
            self._dump('WEIGHTS: %s' % chooser.weights, 2)
        
        self._dump_footer()

        return (return_error, return_similarity)
        
def main():
    dump = 9
    atts = 6
    choices = 5
    steps = 100
    runs = 1
    init = 0
    opts, args = getopt.getopt(sys.argv[1:], 'd:a:c:s:r:i',
                               ['dump=', 'atts=', 'choices=', 'steps=', 'runs=', 'init'])
    for (opt, val) in opts:
        if opt == '-d' or opt == '--dump':
            dump = int(val)
        elif opt == '-a' or opt == '--atts':
            atts = int(val)
        elif opt == '-c' or opt == '--choices':
            choices = int(val)
        elif opt == '-s' or opt == '--steps':
            steps = int(val)
        elif opt == '-r' or opt == '--runs':
            runs = int(val)
        elif opt == '-i' or opt == '--init':
            init = 1
        else:
            print 'Undefined option', `opt`, 'ignored'

    results_e = []
    results_s = []
    results_e_av = []
    for i in range(runs):
        tester = WeightedChooserTester(SimpleInteractionChooser, atts, choices,
                                       dump_level = dump)
        if init:
            (result_e, result_s) = tester.test(steps, init='random')
        else:
            (result_e, result_s) = tester.test(steps)

        # Transform the error results into average error over whole run
        sum = 0.0
        result_e_av = []
        for i in range(len(result_e)):
            sum = sum + result_e[i]
            result_e_av.append(sum / (i + 1))

        results_e.append(result_e)
        results_s.append(result_s)
        results_e_av.append(result_e_av)

    new_results_e = []
    new_results_s = []
    new_results_e_av = []
    for i in range(len(results_e[0])):
        new_results_e.append([])
        new_results_s.append([])
        new_results_e_av.append([])

    for i in range(len(results_e)):
        for j in range(len(results_e[i])):
            new_results_e[j].append(results_e[i][j])
            new_results_s[j].append(results_s[i][j])
            new_results_e_av[j].append(results_e_av[i][j])

    print 'Attributes:', atts, '\tChoices:', choices, '\tSteps:', steps, '\tRuns:', runs
    print

    print 'Normalized error',
    for i in range(runs):
        print '\t',
    print '\t',
    print 'Cosine similarity',
    #for i in range(runs):
    #    print '\t',
    #print '\t',
    #print 'Normalized average error',
    print
    
    for i in range(runs):
        print i + 1, '\t',
    print '\t',
    for i in range(runs):
        print i + 1, '\t',
    #print '\t',
    #for i in range(runs):
    #    print i + 1, '\t',
    print
    
    for i in range(len(new_results_e)):
        av_errors = new_results_e_av[i]
        errors = new_results_e[i]
        sims = new_results_s[i]
        for error in errors:
            print error, '\t',
        print '\t',
        for sim in sims:
            print sim, '\t',
        #print '\t',
        #for error in av_errors:
        #    print error, '\t',
        print

if __name__ == '__main__':
    main()
