# File: interaction.py
#
# Defines the set of interactions available to SMARTedit through the
# DIAManD framework.

# The vital components in this file are the Interaction classes and
# the InteractionController classes. Together, each pair of
# Interaction and InteractionController defines the back end
# calculations/attribute measures and the front end interface of each
# interaction.
#
# The InteractionManager selects the next interaction and handles
# transitions using a SimpleInteractionChooser to model the user.
#
# This file and all others in this distribution are covered by the
# license described in the LICENSE file.
#


#TODO!!!
# thoughts: move interactions up to menu and stress that person should
#  try to run with what they're given (using the spank button to switch)
#  let "do nothing" interaction pop up menu of interactions in command space.
# make run do full backoff

from Tkinter import *
import operator
import math
import vs,textedit,collect
import entropy
import pickle
from threading import *

FULL_ITERATION_LENGTH = 3


##################################################################
#  Generally useful helper functions						#
##################################################################

# Translate a Tk color to an rgb triple
def _to_rgb(self, color):
    return tuple(map(lambda c: c & 0xFF, self.winfo_rgb(color)))

# Translate an rgb triple to a Tk acceptable color
def _to_tk_rgb(self, color):
    return '#%02x%02x%02x' % color

# Return a valuable usable as the length of a single iteration in the
# given SMARTedit app's current macro.
def iteration_count(app):
    if app.vs and app.vs.growth_ended:
	   return app.vs.action_count()
    else:
	   return FULL_ITERATION_LENGTH

def project_best_guess(app, length, instate = None, action_num = None):
    """List the output state probability of making length best guesses
    along with the probabilities of the other possibilities -- e.g.,
    going three steps with the best guess and then choosing something
    other than the best guess -- best guess goes first.  """
    
    assert(length >= 1)
    assert(app.vs)
    
    if action_num is None:
	   action_num = app.vs.get_action_number()
	   
    if instate is None:
	   instate = app.get_text_state('insert')

    current_prob = 1.0
    results = []

    # Stop when length is 1 after handling current step's probability.
    while 1:
	   # Get the available action groups
	   result_set = app.vs.execute_nth(instate, action_num)
	   if not result_set:
		  # No more predictions
		  results.append(current_prob)
		  break

	   # Group the results into equivalent output state sets.
	   result_set.initialize()
	   (hyplist, total_prob) = result_set.pick_next_best()

	   # Add on the lower probability choices.
	   results.append(current_prob - current_prob * total_prob)
	   current_prob = current_prob * total_prob

	   # Check if last action's probability has been calculated
	   if length <= 1:
		  results.append(current_prob)
		  break

	   # Construct the state resulting from taking the highest probability
	   # action.
	   hypothesis = hyplist[0][0].hyp
	   instate = app.simulate_action(hypothesis, instate)
	   action_num = action_num + 1
	   length = length - 1

	   # Action predicted termination.
	   if instate is None:
		  results.append(current_prob)
		  break

    # Put the high probability action sequence first.
    results.reverse()
    return results

# Functions for comparison that allow for some numerical instability.
DEFAULT_EPSILON = 0.00001
def eps_eq(a, b, epsilon = None):
    if epsilon is None:
	   epsilon = DEFAULT_EPSILON
    return abs(a - b) <= epsilon

def eps_lt(lesser, greater, epsilon = None):
    if epsilon is None:
	   epsilon = DEFAULT_EPSILON
    return greater - lesser > -epsilon

# Dot product of two vectors (represented as sequences).
def dot_product(v, w):
    assert(len(v) == len(w))

    return reduce(operator.add, map(operator.mul, v, w), 0)


##################################################################
#  Classes for modelling the user and selecting interactions	#
##################################################################

# Base class interaction chooser.
class InteractionChooser:
    def __init__(self, num_attributes, num_choices):
	   assert(num_attributes >= 1)
	   assert(num_choices >= 1)
	   
	   self.num_attributes = num_attributes
	   self.num_choices = num_choices
    
    def load_profile(self, profile):
	   """Loads an initial set of weights and a step count from a file."""
	   
	   # Should always include a comment string as the first element.
	   raise 'abstract'
    
    def save_profile(self, profile, comment = ''):
	   """Writes an initial set of weights and a step count to a file."""
	   
	   # Should always include a comment string as the first element.
	   raise 'abstract'
    
    def feedback(self, choice, value):
	   """Give feedback on a choice. The value should be between -1 and 1.
	   
	   -1: very, very bad
	   1: very, very good
	   
	   choice should be a number corresponding to the appropriate choice.
	   
	   Cumulative feedback outside of this range will be clipped."""
	   
	   raise 'abstract'
    
    def score(self, choice):
	   """Returns the calculated score of this choice.
	   
	   DOES NOT CALCULATE THE SCORE! Call rescore for that."""
	   
	   raise 'abstract'
    
    def rescore(self, attribute_vectors):
	   """Recalculates the scores for the various choices.
	   
	   An attribute vector of None indicates a disallowed choice."""
	   
	   raise 'abstract'
    
    def score_list(self):
	   return map(self.score, range(self.num_choices))
    
    def complete_step(self):
	   """Finish activity for this particular learning episode."""

	   raise 'abstract'

# Chooser based on a weighted linear combination utility function in
# which the user is modelled as a set of weights.
class WeightedChooser(InteractionChooser):
    STEP_MOD = 1.0

    def __init__(self, num_attributes, num_choices, weights = ()):
	   InteractionChooser.__init__(self, num_attributes, num_choices)
	   
	   # Set the initial weights to 0 unless otherwise specified
	   if not weights:
		  self.weights = (0,) * num_attributes
	   else:
		  self.weights = weights

	   self.step = 0

    def load_profile(self, profile):
	   """Loads an initial set of weights and a step count from a file."""

	   try:
		  f = open(profile + '.profile', 'rb')
	   except IOError:
		  # No profile to load, just skip it.
		  return

	   pickle.load(f) # Run past comment
	   self.weights = pickle.load(f)
	   self.step = pickle.load(f)
	   self.step = int(self.step * self.STEP_MOD)

	   #DEBUG!!!
	   print 'New weights are:', self.weights
	   print 'New step count is:', self.step

    def save_profile(self, profile, comment = ''):
	   """Writes an initial set of weights and a step count to a file."""

	   f = open(profile + '.profile', 'wb')
	   pickle.dump(comment, f)
	   pickle.dump(self.weights, f)
	   pickle.dump(self.step, f)

	   #DEBUG!!!
	   print 'Saving weights:', self.weights
	   print 'Saving step count:', self.step

    # Account for accumulated feedback in the weight vector.
    def _respond_feedback(self):
	   raise 'abstract'

    # Finish a choice step (wrap up the outstanding feedback and set
    # up for the next step). 
    def complete_step(self):
	   self._respond_feedback()
	   self.step = self.step + 1

class SimpleInteractionChooser(WeightedChooser):
    """Helps to score and choose interactions.

    Sequence should go something like:

    rescore interactions
    feedback (as many times as you like on different actions)
    complete_step
    repeat

    Scores can be accessed through the score function or through the
    score_list function. Best choice can be accessed through
    get_best_choice."""

    # The "steps" taken to respond to feedback are weighted by a learning
    # rate that decays over time.

    # Initial learning rate
    INITIAL_UPDATE = 0.25

    # Asymptotic learning rate
    MINIMUM_UPDATE = 0.01

    # Half-life of the learning rate.
    DECAY_RATE = 30.0
    
    def __init__(self, num_attributes, num_choices, weights = ()):
	   WeightedChooser.__init__(self, num_attributes, num_choices,
						   weights)

	   self.scores = [None] * self.num_choices
	   self.feedback_values = [0.0] * self.num_choices

    def feedback(self, choice, value):
	   """Give feedback on a choice. The value should be between -1 and 1.

	   -1: very, very bad
	   1: very, very good

	   choice should be a number corresponding to the appropriate choice.

	   Cumulative feedback outside of this range will be clipped."""

	   # Note: feedback on inapplicable choices is discarded.

	   #DEBUG!!
	   #print 'Accepting feedback of <<', value, '>> on <<', choice, '>>'

	   if self.scores[choice] is not None:
		  self.feedback_values[choice] = self.feedback_values[choice] + value

    def score(self, choice):
	   """Returns the calculated score of this choice.

	   DOES NOT CALCULATE THE SCORE! Call rescore for that."""

	   return self.scores[choice]

    def _rescore_choice(self, choice, attribute_vector):
	   # Calculates the score for the given attribute_vector.
	   # Currently, just takes the dot product of the attributes and the
	   # global weights.
	   if attribute_vector is None:
		  return None
	   else:
		  return dot_product(self.weights, attribute_vector)

    def rescore(self, attribute_vectors):
	   """Recalculates the scores for the various choices.

	   An attribute vector of None indicates a disallowed choice."""
	   
	   self.scores = map(self._rescore_choice, range(self.num_choices),
					 attribute_vectors)
	   self.attribute_vectors = attribute_vectors

    # Calculate and return the current learning rate.
    def _current_factor(self):
	   min = float(self.MINIMUM_UPDATE)
	   max = float(self.INITIAL_UPDATE)
	   rate = float(self.DECAY_RATE)
	   x = float(self.step)

	   return min + (max - min)*2.0**(-x/rate)

    # Return a vector which represents the appropriate response to the
    # feedback on the given choice (i.e., this vector points in the right
    # direction to shift the weights based on the feedback given).
    def _respond_feedback_choice(self, choice):
	   assert(self.scores[choice] is not None)

	   # Truncate feedback to [-1, 1]
	   if not (-1.0 <= self.feedback_values[choice] <= 1.0):
		  if not (-1.0 <= self.feedback_values[choice]):
			 self.feedback_values[choice] = -1.0
		  else:
			 self.feedback_values[choice] = 1.0

	   #DEBUG!!
	   #print 'Updating on <<', choice, '>> to reflect feedback of <<', \
	   #		self.feedback_values[choice], '>>'

	   # Calculate how strongly to respond to this choice.
	   factor = self._current_factor() * self.feedback_values[choice]

	   # Return a vector pointing parallel to this choice's attribute vector
	   # and weighted by the factor (to go in the right direction with the
	   # right magnitude).
	   return map(operator.mul, [factor] * self.num_attributes,
			    self.attribute_vectors[choice])

    def _respond_feedback(self):
	   #Assimilate all pending feedback.

	   # Accumulate the feedback from all the choices.
	   update_vector = [0] * self.num_attributes
	   for choice in range(self.num_choices):
		  if self.scores[choice] is not None:
			 update_vector = map(operator.add, update_vector,
							 self._respond_feedback_choice(choice))

	   # Add the accumulated feedback back into the weight model.
	   self.weights = map(operator.add, self.weights, update_vector)
	   #DEBUG!!
	   #print 'New weights are:', self.weights

	   # Zero the feedback.
	   self.feedback_values = [0.0] * self.num_choices
	   
##################################################################
#  The manager that sets up the current interaction and manages	#
#  the user model and interaction transitions.				#
##################################################################

class InteractionManager:
    """Selects among a variety of interactions to perform at a given point."""
    
    def __init__(self, interaction_types, measures, maxima, app, weights = ()):
	   # interaction_types is a tuple of Interaction classes from which
	   # the interaction manager gets to choose. App is the application
	   # the interaction manager is controlling. measures is a set of
	   # attribute names to measure on the interactions.
	   #
	   # maxima provide upper bounds on the values of the measures (and
	   # lower bounds on the negatives). measures will be clipped to
	   # [-maximum, maximum] and then normalized to [-1,1].
	   
	   self.interactions = interaction_types
	   self.app = app
	   self.measures = measures
	   self.maxima = maxima
	   self.chooser = SimpleInteractionChooser(len(self.measures),
									   len(self.interactions),
									   weights)

    def load_profile(self, profile):
	   self.chooser.load_profile(profile)

    def save_profile(self, profile):
	   self.chooser.save_profile(profile,
						    ("Pickled profile for %s.\n\n" +
							"Available measures are: %s") %
						    (profile, self.measures))

    def reset(self):
	   # Reset the interaction record to NULL.
	   Interaction().RECORD[:] = []

    def _trim_and_norm_vector(self, vector):
	   if vector is None:
		  return vector
	   # Enforce the maxima and normalize
	   def trim_and_norm(value, max):
		  if callable(max):
			 max = max()
		  if value > max:
			 value = max
		  elif value < -max:
			 value = -max
		  return float(value) / max

	   return map(trim_and_norm, vector, self.maxima)
	   
    def score_interactions(self):
	   # Retuns a list of scores in the same order as the interactions.
	   # A score of None indicates an inapplicable interaction.
	   vectors = map(self.score, self.interactions)
	   vectors = map(self._trim_and_norm_vector, vectors)
	   self.chooser.rescore(vectors)
	   scores = self.chooser.score_list()
	   return scores

    def respond_feedback(self):
	   self.chooser.complete_step()

    def score(self, interaction):
	   # Return the vector of attribute score of the given interaction.

	   def get_measure(attribute, i = interaction, a = self.app):
		  # Helper function to actually access the measure.
		  try:
			 return i.measure(attribute, a)
		  except InteractionError:
			 return 0

	   if interaction.applicable(self.app):
		  # Prepare the interaction
		  interaction.prepare_measures(self.app)
		  scores = map(get_measure, self.measures)
	   else:
		  scores = None

	   return scores

    # Give feedback of value on the given choice.
    def feedback(self, choice, value):
	   self.chooser.feedback(choice, value)


##################################################################
#  The Interactions. These are essentially the guts of each		#
#  interaction, calculating the effects and attribute values	#
#  for each interaction. The user interface of the interaction	#
#  is defined by its interaction controller.					#
##################################################################

class InteractionError(RuntimeError):
    pass

class Interaction:

    # A record of the interactions performed.
    RECORD = []

    # Mapping from measure names to functions that calculate the
    # measure's (attribute's) value.
    measures = {}

    # The hand-picked cost of transitioning from the named interaction
    # to this interaction
    follow_ratings = {}

    # The type of this interaction, used for measuring follows and for
    # some debugging/interface text.
    type = 'Generic'

    # Time limit on attribute calculations.
    TIME_LIMIT = 0.1
    
    def __init__(self):
	   pass

    def mark_as_executed(self):
	   self.RECORD.append(self.type)

    def prepare_measures(self, app):
	   # Perform any necessary preparation for the measures.
	   # Should be called once before accessing any measures of
	   # a given interaction on a given state.
	   pass
	   
    def measure(self, attribute, app):
	   """Returns the measured value of the given attribute.
	   
	   attribute is a string naming an attribute. If the attribute
	   is unmeasurable on this Interaction, raises an InteractionError."""

	   if hasattr(self, 'measure_' + attribute):
		  return apply(getattr(self, 'measure_' + attribute),
					(attribute, app))
	   elif self.measures.has_key(attribute):
		  return apply(self.measures[attribute], (attribute, app))
	   else:
		  # No facility for handling this measure
		  raise InteractionError, 'cannot handle attribute: ' + attribute

    def measure_do_nothing(self, attribute, app):
	   # Just to distinguish the special NULL interaction
	   # Only DoNothing should override.
	   return 0

    def measure_follows(self, attribute, app):
	   """Return a measurement of how well this interaction follows the last.
	   
	   Measurement ranges from 0 (default; doesn't follow) to 1 (follows
	   logically)."""
	   assert(attribute == 'follows')
	   
	   # Get the type of the previous action; empty string if
	   # no previous action exists
	   if len(self.RECORD) == 0:
		  previous = ''
	   else:
		  previous = self.RECORD[-1][0]
		  
	   if self.follow_ratings.has_key(previous):
		  return self.follow_ratings[previous]
	   
	   # Return the rating of 'Generic' as the customized default or
	   # 0 as the general default.
	   return self.follow_ratings.get('Generic', 0)
    
    def applicable(self, app):
	   """Return true iff the interaction is applicable.
	   
	   Check the current app state to see if it suits the interaction's
	   needs."""
	   
	   raise InteractionError, 'must be overridden by subtypes.'
    

# Interaction which returns control of the interaction choice to the user.
# This acts as a "bar" for other interactions. If none of them scores better
# than this interaction, then DIAManD will *not* select any interaction,
# leaving the choice in such an uncertain situation to the user.
#
# The actual value of the bar is set either in the profile itself or
# in InteractionInterface.
class DoNothingInteraction(Interaction):
    type = 'DoNothing'
    follow_ratings = {'' : 1,
				  'Generic' : 0.1}

    def __init__(self):
	   Interaction.__init__(self)
	   
    # ONLY THIS INTERACTION SHOULD RETURN A NON-ZERO VALUE FOR THIS ATTRIBUTE
    def measure_do_nothing(self, attribute, app):
	   return 1

    def measure_probability_of_error(self, attribute, app):
	   return 0

    def measure_user_burden(self, attribute, app):
	   return 0

    def measure_value(self, attribute, app):
	   return 0

    def measure_progress(self, attribute, app):
	   return 0

    def applicable(self, app):
	   return 1

# The interaction for recording a single step.
class RecordActionInteraction(Interaction):
    type = 'RecordAction'
    follow_ratings = {'' : 0.3,
				  'RecordIteration' : 0.7,
				  'RecordAction' : 1,
				  'Jump' : 0.15,
				  'Playback' : 0.4,
				  'Suggest' : 0.5,
				  'RunWhileConfident' : 0.7}
    
    def measure_probability_of_error(self, attribute, app):
	   return 0

    def measure_value(self, attribute, app):
	   # Find and return the entropy/information gain of disambiguating
	   # the next action. 
	   calc = app.get_entropy_calculator('insert', 1)
	   calc.timed_approx_calc(tolerance_percent = 0,
						 time_limit = self.TIME_LIMIT)
	   return calc.get_entropy_estimate()

    def measure_user_burden(self, attribute, app):
	   return 1

    def measure_progress(self, attribute, app):
	   return 1

    def applicable(self, app):
	   # Probably should check to see whether we have any prediction
	   # for the next step.
	   try:
		  return app.vs and app.vs.growth_ended
	   except vs.VSError:
		  return 0

# The interaction in which the system plays back the rest of this iteration.
class PlaybackInteraction(Interaction):
    type = 'Playback'
    follow_ratings = {'' : 0,
				  'SuggestAction' : 0.9,
				  'Playback' : 1,
				  'RecordAction' : 0.8,
				  'RecordIteration' : 0.8,
				  'Jump' : 0.8,
				  'RunWhileConfident' : 0.4}

    # Calculate the number of steps to the last step of this iteration.
    def _steps_to_take(self, app):
	   i = iteration_count(app)
	   if i > 0:
		  current = app.vs.get_action_index()
		  target = i - current
	   else:
		  target = i
	   return target
    
    def measure_probability_of_error(self, attribute, app):
	   return 1 - project_best_guess(app, self._steps_to_take(app))[0]

    def measure_user_burden(self, attribute, app):
	   # A small fixed amount plus a small amount per step.
	   return 0.4 + 0.1 * self._steps_to_take(app)

    def measure_value(self, attribute, app):
	   # Entropy of getting to the end or stopping at one of the
	   # steps in the middle. (Note that this isn't quite right since
	   # we don't actually throw out the guess at the step where the
	   # user interrupts).
	   guess_array = project_best_guess(app, self._steps_to_take(app))
	   return entropy.calc_entropy(guess_array)
	   
    def measure_progress(self, attribute, app):
	   return self._steps_to_take(app)

    def applicable(self, app):
	   # Probably should check to see whether we have any prediction
	   # for the next step.
	   try:
		  return app.vs and app.vs.growth_ended
	   except vs.VSError:
		  return 0


# Interaction in which the system applies its best guess until its
# confidence drops below a fixed threshold.
class RunWhileConfidentInteraction(Interaction):
    type = 'RunWhileConfident'
    follow_ratings = {'' : 0,
				  'SuggestAction' : 0.9,
				  'Playback' : 1,
				  'RecordAction' : 0.8,
				  'RecordIteration' : 0.8,
				  'Jump' : 0.8,
				  'RunWhileConfident' : 0.0}

    # Threshold for stopping the run.
    CONFIDENCE_THRESHHOLD = 0.99

    # Maximum number of steps to take (regardless of confidence).
    MAX_STEPS = 200

    def prepare_measures(self, app):
	   self.steps = self._steps_to_take(app)

    # Calculate the number of steps to take until the confidence
    # drops below threshold.
    def _steps_to_take(self, app):
	   running_prob = 1.0
	   current_prob = 1.0
	   steps = 0

	   state = app.get_text_state('insert')
	   current_step = app.vs.get_action_number()
	   while current_prob >= self.CONFIDENCE_THRESHHOLD and \
		    steps < self.MAX_STEPS:
		  self.final_prob = running_prob
		  result_set = app.vs.execute_nth(state, current_step + steps)
		  if not result_set:
			 #DEBUG!!!
			 print 'Empty result set stopping lookahead.'
			 break
		  steps = steps + 1
		  result_set.initialize()
		  (hyplist, current_prob) = result_set.pick_next_best()
		  running_prob = running_prob * current_prob
		  state = app.simulate_action(hyplist[0][0].hyp, state)
		  if not state:
			 #DEBUG!!!
			 print 'Termination (probably) ending lookahead.'
			 break

	   if current_prob < self.CONFIDENCE_THRESHHOLD:
		  steps = steps - 1 # Don't take that fatal last step!
	   else:
		  self.final_prob = running_prob

	   return steps
    
    def measure_probability_of_error(self, attribute, app):
	   return 1 - self.final_prob

    def measure_user_burden(self, attribute, app):
	   return 0.005 * self.steps + 0.3

    def measure_value(self, attribute, app):
	   return 0
	   
    def measure_progress(self, attribute, app):
	   return self.steps

    def applicable(self, app):
	   try:
		  if not (app.vs and app.vs.growth_ended):
			 return 0
		  result_set = app.vs.execute_nth(app.get_text_state('insert'))
		  if not result_set:
			 return 0
		  result_set.initialize()
		  (hyplist, totalprob) = result_set.pick_next_best()
		  return totalprob >= self.CONFIDENCE_THRESHHOLD
	   except vs.VSError:
		  return 0


# Interaction in which the user solves the entire upcoming iteration.
class RecordIterationInteraction(Interaction):
    type = 'RecordIteration'
    follow_ratings = {'' : 1,
				  'RecordIteration' : 1,
				  'RecordAction' : 0.2,
				  'Jump' : 0.15,
				  'Playback' : 0.05,
				  'Suggest' : 0.1}

    def measure_user_burden(self, attribute, app):
	   # Handle case where this is called before actually ending growth.
	   if not app.vs or not app.vs.growth_ended:
		  return 0

	   return 1 * app.vs.action_count()

    def measure_probability_of_error(self, attribute, app):
	   return 0
    
    def measure_value(self, attribute, app):
	   # Handle case where this is called before actually ending growth.
	   if not app.vs or not app.vs.growth_ended:
		  return 0
	   
	   calc = app.get_entropy_calculator('insert', iteration_count(app))
	   calc.timed_approx_calc(tolerance_percent = 0.1,
						 time_limit = self.TIME_LIMIT)
	   return calc.get_entropy_estimate()

    def measure_progress(self, attribute, app):
	   return app.vs.action_count()

    def applicable(self, app):
	   return 1

# Interaction in which the system proposes a solution for the next action
# and the user cycles through the system's guesses until she finds the right
# one.
class SuggestActionInteraction(Interaction):
    type = 'SuggestAction'
    follow_ratings = {'' : 0,
				  'RecordAction' : 0.7,
				  'RecordIteration' : 0.7,
				  'SuggestAction' : 1,
				  'Jump' : 0.8,
				  'Playback' : 0.9}
	
    def measure_user_burden(self, attribute, app):
	   return 0.2

    def measure_probability_of_error(self, attribute, app):
	   return 1 - project_best_guess(app, 1)[0]

    def measure_value(self, attribute, app):
	   calc = app.get_entropy_calculator('insert', 1)
	   calc.timed_approx_calc(tolerance_percent = 0.1,
						 time_limit = self.TIME_LIMIT)
	   return calc.get_entropy_estimate()

    def measure_progress(self, attribute, app):
	   return 1

    def applicable(self, app):
	   # Probably should check to see whether we have any prediction
	   # for the next step.
	   try:
		  return app.vs and app.vs.action_count() > 0
	   except vs.VSError:
		  return 0

# Interaction in which the system selects a high entropy example in the
# file and asks the user to solve that example.
class JumpInteraction(Interaction):
    type = 'Jump'
    follow_ratings = {'' : 0,
				  'SuggestAction' : 0.3,
				  'Playback' : 0.1,
				  'RecordAction' : 0.8,
				  'RecordIteration' : 0.8,
				  'Jump' : 0.5}

    def __init__(self):
	   Interaction.__init__(self)
	   
    def measure_probability_of_error(self, attribute, app):
	   return 0

    def measure_user_burden(self, attribute, app):
	   return 1 * iteration_count(app) + 1

    def measure_value(self, attribute, app):
	   """Jump to new example and record that example.
	   
	   Returns to current state on stop record but retains learned
	   information. Attempts to pick best available example.
	   """

	   NUM_POINTS_PER_TYPE = 5
	   TOLERANCE = 0.5
	   
	   # Pick a swath of points to consider
	   choices = self.generate_choices(app, NUM_POINTS_PER_TYPE)
	   
	   # Select the point to jump to.
	   (choice, calc) = self.select_point(app, choices, TOLERANCE)
	   
	   self.calc = calc
	   self.choice = choices[choice]
	   calc.timed_approx_calc(tolerance_percent = 0.1,
						 time_limit = self.TIME_LIMIT)
	   return calc.get_entropy_estimate()

    def measure_progress(self, attribute, app):
	   return 0

    def get_choice(self):
	   return self.choice
    
    def applicable(self, app):
	   try:
		  return app.vs and app.vs.growth_ended and \
			    app.vs.action_type() == 'move'
	   except vs.VSError:
		  return 0
	   
    def generate_choices(self, app, num_per_choice_type):
	   import ent_util
	   state = app.get_text_state('insert')
	   generators = [
		  #ent_util.RandomExamplePicker(app.vs, state),
		  ent_util.LinewiseExamplePicker(app.vs, state),
		  ent_util.MoveBasedExamplePicker(app.vs, state),
		  ]
	   
	   # Create a list of num_per.. sets of points from each generator,
	   # throwing out points that come back as 'None'.
	   examples = filter(None, map(lambda x: x.next_example(),
							 generators * num_per_choice_type))
	   
	   # Sort and remove duplicates.
	   examples.sort(lambda x, y: cmp(textedit.mark_split(x),
							    textedit.mark_split(y)))
	   i = 1
	   while i < len(examples):
		  if examples[i] == examples[i-1]:
			 del examples[i]
		  else:
			 i = i + 1

	   return examples
		
    def select_point(self, app, locations, tolerance = None):
	   """Select the point with the highest entropy of the given points."""
	   action_num = app.vs.get_action_number()
	   action_count = app.vs.action_count()
	   
	   # Calculate log leaf counts:
	   branch_counts = map(app.vs.num_hyps_nth,
					   range(action_num, action_num +
						    app.vs.action_count()))
	   log_branch_counts = map(entropy.ent_log, branch_counts)
	   # Initialize counts to 1 with extra slot for convenience.
	   log_leaf_counts = [0] * (len(log_branch_counts) + 1)
	   for i in range(len(log_branch_counts), 0, -1):
		  log_leaf_counts[i - 1] = log_leaf_counts[i] + \
							  log_branch_counts[i - 1]
	   del log_leaf_counts[-1]

	   num = len(locations)
	   executor = lambda i, s, b=action_num, v=app.vs: v.execute_nth(i,s+b)
	   ecs = map(entropy.EntropyCalculator,
			   [log_leaf_counts] * num,
			   [action_count] * num,
			   [executor] * num,
			   [app.simulate_action] * num,
			   map(app.get_text_state, locations))
	   
	   if tolerance is None:
		  tolerance = 0.5
	   choice = entropy.choose_highest_entropy(ecs, tolerance)
	   return (choice, ecs[choice])



##################################################################
#  The user interface portion of the interaction manager.		#
##################################################################

# A class which sets up the interactions, displays interaction choices,
# and controls transitions among interactions.
class InteractionInterface(Frame):
    # Note that a rebuked action will likely also suffer the SILENT_UPDATE
    # as another action will be selected over it.
    SILENT_UPDATE = 0.1
    REBUKE_UPDATE = 1

    # The 'bar' for interactions. If they don't rate higher than this,
    # we'll pick 'DoNothing' instead.
    BAR = 0.0
    
    def __init__(self, master):
	   Frame.__init__(self, master, relief='raised', borderwidth=2)

	   self.master = master
	   self.manager = None
	   self.interactions = []
	   self.buttons = []
	   self.controllers = []

	   # The currently executing interaction.
	   self.current = None

	   # The interactions that have been touched this cycle including
	   # the current one.
	   self.used_interactions = []

	   # True when we are forcing continuation of this interaction cycle.
	   self.force_continue = 0

	   # The currently set radio button
	   self.active_button = IntVar()
	   self.active_button.set(-1)

	   # A record of this round's scores
	   self.raw_scores = None
	   self.scores = None

	   # Create our widgets
	   self._create_widgets()

	   # The measures to use and their maximum values.
	   self.MEASURES = ['follows', 'probability_of_error', 'user_burden',
					'value', 'progress']
	   ibound = lambda a=self.master: iteration_count(a)
	   ENTROPY_CAP = 4
	   self.BOUNDS = [1, 1, lambda f=ibound: f()+1, ENTROPY_CAP,
				   lambda f=ibound: f() * 3]
    
	   #KEY = 'F1'
	   #BOUND_KEY = '<Key-%s>' % KEY
	   #self.binding = self.master.text.bind(self.BOUND_KEY,
	   #								 lambda event, s=self:
	   #								 s.go_default())


    def rebuke(self):
	   if self.current is not None:
		  assert(self.scores is not None)

		  # One possible rebuke is an update increment off the amount
		  # necessary to lower this to second best.
		  second_rebuke = 0
		  try:
			 best_score = max(filter(lambda x: x is not None,
								self.raw_scores))
			 best_index = self.raw_scores.index(best_score)
			 second_best = max(filter(lambda x: x is not None,
								 self.raw_scores[:best_index] + \
								 self.raw_scores[best_index + 1:]))
			 second_rebuke = second_best - \
						  self.raw_scores[self.current] - \
						  self.REBUKE_UPDATE
		  except ValueError:
			 pass

		  # Another possibility is simply the update increment
		  update_rebuke = -self.REBUKE_UPDATE

		  if second_rebuke < update_rebuke:
			 rebuke = second_rebuke
		  else:
			 rebuke = update_rebuke

		  self.manager.feedback(self.current, rebuke)

    def go_default(self):
	   if self.current is None:
		  self.buttons[self.best].invoke()
	   else:
		  self.buttons[self.current].invoke()

    def reset(self):
	   if self.current is not None:
		  # End the other interaction without ending the interaction
		  # cycle.
		  self.force_continue = 1
		  self.controllers[self.current].reset()
		  self.force_continue = 0
	   self._erase_current()
	   self._rescore_interactions()
	   self.do_nothing_button.invoke()
	   self.used_interactions = []

    def initialize(self):
	   """Sets up the first interaction.

	   Should be called after adding all interactions but before giving
	   user access to panel."""

	   # Add on the do nothing button purely for the purpose of
	   # having a 'bar' to measure against.
	   do_nothing = DoNothingInteraction()
	   self.add_interaction(do_nothing,
					    DoNothingController(self.master, do_nothing),
					    invisible=1)
	   self.do_nothing_button = self.buttons[-1]
	   weights = [0] * len(self.MEASURES) + [self.BAR]
	   
	   self.manager = InteractionManager(self.interactions,
								  self.MEASURES + ['do_nothing'],
								  self.BOUNDS + [1],
								  self.master, weights)

	   self.reset()
	   
    def _create_widgets(self):
	   # Create the widgets for this panel.
	   label = Label(self, text='Interaction Control Panel',
				  font='Helvetica 12')
	   label.pack(side=TOP, expand=0, fill=X)
	   self.choicebar = Frame(self, relief='sunken', borderwidth=2)
	   self.choicebar.pack(side=TOP, expand=0, fill=X)

	   self.default_controller = InteractionController(None, None)
	   self.default_controller.spankbut.configure(state='disabled')
	   self.default_controller.configure(relief='sunken', borderwidth=2)
	   self._erase_current()

    def _get_full_button_command(self, index):
	   def full_button_command(s=self, i=index):
		  if s.current is not None:
			 # End the other interaction without ending the interaction
			 # cycle.
			 s.force_continue = 1
			 s.controllers[s.current].reset()
			 s.force_continue = 0
		  s._activate_button(i)
		  s._execute(i)

	   return full_button_command

    def add_interaction(self, interaction, subcontroller, name=None,
				    invisible=0):
	   """Add the given interaction with the given interface (controller).

	   Name defaults to the interaction's type."""
	   
	   if name is None:
		  name = interaction.type
	   
	   index = len(self.interactions)
	   self.interactions.append(interaction)

	   # The response to an interaction button's press.
	   button_command = self._get_full_button_command(index)

	   but = self._construct_button(name, button_command, index,
							  invisible)

	   # Set this aside to restore it later.
	   assert(not hasattr(but, 'saved_command'))
	   but.saved_command = button_command
	   self.buttons.append(but)

	   # Function to be called as the interaction winds itself down.
	   def end_command(finished = 0, s=self, button=index):
		  if s.force_continue:
			 # Put the button and control panel back into fallow state.
			 s._erase_current()
		  else:
			 # Unless we are forcing continuation of the interaction cycle,
			 # end the cycle, responding to feedback and choosing a new
			 # interaction.
			 s._respond_interaction()
			 s._erase_current()

			 if finished:
				s._disable_interactions()
			 else:
				s._rescore_interactions()
				s._begin_best()

	   def spank_command(s = self):
		  # Rebuke the current interaction and remove it from the running
		  # as one of the top choices.
		  assert(s.current is not None)
		  spank_index = s.current
		  s.rebuke()
		  s.scores[spank_index] = 0
		  s.force_continue = 1
		  s.controllers[spank_index].reset()
		  s.force_continue = 0

		  # Find the next best interaction
		  try:
			 mod_scores = s.scores[:]
			 mod_scores[spank_index] = None
			 best_score = max(filter(lambda x: x is not None, mod_scores))
			 best_index = mod_scores.index(best_score)
		  except ValueError:
			 s.set_status('Repeating interaction; no other choice!',
					    0.5)
			 best_index = spank_index

		  s._begin_choice(best_index)

	   # Prepare the control panel
	   subcontroller.configure(relief='sunken', borderwidth=2)
	   subcontroller.end_callback(end_command)
	   subcontroller.spank_callback(spank_command)

	   #TEST!!! Temporarily disabling this functionality to play with
	   # it for user testing.
	   #if subcontroller.finished_button is not None:
	   #    highlightcolor = 'white'
	   #    subcontroller.finished_button.configure(background=highlightcolor)
	   self.controllers.append(subcontroller)
	   
    def _respond_interaction(self):
	   # Learn from the lessons of this interaction cycle.
	   assert(self.current is not None)

	   # Perform silent updates (rewarding the current interaction
	   # and penalizing all others that were used in this cycle).
	   self.used_interactions.remove(self.current)

	   # Punish all other interactions that were used
	   best_score = None
	   second_best = None
	   try:
		  best_score = max(filter(lambda x: x is not None, self.raw_scores))
		  best_index = self.raw_scores.index(best_score)
		  second_best = max(filter(lambda x: x is not None,
							  self.raw_scores[:best_index] + \
							  self.raw_scores[best_index + 1:]))
	   except ValueError:
		  pass

	   for choice in self.used_interactions:
		  if self.interactions[choice].type == 'DoNothing':
			 continue

		  if second_best is not None and \
			self.raw_scores[choice] > second_best:
			 self.manager.feedback(choice, second_best - \
							   self.raw_scores[choice] - \
							   self.SILENT_UPDATE)
		  else:
			 self.manager.feedback(choice, -self.SILENT_UPDATE)

	   # Reward the current interaction if it wasn't already rating best.
	   if self.raw_scores[self.current] != best_score:
		  self.manager.feedback(self.current, best_score - \
						    self.raw_scores[self.current] + \
						    self.SILENT_UPDATE)
	   
	   self.manager.respond_feedback()

	   # For the 'follows' rating.
	   self.interactions[self.current].mark_as_executed()
	   self.used_interactions = []

    def _begin_choice(self, choice):
	   # Execute the chosen button
	   #self.buttons[choice].flash()
	   self.active_button.set(choice)
	   self.master.set_status('Initiating the <<%s>> interaction!' % \
						 self.buttons[choice]['text'])
	   self.buttons[choice].invoke()

    def _begin_best(self):
	   # Start up the best interaction automatically.
	   # Assumes that scores have already been calculated!

	   # Find the best interaction
	   try:
		  best_score = max(filter(lambda x: x is not None, self.scores))
		  best_index = self.scores.index(best_score)
	   except ValueError:
		  raise 'no interaction to perform'

	   self._begin_choice(best_index)


    def _erase_current(self):
	   if self.current is not None:
		  self._unenable()
	   self.default_controller.pack(in_=self, side=TOP,
							  expand=1, fill=BOTH)

    def _unenable(self):
	   if self.current is None:
		  return
	   self.controllers[self.current].pack_forget()
	   button = self.buttons[self.current]
	   button.configure(command=button.saved_command)
	   self._normalize_button(self.current)
	   if self.scores is not None:
		  if self.scores[self.current] is not None:
			 self._score(self.current, self.scores[self.current])
		  else:
			 self._disable(self.current)
	   self.current = None
    
    def _enable(self, index):
	   if self.current is not None:
		  self._unenable()
		  self.current = None
	   self.default_controller.pack_forget()
	   self.current = index
	   controller = self.controllers[index]
	   controller.pack(in_=self, side=TOP, expand=1, fill=BOTH)

	   def new_button_command(b=controller.finished_button):
		  if b is not None:
			 b.flash()
			 b.invoke()

	   # TEST!!! Temporarily disabling this capability to see if
	   # it helps with the user interface.
	   #self.buttons[index].configure(command=new_button_command)
	   self.buttons[index].configure(command=lambda : None)

    def _execute(self, index):
	   self._enable(index)
	   if index not in self.used_interactions:
		  self.used_interactions.append(index)
	   self.controllers[index].execute()

    def _normalize_scores(self, scores):
	   nonnone_scores = filter(lambda x: x is not None, scores)
	   top_score = max(nonnone_scores)
	   bottom_score = min(nonnone_scores)
	   if top_score - bottom_score == 0:
		  for i in range(len(scores)):
			 if scores[i] is not None:
				scores[i] = 1
	   else:
		  factor = top_score - bottom_score
		  for i in range(len(scores)):
			 if scores[i] is not None:
				scores[i] = (scores[i] - bottom_score) / factor

    def _disable_interactions(self):
	   self.active_button.set(-1)
	   for i in range(len(self.interactions)):
		  self._disable(i)

    def _rescore_interactions(self):
	   self._disable_interactions()
	   self.raw_scores = self.manager.score_interactions()
	   self.scores = self.raw_scores[:]
	   self._normalize_scores(self.scores)
	   for i in range(len(self.interactions)):
		  if self.scores[i] is not None:
			 self._score(i, self.scores[i])
		  else:
			 self._disable(i)

    #####################################################
    ## The following functions along with _create_widgets
    ## and __init__ should be overriden to give the panel
    ## a different look.

    def _activate_button(self, index):
	   # Make the button look activated (called when the button is pressed).
	   raise 'abstract method'	    

    def _construct_button(self, name, command, index, invisible = 0):
	   # Create, pack, and return the button. Note that
	   # EVEN if more than just the button is created, only the
	   # radiobutton itself should be returned and there should
	   # be a radiobutton.
	   but = Radiobutton(self.choicebar, text=name,
					 command=command, anchor='w',
					 value=index, variable=self.active_button)
	   if not invisible:
		  but.pack(side=TOP, expand=0, fill=X, pady=3, padx=3)
	   return but

    def _normalize_button(self, index):
	   # Return the button to its 'usual' appearance and action.
	   raise 'abstract method'

    def _disable(self, index):
	   # Make the button appear disabled (and act disabled).
	   raise 'abstract method'

    def _score(self, index, score):
	   # Adjust the button/display to reflect the given score.
	   raise 'abstract method'

class IIColorButtons(InteractionInterface):
    def __init__(self, master):
	   InteractionInterface.__init__(self, master)
	   
	   self.STD_BG = _to_rgb(self, Button(self)['background'])
	   self.BAD_BG = (0x80, 0x80, 0xFF)
	   self.REC_BG = self.STD_BG
	   self.STD_FG = _to_rgb(self, Button(self)['foreground'])
	   self.REC_FG = (0x00, 0x00, 0x00)
	   self.BAD_FG = (0x00, 0x00, 0x00)
	   self.ACT_BG = (0xFF, 0xFF, 0xFF)
	   self.ACT_FG = (0x00, 0x00, 0x00)

    def _create_widgets(self):
	   InteractionInterface._create_widgets(self)

    def _activate_button(self, index):
	   self.buttons[index].configure(background=_to_tk_rgb(self, self.ACT_BG),
							   foreground=_to_tk_rgb(self, self.ACT_FG))
    
    def _score(self, index, score):
	   # Adjust the button/display to reflect the given score.
	   def calc_color(sc, top=(0xff, 0xff, 0xff),
				   bottom=(0x20, 0x20, 0x80)):
		  return ((int)((top[0] - bottom[0]) * sc + bottom[0]),
				(int)((top[1] - bottom[1]) * sc + bottom[1]),
				(int)((top[2] - bottom[2]) * sc + bottom[2]))

	   if self.buttons[index]['state'] == 'disabled':
		  self.buttons[index].configure(state='normal')
	   bg_color = calc_color(score, self.REC_BG, self.BAD_BG)
	   fg_color = calc_color(score, self.REC_FG, self.BAD_FG)
	   self.buttons[index].configure(background=_to_tk_rgb(self, bg_color),
							   foreground=_to_tk_rgb(self, fg_color))

    def _normalize_button(self, index):
	   # Return the button to its 'usual' appearance and action.
	   self.buttons[index].configure(state='normal',
							   background=_to_tk_rgb(self, self.STD_BG),
							   foreground=_to_tk_rgb(self, self.STD_FG))

    def _disable(self, index):
	   # Make the button appear disabled (and act disabled).
	   button = self.buttons[index]
	   button.configure(background=_to_tk_rgb(self, self.STD_BG),
					foreground=_to_tk_rgb(self, self.STD_FG))
	   button.configure(state='disabled')

class IIBarsAndFading(InteractionInterface):
    class Bar(Canvas):
	   def __init__(self, master, *args, **kwargs):
		  apply(Canvas.__init__, [self, master] + list(args), kwargs)
		  self.bar = None
		  self.color = None
		  self.configure(relief='sunken', borderwidth=1)

		  zero_color = _to_rgb(self, 'black')
		  one_color = _to_rgb(self, 'green')
		  height = int(self['height'])
		  width = int(self['width'])
		  for line in range(width + 1):
			 color = [0] * len(zero_color)
			 for i in range(len(zero_color)):
				color[i] = (one_color[i] - zero_color[i]) * \
						 float(line) / width + zero_color[i]
				color[i] = int(color[i])
			 color = tuple(color)
			 self.create_line(line, -1, line, height+1,
						   fill = _to_tk_rgb(self, color))

		  self.hider = None

	   def hide(self):
		  if self.hider is None:
			 self.hider = self.create_rectangle(-1,-1,int(self['width'])+1,
										 int(self['height']) + 1,
										 fill=self['background'],
										 outline='')
		  
	   def show(self):
		  if self.hider is not None:
			 self.delete(self.hider)
			 self.hider = None
		  
	   def set_bar_size(self, value):
		  # Value should be between 0 and 1 and it sets the bar's size
		  # appropriately.
		  if self.bar is not None:
			 self.delete(self.bar)
			 self.bar = None

		  cheight = int(self['height']) + 1
		  cwidth = int(self['width']) + 1
		  value = float(value)

		  self.bar = self.create_rectangle(value * (cwidth + 1),
									-1,
									cwidth,
									cheight + 1,
									fill=self['background'],
									outline='')
			 
    def __init__(self, master):
	   InteractionInterface.__init__(self, master)
	   self.bars = []

    def _create_widgets(self):
	   InteractionInterface._create_widgets(self)
#	   self.fade_to_color = _to_rgb(self, self['background'])
#	   self.fade_from_color = _to_rgb(self, self['background'])

    def _construct_button(self, name, command, index, invisible = 0):
	   # Create, pack, and return the button. Note that
	   # EVEN if more than just the button is created, only the
	   # radiobutton itself should be returned and there should
	   # be a radiobutton.
	   f = Frame(self, relief='flat')
	   self.bars.append(self.Bar(f, width='50', height='20'))
	   assert(len(self.bars) == index + 1)
	   self.bars[-1].set_bar_size(0)
	   but = Radiobutton(f, text=name, command=command, anchor='w',
					 value=index, variable=self.active_button)
	   but.pack(side=LEFT, expand=1, fill=X, pady=3, padx=3)
	   self.bars[-1].pack(side=RIGHT, expand=0, fill=NONE, pady=3, padx=5)

	   if not invisible:
		  f.pack(side=TOP, expand=0, fill=X, pady=3, padx=3)
	   
#	   self.fade_to_color = _to_rgb(self, but['background'])
#	   self.fade_from_color = _to_rgb(self, but['foreground'])
	   return but		 

    def _activate_button(self, index):
	   self.bars[index].show()
#	   self.buttons[index].configure(foreground = \
#							   _to_tk_rgb(self, self.fade_from_color))

    def _score(self, index, score):
	   if self.buttons[index]['state'] == 'disabled':
		  self.buttons[index].configure(state='normal')

	   self.bars[index].set_bar_size(score)
	   self.bars[index].show()
	   # all button fonts appear same color (black)
#	   new_font_color = [0,0,0]
#	   for i in range(len(self.fade_from_color)):
#		  new_font_color[i] = self.fade_from_color[i] - self.fade_to_color[i]
#		  new_font_color[i] = (new_font_color[i] * (score*score*0.9 + 0.1))
#		  new_font_color[i] = int(new_font_color[i] + self.fade_to_color[i])
#	   new_font_color = tuple(new_font_color)
#	   self.buttons[index].configure(foreground = \
#							   _to_tk_rgb(self, new_font_color))

    def _normalize_button(self, index):
#	   self.buttons[index].configure(foreground = \
#							   _to_tk_rgb(self, self.fade_from_color))
	   self.buttons[index].configure(state='normal')
	   self.bars[index].show()

    def _disable(self, index):
	   self.buttons[index].configure(state='disabled')
	   self.bars[index].hide()

class InteractionController(Frame):
    def __init__(self, app, interaction, name = ''):
	   Frame.__init__(self, app)

	   self.name = name
	   self.app = app
	   self.interaction = interaction
	   # SUBCLASSES MUST ASSIGN TO DEFAULT BUTTON!
	   self.finished_button = None

	   self._create_widgets()
	   self.end = lambda x=0 : None
	   self.spank = lambda : None

    def execute(self):
	   pass

    # Each subclass is responsible for calling self.end() when it finishes
    # control.
    def end_callback(self, callback):
	   self.end = callback

    def spank_callback(self, callback):
	   self.spank = callback

    def _create_widgets(self):
	   if self.name:
		  self.label = Label(self, text=self.name, font='Helvetica 14',
						 background='red', foreground='black')
	   else:
		  self.label = Label(self, font='Helvetica 14', text='')
	   self.label.pack(side=TOP, expand=0, fill=X, pady=5, padx=3)
	   self.spankbut = Button(self, text="Suggest different interaction",
						 command=self._spank)
	   self.spankbut.pack(side=TOP, expand=0, fill=X, pady=5, padx=3)

    def reset(self):
	   # Stop whatever is happening, clean up, and end.
	   # must call end
	   self.end()

    def _spank(self):
	   self.spank()

    def _standard_pack(self, button):
	   # Pack the given button in the standard style.
	   button.pack(side=TOP, expand=0, fill=X, pady=3, padx=3)

class RecordController(InteractionController):
    def __init__(self, app, interaction, name = 'Recording'):
	   InteractionController.__init__(self, app, interaction, name)

    def execute(self):
	   self.app.start_collect()
	   
    def _create_widgets(self):
	   InteractionController._create_widgets(self)
	   
	   donebut = Button(self, text='Done recording',
					command=self.done)
	   nomorebut = Button(self, text='No more examples',
					  command=self.nomore)
	   self._standard_pack(donebut)
	   self._standard_pack(nomorebut)
	   self.finished_button = donebut

    def reset(self):
	   #self.app.cancel_collect()
	   self.app.system_stop_collect()
	   self.end()

    def nomore(self):
	   self.app.no_more_examples()
	   self.done()

    def done(self):
	   self.app.stop_collect()
	   self.end()

class JumpController(RecordController):
    def __init__(self, app, interaction, name = 'Jumping'):
	   RecordController.__init__(self, app, interaction, name)
	   
    def execute(self):
	   self.app.set_status("Jumping to a new example. " + \
					   "Record on that then hit done recording" + \
					   " to finish.",
					   2.25)
	   
	   self.undostate = self.app.getundostate()
	   self.action_num = self.app.vs.get_action_number()

	   self.app.text.mark_set('insert', self.interaction.get_choice())
	   self.app.text.tag_remove('sel', '1.0', 'end')
	   self.app.text.see('insert')
	   self.app.text.tag_configure('fadeout', foreground='darkgray')
	   self.app.text.tag_add('fadeout', '1.0', self.app.text.index('insert'))

	   self.app.undostate = None
	   RecordController.execute(self)

    def _rollback_before_jump(self):
	   self.app.loadundostate(self.undostate)
	   self.app.vs.action_number = self.action_num	

    def reset(self):
	   self.app.cancel_collect()
	   self._rollback_before_jump()
	   self.app.text.tag_remove('fadeout', '1.0', 'end')
	   self.end()

    def done(self):
	   self.app.stop_collect()
	   self._rollback_before_jump()
	   self.end()

class RunWhileConfidentController(InteractionController):
    def __init__(self, app, interaction):
	   InteractionController.__init__(self, app, interaction, 'Running')
	   self.running = 0

    def _create_widgets(self):
	   InteractionController._create_widgets(self)
	   self.interrupt_button = Button(self, text='Begin run',
							    command=self.int_callback)
	   self._standard_pack(self.interrupt_button)
	   self.finished_button = self.interrupt_button

    def execute(self):
	   self.target = self.app.vs.get_action_number() + \
				  self.interaction.steps - 1
	   self.interrupt_button.configure(text='Begin run')
	   self.running = 0

    def int_callback(self):
	   if self.running:
		  self.running = 0
		  self.interrupt_button.configure(text='Begin run')
	   else:
		  self.running = 1
		  self.interrupt_button.configure(text='Interrupt run')
		  self.go()

    def go(self):
	   while self.running and self.app.vs.get_action_number() <= self.target:
		  if not self.app.show_step():
			 self.running = 0
			 self.end(1)
			 return
		  try:
			 self.app.finish_step()
		  except collect.EndProgram:
			 self.running = 0
			 self.end(1)
			 return
	   self.running = 0
	   if self.app.vs.get_action_number() >= self.target:
		  self.end()

    def reset(self):
	   self.running = 0
	   self.end()


class PlaybackController(InteractionController):
    def __init__(self, app, interaction):
	   InteractionController.__init__(self, app, interaction, 'Running')
	   self.terminated = 0
	   self.running = 0

    def _create_widgets(self):
	   InteractionController._create_widgets(self)
	   self.all_button = Button(self, text='Continue run',
						   command=self.button_callback)
	   self._standard_pack(self.all_button)
	   if not hasattr(self.app, 'playback_speed'):
		  self.app.playback_speed = DoubleVar()
	   self.app.playback_speed.set(2.0)
	   self.speed_control = Scale(self, label='Delay on playback (seconds)',
							from_=0.0, to=4.0, resolution=0.2,
							showvalue=1, orient='horizontal',
							variable=self.app.playback_speed,
							tickinterval=1)
	   self._standard_pack(self.speed_control)
	   self.finished_button = self.all_button
	   
    def execute(self):
	   self.target = self.app.vs.get_action_number() + \
				  self.interaction._steps_to_take(self.app)
	   self.switch_to('start')

    def switch_to(self, mode):
	   assert(mode in ['start', 'continue', 'finish', 'interrupt'])
	   self.mode = mode

	   special = ['start']
	   if mode in special:
		  if mode == 'start':
			 self.switch_to_continue()
			 self.all_button.configure(text='Start run')
		  else:
			 raise 'unrecognized special mode %s' % mode
	   else:
		  eval('self.switch_to_' + mode)()

    def switch_to_continue(self):
	   self.all_button.configure(state='normal', text='Continue run')

    def switch_to_interrupt(self):
	   self.all_button.configure(text='Interrupt run', state='normal')

    def switch_to_finish(self):
	   self.all_button.configure(text='Finish run', state='normal')

    def button_callback(self):
	   assert(self.mode in ['interrupt', 'finish', 'start', 'continue'])
	   if self.mode == 'continue' or self.mode == 'start':
		  self.run()
	   elif self.mode == 'interrupt':
		  self.interrupt()
	   elif self.mode == 'finish':
		  self.finish()
	   else:
		  raise 'unrecognized mode %s' % self.mode

    def run(self):
	   self.switch_to('interrupt')
	   count = self.target - self.app.vs.get_action_number()
	   # Delays should not matter.. overridden by slider.
	   self.app.interruptable_run(None, self.done, count)

    def interrupt(self):
	   self.app.interrupt_run()

    def finish(self):
	   self.end(self.terminated)

    def done(self, terminated = 0):
	   self.terminated = terminated
	   if self.app.vs.get_action_number() == self.target or terminated:
		  self.end(self.terminated)
	   else:
		  self.switch_to('continue')

    def reset(self):
	   self.interrupt()
	   self.finish()


class SuggestController(InteractionController):
    def __init__(self, app, interaction):
	   InteractionController.__init__(self, app, interaction, 'Suggesting')
	   
    def _create_widgets(self):
	   InteractionController._create_widgets(self)
	   
	   acceptbut = Button(self, text='Accept',
					  command=self.accept)
	   againbut = Button(self, text='Guess again',
					 command=self.guess_again)
	   self._standard_pack(acceptbut)
	   self._standard_pack(againbut)
	   self.finished_button = acceptbut

    def reset(self):
	   self.app.loadundostate()
	   self.app.current_action = None
	   self.end()

    def execute(self):
	   self.app.show_step()

    def accept(self):
	   finished = 0
	   try:
		  self.app.finish_step()
	   except collect.EndProgram:
		  finished = 1
	   self.end(finished)

    def guess_again(self):
	   self.app.guess_again()
    
class SuggestIterationController(SuggestController):
    def execute(self):
	   pass

    def accept(self):
	   pass

    def guess_again(self):
	   pass

class DoNothingController(InteractionController):
    def __init__(self, app, interaction):
	   InteractionController.__init__(self, app, interaction,
							    'Please select\nan interaction')
    def _create_widgets(self):
	   InteractionController._create_widgets(self)
	   self.spankbut.pack_forget()
	   
    def execute(self):
	   self.app.set_status('Please select an interaction to perform.')

