#!/usr/bin/env python

# Version space algebra supporting code
# This file contains abstract version space classes.  It includes the base
# classes (VersionSpace, Join, and Union) and some domain-independent
# classes such as ConstInt and LinearInt.
# Domain-dependent version space code should go in a different file.

import sys, string, operator, re, os
from types import *
from math import exp
import math
from copy import copy
import entropy

WINDOWSIZE = 15

# Whether or not to print traces
TRACE = 1

# Whether or not to adjust inapplicable hypothesis prediction probabilities
# according to the current iteration number.
# 0 means fudge probabilities, 1 means don't fudge
NO_INAPP_ADJUSTMENT = 0
# Whether to allow inapplicable hypotheses, or discard them.  0 means
# include, 1 means discard.
NO_INAPP_ALLOWED = 0

# Longest that a segmentless program is allowed to be. If set to -1, a
# program can be as long as we have space for.
# This is the default value
MAX_PROGRAM_LENGTH_DEFAULT = 10
# This is the actual value
MAX_PROGRAM_LENGTH = -1

## Some constants to define the prior probability of a
## given hypothesis. They are used according to the formula:
##
## P = e^-(SCALE*((len(hypothesis) + BASE)^2))
GAUSS_SCALE = 0.5
GAUSS_BASE = -5.0

# Whether to normalize probabilities at each stage in the version space
ALWAYS_NORMALIZE = 0

# Whether to force selections to always leave the cursor in the same
# place it started. PURELY FOR SMEMACS!
FORCE_SAME = 0

def smemacs_mode():
  global FORCE_SAME
  FORCE_SAME = 1

def clip(num, left = 0, right = None):
  """Clip num to the range [left, right].

  If right is None, right = infinity.

  Precondition: left <= right"""

  assert(right is None or left <= right)
  
  if right is None:
    right = num

  if left <= num <= right:
    return num
  elif num < left:
    return left
  else:
    return right

# String formatting of the return result of "getpriors"
def format_priors(priors, indent=0):
  if not priors:
    return ''

  s = ''
  for p in priors:
    # self
    s = s + ('  ' * indent) + '%s -- ' % p[0]
    if p[1] is None:
      s = s + '***'
    else:
      s = s + `int(p[1]*100)`
    s = s + '\n'
    # children
    s = s + format_priors(p[2], indent+1)
  return s

# A single hypothesis, along with an associated explanation why
class Hypothesis:
  def __init__(self, hyp, reason):
    # the hypothesis to be performed (of whatever type)
    self.hyp = hyp
    # a string representing the explanation
    self.reason = reason
  
  def __repr__(self):
    return '<' + `self.hyp` + ' (Reason: %s)' % self.reason + '>'
  
  def isinapplicable(self):
    return isinstance(self.hyp, Inapplicable)

# Utilities to create a distinguished value representing a hypothesis's
# prediction that the current action is inapplicable.
class Inapplicable:
  def __cmp__(self, other):
    if isinstance(other, Inapplicable):
      return 0
    else:
      return 1

  def __hash__(self):
    return 0

def hyp_inapplicable(hyp_to_check):
  return isinstance(hyp_to_check, Inapplicable)

def hyp_make_inapplicable():
  return Inapplicable()

class Reason:
  def __init__(self, reason):
    self.reason = reason
  def __repr__(self):
    return self.reason

# Class for collecting hypotheses and their associated probabilities
# The interface ought to provide a way to retrieve the set of all
# consistent hypotheses, the reasons supporting each such hypothesis,
# choose the most likely hypothesis, and choose the hypothesis which is
# maximally likely but less likely than another hypothesis (for cycling
# through the guesses in the UI).
class ResultSet:
  HUGE_PROB = 99999
  
  def __init__(self):
    self.hypotheses = []
    self.actionlist = None

  def normalize(self):
    # Normalize the total of the probabilities to 1.

    total_prob = 0.0
    for (hyp, prob) in self.hypotheses:
      total_prob = total_prob + prob

    # No normalization if everything is 0 probability
    # None necessary if total_prob is already 1.
    if total_prob == 0 or total_prob == 1:
      return

    new_hyps = self.hypotheses
    self.hypotheses = []
    for (hyp, prob) in new_hyps:
      self.add(hyp, prob / total_prob)
  
  def add(self, hyp, prob):
    # Ignore inapplicable hypotheses if they're not allowed.
    if NO_INAPP_ALLOWED:
      try:
        complex_inapp = hyp_inapplicable(hyp.hyp[1])
      except:
        complex_inapp = 0
      if hyp_inapplicable(hyp) or \
         (hasattr(hyp, 'hyp') and hyp_inapplicable(hyp.hyp)) or \
         complex_inapp:
        return
      
    self.hypotheses.append((hyp, prob))

  def merge_with_prob_and_cb(self, rs, weight=1.0, cb=None):
    new_hyps = []
    for (hyp, prob) in rs.hypotheses:
      hyp.hyp = (hyp.hyp[0], hyp.hyp[1], cb)
      new_hyps.append((hyp, prob*weight))
    self.hypotheses = self.hypotheses + new_hyps
  
  def __repr__(self):
    if not self.hypotheses:
      return 'No hypotheses'
    s = ''
    for (h, prob) in self.hypotheses:
      s = s + '\t' + `h` + ' [Prob %0.3f]' % prob + '\n'
    return 'Hypotheses:\n' + string.rstrip(s)
  
  def __len__(self):
    return len(self.hypotheses)
  
  # Returns a tuple of (hyp, probability)
  def __getitem__(self, index):
    return self.hypotheses[index]

  # Initialization for just grouping (no sorting according to probability)
  def simple_init(self):
    self.last_prob_chosen = self.HUGE_PROB
    self.chosen_actions = []

    # Use a hashed dictionary to group together similar hypotheses
    actions = {}
    for (hyp, prob) in self.hypotheses:
      newlist = actions.get(hyp.hyp, [])
      newlist.append((hyp, prob))
      if hyp_inapplicable(hyp.hyp[1]): continue # PERCY (don't add impossible stuff because for some reason, it will get chosen!)
      actions[hyp.hyp] = newlist
    self.num_distinct_actions = len(actions)

    #print "----- simple_init", len(actions)
    #for a, l in actions.items():
      #print "  ", a, l

    self.actionlist = []
    for hyplist in actions.values():
      # hyplist is a list of (hyp, prob) pairs
      # use it to construct a list of (totalprob, supp_hyps) pairs
      totalprob = reduce(operator.add, map(lambda x: x[1],
                      hyplist))
      self.actionlist.append((totalprob, hyplist))

  # Initialize for choosing an action (collection of hypotheses)
  def initialize(self):
    """Initialize resultset before choosing an action.  This method
    must be called before pick_next_best() or
    num_distinct_actions."""
    self.simple_init()

    # Sort within each hypothesis list to present the "best"
    # hypothesis as the gloss
    for item in self.actionlist:
      item[1].sort(lambda x, y: cmp(y[1], x[1]))
    # Sort by total probability, descending
    self.actionlist.sort(lambda x, y: cmp(y[0], x[0]))

  # Return a tuple of the entropy of a binary choice on the current
  # hypothesis and having the user show us which hypothesis is
  # correct.  This should actually be over output states (not
  # hypotheses), but let's not worry about that for the moment. It
  # might not even make a difference.
  def choice_entropies(self):
    binary_ent = entropy.calc_entropy([self.last_prob_chosen])
    allway_ent = entropy.calc_entropy(map(lambda x: x[0],
              filter(lambda x, done = self.chosen_actions:
                  x[1] not in done, self.actionlist)))
    return (binary_ent, allway_ent)

  # Return the grouped actions
  def get_grouped_actions(self):
    if self.actionlist is None:
      self.simple_init()
    
    return self.actionlist
  
  # Choose the most likely next action
  # Return a set of hypotheses ranked in probability order, such that
  # their output state is most likely
  # Do this in a braindead way now, and maybe improve it later
  # supp_hyps is a list of (hyp, prob) tuples
  # Return (list of actions, total probability)
  # Yecch!
  def pick_next_best(self):
    for (totalprob, supp_hyps) in self.actionlist:
      if totalprob <= self.last_prob_chosen and \
         supp_hyps[0][0].hyp not in self.chosen_actions:
        # Return this output state
        self.last_prob_chosen = totalprob
        self.chosen_actions.append(supp_hyps[0][0].hyp)

        # Added to show the value of this choice vs. having
        # the user tell us the right answer.
#        (bin_ent, all_ent) = self.choice_entropies()
#        print 'Entropy of binary choice: %s' \
#           '\nEntropy of all choices: %s' %\
#           (bin_ent, all_ent)
#        print 'Entropy ratio = %s' %\
#           ((all_ent+0.001)/(bin_ent+0.001),)

        return (supp_hyps, totalprob)

    # If we get here, there was nothing left; cycle around
    ret_actionlist = self.actionlist[0]
    self.last_prob_chosen = ret_actionlist[0]
    self.chosen_actions = [ret_actionlist[1][0][0].hyp]

    return (ret_actionlist[1], ret_actionlist[0])
  
  def sort(self):
    self.hypotheses.sort(lambda x, y: cmp(x[1], y[1]))

# ----------------------------------------------------------------------
# Class implementation of vs algorithm
class VSError(RuntimeError): pass

# Abstract base class illustrates the interface VS classes must support.
# Unions and joins have some of these methods implemented generically by
# the VersionSpaceUnion and VersionSpaceJoin classes below.
class VersionSpace:
  def __init__(self):
    pass

  def feed(self, instate, outstate):
    """Feed the version space with this instate/outstate pair"""
    pass
  
  # Return a ResultSet instance. If ALWAYS_NORMALIZE is true, the
  # ResultSet should be normalized before being returned.
  def execute(self, instate):
    pass
  
  def collapsed(self):
    """Returns true iff the version space has collapsed and is empty."""

    raise VSError, 'class %s must implement its own collapsed()' %\
      `self.__class__`

  def num_hyps(self):
    raise 'abstract function must be implemented!'

# Several functions are associated with particular version spaces:
#
#   inputfunc is called with the original input state and must return the
#   transformed input state suitable for feeding to the sub-version space.
#
#   outputfunc is called with the original input/output states and must
#   return the transformed output state suitable for feeding to the
#   sub-version space.
#
#  outputfunc_up is called with the sub version space's output and the
#  input state, and transforms the output into the output for the parent
#  version space.
#
#   Reasonfunc is called with the original input state, and gives a reason
#   for this particular hypothesis.
class VersionSpaceUnion(VersionSpace):
  """Provides support functions for version space unions."""
  initialized = 0

  def __getstate__(self):
    odict = copy(self.__dict__)
    if odict.has_key('vses') and odict.has_key('initialized') and \
       odict['initialized']:
      odict['initialized'] = 0
      odict['vses'] = map(lambda (x,y,z,trash): (x,y,z),
              odict['vses'])
    return odict

  def uninitialized(self):
    return not self.initialized

  def initialize(self):
    self.initialized = 1
    new_vses = []
    for (vs, prob, name) in self.vses:
      if hasattr(vs, 'inputfunc'):
        in_func = getattr(self, vs.inputfunc)
      else:
        in_func = lambda instate: instate
      if hasattr(vs, 'reasonfunc'):
        reason_func = getattr(self, vs.reasonfunc)
      else:
        reason_func = lambda hyp, instate: hyp.reason
      if hasattr(vs, 'outputfunc'):
        out_func = getattr(self, vs.outputfunc)
      else:
        out_func = lambda instate, outstate: outstate
      if hasattr(vs, 'outputfunc_up'):
        out_func_up = getattr(self, vs.outputfunc_up)
      else:
        out_func_up = lambda hyp, instate: hyp.hyp
      new_vses.append((vs, prob, name, (in_func,
                      out_func,
                      reason_func,
                      out_func_up)))
    self.vses = new_vses
  
  def execute(self, instate):
    rs = ResultSet()
    add = rs.add

    if self.uninitialized():
      self.initialize()

    cum_prob = 0.0
    cum_results = []
    for (vs, prob, name, funcs) in self.vses:
      in_t = funcs[0](instate)
      results = vs.execute(in_t)
      assert isinstance(results, ResultSet)
      if results:
        cum_prob = cum_prob + prob
      cum_results = cum_results + [(results, prob, vs, funcs)]
    if cum_prob == 0:
      return rs
    for (results, vs_prob, vs, funcs) in cum_results:
      factor = vs_prob / cum_prob
      (in_f, out_f, reason_f, out_f_up) = funcs
      for (hyp, hyp_prob) in results.hypotheses:
        reason = reason_f(hyp, instate)
        newhyp = out_f_up(hyp, instate)

        hypobj = Hypothesis(newhyp, reason)
        add(hypobj, hyp_prob * factor)

    # Should already be normalized.
    #rs.normalize()
    return rs

  def feed(self, instate, outstate):
    if self.uninitialized():
      self.initialize()

    for (vs, prob, name, (in_f, out_f, reason_f, out_f_up)) in self.vses:
      in_t = in_f(instate)
      out_t = out_f(instate, outstate)
      vs.feed(in_t, out_t)

  def collapsed(self):
    for vs in self.vses:
      if not vs[0].collapsed():
        return 0

    return 1

  def num_hyps(self):
    return reduce(operator.add, map(lambda x: x[0].num_hyps(),
                  self.vses), 0)

# Define a member function called "joinreasonfunc" to special-process reasons.
# Define a member function called "joinfunc" to implement not-independent
# joins, or turn our children's outputs into more than simply a tuple.
class VersionSpaceJoin(VersionSpace):
  """Provides support functions for version space joins."""
  def execute(self, instate):
    # Compute cross product here
    executions = []
    for (vs, name) in self.vses:
      if hasattr(vs, 'inputfunc'):
        in_t = getattr(self, vs.inputfunc)(instate)
      else:
        in_t = instate
      # vs.execute() returns a ResultSet instance
      executions.append(vs.execute(in_t))

    cum_results = map(lambda x: [x], executions[0].hypotheses)
    for results in executions[1:]:
      new_cum = []
      for r1 in cum_results:
        for r2 in results.hypotheses:
          new_cum.append(r1 + [r2])
      cum_results = new_cum

    rs = ResultSet()
    for result in cum_results:
      # result is a list of (hyp, prob) pairs
      # Check for inapplicability
      hyps = map(lambda x: x[0], result)
      probs = map(lambda x: x[1], result)
      inap = filter(lambda x: x.isinapplicable(), hyps)
      reasons = map(lambda x: x.reason, hyps)
      printable_reasons = filter(None, reasons)
      if not printable_reasons:
        printable_reasons = ['no reason']
      if inap:
        hyp = Inapplicable()
        reason = string.join(printable_reasons, ' and ')
      else:
        # if all hypotheses are applicable
        assert filter(lambda x: not hyp_inapplicable(x.hyp), hyps)
        if hasattr(self, 'joinreasonfunc'):
          reason = self.joinreasonfunc(hyps, instate)
        else:
          reason = string.join(printable_reasons, ' and ')
        if hasattr(self, 'joinfunc'):
          hyp = self.joinfunc(hyps, instate)
        else:
          hyp = tuple(map(lambda x: x.hyp, hyps))
      newhyp = Hypothesis(hyp, reason)
      rs.add(newhyp, reduce(operator.mul, probs, 1.0))

    if ALWAYS_NORMALIZE:
      rs.normalize()
    return rs

  def feed(self, instate, outstate):
    for (vs, name) in self.vses:
      if hasattr(vs, 'inputfunc'):
        in_t = getattr(self, vs.inputfunc)(instate)
      else:
        in_t = instate
      if hasattr(vs, 'outputfunc'):
        out_t = getattr(self, vs.outputfunc)(instate, outstate)
      else:
        out_t = outstate
      vs.feed(in_t, out_t)

  def collapsed(self):
    for vs in self.vses:
      if vs[0].collapsed():
        return 1

    return 0

  def num_hyps(self):
    return reduce(operator.mul, map(lambda x: x[0].num_hyps(),
                  self.vses), 1)

class VSConstInt(VersionSpace):
  def __init__(self, inval=None, outval=None):
    # hypothesis has possible values: 'everything' (or any other
    # string) representing every possible pair of integers, (x, y)
    # representing a single pair, and None, representing bottom
    # Ignore the in-state
    if type(outval) is IntType:
      self.hypothesis = outval
    else:
      self.hypothesis = '*all integers*'

  def is_everything(self):
    return type(self.hypothesis) is StringType

  def is_bottom(self):
    return self.hypothesis is None

  # Classif is 0 or 1; instance is a tuple of two integers
  def feed(self, inval, outval):
    # Assume positive example
    if self.is_everything():
      self.hypothesis = outval
    elif not self.is_bottom():
      if self.hypothesis == outval:
        return
      else:
        self.hypothesis = None
    else:
      return

  def execute(self, inval):
    resultset = ResultSet()
    if type(self.hypothesis) is StringType:
      raise VSError, 'executing complete VSConstInt space!'
    elif type(self.hypothesis) is IntType:
      hyp = Hypothesis(self.hypothesis, '%i' % self.hypothesis)
      resultset.add(hyp=hyp, prob=1)
    return resultset
  
  def num_hyps(self):
    if self.is_everything():
      raise VSError, 'infinite number of hypotheses'
    elif self.is_bottom():
      return 0
    else:
      return 1

  def collapsed(self):
    """Returns true iff the version space has collapsed and is empty."""
    return self.is_bottom()

class VSLinearInt(VersionSpace):
  def __init__(self, inval=None, outval=None):
    # hypothesis has possible values: 'everything' (or any other
    # string) representing every possible pair of integers, (x, y)
    # representing a single pair, and None, representing bottom
    if type(outval) is IntType and type(inval) is IntType:
      self.hypothesis = outval - inval
    else:
       self.hypothesis = '*all integers*'

  def is_everything(self):
    return type(self.hypothesis) is StringType

  def is_bottom(self):
    return self.hypothesis is None

  # Classif is 0 or 1; instance is a tuple of two integers
  def feed(self, inval, outval):
    if outval is None:
      # Indicates we should collapse
      self.hypothesis = None
      return

    rel = outval - inval
    if self.is_everything():
      self.hypothesis = rel
    elif not self.is_bottom():
      if self.hypothesis == rel:
        return
      else:
        self.hypothesis = None
    else:
      return

  def execute(self, inval):
    rs = ResultSet()
    if self.is_everything():
      raise VSError, 'executing complete VSLinearInt space!'
    elif not self.is_bottom():
      #hyp = Hypothesis(hyp=self.hypothesis+inval, reason=
      #        'f(%i)=%i+%i' % (inval,inval,self.hypothesis))
      hyp = Hypothesis(hyp=self.hypothesis+inval, reason=
              '%i' % self.hypothesis)
      rs.add(hyp, prob=1)
    return rs

  def num_hyps(self):
    if self.is_everything():
      raise VSError, 'infinite number of hypotheses'
    elif self.is_bottom():
      return 0
    else:
      return 1

  def collapsed(self):
    """Returns true iff the version space has collapsed and is empty."""
    return self.is_bottom()

# ----------------------------------------------------------------------
# Classes to represent bottom and top in string version spaces
class StringTop: pass
class StringBottom: pass

# Version space that represents a set of strings, partially ordered by a
# substring relation
class VSString(VersionSpace):
  # self.hypothesis = StringTop() means "everything"
  # self.hypothesis = StringBottom() means "bottom"
  def __init__(self, s=None, lower_bound=''):
    if s is None:
      self.hypothesis = StringTop()
    else:
      self.hypothesis = s
    self.lower_bound = lower_bound
  
  def is_everything(self):
    return isinstance(self.hypothesis, StringTop)
  
  def is_bottom(self):
    return isinstance(self.hypothesis, StringBottom)

  def collapsed(self):
    return self.is_bottom()
  
  def feed(self, instance):
    pass

  def weight(self, str):
    """Returns the weight of the given hypothesis according to the
    equation:

    P = e^(-(SCALE*(len(hypothesis) + BASE))^2)"""

    # DEBUG!!
#    print 'Calculating weight of hyp:', str
    try:
      wt = exp(-(GAUSS_SCALE * (len(str) + GAUSS_BASE))**2)
    except OverflowError:
      return 0
    else:
      return wt

class VSConstString(VSString):
  def feed(self, instate, outstate):
    if self.is_everything():
      if outstate is not None:
        self.hypothesis = outstate
      else:
        # Bottom
        self.hypothesis = StringBottom()
    elif outstate == self.hypothesis:
      pass
    else:
      self.hypothesis = StringBottom()
  
  def execute(self, instate):
    rs = ResultSet()
    if self.is_everything():
      raise VSError, 'executing infinite VSConstString space'
    elif self.is_bottom():
      return rs
    else:
      if self.hypothesis:
        rs.add(Hypothesis(self.hypothesis, 'the string %s' %
          `self.hypothesis`), prob=1.0)
      else:
        rs.add(Hypothesis(self.hypothesis, ''), prob=1.0)
    return rs
  
  def num_hyps(self):
    if self.is_everything():
      raise VSError, 'infinite VSConstString space'
    elif self.is_bottom():
      return 0
    else:
      return 1

# This class is obsolete
class VSUserInput(VersionSpace):
  def __init__(self, str=None):
    if str is None:
      self.input_strings = []
    else:
      self.input_strings = [str]
  def execute(self, instate):
    assert(self.input_strings)
    rs = ResultSet()
    hyp = Hypothesis(hyp='', reason='USER INPUT [%s]' %
      self.input_strings[0])
    rs.add(hyp, prob=1.0)
    return rs
  def feed(self, instate, outstate):
    self.input_strings.insert(0, outstate)
  def collapsed(self):
    return 0
  def num_hyps(self):
    return 1

# ----------------------------------------------------------------------
# Test program

def main():
#  import Tkinter
#  text = Tkinter.Text()
#  text.insert('end', 'foo\nbar\nbaz')
  vs = VSLinearInt()
  vs.feed(5, 10)
  print vs.execute(10)
#  print rc.execute((2,0), text)

if __name__ == '__main__':
  main()

