#!/usr/bin/env python

# All the different kinds of insertion functions used by VSInsert

import re, string, operator
import vs, textedit

class VSNumber(vs.VersionSpaceUnion):
	# Union of row number, iteration number -> string functions
	def __init__(self, instate=None, outstate=None):
		row = vs.VSLinearInt()
		row.inputfunc = 'extractrow'
		row.reasonfunc = 'row_reason'
		row.outputfunc_up = 'tostring'
		iteration = vs.VSLinearInt()
		iteration.inputfunc = 'extractiter'
		iteration.reasonfunc = 'iter_reason'
		iteration.outputfunc_up = 'tostring'
		self.vses = [
			(row, 0.50, 'Row'),
			(iteration, 0.50, 'Iteration')
			]
		# All probabilities must sum to one
		assert(reduce(operator.add, map(lambda x: x[1], self.vses))==1.0)
		if instate is not None and outstate is not None:
			self.feed(instate, outstate)
	
	def toint(self, instate, outstate):
		return int(outstate)
	
	def extractrow(self, instate):
		inxy = textedit.rowColTuple(instate.contents, instate.pos)
		return inxy[0]

	def extractiter(self, instate):
		return instate.iteration_number
	
	def tostring(self, hyp, instate):
		return `hyp.hyp`

	def row_reason(self, hyp, instate):
		inxy = textedit.rowColTuple(instate.contents, instate.pos)
		row = inxy[0]
		delta = hyp.hyp - row
		if delta == 0:
			return 'row number'
		elif delta > 0:
			return 'row number+%i' % delta
		else:
			return 'row number-%i' % -delta

	def iter_reason(self, hyp, instate):
		delta = hyp.hyp - instate.iteration_number
		if delta == 0:
			return 'iteration number'
		elif delta > 0:
			return 'iteration number+%i' % delta
		else:
			return 'iteration number-%i' % -delta

class VSPythonIndent(vs.VersionSpaceJoin):
	"""A join of some amount of whitespace, followed by a constant
	string."""

	# Some useful regexps
	indent_regex = re.compile('^(\s+)(.*)')

	def __init__(self, instate=None, outstate=None):
		str = vs.VSConstString()
		str.outputfunc = 'split_code'
		indentstr = VSIdentity()
		indentstr.outputfunc = 'split_indent'
		indentstr.inputfunc = 'get_indent'
		self.vses = [
			(indentstr, 'Indentation'),
			(str, 'Code string')
		]
		if instate is not None and outstate is not None :
			self.feed(instate, outstate)

	def split_code(self, instate, outstate):
		g = self.indent_regex.match(outstate)
		if g is not None:
			return g.group(2)
		else:
			return None

	def split_indent(self, instate, outstate):
		g = self.indent_regex.match(outstate)
		if g is not None:
			return g.group(1)
		else:
			return None

	# Both joinfunc and joinreasonfunc are called with the untransformed
	# state; they can transform it if necessary by calling the transform
	# functions.
	# These are special member functions that are used by execute.
	def joinfunc(self, hyps, instate):
		assert len(hyps) == 2
		return hyps[0].hyp + hyps[1].hyp

	# Both joinfunc and joinreasonfunc are called with the untransformed
	# state; they can transform it if necessary by calling the transform
	# functions.
	# These are special member functions that are used by execute.
	def joinreasonfunc(self, hyps, instate):
		"Provide gloss for joined reason."
		assert(len(hyps) == 2)
		return 'indentation and %s' % hyps[1].reason

	def get_indent(self, instate):
		# Extract the indent from the instate
		g = self.indent_regex.match(instate)
		if g is not None:
			return g.group(1)
		else:
			return None

class VSIdentity(vs.VersionSpace):
	"""The only function in this VS is the identity function."""
	def __init__(self, instate=None, outstate=None):
		# Whether the identity function is consistent or not
		self.hypothesis = 1
		if instate is not None and outstate is not None:
			self.feed(instate, outstate)

	def feed(self, instate, outstate):
		if not self.hypothesis:
			return

		if instate != outstate:
			self.hypothesis = 0

	def execute(self, instate):
		rs = vs.ResultSet()
		if self.hypothesis and instate is not None:
			hyp = vs.Hypothesis(hyp=instate, reason='identity')
			rs.add(hyp, prob=1.0)
		return rs

	def num_hyps(self):
		if self.hypothesis:
			return 1
		else:
			return 0

	def collapsed(self):
		return self.hypothesis == 0

class VSStringNumberString(vs.VersionSpaceJoin):
	# Join of a VSConstString, a VSNumber, and a VSConstString
	# Handles strings of the form "[1]" where the 1 varies by iteration
	# Instate is a program state, outstate is a string
	def __init__(self, instate=None, outstate=None):
		leftstr = vs.VSConstString()
		leftstr.outputfunc = 'split_string_left'
		number = VSNumber()
		number.outputfunc = 'split_number'
		rightstr = vs.VSConstString()
		rightstr.outputfunc = 'split_string_right'
		self.vses = [
			(leftstr, 'Left constant string'),
			(number, 'Number'),
			(rightstr, 'Right constant string')
			]
		if instate is not None and outstate is not None :
			self.feed(instate, outstate)

	def joinfunc(self, hyps, instate):
		assert len(hyps) == 3
		return hyps[0].hyp + hyps[1].hyp + hyps[2].hyp

	def split_string_left(self, instate, outstate):
		digits = map(None, string.digits)
		i = 0
		for i in range(len(outstate)):
			if outstate[i] in digits: break
		else:
			i = i + 1
		return outstate[:i]

	def split_number(self, instate, outstate):
		digits = map(None, string.digits)
		i = 0
		for i in range(len(outstate)):
			if outstate[i] in digits: break
		else:
			i = i + 1
		j = i
		for j in range(i, len(outstate)):
			if outstate[j] not in digits: break
		else:
			j = j + 1
		if outstate[i:j]:
			return int(outstate[i:j])
		else:
			return None

	def split_string_right(self, instate, outstate):
		digits = map(None, string.digits)
		i = 0
		for i in range(len(outstate)):
			if outstate[i] in digits: break
		else:
			i = i + 1
		j = i
		for j in range(i, len(outstate)):
			if outstate[j] not in digits:
				break
		else:
			j = j + 1
		return outstate[j:]

#class VSNumberString(vs.VersionSpaceJoin):
#	# Join of VSNumber and VSConstString
#	# This class is obsolete; it's superceded by VSStringNumberString
#	# Instate is a program state, outstate is a string
#	def __init__(self, instate=None, outstate=None):
#		number = VSNumber()
#		number.outputfunc = 'split_number'
#		conststr = vs.SConstString()
#		conststr.outputfunc = 'split_string'
#		self.vses = [
#			(number, 'Number'),
#			(conststr, 'Constant string')
#			]
#		if instate is not None and outstate is not None :
#			self.feed(instate, outstate)
#
#	def joinfunc(self, hyps, instate):
#		assert len(hyps) == 2
#		return hyps[0].hyp + hyps[1].hyp
#
#	def split_string(self, instate, outstate):
#		digits = map(None, string.digits)
#		i = 0
#		for i in range(len(outstate)):
#			if outstate[i] not in digits: break
#		else:
#			i = i + 1
#		number = outstate[:i]
#		s = outstate[i:]
#		return s
#
#	def split_number(self, instate, outstate):
#		digits = map(None, string.digits)
#		i = 0
#		for i in range(len(outstate)):
#			if outstate[i] not in digits: break
#		else:
#			i = i + 1
#		number = outstate[:i]
#		try:
#			ret = int(number)
#		except ValueError:
#			ret = None
#		return ret
#
