#!/usr/bin/env python

"""Program version spaces used in SMARTedit"""

import vs, vsaction
import textedit

DEBUG = 0

INNER_LOOP_PROBABILITY = 0.80
OUTER_LOOP_PROBABILITY = 0.20

class VSProgram(vs.VersionSpace):
  """Program that learns from segmented sequences of actions"""

  def __init__(self, actions=None):
    """Initialize VSProgram.  actions may be a list of textedit
    instances, if desired."""
    # The number of iterations before inapp reaches 50% of its original
    # probability
    self.SCALE = 10
    self.act_vses = []
    self.growth_ended = 0
    # which action we're currently on (execution or feeding)
    self.action_number = 0
    # the currently active inner loop, or None
    self.inner_loop = None
    if actions:
      self.initialize(actions)

    # Define the boundaries of the text.
    self.region = (None, None)

  # Reset the macro to re-run from the start.
  def reset(self):
    self.action_number = 0
    self.region = (None, None)
    self.inner_loop = None
  
  # Useful for testing truth value of VSProgram instance
#  def __len__(self):
#    """Return the number of actions in this Program"""
#    return self.action_count()

  def end_growth(self):
    """Tell this Program that it's done growing"""
    self.growth_ended = 1

  def _add_action(self, act):
    """Add this action as a sub-vs of Program"""
    assert(not self.growth_ended)
    if DEBUG: print 'adding action', act
    if isinstance(act, VSProgram):
      a = act
      if DEBUG: print 'Adding program as action'
    else:
      a = vsaction.VSActionUnion(act)
      if DEBUG: print 'Adding action type', act.type
    self.act_vses.append(a)

  def initialize(self, actions):
    """Initialize our sub-vses with these actions"""
    self.act_vses = []
    numacts = len(actions)
    for i in range(numacts):
      self._add_action(actions[i])
    self.feed_end()
    self.action_number = len(actions)

  def feed_end(self):
    """Indicate that the program has reached its final length"""
    self.end_growth()

  def feed_terminate(self, instate):
    """Indicate that no further examples exist past this point."""
    
    # BUG!!!: currently assumes that we're moving forward and
    # this is the end of the region.  Also, does *not* update
    # the boundary acording to changes in the text.
    if self.region[1] is None:
      self.region = (self.region[0], instate.pos)
    else:
      self.region = min((self.region[0], instate.pos),
               self.region)

  def get_action_number(self):
    """Return the count of all actions ever performed in this
    program.  If the Program is still growing, return None."""
    if not self.growth_ended:
      return None
    else:
      return self.action_number
  
  def get_action_index(self):
    """Return the index of the current action, as a number between 0
    and the number of actions in the program - 1.  If the Program is
    still growing, return None."""
    if not self.growth_ended:
      return None
    else:
      return self.action_number % len(self.act_vses)
  
  def action_type(self, index = None):
    assert(self.growth_ended)

    # Wrap to an existing action
    if index is None:
      index = self.action_number
    index = index % self.action_count()
    return self.act_vses[index].type

  def iteration_number(self, index=None):
    """Return the number of the iteration corresponding to this
    index.  If index is None, use our instance's counter.  If the
    Program is still growing, return 1."""
    if index is None:
      index = self.action_number

    if not self.growth_ended:
      return 1
    elif self.action_count() == 0:
      return 1
    else:
      return (index / self.action_count()) + 1

  # action is a textediting action from the textedit module
  # index is the number of the action in the sequence.
  # tlau: I removed the classif argument since it must always be 1
  def feed_nth(self, action):
    """Update Program such that action is an instance of the next
    program action"""
    if DEBUG: print 'Beginning feed_nth with %s action' % action.type

    if self.inner_loop:
      # add it to the inner loop's VSProgram
      if DEBUG: print 'feeding to innerloop VS'
      if DEBUG: print 'outer loop action num increased to',\
          self.action_number
      return self.inner_loop.feed_nth(action)
    if DEBUG: print self, 'action number:', self.action_number
    if DEBUG: print self, 'vses:', self.act_vses
    if self.action_number < len(self.act_vses):
      print 'ERROR: action number is less than len(act_vses)'
      print '   ', self, 'action number:', self.action_number
      print '   ', self, 'vses:', self.act_vses

    iter = self.iteration_number()
    action.instate.iteration_number = iter
    action.outstate.iteration_number = iter    

    if self.growth_ended:
      # Update existing action
      index = self.action_number % self.action_count()
      act = self.act_vses[index]
      if isinstance(act, VSProgram):
        return act.feed_nth(action)
      else:
        act.feed(action)
    elif  self.action_number == len(self.act_vses):
      # Adding a new action to the end
      self._add_action(action)
    else:
      #DEBUG!!!
      print "Action number is:", self.action_number
      print 'Act vses array is:', self.act_vses
      raise vs.VSError, 'action index is greater than number of actions'
    self.action_number = self.action_number + 1

  def _adjust_inapp_prob(self, prob, state):
    """Return a new value for this probability accounting for the
    fact that it predicts 'inapplicable'"""

    iter = state.iteration_number
    if iter <= 1:
      iter = 1.5
    factor = (iter - 1.0) / (self.SCALE - 1.0) + 1
    return prob * (1.0 - 1.0 / factor)

  def _test_region(self, pos):
    if self.region[0]:
      if pos < self.region[0]:
        return 0
    if self.region[1]:
      if pos > self.region[1]:
        return 0
    return 1
  
  def execute_nth(self, instate, index=None):
    """Return a ResultSet indicating the result of executing the
    next action in this Program on this instate"""

    assert(self.growth_ended)
    iter = self.iteration_number(index)
    instate.iteration_number = iter
    # Wrap to an existing action
    if index is None:
      index = self.action_number
    index = index % self.action_count()
    next_index = (index + 1) % self.action_count()
    act = self.act_vses[index]
    nextact = self.act_vses[next_index]

    # first check for whether we're executing an inner loop
    inner_loop_rs = None
    outer_loop_rs = None
    if self.inner_loop:
      if DEBUG: print 'outer loop: passing thru inner loop exec'
      # Note: we currently have no way of specifying which action
      # in the inner loop to execute

      # TODO: debug
      inner_loop_rs = self.inner_loop.execute_nth(instate)
#      if self.inner_loop.get_action_index() == 0:
        # if the next innerloop action is the first action in
        # the loop, we may choose to exit the inner loop and
        # execute the next action in the outer loop instead.
#        if isinstance(nextact, VSProgram):
          # next outerloop action is another innerloop
          # TODO!!! fix!!!
#          print 'ERROR! adjacent innerloops!'
#        else:
          # The reason string for outer_loop_rs is added
          # both here and in the last else clause below.
          # This is because the index of the step in the
          # action sequence depends on whether it's the next
          # action in the outer loop, or the next next
          # action.
      outer_loop_rs = nextact.execute(instate)
      reason_prefix='Step %i of %i is %%s' %\
        (next_index+1, self.action_count())
      for (hyp, prob) in outer_loop_rs:
        hyp.reason = reason_prefix % hyp.reason
      #return self.inner_loop.execute_nth(instate)
    elif isinstance(act, VSProgram):
      # if we get here, we're beginning an inner loop
      assert(self.inner_loop is None)
      self.inner_loop = act
      # Note: we currently have no way of specifying which action
      # in the inner loop to execute
      inner_loop_rs = self.inner_loop.execute_nth(instate)
    else:
      outer_loop_rs = act.execute(instate)
      reason_prefix='Step %i of %i is %%s' %\
        (index+1, self.action_count())
      for (hyp, prob) in outer_loop_rs:
        hyp.reason = reason_prefix % hyp.reason

    # Add gloss indicating which step we're on for inner_loop
    # hypotheses
    if inner_loop_rs is not None:
      reason_prefix = 'Step %i of %i is %%s' % \
           (index + 1, self.action_count())
      for (hyp, prob) in inner_loop_rs:
        hyp.reason = reason_prefix % hyp.reason

    # Merge in inner and outer loop hypothesis, and add in callbacks.
    # At this point, the hypotheses member variable of a ResultSet
    # contains a tuple of (Hypothesis, callback), where Hypothesis is
    # a 
    rs = vs.ResultSet()
    if inner_loop_rs is not None:
      rs.merge_with_prob_and_cb(inner_loop_rs,
        INNER_LOOP_PROBABILITY, None)
    if outer_loop_rs is not None:
      if inner_loop_rs is not None:
        rs.merge_with_prob_and_cb(outer_loop_rs,
          OUTER_LOOP_PROBABILITY,
          self.terminate_innerloop_execution)
      else:
        rs.merge_with_prob_and_cb(outer_loop_rs,
          OUTER_LOOP_PROBABILITY, None)

    # Adjust for the region
    if self.region[0] is not None or self.region[1] is not None:
      new_rs = vs.ResultSet()
      add = new_rs.add
      inapp = vs.hyp_inapplicable
      hypfunc = vs.Hypothesis
      for (hyp, prob) in rs:
        true_hyp = hyp.hyp
        if inapp(true_hyp[1]):
          add(hyp, prob)
          continue

        (command, arg, cb) = true_hyp
        if command in ('move', ):
          pos = textedit.positionFromTuple(instate.contents,
                         arg)
          safe = self._test_region(pos)
        elif command in ('delete', 'select'):
          pos0 = textedit.positionFromTuple(instate.contents,
                          arg[0])
          pos1 = textedit.positionFromTuple(instate.contents,
                          arg[1])
          safe = self._test_region(pos0) and \
              self._test_region(pos1)
        else:
          safe = 1

        if not safe:
          true_hyp = (true_hyp[0],
            vs.hyp_make_inapplicable(), cb) 
        hyp = hypfunc(hyp=true_hyp,
                reason=hyp.reason)
        
        add(hyp, prob)

      rs = new_rs

    # Adjust for inapplicability
    if vs.NO_INAPP_ADJUSTMENT:
      return rs

    new_rs = vs.ResultSet()
    add = new_rs.add
    adjust = self._adjust_inapp_prob
    inapp = vs.hyp_inapplicable
    for (hyp, prob) in rs:
      if inapp(hyp.hyp[1]):
        prob = adjust(prob, instate)
      add(hyp, prob)
    new_rs.normalize()
    return new_rs

  # How many actions?
  def action_count(self):
    """Return the number of actions in this Program"""
    return len(self.act_vses)
  
  def collapsed(self):
    """Return 1 iff any of the Action vses has collapsed"""
    for vs in self.act_vses:
      if vs.collapsed():
        return 1
    return 0

  # How many hypotheses are in the space?
  def num_hyps(self):
    """Return the number of hypotheses in the Program VS"""
    return reduce(operator.add, map(lambda x: x.num_hyps(),
      self.act_vses))

  def num_hyps_nth(self, index):
    """Return the number of hypotheses in the indexth action"""
    # Wrap to an existing action
    assert(len(self.act_vses) > 0)
    index = index % len(self.act_vses)

    return self.act_vses[index].num_hyps()

  def start_innerloop(self):
    """Begin recording the inner loop"""
    if DEBUG: print 'starting inner loop'
    if self.inner_loop:
      raise vs.VSError, 'doubly nested loops not supported'
    self.inner_loop = VSSegmentlessProgram()
    if DEBUG: print 'inner loop is', self.inner_loop
    # create a new program instance, and add it as an action
    self._add_action(self.inner_loop)
    self.action_number = self.action_number + 1
  def end_innerloop(self):
    """End recording of the inner loop"""
    if DEBUG: print 'ending inner loop'
    self.inner_loop.end_growth()
    self.inner_loop = None
  def iterate_innerloop(self):
    """Iterate the inner loop"""
    if DEBUG: print 'iterating inner loop'
    assert(self.inner_loop)
    self.inner_loop.feed_end()
  def terminate_innerloop_execution(self):
    """Terminate the inner loop's execution"""
    if DEBUG: print '!!! Terminating inner loop execution'
    self.inner_loop = None
    if DEBUG: print 'action number was %i, increased to %i' %\
      (self.action_number, self.action_number + 1)
    self.action_number = self.action_number + 1

class VSSegmentlessProgram(vs.VersionSpace,VSProgram):
  # A program which does not require explicit segmentation (but can
  # use some segmentation hints if provided).
  
  def __init__(self):
    self.program_vses = []
    self.actions = []
    # which action we're currently on (execution or feeding)
    self.action_number = 0

    # The programs under this node can keep getting longer
    # until this becomes true.
    self.growth_ended = 0

  def action_count(self):
    """Temporary patch to return some number of actions.
    
    WORKAROUND!!!"""

    if not self.growth_ended:
      raise vs.VSError, "can't even guess the length of growing program."
    else:
      try:
        # Find the first occurrence of an uncollapsed VS. 
        collapsed_vses = map(lambda p: p.collapsed(),
                 self.program_vses)
        index = collapsed_vses.index(0)
      except ValueError:
        return 0

      return index + 1
  
  def get_action_index(self):
    """Return the index of the current action, as a number between 0
    and the number of actions in the program - 1.  If the Program is
    still growing, return None."""
    raise VSError, 'no action index available for SegmentlessProgram'

  def end_growth(self):
    self.growth_ended = 1

  def feed_end(self):
    # The given index is >= the maximum iteration length.
#    self.program_vses = self.program_vses[0:self.action_number]
    self.end_growth()
    map(lambda p: p.feed_end(), self.program_vses)

  # PERCY: THIS IS NOT USED: see other feed_nth(self, action)
  def feed_nth(self, action):
    """Add this action to the action sequence.

    Feed it to all sub-programs."""

    #DEBUG!!
    #print 'Feeding action #', index, '::', action

    if DEBUG: print self, 'action number:', self.action_number

    new_program = None
    assert(self.action_number >= len(self.actions))

    if not self.growth_ended:
      self.actions.append(action)
      new_program = VSProgram(self.actions)
    else:
      new_program = None

#    if index < len(self.actions):
#      # This is an overlap of a previous action, guaranteed.
#      self.end_growth()
#      print 'Overlapping previously existing action, number:', index
#    elif self.growth_ended:
#      # Have to wrap this back around into existing programs
#      pass
#    elif index > len(self.actions):
#      # Tried to grow past the end
#      raise vs.VSError, 'tried to grow segmentless program improperly.'
#    elif vs.MAX_PROGRAM_LENGTH != -1 and \
#      index >= vs.MAX_PROGRAM_LENGTH:
#      # This would be a potentially brand new action (requiring
#      # a longer program), but we've reached the program length
#      # limit. So, start wrapping.
#      self.end_growth()
#    else:
#      # This is a brand new action (potentially) and a new
#      # program should be created.

#      self.actions.append(action)
#      new_program = VSProgram(self.actions)
    
    for i in range(len(self.program_vses)):
      if not self.program_vses[i].collapsed():
        self.program_vses[i].feed_nth(action)

    if new_program:
      self.program_vses.append(new_program)

    self.action_number = self.action_number + 1

  def _calculate_weights(self, result_sets):
    """Returns a list of normalized weights for each program subspace.

    Currently each successive space gets half the weight of the previous
    space."""

    init = 1.0
    total = 0.0
    weights = []
    for i in range(len(self.program_vses)):
      init = init / 2.0

      if not result_sets[i]:
        weights.append(0.0)
      else:
        weights.append(init)
        total = total + init

    if total:
      weights = map(lambda x, den = total: x / den, weights)

    return weights

  def execute_nth(self, instate, index=None):
    sub_rsets = []
    for i in range(len(self.program_vses)):
      program = self.program_vses[i]
      if self.program_vses[i].collapsed():
        sub_rsets.append([])
      else:
        sub_rs = program.execute_nth(instate, index)
        sub_rsets.append(sub_rs)

    weights = self._calculate_weights(sub_rsets)

    rs = vs.ResultSet()
    for i in range(len(sub_rsets)):
      sub_rs = sub_rsets[i]
      for (hyp, prob) in sub_rs:
        rs.add(hyp=hyp, prob=prob * weights[i])

    rs.normalize()
    return rs

  def collapsed(self):
    # Lazily created spaces don't collapse until their growth is ended.
    if not self.growth_ended:
      return 0
    
    for vs in self.program_vses:
      if not vs.collapsed():
        return 0

    return 1

#
# The Startless Program is somewhat dubious and should only be used with
# caution.
#
class VSStartlessProgram(vs.VersionSpace):
  # A program which detects when repetition begins.

  def __init__(self):
    self.program_vses = []
    self.next_action = 0

    self.growth_ended = 0

  def end_growth(self):
    self.growth_ended = 1

  def action_count(self):
    """Temporary patch to return some number of actions.
    
    WORKAROUND!!!"""

    for program in self.program_vses:
      count = program.action_count()
      if count > 0:
        return count
    
    return 0

  def feed_begin(self, index):
    # There's a program that starts at this point or before.
    self.end_growth()
    self.program_vses = self.program_vses[0:index+1]

  def feed_end(self, index):
    # Index is > the last action number at which any program might
    # start.
    # Not sure what this means in this context, but certainly, we
    # can stop spawning programs.

    if not self.growth_ended:
      self.end_growth()
      self.program_vses = self.program_vses[0:index]
      for i in range(len(self.program_vses)):
        self.program_vses[i].feed_end(index - i)

    #DEBUG!!
    #print 'Ending growth at space:', index

  def feed_nth(self, action, index):
    """Add this action to the action sequence by feeding it to all
    sub-programs"""

    #DEBUG!!
    #print '-----------------------------------------------------'
    #print 'Next expected action is:', self.next_action
    #print 'Adding action #', index
    #print 'Type is:', action.type

    if not self.growth_ended:
      if index < self.next_action:
        # This is an overlap of a previous action.
        raise vs.VSError, 'No overlapping action numbers allowed'
      elif index > self.next_action:
        # Tried to grow past the end
        raise vs.VSError, 'tried to grow program improperly.'
      else:
        # This is a brand new action (potentially) and a new
        # program should be created

        #DEBUG!!
        #print 'Creating space:', len(self.program_vses)
      
        self.program_vses.append(VSSegmentlessProgram())
      
    self.next_action = self.next_action + 1

    for i in range(len(self.program_vses)):
      if not self.program_vses[i].collapsed():
        #DEBUG!!
        #print 'Feeding action', index, 'to space', i, 'as action', index - i
        true_index = index - i
        self.program_vses[i].feed_nth(action, true_index)

  def _calculate_weights(self, index, result_sets):
    "Returns a list of normalized weights for each program space."

    weights = []
    total = 0.0
    for i in range(len(self.program_vses)):
      if not result_sets[i]:
        weights.append(0.0)
      else:
        # Weights are (arbitrarily) proportional to the
        # square of the number of matching actions.
        # The 1.0 makes this a float computation and
        # accounts for cases where this is action n for
        # a program with n steps (0 .. n - 1)
        true_index = index - i + 1.0
        matches = true_index -\
          self.program_vses[i].get_action_number()
        weight = matches * matches
        weights.append(weight)
        total = total + weight

    if total:
      weights = map(lambda x, den = total: x / den, weights)
    return weights

  def execute_nth(self, instate, index):
    #DEBUG!!
    #print 'Executing action', index

    sub_rsets = []
    for i in range(len(self.program_vses)):
      program = self.program_vses[i]
      true_index = index - i
      if self.program_vses[i].collapsed():
        sub_rsets.append([])
      else:
        sub_rs = program.execute_nth(instate, true_index)
        sub_rsets.append(sub_rs)

    weights = self._calculate_weights(index, sub_rsets)
    rs = vs.ResultSet()
    for i in range(len(sub_rsets)):
      sub_rs = sub_rsets[i]
      for (hyp, prob) in sub_rs:
        rs.add(hyp=hyp, prob=prob * weights[i])

    rs.normalize()
    return rs

  def collapsed(self):
    # Lazily created spaces don't collapse until their growth is ended.
    if not self.growth_ended:
      return 0
    
    for vs in self.program_vses:
      if not vs.collapsed():
        return 0

    return 1
  
def main():
  pass

if __name__ == '__main__':
  main()

