package programinduction

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,fails}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}

/*
A type is either a
 - Base type: defined by user (e.g., boolean, int, world)
 - Function type: source -> target
Note the type system is monomorphic (simply typed).

A combinator is a function with some arity and type.  A combinator is either a
- Primitive combinator: defined by user; there are some standard ones in the standard library
  Given the appropriate number of arguments, it knows how to evaluate itself.
- Composite combinator: consists of a list of router and two combinators (function and argument).
  Think of a composite combinator as a binary tree.
  Each router specifies how to pass an argument down the tree (left, right, or both).
All combinators have an associated type, which is determined bottom up from the leaves.

Nodes encapsulate combinators.
*/

////////////////////////////////////////////////////////////
// Monomorphic type system

trait Type {
  def -:(that:Type) = FuncType(that, this) // Note that -: is right associative and thus binds to the right
  def reduce(n:Int) : Type // If apply n arguments, what's the type?
  def finalOutType : Type
  def arity : Int // Number of arguments that a type takes
}
trait BaseType extends Type { // Override this
  def reduce(n:Int) = if (n == 0) this else throw impossible
  def finalOutType : Type = this
  def arity = 0
}
case class FuncType(inType:Type, outType:Type) extends Type {
  def reduce(n:Int) = if (n == 0) this else outType.reduce(n-1)
  def finalOutType : Type = outType.finalOutType
  def arity : Int = 1 + outType.arity
  override def toString = {
    (if (inType.isInstanceOf[FuncType]) "("+inType.toString+")" else inType.toString) + " -> " + outType.toString
  }
}
object Type {
  //case class InvalidException extends Exception
  def formType(routers:List[Router], funcType:Type, argType:Type) : Type = {
    routers match {
      case Nil =>
        val tf = funcType.asInstanceOf[FuncType]
        if (tf.inType == argType) tf.outType // Apply on argument
        else throw fails("Function takes %s, argument is %s", tf.inType, argType)
      case r :: rs =>
        if (r == Router.B) {
          val ta = argType.asInstanceOf[FuncType]
          ta.inType -: formType(rs, funcType, ta.outType)
        }
        else if (r == Router.C) {
          val tf = funcType.asInstanceOf[FuncType]
          tf.inType -: formType(rs, tf.outType, argType)
        }
        else if (r == Router.S) {
          val tf = funcType.asInstanceOf[FuncType]
          val ta = argType.asInstanceOf[FuncType]
          if (tf.inType == ta.inType)
            tf.inType -: formType(rs, tf.outType, ta.outType)
          else
            throw fails("For S, function and argument need to take same type, but got %s and %s", tf.inType, ta.inType)
        }
        else
          throw impossible
    }
  }
}

////////////////////////////////////////////////////////////
// Combinators and nodes

trait Callable {
  val myType : Type
  def call(args:List[Combinator], forceEval:Boolean) : Combinator
  def valueString : String
  //override def toString = valueString+" : "+myType // Too verbose
  override def toString = valueString
}

// A combinator will never be called with the wrong number of arguments.
// Nodes might though, creating partial combinators.
trait Combinator extends Callable {
  protected def badArgs(revArgs:List[Combinator]) = fails("TYPE ERROR: " + this + " CALLED WITH ARGS " + revArgs.mkString(" | "))
}

// (This could change for CCG in natural language)
case class Router(goesFunc:Boolean, goesArg:Boolean) {
  override def toString = {
    if (goesFunc && goesArg) "S"
    else if (goesFunc) "C"
    else if (goesArg) "B"
    else throw impossible
  }
}
object Router {
  val B = Router(false, true)
  val C = Router(true, false)
  val S = Router(true, true)

  def parse(c:Char) = Router(c == 'S' || c == 'C', c == 'S' || c == 'B')

  // Example: s = "BBCS"
  def parse(s:String) : List[Router] = {
    if (s.size == 0) Nil
    else parse(s.charAt(0)) :: parse(s.substring(1))
  }
}

// Abstract versions.  Remember a node wraps a combinator.
// If not enough arguments are specified, its the node's job to create a
// partial combinator that remembers the partial arguments and is willing to accept more.
trait AbstractNode extends Callable {
  def lazyEval : Boolean // When we call this node, do we immediately evaluate?
  val myType = computeType // Only need to do it once (assumes never changes, even if node is mutable)
  def computeType : Type
  def computeDepth : Int

  //def isSame(node:AbstractNode)
}

// A combinator that has a subset of arguments applied to it.
// Corresponds to a node with a list of arguments.
// Partial combinators are only created at evaluation time.
// Nodes implicit correspond to partial combinators, but the node does all the heavy lifting with routing.
case class PartialCombinator(func:Callable, extraArgs:List[Combinator]) extends Combinator {
  val myType = func.myType.reduce(extraArgs.size)
  def call(args:List[Combinator], forceEval:Boolean) = func.call(extraArgs ++ args, forceEval)
  //def valueString = "PARTIAL"
  def valueString = "PARTIAL("+func.valueString+" "+extraArgs.map(_.valueString).mkString(" ")+")"
}

trait AbstractPrimitiveNode extends AbstractNode {
  def value : Combinator
  def computeType = value.myType

  def call(extraArgs:List[Combinator], forceEval:Boolean) = {
    if ((!forceEval && lazyEval) || extraArgs.size < myType.arity) // Not enough arguments
      new PartialCombinator(this, extraArgs)
    else
      value.call(extraArgs, true)
  }

  def valueString = value.valueString

  def computeDepth = 0
}

trait AbstractCompositeNode[NodeType <: AbstractNode] extends AbstractNode {
  def routers : List[Router]
  def func : NodeType
  def arg : NodeType

  def computeType = Type.formType(routers, func.myType, arg.myType)

  def call(extraArgs:List[Combinator], forceEval:Boolean) : Combinator = {
    if ((!forceEval && lazyEval) || extraArgs.size < myType.arity) // Not enough arguments
      new PartialCombinator(this, extraArgs)
    else {
      def recurse(routers:List[Router], extraArgs:List[Combinator], funcRevExtraArgs:List[Combinator], argRevExtraArgs:List[Combinator]) : Combinator = {
        routers match {
          case Nil =>
            // Send three things in this order:
            //  (1) arguments that were routed to func
            //  (2) arg applied to things that were routed to it
            //  (3) leftover arguments
            // Note: don't need to forceEval on arguments
            func.call(funcRevExtraArgs.reverse ++ List(arg.call(argRevExtraArgs.reverse, false)) ++ extraArgs, true)
          case r :: rs => 
            val a :: as = extraArgs // Shouldn't fail because we check arity up there
            recurse(rs, as,
              if (r.goesFunc) a :: funcRevExtraArgs else funcRevExtraArgs,
              if (r.goesArg) a :: argRevExtraArgs else argRevExtraArgs)
        }
      }
      recurse(routers, extraArgs, Nil, Nil)
    }
  }

  def valueString = {
    "(" +
      (if (routers == Nil) "" else routers.mkString("") + " ") +
      func.valueString + " " + arg.valueString +
    ")"
  }

  def computeDepth = 1 + (func.computeDepth max arg.computeDepth)
}

// Basic versions just for testing
trait BNode extends AbstractNode {
  var lazyEval = false
  def setLazyEval = { lazyEval = true; this }
}
case class BPrimitiveNode(value:Combinator) extends BNode with AbstractPrimitiveNode
case class BCompositeNode(routers:List[Router], func:BNode, arg:BNode) extends BNode with AbstractCompositeNode[BNode]
object BNode {
  def __(s:String, func:BNode, arg:BNode) = BCompositeNode(Router.parse(s), func, arg)
  def __(s:String, func:BNode, arg:Combinator) = BCompositeNode(Router.parse(s), func, BPrimitiveNode(arg))
  def __(s:String, func:Combinator, arg:BNode) = BCompositeNode(Router.parse(s), BPrimitiveNode(func), arg)
  def __(s:String, func:Combinator, arg:Combinator) = BCompositeNode(Router.parse(s), BPrimitiveNode(func), BPrimitiveNode(arg))

  def __(func:BNode, arg:BNode) = BCompositeNode(Nil, func, arg)
  def __(func:BNode, arg:Combinator) = BCompositeNode(Nil, func, BPrimitiveNode(arg))
  def __(func:Combinator, arg:BNode) = BCompositeNode(Nil, BPrimitiveNode(func), arg)
  def __(func:Combinator, arg:Combinator) = BCompositeNode(Nil, BPrimitiveNode(func), BPrimitiveNode(arg))
}
