package programinduction

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import fig.record.Record
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,fails}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}
import scala.reflect.Manifest

import StandardLibrary._
import Router.{B,C,S}
import programinduction.{MPrimitiveNode => __}
import programinduction.{MCompositeNode => ___}
import MyUtils._

/**
The object of inference is a combinator tree.  The DP prior specifies
preferences between various programs and the likelihood is given by a set of
test cases.

Now let's focus on inference.  A correct sampler is to sample from the prior
and reject if program doesn't pass test cases (likelihood 0).  This is clearly
inefficient.

Let's construct a sampler that visits only programs that pass the test cases.
From one program, we can easily explore many extensionally equivalent programs
(those having the same denotation---function f) by refactoring.  The challenge
is jumping between programs that are extensionally equivalent on the test cases
(have the same empirical denotation).

Define the "empirical denotation" of every subtree to be the coarsest partial
function f such that any combinator whose denotation is a completion of f can
be grafted in (without affecting the behavior of the overall program).

The user specifies: a set of *focused* empirical denotations and for each
e-denotation, a finite set of candidate combinator trees.  Now we can sample
over the space of combinators where for each combinator c with a focused
e-denotation, some refactoring of c appears in the candidate set for that
e-denotation.

A valid sampler is gotten by refactoring moves and when a combinator belongs to
the candidate set of a e-denotation, we randomly sample another candidate.

However, this sampler might not mix well, because in order to make progress,
one has to go back to the exact candidate combinators, which might have low
probability.  The heuristic solution is to mutate the candidates as we
refactor.

Syntactically, _candidateNodes_ specifies a set of nodes (which are allowed to change).
Semantically, _examples_ specifies the empirical denotation (partial function).
*/

case class Example(input:EDenotation, output:EDenotation) {
  override def toString = input+"=>"+output
}

trait EDenotation {
  def call(args:List[Combinator]) : EDenotation
}
case class PrimitiveEDenotation(value:Combinator) extends EDenotation {
  require (value.myType.isInstanceOf[BaseType])
  override def toString = value.valueString
  def call(args:List[Combinator]) = {
    if (args.size == 0) this
    else throw fails("Tried to call a primitive %s with arguments %s", this, args)
  }
}
case class CompositeEDenotation(examples:List[Example]) extends EDenotation {
  // _examples_ should be HashSet, but that could be expensive (because most of the time there's one example),
  // so just make sure examples are in order.
  override def toString = "["+examples.mkString(";")+"]"
  def call(args:List[Combinator]) = {
    args match {
      case Nil => this
      case arg :: rest_args =>
        examples.find(_.input match { case PrimitiveEDenotation(value) => value == arg; case _ => false }) match {
          case None => throw fails("Don't know how to apply function %s to %s (not in table)", this, args)
          case Some(ex) => ex.output.call(rest_args)
        }
    }
  }
}

// Note that this structure only exists for the split e-denotations.
class EDenotationInfo {
  // These nodes are more specific than KNodes because they contain denotation information,
  // but less specific than MNodes because they can still be shared (no mutable pointer nonsense)
  val candidateNodes = new ArrayBuffer[DNode] // The candidate set (elements can change in the heuristic version)
  var probs : Array[Double] = null // Posterior probability (for initialization)

  // Assume number of candidates is not large so we step through the list.  Hash?
  def replace(node:DNode, newNode:DNode) = {
    foreach(candidateNodes.size, { i:Int =>
      if (node == candidateNodes(i))
        candidateNodes(i) = newNode
    })
  }
}

object EDenotation {
  val invalid : EDenotation = {
    object Invalid { override def toString = "INVALID" }
    PrimitiveEDenotation(WrappedValue[AnyRef](Invalid))
  }
}

// Specifies a combinator with a particular empirical denotation.
trait DNode extends AbstractNode {
  def lazyEval = false
  def edenotation : EDenotation

  def toKNode : KNode

  def replaceInfo(params:Params, newNode:DNode) : Unit
}

case class DPrimitiveNode(value:Combinator, edenotation:EDenotation) extends DNode with AbstractPrimitiveNode {
  def replaceInfo(params:Params, newNode:DNode) = {
    if (edenotation == newNode.edenotation) {
      params.edenotations.get(edenotation) match {
        case Some(info) => info.replace(this, newNode) // This is a focused e-denotation
        case None => // Do nothing
      }
    }
  }

  def toKNode = KPrimitiveNode(value)

  override def valueString = super.valueString+"_"+edenotation
}

case class DCompositeNode(routers:List[Router], func:DNode, arg:DNode) extends DNode with AbstractCompositeNode[DNode] {
  def toKNode = KCompositeNode(routers, func.toKNode, arg.toKNode)

  // Compute from func and arg
  val edenotation = {
    //dbgs("EDenotation: (%s %s %s)", routers, this.func, this.arg)
    // Essentially do evaluation (abstract interpretation).
    // Don't handle any subtyping of arguments (e.g., finer function can be used in place where a coarser function is expected):
    // e-denotations of arguments must match *exactly*.
    // If we don't know the answer to an evaluation, just leave it out.
    def recurse(routers:List[Router], funcEDenotation:EDenotation, argEDenotation:EDenotation) : Option[EDenotation] = {
      routers match {
        case Nil => // Done routing, just regular function application
          val func = funcEDenotation.asInstanceOf[CompositeEDenotation]
          func.examples.find(_.input == argEDenotation).flatMap { ex => Some(ex.output) }
        case r :: rs =>
          r match {
            case Router.B =>
              val arg = argEDenotation.asInstanceOf[CompositeEDenotation]
              val newExamples = arg.examples.flatMap { ex => // For each possible input to arg
                recurse(rs, funcEDenotation, ex.output).flatMap{out => Some(Example(ex.input, out))}
              }
              if (newExamples.size == 0) None else Some(CompositeEDenotation(newExamples))

            case Router.C =>
              val func = funcEDenotation.asInstanceOf[CompositeEDenotation]
              val newExamples = func.examples.flatMap { ex => // For each possible input to func
                recurse(rs, ex.output, argEDenotation).flatMap{out => Some(Example(ex.input, out))}
              }
              if (newExamples.size == 0) None else Some(CompositeEDenotation(newExamples))

            case Router.S =>
              val func = funcEDenotation.asInstanceOf[CompositeEDenotation]
              val arg = argEDenotation.asInstanceOf[CompositeEDenotation]
              val newExamples = func.examples.flatMap { ex1 => // For each possible input to func
                arg.examples.flatMap { ex2 => // For each possible input to arg
                  if (ex1.input == ex2.input)
                    recurse(rs, ex1.output, ex2.output).flatMap{out => Some(Example(ex1.input, out))}
                  else
                    None
                }
              }
              // They won't be aligned, so the below is wrong
              // Assume that the examples for both line up (we've set it up this way);
              // Otherwise, we would have to pair the inputs
              /*require (func.examples.size == arg.examples.size, "Not aligned: "+func+" AND "+arg)
              val newExamples = (func.examples zip arg.examples).flatMap { ex_pair =>
                require (ex_pair._1.input == ex_pair._2.input, "EDenotations not aligned: "+ex_pair)
                recurse(rs, ex_pair._1.output, ex_pair._2.output).flatMap{out => Some(Example(ex_pair._1.input, out))}
              }*/
              if (newExamples.size == 0) None else Some(CompositeEDenotation(newExamples))

            case _ => throw impossible
          }
      }
    }
    recurse(routers, func.edenotation, arg.edenotation) match {
      case Some(edenotation) => edenotation
      case None => throw fails("Can't form e-denotation from %s | %s | %s", this.routers, this.func, this.arg)
    }
  }

  def replaceInfo(params:Params, newNode:DNode) = {
    if (edenotation == newNode.edenotation) {
      params.edenotations.get(edenotation) match {
        case Some(info) => info.replace(this, newNode) // This is a focused e-denotation
        case None => // Do nothing
      }

      // Recurse
      newNode match {
        case newNode:DCompositeNode =>
          if (routers == newNode.routers) {
            func.replaceInfo(params, newNode.func)
            arg.replaceInfo(params, newNode.arg)
          }
        case _ =>
      }
    }
  }

  override def valueString = super.valueString+"_"+edenotation
}

object DNode {
  val invalid : DNode = {
    object Invalid { override def toString = "INVALID" }
    DPrimitiveNode(WrappedValue[AnyRef](Invalid), EDenotation.invalid)
  }
}
