package programinduction

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import fig.record.Record
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,fails}
import tea.Utils.{track,begin_track,end_track,logs,dbgs,warnings}

import java.util.{Set,ArrayList,Random}
import scala.reflect.Manifest

import StandardLibrary._
import Router.{B,C,S}
import programinduction.{MPrimitiveNode => __}
import programinduction.{MCompositeNode => ___}
import MyUtils._

/**
This file deals with learning combinator programs from examples.

Two types of data structures are maintained for combinators:
 - A mutable version (which we can apply MCMC on to explore gradually).
   Store the values of this combinator applied on every test case (once routing has been completed).
 - An immutable version (which is used for hash table lookups).
Not all mutable combinators (especially top-level ones are cached, especially top-level ones).
*/

/*
A world consists of a list of scenarios.
A scenario consists of a latent program and a set of test cases.
*/
trait AbstractWorld {
  def worldOpts : WorldOptions // ABSTRACT

  val primitives = new HashMap[Type, ArrayBuffer[Combinator]] // SET IN initialize()
  val scenarios = new ArrayBuffer[Scenario] // SET IN initialize()
  val params = new Params(worldOpts) // SET in initialize()
  val transforms = new ArrayBuffer[Transform]

  val fullSamples = new CounterMap[List[KNode]]
  val fullSampleLogLikelihoods = new HashMap[List[KNode], Double]

  // What did we do:
  //   Refactor (transform)
  //   Change candidate
  // Was a move even possible?
  // Did we accept the move?
  case class Event(description:String, possible:Boolean, acceptRatio:Double, accept:Boolean)
  val samplingStats = new HashMap[String, CounterMap[String]]

  val eventsOut = fig.basic.IOUtils.openOutHard(Execution.getFile("events"))
  def recordEvent(event:Event) = {
    if (event.possible) {
      val s = if (event.possible) { if (event.accept) "accept" else "reject" } else "impossible"
      samplingStats.getOrElseUpdate(event.description, new CounterMap[String]).incr(s, 1)
      if (worldOpts.outputEvents) {
        if (event.accept)
          eventsOut.println("STATE\t"+scenarios.map(_.currRoot.valueString).mkString("\t"))
        eventsOut.println(event.description+"\t"+s+"\t"+event.acceptRatio)
      }
    }
  }

  def allowCombination_?(func:KNode, arg:KNode) : Boolean

  ////////////////////////////////////////////////////////////

  type TestCase <: AbstractTestCase
  trait AbstractTestCase {
    var enabled = true // Some training cases be rejected because they don't align, but we can still test on testing (TextEdit).

    val predOutputs = new CounterMap[Combinator]

    def inputs : List[Combinator]
    def output : Combinator
    def passes(node:AbstractNode) = !enabled || node.call(inputs, true) == output
    //def passes(node:AbstractNode) = { val a = node.call(inputs, true); dbgs("COMPARE %s %s => %s", a, output, a == output); a == output }
    def getLoss(node:AbstractNode) : Double = getLoss(node.call(inputs, true)) // Override if want to evaluate on something else
    def getLoss(predOutput:Combinator) = if (predOutput == output) 0.0 else 1.0 // Override if want to evaluate on something else
    lazy val FType = {
      def recurse(inputs:List[Combinator]) : Type = {
        inputs match {
          case Nil => output.myType
          case c :: cs => c.myType -: recurse(cs)
        }
      }
      recurse(inputs)
    }
    override def toString = inputs.map(_.valueString).mkString(",")+"=>"+output.valueString
  }
  type Scenario <: AbstractScenario

  def avgTestError = {
    (0.0 /: scenarios) { (sum:Double,scenario:AbstractScenario) => sum + scenario.testError * scenario.testTestCases.size } /
    (0 /: scenarios)(_ + _.testTestCases.size)
  }

  trait AbstractScenario {
    def name : String
    def testCases : Seq[TestCase] // SET by init()
    val roots = new ArrayBuffer[MNode] // Latent program; SET by init()
    var currRootIdx = 0
    def currRoot = roots(currRootIdx)
    def alwaysNeedsToPassTestCases = true // Whether we should expect root to always pass (e.g., recipes, editing)
    def FType = testCases.head.FType
    def retType = FType.finalOutType

    // My params is to get a baseline
    lazy val myParams = {
      val params = new Params(worldOpts)
      params.initialize(primitives, computeRogueProb _)
      params
    }
    def params = {
      if (worldOpts.multitaskSharing) AbstractWorld.this.params
      else myParams
    }

    def createInitRootNodes : Unit // ABSTRACT

    def initialize : Unit = track("AbstractScenario.initialize: %s", name) {
      createInitRootNodes
      require (roots.size > 0, "No roots to start with for scenario "+name+" with "+testCases.size+" test cases")
      currRootIdx = worldOpts.random.nextInt(roots.size) // Start with a random root
      Utils.track_printAll("%s roots", roots.size) {
        //roots.foreach { root => logs("%s%s", root, {if(root eq currRoot) " *" else ""}) }
        roots.foreach { root => prettyPrintLog(root) }
      }
      logs("oracle test error = %s", (1.0 /: roots)(_ min testError(_)))
      roots.foreach { root => initializeNode(root, root eq currRoot) }
    }

    ////////////////////////////////////////////////////////////
    // For computing DNodes and ednotations
    case class InterceptionCombinator(value:Combinator) extends Combinator { // To record function behavior
      // Record input/output behavior which will be used to form an edenotation.
      val uses = new ArrayBuffer[Use]
      def call(wrappedArgs:List[Combinator], forceEval:Boolean) = {
        val ret = value.call(wrappedArgs, forceEval) // Call our friendly combinator
        //dbgs("%s : %s => %s", this, wrappedArgs.mkString(" | "), ret)
        val wrappedRet = {
          if (ret.myType.isInstanceOf[BaseType]) ret
          else InterceptionCombinator(ret) // Only wrap function types
        }
        uses += Use(wrappedArgs, wrappedRet)
        wrappedRet
      }

      // Recursively compute edenotation (hope there are no cycles).
      lazy val edenotation = usesToEDnotation(uses.toList)

      val myType = value.myType
      def valueString = "INTERCEPT("+value.valueString+")"
    }

    // A use is essentially a test case and specifies the same information as a EDenotation.
    // The information is very linear and flat, so we need binarize it.
    case class Use(inputs:List[Combinator], output:Combinator) {
      def strip = Use(inputs.tail, output)
    }

    def combinatorToDenotation(c:Combinator) : EDenotation = {
      c match {
        case c:InterceptionCombinator => c.edenotation
        case PartialCombinator(node:BNode, args) => createDNode(node).edenotation.call(args) // This happens when functions are arguments (e.g., While loops)
        case _ => PrimitiveEDenotation(c)
      }
    }

    def usesToEDnotation(uses:List[Use]) : EDenotation = {
      // Convert linearized uses into a set of examples
      def extractExamples(uses:List[Use]) : List[Example] = {
        uses match {
          case Nil => Nil
          case use :: rest_uses =>
            val input = combinatorToDenotation(use.inputs.head)
            val (same_uses, other_uses) = rest_uses.partition { u =>
              combinatorToDenotation(u.inputs.head) == input
            }
            val output = convert((use::same_uses).map(_.strip))
            Example(input, output) :: extractExamples(other_uses)
        }
      }
      def convert(uses:List[Use]) : EDenotation = {
        val use = uses.head
        if (use.inputs.size == 0) {
          // Assume all outputs are the same
          uses.foreach { u => require (u.output == use.output) }
          combinatorToDenotation(use.output)
        }
        else
          CompositeEDenotation(extractExamples(uses))
      }
      convert(uses)
    }

    // Create the actual DNodes (will lazily create the edenotations as we do this)
    def createDNode(node:BNode) : DNode = {
      node match {
        case node:BPrimitiveNode =>
          node.value match {
            case interceptor:InterceptionCombinator =>
              DPrimitiveNode(interceptor.value, interceptor.edenotation)
            case _ =>
              DPrimitiveNode(node.value, PrimitiveEDenotation(node.value))
          }
        case node:BCompositeNode =>
          DCompositeNode(node.routers, createDNode(node.func), createDNode(node.arg))
      }
    }

    // Run program to compute all edenotations.
    // Idea: Setup interception combinators to catch all the calls and record how the combinators are used at the leaves.
    def computeDNode(myTestCases:List[AbstractTestCase], node:AbstractNode) : DNode = {
      // Replace leaves (primitives) that have function types with interception combinators
      def createInterceptedVersion(node:AbstractNode) : BNode = {
        node match {
          case node:AbstractPrimitiveNode =>
            val c = {
              if (node.myType.isInstanceOf[BaseType]) node.value
              else InterceptionCombinator(node.value)
            }
            BPrimitiveNode(c)
          case node:AbstractCompositeNode[_] =>
            BCompositeNode(node.routers, createInterceptedVersion(node.func), createInterceptedVersion(node.arg))
        }
      }

      val wrappedNode = createInterceptedVersion(node)
      //myTestCases.foreach { testCase => dbgs("%s => %s", testCase.inputs, wrappedNode.call(testCase.inputs, true)) }
      val ok = myTestCases.forall(_.passes(wrappedNode))
      require (ok, "Failed test cases: "+node+" | "+myTestCases)

      createDNode(wrappedNode)
    }

    // Called when we initialize the focused e-denotations
    def addEDenotationCandidate(myTestCases:List[AbstractTestCase], candidate:AbstractNode) = {
      val dnode = computeDNode(myTestCases, candidate)
      val info = params.edenotations.getOrElseUpdate(dnode.edenotation, new EDenotationInfo)
      info.candidateNodes += dnode
    }
    def addEDenotationCandidates(myTestCases:List[AbstractTestCase], candidates:Seq[AbstractNode]) = {
      candidates.foreach(addEDenotationCandidate(myTestCases, _))
    }

    // This is used to setup edenotation candidates
    // A subproblem involves mapping an input (e.g., (initial state, current state)) to a desired value
    // (e.g., the next position that the user moves to).
    // The initial state is needed to get things like a block which is to be deleted and inserted elsewhere.
    // This class defines a set of test cases (which specifies a e-denotation).
    // For this e-denotation, one specifies a set of candidate combinators.
    abstract class SubProblem[I,O](inputs:List[List[I]], outputs:List[O])(implicit mI:Manifest[I], mO:Manifest[O]) {
      var cases = mapList2(inputs, outputs, { (is:List[I],o:O) =>
        new AbstractTestCase {
          val inputs = is.map(WrappedValue(_))
          val output = WrappedValue(o)
        }
      })

      def candidates : List[MNode] // ABSTRACT

      def mustPass = true
      def removeHardCases = false

      // Filter out candidates that don't pass the test cases.
      // Go one test case at a time, pruning away candidates.
      // If we pruned them all away, then just use the last batch of candidates and raise a warning.
      def findValidCandidates(cases:List[AbstractTestCase], candidates:List[MNode]) : List[MNode] = {
        cases match {
          case Nil => candidates
          case c :: cs =>
            val validCandidates = candidates.filter(c.passes(_))
            if (validCandidates.size == 0) {
              if (removeHardCases) {
                warnings("No valid candidates among\n%s\nfor subproblem:\n%s\nremoving last %s/%s cases",
                  candidates.map(_.valueString).mkString("\n"), cases.mkString("\n"), cases.size, this.cases.size)
                this.cases = this.cases.slice(0, this.cases.size-cases.size) // Remove the cases we couldn't do
              }
              else if (mustPass) {
                throw fails("No valid candidates among\n%s\nfor subproblem:\n%s",
                  this.candidates.map(_.valueString).mkString("\n"), cases.mkString("\n"))
              }
              candidates
            }
            else
              findValidCandidates(cs, validCandidates)
        }
      }
      val validCandidates = findValidCandidates(cases, candidates)
      if (worldOpts.verbose >= 3) candidates.foreach { candidate => dbgs("CAND %s %s", validCandidates.find(_ == candidate) != None, candidate) }
      if (worldOpts.verbose >= 2) Utils.track_printAll("validCandidates for %s", cases.mkString(" ; ")) {
        validCandidates.foreach { node => logs("%s", node.valueString) }
      }
      addEDenotationCandidates(cases, validCandidates)
      def defaultCandidate = validCandidates(worldOpts.random.nextInt(validCandidates.size))
    }

    // Run on test cases and record the empirical denotation of every subtree.
    // initializeDNodes(node) is needed rather than node.updateEDenotationsDown() because
    // the latter assumes leaves have their DNode's filled in.
    // To fill in those, we need something heavier duty.
    // This function can be afford to be a bit slow.
    def initializeDNodes(node:MNode) = {
      def set(node:MNode, dnode:DNode) : Unit = {
        node.dnode = dnode
        (node, dnode) match {
          case (node:MCompositeNode, dnode:DCompositeNode) =>
            require (node.routers == dnode.routers)
            set(node.func, dnode.func)
            set(node.arg, dnode.arg)
          case (node:MPrimitiveNode, dnode:DPrimitiveNode) =>
          case _ => throw impossible
        }
      }
      val dnode = computeDNode(enabledTrainTestCases, node)
      set(node, dnode)
    }

    def initializeNode(node:MNode, add:Boolean) : Unit = {
      node.updateKeysDown(params)
      initializeDNodes(node)
      node.updateParentsDown(None)

      def recurse(node:MNode) : Unit = { // Similar to addToParams
        //dbgs("recurse %s", node)
        if (node.key.addCustomer(worldOpts.random, params).numTablesChanged) {
          node match {
            case node:MCompositeNode =>
              recurse(node.func) 
              recurse(node.arg) 
            case _ =>
          }
        }
      }

      if (add) recurse(node)
    }

    def passesAllTrainTestCases : Boolean = passesAllTrainTestCases(currRoot)
    def passesAllTrainTestCases(node:AbstractNode) = trainTestCases.forall(_.passes(node))
    def trainTestCases = testCases.slice(0, testStart).toList
    def enabledTrainTestCases = trainTestCases.filter(_.enabled)

    // When we evaluate, we might want to consider
    def testTestCases = testCases.slice(testStart, testCases.size).toList
    def testError : Double = testError(currRoot)
    def testError(node:AbstractNode) = {
      if (testTestCases.size == 0) 0.0
      else 1.0 * (0.0 /: testTestCases)(_ + _.getLoss(node)) / testTestCases.size
    }
    def testStart = worldOpts.numTrainTestCases min testCases.size // Index where test data starts

    trait ExhaustiveSearch {
      val validNodes = new ArrayBuffer[KNode] // Ones that pass the test cases
      var nodes = new HashMap[Type, ArrayBuffer[KNode]]
      def done = validNodes.size >= worldOpts.exhaustiveMaxSolutions

      def evaluate(node:KNode) : Unit // ABSTRACT
      def search : Unit // ABSTRACT

      // Add primitives (depth 0)
      primitives.valuesIterator.foreach(_.foreach { c:Combinator =>
        val node = KPrimitiveNode(c)
        evaluate(node)
        append(nodes, node.myType, node)
      })

      search
      logs("%s valid nodes", validNodes.size)
      validNodes
    }

    class BaseFormExhaustiveSearch extends ExhaustiveSearch {
      // Construct a program in base form (without routers); and the pull up all the arguments
      // (this reduces the number of extensionally equivalent programs)
      // Later, we pull those arguments out

      case class InputCombinator(i:Int, myType:Type) extends Combinator {
        def call(args:List[Combinator], forceEval:Boolean) = this
        //def call(args:List[Combinator], forceEval:Boolean) = throw impossible
        def valueString = "INPUT"+i
      }
      lazy val inputNodes = {
        def recurse(myType:Type) : List[Type] = {
          myType match {
            case myType:BaseType => Nil
            case FuncType(inType, outType) => inType :: recurse(outType)
          }
        }
        val argTypes = recurse(FType).toArray
        map(argTypes.size, { i:Int => KPrimitiveNode(new InputCombinator(i, argTypes(i))) })
      }

      // This is is equivalent to but faster than abstractArgs and then calling passesTestCases
      def directPassesTestCases(node:KNode) = {
        //dbgs("directPassesTestCases: %s", node)
        // Replace InputCombinator's with the real thing and evaluate (no routing!)
        def eval(node:KNode, args:List[Combinator], testCase:TestCase) : Combinator = {
          node match {
            case node:KPrimitiveNode =>
              node.value match {
                case InputCombinator(i, _) => testCase.inputs(i)
                case _ => node.value.call(args, true)
              }
            case KCompositeNode(Nil, func, arg) =>
              eval(func, eval(arg, Nil, testCase) :: args, testCase)
            case _ => throw impossible
          }
        }
        trainTestCases.forall { testCase => 
          eval(node, Nil, testCase) == testCase.output
        }
      }

      type Mask = Int

      // Return abstracted node and the bit mask of the arguments that should be passed down this way
      def abstractArgs(node:KNode) : (KNode,Mask) = {
        node match {
          case node:KPrimitiveNode =>
            node.value match {
              case InputCombinator(i, myType) => (KPrimitiveNode(Id(myType)), 1<<i)
              case _ => (node, 0)
            }
          case node:KCompositeNode =>
            val (newFunc, funcMask) = abstractArgs(node.func)
            val (newArg, argMask) = abstractArgs(node.arg)
            def computeRouters(i:Int) : List[Router] = {
              if (i == inputNodes.size) Nil
              else {
                val goesFunc = (funcMask & (1 << i)) != 0
                val goesArg = (argMask & (1 << i)) != 0
                if (goesFunc && goesArg) Router.S :: computeRouters(i+1)
                else if (goesFunc) Router.C :: computeRouters(i+1)
                else if (goesArg) Router.B :: computeRouters(i+1)
                else computeRouters(i+1)
              }
            }
            val routers = computeRouters(0)
            (routers.reverse, newArg) match { // (C*B f I) => f
              case (B::cs, KPrimitiveNode(_:Id)) if cs.forall(_ == C) => (newFunc, funcMask|argMask) // Avoid composition with identity
              case _ => (KCompositeNode(routers, newFunc, newArg), funcMask|argMask)
            }
        }
      }

      def evaluate(node:KNode) = {
        if (node.myType == retType && directPassesTestCases(node)) {
          val (finalNode, mask) = abstractArgs(node)
          if (mask == (1<<inputNodes.size)-1) { // Only valid if all arguments used
            logs("Found valid node using direct testing: %s ==> %s", node, finalNode)
            require (passesAllTrainTestCases(finalNode))
            validNodes += finalNode
          }
        }
      }

      // Our goal is to find all programs that have a given empirical denotation.
      // This space can be decomposed into subproblems.

      // We first find programs in base form.  Each combinator then is associated with a Behavior,
      // a vector of outputs, one for each test case.
      // Construct a hypergraph where nodes are behaviors and hyperedges indicate
      // reductions from one problem other problems (via application of some primitive function).
      // Construct this hypergraph by iterating to a fixed point or until we're fed up.
      def searchSemantics = {
        val C = trainTestCases.size // Number of test cases
        val K = FType.arity // Number of inputs

        def maskToString(mask:Mask) = map(K, { k:Int => if ((mask & (1<<k)) != 0) 1 else 0 }).mkString("") 

        val targetBehavior = trainTestCases.map(_.output)
        type Behavior = List[Combinator] // List of outputs which are produced for each of the test cases
        def behaviorToString(behavior:Behavior) = {
          behavior.map(_.valueString).mkString(",")+(if (behavior == targetBehavior) ":TARGET" else "")
        }

        val mincost = new HashMap[Behavior,Int] // behavior -> minimum cost (computed from hyperedges)
        // How we go from one node (behavior) to another
        case class Hyperedge(func:Combinator, args:List[Behavior]) {
          override def toString = "("+func.valueString+(if (args.size == 0) "" else " ")+args.map(behaviorToString).mkString(" ")+")=>"+
            map(C, { i:Int => func.call(args.map(_(i)), true).valueString }).mkString(",")
          def computeCost = 1 + (0 /: args)(_ + mincost(_))
        }
        val hypergraph = new HashMap[Type,HashMap[Behavior,HashSet[Hyperedge]]] // type -> behavior -> hyperedges leaving it

        // Return whether anything was changed
        def add(behavior:Behavior, edge:Hyperedge) = {
          val set = hypergraph.getOrElseUpdate(behavior.head.myType, new HashMap[Behavior,HashSet[Hyperedge]]).
            getOrElseUpdate(behavior, new HashSet[Hyperedge])
          val newCost = edge.computeCost
          val costChanged = mincost.get(behavior) match {
            case Some(oldCost) if oldCost <= newCost => false
            case _ => mincost(behavior) = newCost; true
          }
          val edgesChanged = {
            if (set(edge)) false
            else {
              set += edge
              if (worldOpts.verbose >= 5) dbgs("HYPEREDGE\t%s\t%s\t%s", behaviorToString(behavior), newCost, edge)
              true
            }
          }
          costChanged || edgesChanged
        }

        //// Initialize behaviors that we can get right off the bat
        // Test case inputs (e.g., ARG0, ARG1)
        foreach(K, { k:Int =>
          val behavior = trainTestCases.map(_.inputs(k))
          add(behavior, Hyperedge(inputNodes(k).value, Nil))
        })
        // Base-typed primitive combinators (e.g., int, bool)
        primitives.keysIterator.foreach { myType =>
          if (myType.isInstanceOf[BaseType]) {
            primitives(myType).foreach { value =>
              val behavior = mapList(trainTestCases.size, { i:Int => value }) // Return same value c
              add(behavior, Hyperedge(value, Nil))
            }
          }
        }

        // Try to get other behaviors by combining behaviors using functional primitives (e.g., +, if, <)
        track("Constructing hypergraph") {
          Utils.track_foreach("Iteration", worldOpts.exhaustiveMaxIters, { iter:Int =>
            // Enumerate new combinations of function-typed primitive combinators with various behaviors
            val newHyperedges = new ArrayBuffer[Hyperedge]
            primitives.keysIterator.foreach{ myType =>
              primitives(myType).foreach { func => // Pick up a functional primitive combinator
                // For each argument to func, choose behavior of the appropriate type to fill in
                def enumerateArgBehaviors(revArgs:List[Behavior], myType:Type) : Unit = {
                  myType match {
                    case myType:BaseType => // Done choosing behaviors for every argument
                      val args = revArgs.reverse
                      newHyperedges += Hyperedge(func, args)
                    case myType:FuncType =>
                      hypergraph.get(myType.inType) match {
                        case None =>
                        case Some(map) =>
                          map.keysIterator.foreach { b => // For each possible behavior of the right type
                            enumerateArgBehaviors(b :: revArgs, myType.outType)
                          }
                      }
                  }
                }
                enumerateArgBehaviors(Nil, myType)
              }
            }

            // Add these new hyperedges
            var changed = false
            newHyperedges.foreach { hyperedge =>
              val behavior = mapList(C, { i:Int => // For every test case, apply func to the behavior
                hyperedge.func.call(hyperedge.args.map(_(i)), true)
              })
              changed |= add(behavior, hyperedge)
            }
            changed
          })
        }

        // A mask is a bit-vector indicating exactly which of the K inputs are being used.
        // A triple (hyperedge, n, mask) describes the set of combinators that take inputs described by the mask
        // and applies hyperedge.func to the first n arguments which satisfies the appropriate behaviors
        // (e.g., (+ <behavior that returns [3,4]).
        // Note that a triple specifies a unique e-denotation, but a e-denotation could correspond to multiple triples.
        // For each triple that is reachable from the target behavior,
        // compute a single combinator (doesn't matter which)---note that this need not exist.
        case class Triple(hyperedge:Hyperedge, n:Int, mask:Mask) {
          override def toString = fmts("%s(n=%s)[%s]", hyperedge, n, maskToString(mask))
        }
        val tripleMap = new HashMap[Triple, Option[KNode]]
        val behaviorMap = new HashMap[(Behavior,Mask), List[KNode]] // Need to keep many ways of achieving a behavior

        def extractFromTriple(hyperedge:Hyperedge, n:Int, mask:Mask) : Option[KNode] = {
          val triple = Triple(hyperedge, n, mask)

          tripleMap.get(triple) match {
            case Some(nodeOpt) => nodeOpt // Cached version
            //case Some(nodeOpt) => dbgs("cached %s %s", triple, nodeOpt); nodeOpt
            case None =>
              tripleMap(triple) = None // Avoid cycles
              var nodeOpt : Option[KNode] = None

              if (worldOpts.verbose >= 6) begin_track("extractFromTriple %s", triple)
              if (n == 0) { // Primitive
                nodeOpt = hyperedge.func match {
                  case input:InputCombinator => if (mask == (1 << input.i)) Some(KPrimitiveNode(Id(input.myType))) else None
                  case _ => if (mask == 0) Some(KPrimitiveNode(hyperedge.func)) else None
                }
              }
              else if (mask == 0 && n == hyperedge.args.size) {
                // Don't create expressions that evaluate to primitives that don't depend on the inputs (+ 1 1)
              }
              else {
                // For each input k = 0..K-1 that is on in the mask,
                // decide whether to send it to the function or argument or both, but not neither
                def recurse(k:Int, funcMask:Int, argMask:Int, routers:List[Router]) : Unit = {
                  if (k == -1) { // Done processing inputs
                    val localNodeOpt = extractFromTriple(hyperedge, n-1, funcMask) match {
                      case Some(funcNode) =>
                        extractFromBehavior(hyperedge.args(n-1), argMask) match {
                          case Some(argNode) =>
                            // (B f I) => f
                            if (worldOpts.verbose >= 5) dbgs("COMBINE %s AND %s (routers %s); triple %s", funcNode, argNode, routers, triple)
                            routers.reverse match { // (C*B f I) => f
                              case B::cs if cs.forall(_ == Router.C) =>
                                argNode match {
                                  case KPrimitiveNode(id:Id) => Some(funcNode) // Collapse out the argNode
                                  case _ => Some(KCompositeNode(routers, funcNode, argNode))
                                }
                              case _ => Some(KCompositeNode(routers, funcNode, argNode))
                            }
                          case None => None
                        }
                      case None => None
                    }
                    localNodeOpt match {
                      case None =>
                      case Some(node) => 
                        if (nodeOpt == None || node.numNodes < nodeOpt.get.numNodes) nodeOpt = Some(node)
                        //if (nodeOpt == None || hyperedge.computeCost < behaviorCost) nodeOpt = Some(node) // Don't know how to deal with cost
                    }
                  }
                  else {
                    if ((mask & (1<<k)) != 0) { // If input k is actually being uesd
                      recurse(k-1, funcMask, argMask | (1<<k), Router.B::routers)
                      recurse(k-1, funcMask | (1<<k), argMask, Router.C::routers)
                      recurse(k-1, funcMask | (1<<k), argMask | (1<<k), Router.S::routers)
                    }
                    else {
                      recurse(k-1, funcMask, argMask, routers)
                    }
                  }
                  //dbgs("recurse %s k=%s, func[%s], arg[%s], routers=%s", triple, k, maskToString(funcMask), maskToString(argMask), routers.mkString(""))
                }
                recurse(K-1, 0, 0, Nil)
              }

              if (worldOpts.verbose >= 5) {
                if (nodeOpt != None)
                  dbgs("TRIPLE\t%s\t%s", triple, nodeOpt.get.valueString)
              }
              if (worldOpts.verbose >= 6) end_track
              tripleMap(triple) = nodeOpt
              nodeOpt
          }
        }

        def extractFromBehavior(behavior:Behavior, mask:Mask) : Option[KNode] = {
          //dbgs("extractFromBehavior %s %s", behavior, maskToString(mask))
          (behaviorMap.get((behavior,mask)) match {
            case Some(nodeList) => nodeList // Cached version
            case None =>
              behaviorMap((behavior,mask)) = Nil // Avoid cycles

              // Heuristic: sort by increasing cost so that we at least get the 
              /*val hyperedges = hypergraph(behavior.head.myType)(behavior).toList.
                map { hyperedge => (hyperedge, hyperedge.computeCost) }.
                sort { _._2 < _._2 }.
                map { _._1 }*/
              val hyperedges = hypergraph(behavior.head.myType)(behavior)

              val nodeList = {
                if (worldOpts.keepNodePerBehavior) {
                  var nodeOpt : Option[KNode] = None
                  hypergraph(behavior.head.myType)(behavior).foreach { hyperedge =>
                    extractFromTriple(hyperedge, hyperedge.args.size, mask) match {
                      case Some(node) =>
                        if (nodeOpt == None || node.numNodes < nodeOpt.get.numNodes) nodeOpt = Some(node)
                      case None =>
                    }
                  }
                  nodeOpt.toList
                }
                else {
                  hypergraph(behavior.head.myType)(behavior).flatMap { hyperedge =>
                    extractFromTriple(hyperedge, hyperedge.args.size, mask)
                  }.toList
                }
              }

              if (worldOpts.verbose >= 5) {
                if (nodeList != Nil)
                  dbgs("BEHAVIOR/MASK\t%s[%s]\t%s", behavior.map(_.valueString).mkString(","), maskToString(mask), nodeList.map(_.valueString).mkString(","))
              }
              behaviorMap((behavior,mask)) = nodeList
              nodeList
          }) match {
            case Nil => None
            case node :: _ => Some(node)
          }
        }

        track("Extracting triples and behaviors") {
          extractFromBehavior(targetBehavior, (1<<K)-1)
          logs("%s (behavior,mask) pairs, %s triples", behaviorMap.size, tripleMap.size)
        }

        // For each triple, there is a unique denotation.  Add the found node to this e-denotation.
        def extractEDenotation(triple:Triple) = {
          tripleMap(triple) match {
            case None =>
            case Some(node) =>
              node match {
                case KPrimitiveNode(id:Id) => // Ignore this (comes from INPUT0, ...)
                case _ =>
                  val cases = mapList(C, { i:Int =>
                    // The test case includes applying on all but the first hyperedge.n arguments and returning the desired behavior
                    val hyperedge = triple.hyperedge
                        
                    // The inputs are the ones which get routed in followed by the rest of the arguments starting with triple.n
                    def getInputs(k:Int) : List[Combinator] = {
                      if (k == K) mapList(triple.n, hyperedge.args.size, { k:Int => hyperedge.args(k)(i) })
                      else {
                        if ((triple.mask & (1<<k)) != 0) trainTestCases(i).inputs(k) :: getInputs(k+1)
                        else getInputs(k+1)
                      }
                    }
                    new AbstractTestCase {
                      val inputs = getInputs(0)
                      val output = hyperedge.func.call(hyperedge.args.map(_(i)), true)
                    }
                  })
                  if (worldOpts.verbose >= 5) dbgs("ADD\t%s\t%s\t%s", triple, cases, node)
                  addEDenotationCandidate(cases, node)
              }
          }
        }
        track("Extracting e-denotations") {
          tripleMap.keysIterator.foreach(extractEDenotation)
        }

        // Just need one since we have all the e-denotation stuff
        validNodes += behaviorMap((targetBehavior,(1<<K)-1)).head
      }

      // Consider all possible combinators.
      // This only returns a list of flat combinators.
      // Use searchSemantics instead to get factoring using e-denotations.
      def searchSyntax = {
        foreach(1, worldOpts.exhaustiveMaxDepth+1, { depth:Int =>
          // Construct all valid nodes up to depth _depth_ (limiting number of nodes)
          if (!done) track("Constructing nodes of depth %s/%s", depth, worldOpts.exhaustiveMaxDepth) {
            val newNodes = new ArrayBuffer[KNode]
            if (!done) nodes.keysIterator.foreach { funcType =>
              if (!done) funcType match {
                case FuncType(argType, outType) =>
                  nodes(funcType).foreach { func:KNode =>
                    def considerArg(arg:KNode) = {
                      if ((func.computeDepth == depth-1 || arg.computeDepth == depth-1) &&
                          (1+func.numNodes+arg.numNodes) <= worldOpts.exhaustiveMaxNodes &&
                          allowCombination_?(func, arg) &&
                          !done) {
                        if (worldOpts.verbose >= 5) dbgs("%s || %s", func, arg)
                        val newNode = KCompositeNode(Nil, func, arg)
                        evaluate(newNode)
                        newNodes += newNode
                      }
                    }
                    if (!done) nodes.get(argType) match {
                      case Some(args) => args.foreach(considerArg)
                      case None =>
                    }
                    if (!done) inputNodes.foreach { input:KNode =>
                      if (input.myType == argType) considerArg(input)
                    }
                  }
                case _ =>
              }
            }
            logs("%s new nodes", newNodes.size)
            newNodes.foreach { node => append(nodes, node.myType, node) }
          }
        })
      }

      def search = {
        if (worldOpts.searchSemantics) searchSemantics
        else searchSyntax
      }
    }

    def exhaustiveSearch : Seq[KNode] = Utils.track_printAll("exhaustiveSearch") {
      (new BaseFormExhaustiveSearch).validNodes
    }

    def savePredictions = {
      testTestCases.foreach { testCase =>
        testCase.predOutputs.incr(currRoot.call(testCase.inputs, true), 1)
      }
    }

    def sample(iter:Int) : Unit = {
      if (roots.size > 1 && worldOpts.random.nextDouble < worldOpts.changeCurrRootProb) {
        // Try to change roots
        val i = worldOpts.random.nextInt(roots.size)
        if (i != currRootIdx) {
          considerTransformResult("root", currRoot, roots(i), 1) match {
            case Some(newRoot) =>
              if (worldOpts.verbose >= 2) logs("Switch roots from %s to %s", currRoot, newRoot)
              currRootIdx = i
            case None =>
          }
        }
      }
      else {
        // Pick a random node to sample
        val i = worldOpts.random.nextInt(currRoot.numNodes)
        val node = currRoot.getInDFSOrder(i)
        val canRefactor = iter >= worldOpts.allowRefactoringFromIter
        if (worldOpts.random.nextDouble < worldOpts.changeCandidateProb || !canRefactor)
          sampleCandidate(node)
        else
          sampleRefactor(node)
      }
    }

    def sampleCandidate(node:MNode) = {
      params.edenotations.get(node.dnode.edenotation) match {
        case Some(info) =>
          if (worldOpts.edenotationAllowJumpBack || info.candidateNodes.contains(node.dnode)) { // Can change?
            val newDNode = info.candidateNodes(worldOpts.random.nextInt(info.candidateNodes.size))
            val newNode = MNode.fromDNode(newDNode)
            if (worldOpts.verbose >= 3) logs("Proposing new candidate %s | %s", newNode, newDNode)
            val (root, newRoot) = getRoots(node, newNode)
            considerTransformResult("candidate", root, newRoot, 1) match {
              case Some(newRoot) =>
                roots(currRootIdx) = newRoot
              case _ => // No change
            }
          }
        case None => // No candidates available
          recordEvent(Event("candidate", false, Double.NaN, false))
      }
    }

    def sampleRefactor(node:MNode) = {
      // Refactor
      val transform = transforms(worldOpts.random.nextInt(transforms.size))
      transformMove(transform, node) match {
        case Some(newRoot) => 
          // Change e-denotation information if desired
          if (worldOpts.edenotationChangeSet)
            currRoot.dnode.replaceInfo(params, newRoot.dnode)

          roots(currRootIdx) = newRoot
        case None => // No change
      }
    }

    def repCheck = {
      currRoot.repCheck
      if (alwaysNeedsToPassTestCases) require (passesAllTrainTestCases, currRoot+" does not pass all test cases")
    }

    def log = {
      if (worldOpts.verbose >= 3) logs("%s", currRoot.dnode)
      else logs("%s", currRoot)
    }

    // Return current root of _node_
    // Create a new root for _newNode_
    private def getRoots(node:MNode, newNode:MNode) : (MNode, MNode) = {
      if (node.isTopLevel) {
        newNode.updateKeysDown(params)
        newNode.updateDNodesDown(params)
        (node, newNode) 
      }
      else {
        val parent = node.parent
        val newParent = MCompositeNode(parent.routers,
          if (parent.func eq node) newNode else parent.func,
          if (parent.arg eq node) newNode else parent.arg)
        getRoots(parent, newParent)
      }
    }

    // Return new root (None if no change)
    def transformMove(transform:Transform, node:MNode) : Option[MNode] = {
      transform.transform(node) match {
        case None => // Transform not applicable here
          recordEvent(Event(transform.toString, false, Double.NaN, false))
          None
        case Some((newNode, bias)) =>
          if (worldOpts.verbose >= 2) begin_track("Applied transform %s, considering move...", transform)
          val (root, newRoot) = getRoots(node, newNode)
          val result = considerTransformResult(transform.toString, root, newRoot, bias) match {
            case Some(newRoot) =>
              if (worldOpts.verbose >= 2) logs("Accepted %s", transform)
              Some(newRoot)
            case None => None
          }
          if (worldOpts.verbose >= 2) end_track
          result
      }
    }

    // Return new new root (None if reject)
    def considerTransformResult(description:String, root:MNode, newRoot:MNode, bias:Double) : Option[MNode] = {
      if (worldOpts.verbose >= 3) {
        dbgs("root: %s", root)
        dbgs("newRoot: %s", newRoot)
      }

      val actions = new ArrayBuffer[Action]
      //var modelProbRatio = 1.0
      var logModelProbRatio = 0.0

      def removeFromParams(node:MNode) : Unit = {
        val action = node.key.removeCustomer(worldOpts.random, params)
        if (worldOpts.verbose >= 3) logs("REMOVE: %s", action)
        actions += action
        //modelProbRatio *= action.modelPotential
        //Utils.assertValid(modelProbRatio)
        logModelProbRatio += Math.log(action.modelPotential)
        Utils.assertValid(logModelProbRatio)

        if (action.numTablesChanged) {
          node match {
            case node:MCompositeNode =>
              removeFromParams(node.func)
              removeFromParams(node.arg)
            case _ =>
          }
        }
      }
      def addToParams(node:MNode) : Unit = {
        val action = node.key.addCustomer(worldOpts.random, params)
        if (worldOpts.verbose >= 3) logs("ADD: %s", action)
        actions += action
        //modelProbRatio *= action.modelPotential
        //Utils.assertValid(modelProbRatio)
        logModelProbRatio += Math.log(action.modelPotential)
        Utils.assertValid(logModelProbRatio)

        if (action.numTablesChanged) {
          node match {
            case node:MCompositeNode =>
              addToParams(node.func) 
              addToParams(node.arg) 
            case _ =>
          }
        }
      }

      removeFromParams(root)
      addToParams(newRoot)

      val proposalForwProb = 1.0/root.numNodes
      val proposalRevProb = 1.0/newRoot.numNodes

      Utils.assertValid(logModelProbRatio)
      val modelProbRatio = Math.exp(logModelProbRatio) // Okay if this blows up

      val acceptRatio = Math.pow(modelProbRatio, 1.0/worldOpts.temperature) * proposalRevProb / proposalForwProb * bias
      if (worldOpts.verbose >= 3)
        logs("p(new)/p(old) = %s, q(new|old) = %s, q(old|new) = %s, acceptRatio = %s", modelProbRatio, proposalForwProb, proposalRevProb, acceptRatio)

      if (worldOpts.random.nextDouble < acceptRatio) {
        if (worldOpts.verbose >= 4) dbgs("accept")
        newRoot.updateParentsDown(None)
        recordEvent(Event(description, true, acceptRatio, true))
        Some(newRoot)
      }
      else { // Nothing changed
        if (worldOpts.verbose >= 4) dbgs("reject")
        Utils.foreachReverse(actions.size, { i:Int => actions(i).undoAction(params) })
        recordEvent(Event(description, true, acceptRatio, false))
        None
      }
    }

    def crudeInit = {
      considerTransformResult("initialize", currRoot, rechooseCandidates(currRoot), Double.PositiveInfinity) match {
        case Some(newRoot) => roots(currRootIdx) = newRoot
        case None => throw fails("Unable to initialize")
      }
    }
    def rechooseCandidates(node:MNode) : MNode = {
      def recurse(node:MNode) : MNode = {
        // Switch to the best candidate
        val newNode = params.edenotations.get(node.dnode.edenotation) match {
          case Some(info) =>
            if (!info.candidateNodes.contains(node.dnode)) node // Only consider changing if we're already a candidate (side effect: avoid cycles)
            else {
              val newDNode = info.candidateNodes(Utils.argmax(info.probs))
              if (node.dnode == newDNode) node
              else {
                if (worldOpts.verbose >= 1) {
                  dbgs("OLD: %s", node)
                  dbgs("NEW: %s", MNode.fromDNode(newDNode))
                }
                MNode.fromDNode(newDNode)
              }
            }
          case None => node
        }
        // Recurse
        newNode match {
          case node:MPrimitiveNode => node
          case node:MCompositeNode =>
            MCompositeNode(node.routers, recurse(node.func), recurse(node.arg))
        }
      }

      val newNode = recurse(node)
      newNode.updateKeysDown(params)
      newNode.updateDNodesDown(params)
      newNode
    }
  }

  ////////////////////////////////////////////////////////////

  def numPrimitives(myType:Type) = primitives.get(myType) match { case None => 0; case Some(a) => a.size }
  def addPrimitive(c:Combinator) = append(primitives, c.myType, c)

  def computeRogueProb(c:Combinator) : Double // ABSTRACT

  def initialize : Unit = track("AbstractWorld.initialize") {
    initializeTransforms

    populatePrimitives
    params.initialize(primitives, computeRogueProb _)

    populateScenarios
    while (scenarios.size > worldOpts.maxScenarios) scenarios.remove(scenarios.size-1)
    scenarios.foreach(_.initialize)

    track("Cleaning-up e-denotations") {
      params.removeDuplicateEDenotationCandidates
      params.pruneSingletonEDenotationCandidates
      //if (worldOpts.verbose >= 2) params.logEDenotations
      logs("%s e-denotations after pruning", params.edenotations.size)
    }

    // Change the roots
    if (worldOpts.crudeInit) track("Changing candidates using crude heuristic") {
      params.estimateCrudeEDEnotationDistribution
      scenarios.foreach(_.crudeInit)
    }

    // Print statistics
    val numTrain = new fig.basic.FullStatFig
    val numTest = new fig.basic.FullStatFig
    val numOps = new fig.basic.FullStatFig
    scenarios.foreach { scenario =>
      numTrain.add(scenario.enabledTrainTestCases.size)
      numTest.add(scenario.testTestCases.size)
    }
    logs("%s scenarios; #train: %s; #test: %s", scenarios.size, numTrain, numTest)

    if (worldOpts.verbose >= 3) log
    repCheck
  }
  def populatePrimitives : Unit // ABSTRACT
  def populateScenarios : Unit // ABSTRACT
  def initializeTransforms = { // EXTENSIBLE
    transforms ++=
      Transform_B_Abstraction :: Transform_B_Application ::
      Transform_C_Abstraction :: Transform_C_Application ::
      Transform_S_Abstraction :: Transform_S_Application ::
      Transform_CI_Abstraction :: Transform_CI_Application ::
      //Transform_I_Abstraction :: Transform_I_Application :: // Don't need this when we have CI (reduces spurious I's)
      Nil
  }

  def learn = {
    if (worldOpts.exhaustiveSearch)
      learnExhaustiveSearch
    else
      learnSampling
  }

  def learnExhaustiveSearch = Utils.track_printAll("learnExhaustiveSearch") {
    scenarios.foreach { scenario =>
      scenario.exhaustiveSearch
    }
  }

  def prettyPrintLogRightBranching(node:AbstractNode) = {
    def recurse(node:AbstractNode, indent:Int, level:Int) : Unit = {
      node match {
        case node:AbstractPrimitiveNode => logs("%s%s%s", " "*indent, node.valueString, ")"*level)
        case node:AbstractCompositeNode[_] =>
          val routerStr = node.routers.mkString("")
          logs("%s(%s %s", " "*indent, routerStr, node.func.valueString)
          recurse(node.arg, indent+1+routerStr.size+1, level+1)
      }
    }
    recurse(node, 0, 0)
  }
  def prettyPrintLogLeftBranching(node:AbstractNode) = {
    def recurse(node:AbstractNode, prefix:String) : Unit = {
      node match {
        case node:AbstractPrimitiveNode => logs("%s%s", prefix, node.valueString)
        case node:AbstractCompositeNode[_] =>
          val routerStr = node.routers.mkString("")
          recurse(node.func, prefix+"("+routerStr+" ")
          //logs("%s%s)", " "*prefix.size, node.arg.valueString)
          logs("%s%s)", "  ", node.arg.valueString)
      }
    }
    recurse(node, "")
  }
  def prettyPrintLog(node:AbstractNode) = {
    if (worldOpts.rightBranching) prettyPrintLogRightBranching(node)
    else prettyPrintLogLeftBranching(node)
  }

  def outputAndRecordState(iter:Int) = {
    if (iter == worldOpts.numIters || iter % worldOpts.recordStateInterval == 0) {
      Record.begin("iteration", iter)
      Execution.putLogRec("logLikelihood", params.logLikelihood)
      Utils.track_printAll("Evaluation") {
        if (worldOpts.verbose >= 2) {
          scenarios.foreach { scenario =>
            if (scenario.currRoot.numNodes > 30) Utils.track_printAll("%s", scenario.name) {
              prettyPrintLog(scenario.currRoot)
              scenario.testCases.foreach { testCase =>
                val trueOutput = testCase.output
                val predOutput = scenario.currRoot.call(testCase.inputs, true)
                logs("%s => TRUE: %s\tPRED:%s\t%s",
                  testCase.inputs.map(_.valueString).mkString(","), trueOutput.valueString, predOutput.valueString,
                  if (trueOutput == predOutput) "[correct]" else "[WRONG]")
              }
            }
            else {
              logs("%s => %s", scenario.currRoot, scenario.testCases.map { testCase =>
                testCase.inputs.map(_.valueString).mkString(",")+":"+testCase.output.valueString+":"+scenario.currRoot.call(testCase.inputs, true).valueString
              }.mkString(" "))
            }
          }
        }
        Execution.putLogRec("avgTestError", avgTestError)
      }
      Utils.track_printAll("Stats") {
        samplingStats.keysIterator.foreach { description =>
          Utils.track_printAll("%s", description) {
            samplingStats(description).write(Integer.MAX_VALUE, true, "", logs("%s", _))
          }
        }
      }
      Record.end
    }

    if (iter == worldOpts.numIters || iter % worldOpts.outputStateInterval == 0) {
      Utils.writeLines(Execution.getFile("state."+iter), { puts:(String=>Any) =>
        scenarios.foreach { scenario =>
          //puts(scenario.currRoot.toString)
          puts(fmts("%s\t%s => %s", scenario.name, scenario.currRoot, scenario.testCases.map { testCase =>
            testCase.inputs.map(_.valueString).mkString(",")+":"+testCase.output.valueString+":"+scenario.currRoot.call(testCase.inputs, true).valueString
          }.mkString(" ")))
        }
        //puts(fmts("avgTestError = %s", avgTestError))
      })

      params.write(Execution.getFile("params."+iter))

      // Count the occurrences of each combinators
      // Note that this is different from params's counts in dishes, because there's no collapsing due to the HDP.
      Utils.writeLines(Execution.getFile("combinators."+iter), { puts:(String=>Any) =>
        val map = new CounterMap[String]
        def recurse(node:MNode) : Unit = {
          map.incr(node.key.valueString, 1)
          node match {
            case node:MCompositeNode =>
              recurse(node.func)
              recurse(node.arg)
            case _ =>
          }
        }
        scenarios.foreach { scenario => recurse(scenario.currRoot) }
        map.write(Integer.MAX_VALUE, true, "", puts)
      })
    }

    // Bayesian averaging: record predictions
    if (iter >= worldOpts.burnInFrac*worldOpts.numIters)
      scenarios.foreach(_.savePredictions)
  }

  def learnSampling = Utils.track_printAll("learnSampling") {
    logs("Transforms: %s", transforms.mkString(" "))
    track("Sampling") {
      Utils.track_foreach("Iteration", worldOpts.numIters, { iter:Int =>
        outputAndRecordState(iter)

        //dbgs("ROOT %s", scenarios.head.currRoot.dnode)
        foreach(worldOpts.numHitsPerIter, { hit:Int =>
          if (worldOpts.rememberFullSamples) {
            val sample = scenarios.map(_.currRoot.key).toList
            fullSamples.incr(sample, 1)
            fullSampleLogLikelihoods.getOrElseUpdate(sample, params.logLikelihood)
          }
          scenarios.foreach { scenario =>
            //params.clearZeros // Need to invalid KNodes in other roots if this is done
            if (worldOpts.verbose >= 5) track("Before") { log }
            scenario.sample(iter)
            if (worldOpts.verbose >= 5) track("After") { log }
            repCheck
          }
        })
        !Execution.shouldBail
      })
      outputAndRecordState(worldOpts.numIters)
    }

    track("Final samples") {
      scenarios.foreach { scenario => track("%s", scenario.name) { prettyPrintLog(scenario.currRoot) } }
    }

    if (worldOpts.rememberFullSamples) Utils.track_printAll("Full samples") {
      // Allow one to compute the true distribution over the explored space
      var logZ = Double.NegativeInfinity
      fullSampleLogLikelihoods.valuesIterator.foreach { ll => logZ = fig.basic.NumUtils.logAdd(logZ, ll) }

      fullSamples.foreachSorted(Integer.MAX_VALUE, { (sample,count) =>
        val empProb = fullSamples.getProb(sample)
        val trueLL = fullSampleLogLikelihoods(sample)
        val trueProb = Math.exp(trueLL-logZ)
        logs("%s %s %s %s\t%s", count, empProb, trueProb, trueLL, sample.map(_.valueString).mkString("\t"))
      })
    }
  }

  def repCheck = if (worldOpts.repCheck) {
    params.repCheck
    scenarios.foreach(_.repCheck)

    // Check that params is consistent with the scenarios
    // Have everyone enter
    val checkDishes = new HashMap[KNode,Int]
    def recurse(node:KNode) : Unit = {
      checkDishes(node) = checkDishes.getOrElse(node, 0) + 1
      if (checkDishes(node) == 1) {
        node match {
          case node:KCompositeNode =>
            recurse(node.func)
            recurse(node.arg)
          case node:KPrimitiveNode =>
            // Must be identity or a primitive
            if (!worldOpts.allowRoguePrimitives)
              require (node.value.isInstanceOf[Id] || primitives(node.myType).contains(node.value), "Rogue primitive: " + node)
        }
      }
    }
    scenarios.foreach { scenario =>
      recurse(scenario.currRoot.key)
    }

    checkDishes.keysIterator.foreach { node =>
      require (checkDishes(node) == node.n, node)
    }
  }

  def log = Utils.track_printAll("World") {
    Utils.track_foreach("Scenario", scenarios.size, { i:Int =>
      scenarios(i).log
      true
    })
    Utils.track_printAll("Params") {
      params.log
    }
  }

  // Look at the average
  def getTestError = track("getTestError") {
    val errors = new fig.basic.StatFig
    scenarios.foreach { scenario => track("Scenario %s", scenario.name) {
      scenario.testTestCases.foreach { testCase =>
        errors.add(testCase.predOutputs.argmaxAndMax match {
          case (Some(predOutput), _) =>
            val loss = testCase.getLoss(predOutput)
            if (worldOpts.verbose >= 1) {
              Utils.track_printAll("Test case") {
                logs("Input: %s", testCase.inputs.map(_.valueString).mkString(" | "))
                logs("True: %s", testCase.output.valueString)
                logs("Pred: %s", predOutput.valueString)
                logs("Loss: %s", loss)
              }
            }
            loss
          case _ => throw fails("Unable to get best prediction")
        })
      }
    } }
    errors.mean
  }

  ////////////////////////////////////////////////////////////

  abstract class Transform {
    def transform(node:MNode) : Option[(MNode,Double)] // Don't mutate anything

    // Only allow abstracting primitives?
    def allowAbstraction_?(node:MNode) = {
      !isIdMNode(node) &&
      (!worldOpts.onlyAbstractPrimitiveCombinators || node.isInstanceOf[MPrimitiveNode]) &&
      (!worldOpts.onlyAbstractBaseTypes || node.isInstanceOf[BaseType])
    }

    def isIdMNode(node:MNode) = {
      node match {
        case __(id:Id) => true
        case _ => false
      }
    }

    // Create an identity node which can be applied to arg after routers have sent stuff to it.
    // Creating an identity node is hard because we need to fill in its e-denotation.
    def createIdMNode(arg:MNode, routers:List[Router]) = {
      // If an combinator x has myType and is sitting in argument position of a node with routers,
      // what is x's type after everything has been routed to it?
      def reduceIfArg(myType:Type, routers:List[Router]) : Type = {
        routers match {
          case Nil => myType
          case r :: rs =>
            if (r.goesArg) reduceIfArg(myType.reduce(1), rs)
            else reduceIfArg(myType, rs)
        }
      }

      var examples = new ArrayBuffer[Example]
      def getExamples(edenotation:EDenotation, routers:List[Router]) : Unit = {
        routers match {
          case Nil => examples += Example(edenotation, edenotation)
          case r :: rs =>
            if (r.goesArg) {
              edenotation match {
                case edenotation:CompositeEDenotation =>
                  edenotation.examples.foreach { ex => getExamples(ex.output, rs) }
                case _ => throw fails("Can route to non-function %s", edenotation)
              }
            }
            else
              getExamples(edenotation, rs)
        }
      }
      val idNode = __(Id(reduceIfArg(arg.myType, routers)))
      getExamples(arg.dnode.edenotation, routers)
      require (examples.size > 0)
      idNode.dnode = DPrimitiveNode(idNode.value, CompositeEDenotation(examples.toList))
      idNode
    }
  }

  val B_template = new TransformTemplate(worldOpts) {
    def source = Q(Nil, Q(B::vr::Nil, vx, vy), vz)
    def target = Q(vr::Nil, vx, Q(Nil, vy, vz))
  }
  object Transform_B_Application extends Transform {
    def transform(node:MNode) = node match {
      case ___(r0, ___(r1Br, x, y), z) =>
        B_template.transform(r0::r1Br::Nil) match {
          case Some((rr0r::rr1::Nil, bias)) => Some((___(rr0r, x, ___(rr1, y, z)), bias))
          case _ => None
        }
      case _ => None
    }
    override def toString = "B-Apply"
  }
  object Transform_B_Abstraction extends Transform {
    def transform(node:MNode) = node match {
      case ___(rr0r, x, ___(rr1, y, z)) =>
        B_template.untransform(rr0r::rr1::Nil) match {
          case Some((r0::r1Br::Nil, bias)) => Some((___(r0, ___(r1Br, x, y), z), bias))
          case _ => None
        }
      case _ => None
    }
    override def toString = "B-Abstract"
  }

  val C_template = new TransformTemplate(worldOpts) {
    def source = Q(Nil, Q(C::vr::Nil, vx, vy), vz)
    def target = Q(vr::Nil, Q(Nil, vx, vz), vy)
  }
  object Transform_C_Application extends Transform {
    def transform(node:MNode) = node match {
      case ___(r0, ___(r1Cr, x, y), z) =>
        C_template.transform(r0::r1Cr::Nil) match {
          case Some((rr0r::rr1::Nil, bias)) => Some((___(rr0r, ___(rr1, x, z), y), bias))
          case _ => None
        }
      case _ => None
    }
    override def toString = "C-Apply"
  }
  object Transform_C_Abstraction extends Transform {
    def transform(node:MNode) = node match {
      case ___(rr0r, ___(rr1, x, z), y) =>
        C_template.untransform(rr0r::rr1::Nil) match {
          case Some((r0::r1Cr::Nil, bias)) => Some((___(r0, ___(r1Cr, x, y), z), bias))
          case _ => None
        }
      case _ => None
    }
    override def toString = "C-Abstract"
  }

  val S_template = new TransformTemplate(worldOpts) {
    def source = Q(Nil, Q(S::vr::Nil, vx, vy), vz)
    def target = Q(vr::Nil, Q(Nil, vx, vz), Q(Nil, vy, vz))
  }
  // PROBLEM: even when z == z_copy, their e-denotations could be different depending on whether
  // it exists once or twice.
  // ((S (B + (I 3)) (I 4)) f): f has empirical denotation 3=>..., 4=>...
  // (+ (f 3) (f 4)): the first f has empirical denotation 3=>, second has empirical denotation 4=>
  // This is only a problem when we are S-abstracting function combinators, so only allow S-abstraction and S-application of primitives.
  object Transform_S_Application extends Transform {
    def transform(node:MNode) = node match {
      case ___(r0, ___(r1Sr, x, y), z) =>
        S_template.transform(r0::r1Sr::Nil) match {
          case Some((rr0r::rr1::rr2::Nil, bias)) if z.myType.isInstanceOf[BaseType] => Some((___(rr0r, ___(rr1, x, z), ___(rr2, y, z)), bias))
          case _ => None
        }
      case _ => None
    }
    override def toString = "S-Apply"
  }
  object Transform_S_Abstraction extends Transform {
    def transform(node:MNode) = node match {
      case ___(rr0r, ___(rr1, x, z), ___(rr2, y, z_copy)) if z == z_copy && z.myType.isInstanceOf[BaseType] =>
        S_template.untransform(rr0r::rr1::rr2::Nil) match {
          case Some((r0::r1Sr::Nil, bias)) => Some((___(r0, ___(r1Sr, x, y), z), bias))
          case _ => None
        }
      case _ => None
    }
    override def toString = "S-Abstract"
  }

  val CI_template = new TransformTemplate(worldOpts) {
    def source = Q(Nil, Q(C::vr::Nil, vI, vy), vz)
    def target = Q(vr::Nil, vz, vy)
  }
  object Transform_CI_Application extends Transform {
    def transform(node:MNode) = node match {
      case ___(r0, ___(r1Cr, __(id:Id), y), z) =>
        CI_template.transform(r0::r1Cr::Nil) match {
          case Some((rr0r::Nil, bias)) => Some((___(rr0r, z, y), bias))
          case _ => None
        }
      case _ => None
    }
    override def toString = "CI-Apply"
  }
  object Transform_CI_Abstraction extends Transform {
    def transform(node:MNode) = node match {
      case ___(rr0r, z, y) =>
        CI_template.untransform(rr0r::Nil) match {
          case Some((r0::r1Cr::Nil, bias)) if allowAbstraction_?(z) => Some((___(r0, ___(r1Cr, createIdMNode(z, r0), y), z), bias))
          case _ => None
        }
      case _ => None
    }
    override def toString = "CI-Abstract"
  }

  // Useful for abstracting out the func node, but we can do it directly with CI.
  object Transform_I_Application extends Transform {
    def transform(node:MNode) = node match {
      case ___(Nil, __(id:Id), x) => Some((x, 1))
      case _ => None
    }
    override def toString = "I-Apply"
  }
  object Transform_I_Abstraction extends Transform {
    def appliedByIdentity(node:MNode) = {
      !node.isTopLevel && {
        node.parent match {
          case ___(Nil, __(id:Id), x) => true
          case _ => false
        }
      }
    }
    def transform(node:MNode) = node match {
      case __(id:Id) => None // Don't create (I I)
      case ___(Nil, __(id:Id), _) => None // Don't create (I (I x))
      case x if !appliedByIdentity(x) && allowAbstraction_?(x) => Some((___(Nil, __(Id(x.myType)), x), 1))
      case _ => None
    }
    override def toString = "I-Abstract"
  }
}
