package programinduction

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,fails}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}
import scala.reflect.Manifest

import StandardLibrary._
import Router.{B,C,S}
import programinduction.{MPrimitiveNode => __}
import programinduction.{MCompositeNode => ___}
import MyUtils._

case class Action(node:KNode, dn:Int, modelPotential:Double) {
  def numTablesChanged = (node.n == 0 && dn == -1) || (node.n == 1 && dn == +1) // Called after action is executed
  override def toString = fmts("Action(%s, dn=%s, potential=%s)", node.valueString, dn, modelPotential)

  def doAction(params:Params) = {
    node.n += dn
    params.changeN(node.myType, dn)
    //dbgs("  NEW %s %s", node.myType, params.N(node.myType))
  }
  def undoAction(params:Params) = {
    node.n -= dn
    params.changeN(node.myType, -dn)
  }
}

// Non-mutable versions (as a key)
// However, these keys store the actual value as well.
// TODO: compute and cache hash functions
trait KNode extends AbstractNode {
  def lazyEval = false
  val numNodes : Int // Number of nodes in descendents

  var n = 0 // Number of customers eating this dish
  var localPotential = 1.0 // psi (local for generating a new one of these)

  // Factor change in model probability when this action is executed
  // Call this function before executing the action
  def modelPotential(dn:Int, params:Params) : Double = {
    val p1 = {
      if (dn == +1) {
        if (n == 0) params.alpha*localPotential
        else 1.0*n
      }
      else if (dn == -1) {
        if (n == 1) 1.0/(params.alpha*localPotential)
        else 1.0/(n-1)
      }
      else
        throw impossible
    }
    val p2 = {
      val N = params.N.getOrElse(myType, 0)
      if (dn == +1) 1.0/(params.alpha+N)
      else params.alpha+N-1
    }
    require (p1 > 0 && p2 > 0, p1 + " " + p2)
    p1 * p2
  }

  // Return the action that can be used to replay this action
  def change(dn:Int, params:Params) = {
    val action = Action(this, dn, modelPotential(dn, params))
    action.doAction(params)
    action
  }

  def removeCustomer(random:Random, params:Params) = change(-1, params)
  def addCustomer(random:Random, params:Params) = change(+1, params)
}

case class KPrimitiveNode(value:Combinator) extends KNode with AbstractPrimitiveNode {
  require (!value.isInstanceOf[KNode])
  val numNodes = 1
}

case class KCompositeNode(routers:List[Router], func:KNode, arg:KNode) extends KNode with AbstractCompositeNode[KNode] {
  val numNodes = 1 + func.numNodes + arg.numNodes
}

object KNode {
  val invalid : KNode = {
    object Invalid { override def toString = "INVALID" }
    KPrimitiveNode(WrappedValue[AnyRef](Invalid))
  }
}

////////////////////////////////////////////////////////////

object MNode {
  val invalid : MCompositeNode = {
    object Invalid { override def toString = "INVALID" }
    MCompositeNode(Nil, MPrimitiveNode(Id[AnyRef]), MPrimitiveNode(WrappedValue[AnyRef](Invalid)))
  }

  def fromKNode(node:KNode) : MNode = {
    node match {
      case node:KPrimitiveNode => MPrimitiveNode(node.value)
      case node:KCompositeNode => MCompositeNode(node.routers, fromKNode(node.func), fromKNode(node.arg))
    }
  }

  def fromDNode(node:DNode) : MNode = {
    val mnode = node match {
      case node:DPrimitiveNode => MPrimitiveNode(node.value)
      case node:DCompositeNode => MCompositeNode(node.routers, fromDNode(node.func), fromDNode(node.arg))
    }
    // Important to set the dnode, because we can't easily recover the DNode from the MNode (in contrast, KNode from MNode is easy)
    mnode.dnode = node
    mnode
  }
}

// Mutable versions
trait MNode extends AbstractNode {
  def lazyEval = false
  var parent : MCompositeNode = MNode.invalid
  // These should not change once they are set
  var key = KNode.invalid
  var dnode = DNode.invalid

  def numNodes = key.numNodes
  def getInDFSOrder(i:Int) : MNode

  def isTopLevel = parent == MNode.invalid

  def getRoot : MNode = {
    if (isTopLevel) this
    else parent.getRoot
  }

  // Recursively update values
  def updateKeysDown(params:Params) : Unit
  def updateDNodesDown(params:Params) : Unit
  def updateParentsDown(newParent:Option[MCompositeNode]) : Unit

  def repCheck = {
    require (myType == computeType, myType + " doesn't match " + computeType) 
    require (key.myType == myType)
    require (dnode.myType == myType)
  }
}
case class MPrimitiveNode(value:Combinator) extends MNode with AbstractPrimitiveNode {
  require (!value.isInstanceOf[MNode])
  def updateKeysDown(params:Params) = {
    if (key eq KNode.invalid) key = params.canonicalize(KPrimitiveNode(value))
  }
  def updateDNodesDown(params:Params) = {
    require (!(dnode eq DNode.invalid))
  }
  def updateParentsDown(newParent:Option[MCompositeNode]) = {
    newParent match {
      case None => // Top-level
        parent = MNode.invalid
      case Some(newParent) =>
        if (!(parent eq newParent)) parent = newParent
    }
  }

  def getInDFSOrder(i:Int) = { require (0 <= i && i < key.numNodes); this }

  override def repCheck = {
    super.repCheck
    require (key.asInstanceOf[KPrimitiveNode].value == value)
  }
}

case class MCompositeNode(routers:List[Router], func:MNode, arg:MNode) extends MNode with AbstractCompositeNode[MNode] {
  def updateKeysDown(params:Params) = {
    // Assume if key is already valid, then all descendents are also valid
    if (key eq KNode.invalid) {
      func.updateKeysDown(params)
      arg.updateKeysDown(params)
      key = params.canonicalize(KCompositeNode(routers, func.key, arg.key))
    }
  }
  def updateDNodesDown(params:Params) = {
    // Assume if dnode is already valid, then all descendents are also valid

    if (dnode eq DNode.invalid) {
      // This is too hacky - the transform is now reponsible for updating denotations
      /*// HACK: When Identity first appears (via Transform_I_Abstraction),
      // assume it'll appear as (I x), so we can read off its edenotation for I (apply x and return x).
      // Otherwise we have no way of getting it without rerunning the entire program which would be too expensive.
      if (func.dnode eq DNode.invalid) {
        func match {
          case func:MPrimitiveNode if func.value.isInstanceOf[Id] =>
            require (routers == Nil, this)
            func.dnode = DPrimitiveNode(func.value, CompositeEDenotation(Example(arg.dnode.edenotation, arg.dnode.edenotation) :: Nil))
          case _ =>
        }
      }*/

      func.updateDNodesDown(params)
      arg.updateDNodesDown(params)
      dnode = DCompositeNode(routers, func.dnode, arg.dnode)
    }
  }
  def updateParentsDown(newParent:Option[MCompositeNode]) = {
    // Assume if parent is up to date, then all descendents are also valid
    newParent match {
      case None => // Top-level
        parent = MNode.invalid
        func.updateParentsDown(Some(this))
        arg.updateParentsDown(Some(this))
      case Some(newParent) =>
        if (!(parent eq newParent)) {
          parent = newParent
          func.updateParentsDown(Some(this))
          arg.updateParentsDown(Some(this))
        }
    }
  }

  // Return the i-th node in a DFS traversal of this tree
  def getInDFSOrder(i:Int) : MNode = {
    require (0 <= i && i < key.numNodes)
    if (i == 0) this
    else if (i < 1+func.key.numNodes) func.getInDFSOrder(i-1)
    else arg.getInDFSOrder(i-func.key.numNodes-1)
  }

  override def repCheck = {
    super.repCheck
    require (func.parent eq this)
    require (arg.parent eq this)
    key match {
      case key:KCompositeNode =>
        require (routers == key.routers)
        require (func.key eq key.func)
        require (arg.key eq key.arg)
      case _ => require (false)
    }
  }
}

