package programinduction

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,fails}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}
import scala.reflect.Manifest

import StandardLibrary._
import Router.{B,C,S}
import programinduction.{MPrimitiveNode => __}
import programinduction.{MCompositeNode => ___}
import MyUtils._

class Params(worldOpts:WorldOptions) {
  val N = new HashMap[Type, Int] // t -> Number of customers in restaurant type
  val dishes = new HashMap[KNode, KNode] // c -> information about the dish
  val edenotations = new HashMap[EDenotation, EDenotationInfo]

  var alpha = Double.NaN
  var typeStopProb = Double.NaN
  var numBaseTypes = -1
  var primitives : HashMap[Type, ArrayBuffer[Combinator]] = null
  var rogueProbFunc : (Combinator => Double) = null

  def write(path:String) = {
    val map = new CounterMap[KNode]
    dishes.keysIterator.foreach { node =>
      if (node.n > 0) map.incr(node, node.n)
    }
    Utils.writeLines(path, { puts:(String=>Any) =>
      map.write(Integer.MAX_VALUE, true, "", puts)
    })
  }

  // For every edenotation, infer a posterior distribution over candidates (used to initialize)
  def estimateCrudeEDEnotationDistribution = track("estimateCrudeEDEnotationDistribution") {
    // One giant distribution over combinators trees
    var crudeMap = new CounterMap[KNode]

    val set = new HashSet[KNode]
    edenotations.valuesIterator.foreach { info =>
      info.candidateNodes.foreach { node =>
        set += node.toKNode
      }
    }
    set.foreach { node => crudeMap.incr(node, 1) }

    foreach(10, { iter:Int =>
      val newCrudeMap = new CounterMap[KNode]
      // E-step: for each edenotation, estimate posterior distribution
      edenotations.valuesIterator.foreach { info =>
        val nodes = info.candidateNodes.map(_.toKNode).toArray
        val probs = nodes.map { node => crudeMap.getProb(node) }
        Utils.normalizeOrFail_!(probs)
        info.probs = probs
        (nodes zip probs).foreach { pair => val (node, prob) = pair
          newCrudeMap.incr(node, prob)
        }
      }

      // M-step
      crudeMap = newCrudeMap 
    })

    crudeMap.write(100, true, "", logs("%s", _))

    crudeMap
  }

  def changeN(myType:Type, d:Int) = { N(myType) = N.getOrElse(myType, 0) + d }

  def removeDuplicateEDenotationCandidates = {
    edenotations.valuesIterator.foreach { info =>
      val set = new HashSet[DNode]
      //info.candidateNodes.foreach { node => if (node.isInstanceOf[DCompositeNode]) set += node } // TMP
      info.candidateNodes.foreach(set += _)
      info.candidateNodes.clear
      set.foreach(info.candidateNodes += _)
    }
  }

  def pruneSingletonEDenotationCandidates = {
    edenotations.keysIterator.foreach { edenotation =>
      if (edenotations(edenotation).candidateNodes.size <= 1)
        edenotations -= edenotation
    }
  }

  // p(router | arity): Assume uniform over over router length and uniform over routing vectors of that length
  // This encourages shorter routers.
  val routerProbData = map(10, { arity:Int =>
    map(arity+1, { routerLen:Int => // Uniform over router length 0..arity (inclusive)
      1.0/((arity+1)*Math.pow(3, routerLen))
    })
  })
  def routerProb(node:KNode) = {
    val arity = node.myType.arity
    val routerLen = node match {
      case node:KPrimitiveNode => 0
      case node:KCompositeNode => node.routers.size
    }
    routerProbData(arity)(routerLen)
  }
  def intermediateTypeProb(myType:Type) : Double = {
    myType match {
      case myType:BaseType => worldOpts.typeStopProb * 1.0/numBaseTypes
      case myType:FuncType => (1-worldOpts.typeStopProb) * intermediateTypeProb(myType.inType) * intermediateTypeProb(myType.outType)
    }
  }

  // Local potential psi 
  def localPotential(node:KNode) = {
    node match {  
      case node:KPrimitiveNode =>
        val (rogue_?, numPrimitives) = primitives.get(node.myType) match {
          case None => (true, 0)
          case Some(l) => (!l.contains(node.value), l.size)
        }
        if (rogue_?) {
          worldOpts.roguePrimitiveProb * rogueProbFunc(node.value)
        } else {
          // For the identity primitive which only exists for types of the form t -> t
          val numExtraPrimitives = node.myType match {
            case FuncType(inType, outType) => if (inType == outType) 1 else 0
            case _ => 0
          }
          if (numPrimitives+numExtraPrimitives == 0) throw fails("No primitives registered with type %s, but we got %s", node.myType, node)
          (1-worldOpts.roguePrimitiveProb) * 1.0/(numPrimitives+numExtraPrimitives)
        }
      case node:KCompositeNode =>
        routerProb(node) * intermediateTypeProb(node.myType)
    }
  }

  def canonicalize(node:KNode) = {
    dishes.getOrElseUpdate(node, {
      node.localPotential = localPotential(node)
      assertValid(node.localPotential)
      node
    })
  }

  def logLikelihood : Double = {
    import fig.basic.NumUtils.logGamma
    var sum = 0.0
    dishes.keysIterator.foreach { node =>
      if (node.n > 0)
        sum += Math.log(node.localPotential) + logGamma(node.n)
    }
    N.keysIterator.foreach { myType =>
      if (N(myType) > 0)
        sum -= logGamma(alpha+N(myType)) - logGamma(alpha)
    }
    sum/worldOpts.temperature
  }

  // Remove entries that have zero counts (do this from time to time)
  // Dangerous with multiple roots!
  /*def clearZeros = {
    dishes.keysIterator.foreach { node =>
      if (node.n == 0) dishes -= node
    }
    N.keysIterator.foreach { myType =>
      if (N(myType) == 0) N -= myType
    }
  }*/

  def initialize(primitives:HashMap[Type, ArrayBuffer[Combinator]], rogueProbFunc:(Combinator=>Double)) = {
    alpha = worldOpts.alpha 
    typeStopProb = worldOpts.typeStopProb
    this.primitives = primitives
    this.rogueProbFunc = rogueProbFunc
    numBaseTypes = primitives.keysIterator.filter(_.isInstanceOf[BaseType]).toList.size
    if (numBaseTypes == 0) numBaseTypes = 1
    //require (numBaseTypes > 0)
  }

  def repCheck = {
    val checkN = new HashMap[Type,Int]
    dishes.keysIterator.foreach { node =>
      require (dishes(node) eq node)
      checkN(node.myType) = checkN.getOrElse(node.myType, 0) + node.n
    }
    //dbgs("N: %s", N)
    //dbgs("checkN: %s", checkN)
    assertSameCounts(N, checkN)
  }

  def log = {
    logs("alpha = %s", alpha)
    Utils.track_printAll("Types") {
      N.keysIterator.foreach { myType =>
        logs("%s: N=%s", myType, N(myType))
      }
    }
    Utils.track_printAll("Dishes") {
      dishes.keysIterator.foreach { node =>
        logs("%s: n=%s, potential=%s", node, node.n, node.localPotential)
      }
    }
    logEDenotations
  }
  def logEDenotations = {
    Utils.track_printAll("EDenotations") {
      edenotations.keysIterator.foreach { edenotation =>
        Utils.track_printAll("%s", edenotation) {
          edenotations(edenotation).candidateNodes.foreach { node =>
            logs("%s", node)
          }
        }
      }
    }
  }
}
