package programinduction

import scala.collection.mutable.HashMap
import scala.reflect.Manifest

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,fails}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}
import MyUtils._

object StandardLibrary {
  //// Wrap Scala values and functions (no polymorphic functions allowed)
  // Arguments and return values will be wrapped by default.
  // This works for boring monomorphic types (Int, String, any class you wish to define),
  // but for functions and types with type parameters (Tuples), need to do two things:
  //   1) Pre-wrap those arguments and return values as a combinator.
  //   2) Override arg<n>Type or retType to specify the type of the combinator.
  // Only allow wrapped functions to take wrapped values (no other combinators allowed).
  // {02/13/10} Evaluation is really slow.  To make it faster:
  //  - Don't use getWrappedType; just directly call WrappedType
  //  - Call getValue/WrappedValue instead of fromCombinator/toCombinator
  private def getManifest[A](implicit m:Manifest[A]) = m
  private val combinatorManifest = getManifest[Combinator]
  private def isCombinator[A](m:Manifest[A]) = m <:< combinatorManifest

  private def getWrappedType[A](implicit m:Manifest[A]) = { // Check that we're not wraping combinators
    if (isCombinator(m)) throw new RuntimeException("Must specify the type of " + m.toString + " manually; refuse to wrap combinators")
    else WrappedType[A]
  }

  def toCombinator[A](value:A)(implicit m:Manifest[A]) : Combinator = {
    if (isCombinator(m)) value.asInstanceOf[Combinator] // Already a combinator, don't need to wrap it
    else {
      // Can't wrap these, because to do so, we would need to look into the type of A, which I don't know how to do
      if (m.toString.startsWith("scala.Tuple2") || m.toString.startsWith("scala.Function"))
        throw new RuntimeException("Need to manually wrap the argument as a combinator: "+m)
      WrappedValue[A](value)
    }
  }
  private def fromCombinator[A](arg:Combinator)(implicit m:Manifest[A]) : A = {
    if (isCombinator(m)) arg.asInstanceOf[A] // Already is combinator, make it the right type of combinator
    else arg.asInstanceOf[WrappedValue[A]].value // Assume it's a primitive combinator with some value to be unpacked
  }

  def getValue[A](c:Combinator) : A = c.asInstanceOf[WrappedValue[A]].value

  case class WrappedType[A](implicit m:Manifest[A]) extends BaseType {
    val name = {
      val s = m.toString
      val i = s.lastIndexOf('$')
      if (i == -1) s else s.substring(i+1)
    }
    override val hashCode = name.hashCode
    override def equals(that:Any) = {
      that match {
        case that:WrappedType[_] => this.name == that.name
        case _ => false
      }
    }
    /*override val hashCode = m.toString.hashCode // m.hashCode doesn't work! - but this is unreliable for inner classes
    override def equals(that:Any) = {
      that match {
        case that:WrappedType[_] => this.m == that.m
        case _ => false
      }
    }*/
    override def toString = name
  }

  case class WrappedValue[A](value:A)(implicit mA:Manifest[A]) extends Combinator {
    //val myType = getWrappedType[A]
    val myType = WrappedType[A]
    def call(args:List[Combinator], forceEval:Boolean) = {
      args match {
        case Nil => this
        case _ => throw badArgs(args)
      }
    }
    def valueString = {
      value match {
        case value:String => escape(value)
        case _ => value.toString
      }
    }
  }
  case class WrappedFunction1[A,R](name:String, func:Function1[A,R])(implicit mA:Manifest[A], mR:Manifest[R]) extends Combinator {
    def arg1Type : Type = getWrappedType[A]
    def retType : Type = getWrappedType[R]
    val myType : Type = arg1Type -: retType
    def call(args:List[Combinator], forceEval:Boolean) = {
      args match {
        //case a :: Nil => WrappedValue[R](func.apply(fromCombinator[A](a)))
        case a :: Nil => WrappedValue[R](func.apply(getValue[A](a)))
        case _ => throw badArgs(args)
      }
    }
    def valueString = name
  }
  case class WrappedFunction2[A,B,R](name:String, func:Function2[A,B,R])(implicit mA:Manifest[A], mB:Manifest[B], mR:Manifest[R]) extends Combinator {
    def arg1Type : Type = getWrappedType[A]
    def arg2Type : Type = getWrappedType[B]
    def retType : Type = getWrappedType[R]
    val myType : Type = arg1Type -: arg2Type -: retType
    def call(args:List[Combinator], forceEval:Boolean) = {
      args match {
        case a :: b :: Nil => WrappedValue[R](func.apply(getValue[A](a), getValue[B](b)))
        case _ => throw badArgs(args)
      }
    }
    def valueString = name
  }
  case class WrappedFunction3[A,B,C,R](name:String, func:Function3[A,B,C,R])(implicit mA:Manifest[A], mB:Manifest[B], mC:Manifest[C], mR:Manifest[R]) extends Combinator {
    def arg1Type : Type = getWrappedType[A]
    def arg2Type : Type = getWrappedType[B]
    def arg3Type : Type = getWrappedType[C]
    def retType : Type = getWrappedType[R]
    val myType : Type = arg1Type -: arg2Type -: arg3Type -: retType
    def call(args:List[Combinator], forceEval:Boolean) = {
      args match {
        case a :: b :: c :: Nil => WrappedValue[R](func.apply(getValue[A](a), getValue[B](b), getValue[C](c)))
        case _ => throw badArgs(args)
      }
    }
    def valueString = name
  }
  case class WrappedFunction4[A,B,C,D,R](name:String, func:Function4[A,B,C,D,R])(implicit mA:Manifest[A], mB:Manifest[B], mC:Manifest[C], mD:Manifest[D], mR:Manifest[R]) extends Combinator {
    def arg1Type : Type = getWrappedType[A]
    def arg2Type : Type = getWrappedType[B]
    def arg3Type : Type = getWrappedType[C]
    def arg4Type : Type = getWrappedType[D]
    def retType : Type = getWrappedType[R]
    val myType : Type = arg1Type -: arg2Type -: arg3Type -: arg4Type -: retType
    def call(args:List[Combinator], forceEval:Boolean) = {
      args match {
        case a :: b :: c :: d :: Nil => WrappedValue[R](func.apply(getValue[A](a), getValue[B](b), getValue[C](c), getValue[D](d)))
        case _ => throw badArgs(args)
      }
    }
    def valueString = name
  }
  case class WrappedFunction5[A,B,C,D,E,R](name:String, func:Function5[A,B,C,D,E,R])(implicit mA:Manifest[A], mB:Manifest[B], mC:Manifest[C], mD:Manifest[D], mE:Manifest[E], mR:Manifest[R]) extends Combinator {
    def arg1Type : Type = getWrappedType[A]
    def arg2Type : Type = getWrappedType[B]
    def arg3Type : Type = getWrappedType[C]
    def arg4Type : Type = getWrappedType[D]
    def arg5Type : Type = getWrappedType[E]
    def retType : Type = getWrappedType[R]
    val myType : Type = arg1Type -: arg2Type -: arg3Type -: arg4Type -: arg5Type -: retType
    def call(args:List[Combinator], forceEval:Boolean) = {
      args match {
        case a :: b :: c :: d :: e :: Nil => WrappedValue[R](func.apply(getValue[A](a), getValue[B](b), getValue[C](c), getValue[D](d), getValue[E](e)))
        case _ => throw badArgs(args)
      }
    }
    def valueString = name
  }

  // Handy functions
  def wrap[A,R](name:String, func:Function1[A,R])(implicit mA:Manifest[A], mR:Manifest[R]) = {
    WrappedFunction1(name, func)
  }
  def wrap[A,B,R](name:String, func:Function2[A,B,R])(implicit mA:Manifest[A], mB:Manifest[B], mR:Manifest[R]) = {
    WrappedFunction2(name, func)
  }
  def wrap[A,B,C,R](name:String, func:Function3[A,B,C,R])(implicit mA:Manifest[A], mB:Manifest[B], mC:Manifest[C], mR:Manifest[R]) = {
    WrappedFunction3(name, func)
  }
  def wrap[A,B,C,D,R](name:String, func:Function4[A,B,C,D,R])(implicit mA:Manifest[A], mB:Manifest[B], mC:Manifest[C], mD:Manifest[D], mR:Manifest[R]) = {
    WrappedFunction4(name, func)
  }
  def wrap[A,B,C,D,E,R](name:String, func:Function5[A,B,C,D,E,R])(implicit mA:Manifest[A], mB:Manifest[B], mC:Manifest[C], mD:Manifest[D], mE:Manifest[E], mR:Manifest[R]) = {
    WrappedFunction5(name, func)
  }

  val UnitType = WrappedType[Unit]
  type WrappedUnit = WrappedValue[Unit]
  val BooleanType = WrappedType[Boolean]
  type WrappedBoolean = WrappedValue[Boolean]
  val IntType = WrappedType[Int]
  type WrappedInt = WrappedValue[Int]
  val DoubleType = WrappedType[Double]
  type WrappedDouble = WrappedValue[Double]
  val StringType = WrappedType[String]
  type WrappedString = WrappedValue[String]
  val IntListType = WrappedType[List[Int]]
  type WrappedIntList = WrappedValue[List[Int]]

  // Identity: Id a = a
  case class Id(argType:Type) extends Combinator {
    val myType = argType -: argType
    def call(args:List[Combinator], forceEval:Boolean) = {
      args match {
        case a :: extraArgs => a.call(extraArgs, forceEval)
        case _ => throw badArgs(args)
      }
    }
    def valueString = "I"
  }
  object Id { // Allows me to write Id[Int]
    def apply[A](implicit mA:Manifest[A]) : Combinator = Id(getWrappedType(mA))
  }

  // Kill: Kill a b = a
  case class Kill(arg1Type:Type, arg2Type:Type) extends Combinator {
    val myType = arg1Type -: arg2Type -: arg1Type
    def call(args:List[Combinator], forceEval:Boolean) = {
      args match {
        case a :: b :: extraArgs => a.call(extraArgs, forceEval)
        case _ => throw badArgs(args)
      }
    }
    def valueString = "K"
  }

  // Conditional/loop
  case class If(retType:Type) extends Combinator {
    val myType = BooleanType -: retType -: retType -: retType
    def call(args:List[Combinator], forceEval:Boolean) = {
      args match {
        case ifExpr :: thenExpr :: elseExpr :: extraArgs =>
          (if (ifExpr.call(Nil, true).asInstanceOf[WrappedBoolean].value) thenExpr else elseExpr).call(extraArgs, forceEval)
        case _ => throw badArgs(args)
      }
    }
    val valueString = "if"
  }
  object If {
    def apply[A](implicit mA:Manifest[A]) : Combinator = If(getWrappedType(mA))
  }

  // While loop
  val maxWhileIters = 1000
  case class While(retType:Type) extends Combinator {
    val myType = (retType -: BooleanType) -: (retType -: retType) -: retType -: retType
    def call(args:List[Combinator], forceEval:Boolean) = {
      args match {
        case testExpr :: incrExpr :: initExpr :: extraArgs =>
          var expr = initExpr
          var iter = 0
          while (iter < maxWhileIters && testExpr.call(expr :: Nil, true).asInstanceOf[WrappedBoolean].value) {
            expr = incrExpr.call(expr :: Nil, forceEval)
            iter += 1
          }
          expr.call(extraArgs, forceEval)
        case _ => throw badArgs(args)
      }
    }
    val valueString = "while"
  }
  object While {
    def apply[A](implicit mA:Manifest[A]) : Combinator = While(getWrappedType(mA))
  }

  // Until loop (differs from While in just one ! on the test)
  val maxUntilIters = 1000
  case class Until(retType:Type) extends Combinator {
    val myType = (retType -: BooleanType) -: (retType -: retType) -: retType -: retType
    def call(args:List[Combinator], forceEval:Boolean) = {
      args match {
        case testExpr :: incrExpr :: initExpr :: extraArgs =>
          var expr = initExpr
          var iter = 0
          while (iter < maxUntilIters && !testExpr.call(expr :: Nil, true).asInstanceOf[WrappedBoolean].value) {
            expr = incrExpr.call(expr :: Nil, forceEval)
            iter += 1
          }
          expr.call(extraArgs, forceEval)
        case _ => throw badArgs(args)
      }
    }
    val valueString = "until"
  }
  object Until {
    def apply[A](implicit mA:Manifest[A]) : Combinator = Until(getWrappedType(mA))
  }

  // Fixed point combinator (only for functions)
  // Factorial example: 
  //  f g n = if (n == 0) 1 else n * g(n-1)
  //  fact = fix f
  // fact : int -> int
  // f : (int -> int) -> (int -> int)
  // fix : ((int -> int) -> (int -> int)) -> (int -> int)
  // fix needs to call f and pass (fix f) as g
  // Note that we only compute fixed points of functions.
  case class Fix(argType:Type, retType:Type) extends Combinator {
    val funcType = argType -: retType
    val myType = (funcType -: funcType) -: funcType
    def call(args:List[Combinator], forceEval:Boolean) = {
      args match {
        case funcExpr :: argExpr :: extraArgs => // funcExpr = f, argExpr = n
          val fixedPointExpr = new PartialCombinator(this, funcExpr :: Nil) // fixedPointExpr = g = fix f
          funcExpr.call(fixedPointExpr :: argExpr :: extraArgs, forceEval) // call f g n = f (fix f) n
        case _ => throw badArgs(args)
      }
    }
    val valueString = "fix"
  }
}

object ArithmeticLibrary {
  import StandardLibrary._

  // Don't want to make toCombinator implicit, because that's too fast and loose
  implicit def unitToCombinator(value:Unit) = toCombinator(value)
  implicit def boolToCombinator(value:Boolean) = toCombinator(value)
  implicit def intToCombinator(value:Int) = toCombinator(value)
  implicit def intListToCombinator(value:List[Int]) = toCombinator(value)

  // Primitives
  val Eq = wrap("=", {(x:Int,y:Int) => x==y})
  val Lt = wrap("<", {(x:Int,y:Int) => x<y})
  val Gt = wrap(">", {(x:Int,y:Int) => x>y})
  val Lteq = wrap("<=", {(x:Int,y:Int) => x<=y})
  val Gteq = wrap(">=", {(x:Int,y:Int) => x>=y})

  val Add = wrap("+", {(x:Int,y:Int) => x+y})
  val Sub = wrap("-", {(x:Int,y:Int) => x-y})
  val Mul = wrap("*", {(x:Int,y:Int) => x*y})
  val Div = wrap("/", {(x:Int,y:Int) => x/y})

  val Head = wrap("head", {l:List[Int] => l.head})
  val Tail = wrap("tail", {l:List[Int] => l.tail})
  // Don't crash because we're doing call by value
  //val Head = wrap("head", {l:List[Int] => if (l.size == 0) 0 else l.head})
  //val Tail = wrap("tail", {l:List[Int] => if (l.size == 0) Nil else l.tail})
  val Empty_? = wrap("empty?", {l:List[Int] => l.size == 0})
  val Length = wrap("length", {l:List[Int] => l.size})
  val Cons = wrap("cons", {(x:Int, xs:List[Int]) => x :: xs})

  // Test the functions
  def main(args:Array[String]) = {
    fig.basic.LogInfo.msPerLine = 0
    import BNode._
    def disp(c:BNode) = logs("%s ==> %s", c, c.call(Nil, true))

    // Simple
    disp(__(__(Add, 3), 4))
    disp(__("C", Sub, 3))
    disp(__(__("C", Sub, 3), 10))
    disp(__(__("S", Add, Id[Int]), 10))

    // Compute min/max
    val Min = __("CS", __("SC", __("BB", If[Int], Lt), Id[Int]), Id[Int])
    val Max = __("CS", __("SC", __("BB", If[Int], Gt), Id[Int]), Id[Int])
    disp(Min)
    disp(__(__(Min, 3), 5))
    disp(__(__(Min, 5), 3))

    // Make sure if/then is lazy
    disp(__(__(__(If[Int], true), 42), __(Head, Nil).setLazyEval))
    //disp(__(__(__(If[Int], true), 42), __(Head, Nil))) // Crashes

    // Build a list of integers
    disp(__(__(__(While[List[Int]], Nil),
               __("C", __("B", Lt, Length), 5)), // Test
            __("S", __("B", Cons, Length), Id[List[Int]]))) // Incr

    // List[Int] -> Int
    val ListSum = __(Fix(IntListType, IntType),
      __("BS", __("C", __("B", If[Int], Empty_?), 0),
               __("BS", __("B", Add, Head), __("CB", Id(IntListType -: IntType), Tail)).setLazyEval))
    disp(ListSum)
    disp(__(ListSum, Nil))
    disp(__(ListSum, List(1, 4, 5)))
  }
}
