package programinduction

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,fails}
import tea.Utils.{track,begin_track,end_track,logs,dbgs,warnings}

import java.util.{Set,ArrayList,Random}
import scala.reflect.Manifest

import StandardLibrary._
import Router.{B,C,S}
import programinduction.{MPrimitiveNode => __}
import programinduction.{MCompositeNode => ___}
import MyUtils._

import org.yaml.snakeyaml.Yaml

/**
Our data comes from SMARTedit.
For each task/scenario, we have set of test cases (their examples/iterations),
each of which specifies a mapping from state=(cursor position, contents of text buffer) to state.
*/

// caseNum (starts at 1)
case class State(caseNum:Int, contents:String, pos:Int) {
  require (validPos(pos))
  def log = Utils.track_printAll("State") {
    logs("%s__%s", contents.substring(0, pos), contents.substring(pos, contents.size))
  }

  lazy val coarseContents = State.coarsify(contents)

  def validPos(i:Int) = i >= 0 && i <= contents.size

  def move(newPos:Int) = if (validPos(newPos)) State(caseNum, contents, newPos) else this
  def insert(str:String) = State(caseNum, contents.substring(0, pos)+str+contents.substring(pos, contents.size), pos+str.size)
  def delete(newPos:Int) = if (validPos(newPos)) State(caseNum, contents.substring(0, pos)+contents.substring(newPos, contents.size), pos) else this
  def findBeginsWith(str:String) = contents.indexOf(str, pos)
  def findEndsWith(str:String) = contents.indexOf(str, pos)+str.size

  def coarseFindBeginsWith(coarseStr:String) = coarseContents.indexOf(coarseStr, pos)
  def coarseFindEndsWith(coarseStr:String) = coarseContents.indexOf(coarseStr, pos)+coarseStr.size

  def substring(i:Int, j:Int) = { // Return none if out of bounds
    if (!validPos(i) || !validPos(j) || i > j) None
    else Some(contents.substring(i, j))
  }

  private def sanitize(s:String) = s.replaceAll("\n", "\\\\n").replaceAll("\t", "\\\\t")
  val window = 15
  override def toString = {
    "\""+sanitize(contents.substring(0 max (pos-window), pos))+"["+pos+"]"+sanitize(contents.substring(pos, contents.size min (pos+window)))+"\""
  }
}
object State {
  def coarsify(s:String) = {
    val buf = new StringBuilder(s.size)
    foreach(s.size, { i:Int =>
      val c = s(i)
      buf += {
        if (Character.isWhitespace(c)) s(i)
        else if (Character.isLetter(c)) 'a'
        //else if (Character.isUpperCase(c)) 'A'
        //else if (Character.isLowerCase(c)) 'a'
        else if (Character.isDigit(c)) '0'
        else '!'
      }
    })
    buf.toString
  }
}

case class TextEditWorld(worldOpts:WorldOptions, textOpts:TextOptions) extends AbstractWorld {
  val StateType = WrappedType[State]

  case class TestCase(startState:State, endState:State) extends AbstractTestCase {
    val inputs = WrappedValue(startState) :: Nil
    val output = WrappedValue(endState)
    override def getLoss(predOutput:Combinator) = {
      val predState = predOutput.asInstanceOf[WrappedValue[State]].value
      if (predState.contents == endState.contents) {
        if (predState.pos != endState.pos) 0.01 else 0.0 // Tiny error for wrong position
      }
      else 1.0
    }
  }

  def allowCombination_?(func:KNode, arg:KNode) : Boolean = true

  // Very basic stuff
  val ConstInt = Kill(IntType, StateType)
  val ConstString = Kill(StringType, StateType)
  import ArithmeticLibrary.{Add,Sub}
  val Append = wrap("append", {(str1:String,str2:String) => str1+str2})

  // Actions
  val Move = wrap("move", {(pos:Int,state:State) => state.move(pos)})
  val Insert = wrap("insert", {(str:String,state:State) => state.insert(str)})
  val Delete = wrap("delete", {(pos:Int,state:State) => state.delete(pos)})

  // Fetching locations
  //val BeginWord = wrap("begin-word", {state:State => state.contents.indexOf(state.pos})
  val Pos = wrap("pos", {state:State => state.pos})
  val CaseNum = wrap("caseNum", {state:State => state.caseNum.toString})
  val Substring = wrap("substring", {(i:Int,j:Int,state:State) => state.substring(i, j).getOrElse("")})

  // Searching (only forward)
  val FindBeginsWith = wrap("find-begin", {(str:String,state:State) => state.findBeginsWith(str)})
  val FindEndsWith = wrap("find-end", {(str:String,state:State) => state.findEndsWith(str)})
  val CoarseFindBeginsWith = wrap("coarse-find-begin", {(coarseStr:String,state:State) => state.coarseFindBeginsWith(coarseStr)})
  val CoarseFindEndsWith = wrap("coarse-find-end", {(coarseStr:String,state:State) => state.coarseFindEndsWith(coarseStr)})

  def computeRogueProb(c:Combinator) = 1.0

  def populatePrimitives = {
    val primitives = new ArrayBuffer[Combinator]
    
    primitives ++= List(WrappedValue(""), WrappedValue(0))
    primitives ++= List(ConstInt, ConstString)
    primitives ++= List(Add, Sub, Append)
    primitives ++= List(Move, Insert, Delete)
    primitives ++= List(Pos, CaseNum, Substring)
    primitives ++= List(FindBeginsWith, FindEndsWith)
    primitives ++= List(CoarseFindBeginsWith, CoarseFindEndsWith)

    primitives.foreach(addPrimitive)
  }

  def populateScenarios = {
    import java.io._
    track("Loading scenarios from %s", textOpts.inPath) {
      val yaml = new Yaml
      var offset = 0
      val paths : Seq[String] = {
        if (new File(textOpts.inListPath).exists)
          Utils.readLines(textOpts.inListPath).flatMap { x =>
            var y = x.replaceAll("""\s*#.*$""", "")
            if (y.size == 0) None
            else Some(new File(textOpts.inListPath).getParent+"/"+y)
          }
        else
          new File(textOpts.inPath).list.toList.sortWith(_ < _).map(textOpts.inPath+"/"+_)
      }

      paths.foreach { path =>
        val name = new File(path).getName
        if (name.endsWith(".yaml")) {
          if (offset >= worldOpts.scenarioOffset && scenarios.size < worldOpts.maxScenarios) {
            track("%s", name) {
              val reader = fig.basic.IOUtils.openInHard(path)
              scenarios += extractScenario(name, yaml.load(reader))
              reader.close
            }
          }
          offset += 1
        }
      }
    }
  }

  def extractScenario(name:String, scenarioObj:AnyRef) = {
    import java.util.{LinkedHashMap,ArrayList}
    def extractState(caseNum:Int, obj:AnyRef) = {
      val map = obj.asInstanceOf[LinkedHashMap[Object,Object]]
      val contents = map.get("contents").asInstanceOf[String]
      val pos = map.get("pos").asInstanceOf[Int]
      State(caseNum, contents, pos)
    }
    val map = scenarioObj.asInstanceOf[LinkedHashMap[Object,Object]]
    var caseNum = 0
    val testCases = map.get("examples").asInstanceOf[ArrayList[Object]].toArray(new Array[Object](0)).map { exObj =>
      val map = exObj.asInstanceOf[LinkedHashMap[Object,Object]]
      caseNum += 1
      TestCase(extractState(caseNum, map.get("before")), extractState(caseNum, map.get("after")))
    }
    if (worldOpts.verbose >= 5) {
      foreach(testCases.size, { i:Int => val testCase = testCases(i)
        track("Test case") {
          if (i == 0) testCase.startState.log
          testCase.endState.log
        }
      })
    }
    Scenario(name, testCases)
  }

  case class Scenario(name:String, testCases:Seq[TestCase]) extends AbstractScenario {
    // Break up the test case into a series of operations to accomplish the task
    trait Op {
      def perform(state:State) : State
    }
    case class MoveOp(count:Int) extends Op {
      def perform(state:State) = state.move(state.pos+count)
      override def toString = "move("+count+")"
    }
    case class DelOp(count:Int) extends Op {
      def perform(state:State) = state.delete(state.pos+count)
      override def toString = "del("+count+")"
    }
    case class InsOp(str:String) extends Op {
      def perform(state:State) = state.insert(str)
      override def toString = "ins("+escape(str)+")"
    }

    def align(testCase:TestCase) : List[Op] = {
      val startState = testCase.startState
      val endState = testCase.endState
      var s1 = startState.contents
      var s2 = endState.contents
      def n1 = s1.size
      def n2 = s2.size

      // Ignore common prefix and suffix 
      var nstart = 0; while (nstart < s1.size && nstart < s2.size && s1(nstart) == s2(nstart)) nstart += 1
      var nend = 0; while (nend < s1.size-nstart && nend < s2.size-nstart && s1(s1.size-nend-1) == s2(s2.size-nend-1)) nend += 1
      // Give some leeway, because we might want to insert/delete first, not move until later
      nstart = (nstart-10) max 0
      nend = (nend-10) max 0
      s1 = s1.substring(nstart, s1.size-nend)
      s2 = s2.substring(nstart, s2.size-nend)

      // Use dynamic programming to find minimum cost alignment
      // Try to minimize the number of edits (incrCosts), but also minimize the number of switches between types of edits (initCosts)
      case class Step(d:Int, count:Int)
      val STEP_INS = 0
      val STEP_DEL = 1
      val STEP_MOVE = 2
      val N_STEP = 3
      val mincost = Array.ofDim[Int](n1+1, n2+1, N_STEP) // i1, i2, last_step -> minimum cost to convert s1[i1..-1] into s2[i2..-1]
      val next = Array.ofDim[Int](n1+1, n2+1, N_STEP)
      val initCosts = Array(4, 3, 2)
      val incrCosts = Array(3, 2, 1)

      Utils.foreachReverse(n1+1, { i1:Int =>
        Utils.foreachReverse(n2+1, { i2:Int =>
          foreach(N_STEP, { last_step:Int =>
            def cost(step:Int) = incrCosts(step) + { if (step != last_step) initCosts(step) else 0 }
            val (mc, nxt) = {
              if (i1 == n1 && i2 == n2) (0, 0)
              else {
                val moveCost = {
                  if (i1 < n1 && i2 < n2 && s1(i1) == s2(i2)) cost(STEP_MOVE) + mincost(i1+1)(i2+1)(STEP_MOVE)
                  else Integer.MAX_VALUE
                }
                val insertCost = {
                  if (i2 < n2) cost(STEP_INS) + mincost(i1)(i2+1)(STEP_INS)
                  else Integer.MAX_VALUE
                }
                val deleteCost = {
                  if (i1 < n1) cost(STEP_DEL) + mincost(i1+1)(i2)(STEP_DEL)
                  else Integer.MAX_VALUE
                }

                if (insertCost <= moveCost && insertCost <= deleteCost) // Prefer insert
                  (insertCost, STEP_INS)
                else if (deleteCost < moveCost) // Then delete
                  (deleteCost, STEP_DEL)
                else
                  (moveCost, STEP_MOVE)
              }
            }
            mincost(i1)(i2)(last_step) = mc
            next(i1)(i2)(last_step) = nxt
          })
        })
      })

      if (worldOpts.verbose >= 5) dbgs("ALIGN %s %s => %s", n1, n2, mincost(0)(0)(STEP_MOVE))
      def add(d:Int, steps:List[Step]) = { // Collapse duplicates
        steps match {
          case step :: rest_steps if step.d == d => Step(d, step.count+1) :: rest_steps
          case _ => Step(d, 1) :: steps
        }
      }
      def recurse(i1:Int, i2:Int, step:Int) : List[Step] = {
        if (i1 == n1 && i2 == n2) Nil
        else {
          next(i1)(i2)(step) match {
            case STEP_INS => add(STEP_INS, recurse(i1, i2+1, STEP_INS))
            case STEP_DEL => add(STEP_DEL, recurse(i1+1, i2, STEP_DEL))
            case STEP_MOVE => add(STEP_MOVE, recurse(i1+1, i2+1, STEP_MOVE))
            case _ => throw impossible
          }
        }
      }
      var steps = recurse(0, 0, STEP_MOVE)

      // Remove beginning and ending moves if they exist (later we'll replace them with ones from the cursor position)
      if (steps.head.d == STEP_MOVE) {
        s1 = s1.substring(steps.head.count, s1.size)
        s2 = s2.substring(steps.head.count, s2.size)
        nstart += steps.head.count
        steps = steps.tail
      }
      if (steps.last.d == STEP_MOVE) {
        s1 = s1.substring(0, s1.size-steps.last.count)
        s2 = s2.substring(0, s2.size-steps.last.count)
        nend += steps.last.count
        steps = steps.filterNot(_ eq steps.last)
      }

      // Convert steps into operations
      // Make sure there's a move between every insert/delete operation (even if we don't move the cursor),
      // e.g., Move Insert Move Insert Move Delete Move Delete Move
      var (i1, i2) = (0, 0)
      val ops = new ArrayBuffer[Op]
      require (nstart+nend+s1.size == startState.contents.size)
      require (nstart+nend+s2.size == endState.contents.size)
      ops += MoveOp(nstart-startState.pos)
      Utils.track_printAll("%s steps", steps.size) {
        steps.foreach { step =>
          if (!ops.last.isInstanceOf[MoveOp] && step.d != STEP_MOVE) ops += MoveOp(0)
          ops += (step.d match {
            case STEP_MOVE =>
              if (worldOpts.verbose >= 4) logs("MOVE %s at %s", escape(s1.substring(i1, i1+step.count)), nstart+i2)
              i1 += step.count
              i2 += step.count
              MoveOp(step.count)
            case STEP_DEL => MoveOp(step.count)
              if (worldOpts.verbose >= 4) logs("DELETE %s at %s", escape(s1.substring(i1, i1+step.count)), nstart+i2)
              i1 += step.count
              DelOp(step.count)
            case STEP_INS => MoveOp(step.count)
              if (worldOpts.verbose >= 4) logs("INSERT %s at %s", escape(s2.substring(i2, i2+step.count)), nstart+i2)
              i2 += step.count
              InsOp(s2.substring(i2-step.count, i2))
            case _ => throw impossible
          })
        }
      }
      require (i1 == n1 && i2 == n2)
      require (endState.contents.size == nstart+n2+nend)
      //dbgs("%s %s", nstart+n2, endState.pos)
      ops += MoveOp(endState.pos-(nstart+n2))
      ops.toList
    }

    // Turn a function _node_ of type currState -> state
    // into one of type startState -> currState -> state,
    // by ignoring startState
    def ignoreInitState(node:MNode) = {
      if (textOpts.accessToStartState)
        ___(B::C::Nil, ___(B::Nil, __(Kill(node.myType.finalOutType, StateType)), node), __(Id(StateType)))
      else
        node
    }
    def ignoreCurrState(node:MNode) = {
      if (textOpts.accessToStartState)
        ___(B::Nil, __(Kill(node.myType.finalOutType, StateType)), node)
      else
        node
    }

    // currState -> nextState
    def createSimplePosGetter(states:List[State], positions:List[Int]) : MNode = {
      new SubProblem(states.map(_ :: Nil), positions) {
        override def removeHardCases = false
        def candidates = {
          // Generate candidates that might get the first test case right.
          // Then filter based on whether they get all test cases right.
          val state = states.head
          val pos = positions.head

          val nodes = new ArrayBuffer[MNode]
          nodes += ___(Nil, __(ConstInt), __(WrappedValue(pos))) // Absolute position
          nodes += ___(B::Nil, ___(Nil, __(Add), __(WrappedValue(pos-state.pos))), __(Pos)) // Character offset

          // Searching for a string
          Array((-1, 0), (0, +1)).foreach { dirPair => // -1: ends-with: +1: starts-with
            val (d1, d2) = dirPair
            // Search for a string of a certain length in the given direction
            foreach(1, 8+1, { len:Int =>
              state.substring(pos+d1*len, pos+d2*len) match {
                case Some(str) =>
                  nodes += ___(Nil, __(if (d1 == -1) FindEndsWith else FindBeginsWith), __(WrappedValue(str))) // Verbatim 
                  nodes += ___(Nil, __(if (d1 == -1) CoarseFindEndsWith else CoarseFindBeginsWith), __(WrappedValue(State.coarsify(str)))) // Search for coarse
                case None =>
              }
            })
          }

          nodes.toList
        }
      }.defaultCandidate
    }

    // startState -> currState -> nextState (but ignore startState)
    def createPosGetter(states:List[List[State]], positions:List[Int]) : MNode = {
      val node = createSimplePosGetter(states.map(_.last), positions)
      ignoreInitState(node)
    }

    def createStringGetter(states:List[List[State]], strs:List[String]) : MNode = {
      val inputs = {
        if (textOpts.accessToStartState) states
        else states.map(_.last :: Nil)
      }
      new SubProblem(inputs, strs) {
        override def removeHardCases = false
        def candidates = {
          val List(startState, state) = states.head
          val str = strs.head
          val caseNum = state.caseNum.toString

          val nodes = new ArrayBuffer[MNode]
          nodes += ignoreInitState(___(Nil, __(ConstString), __(WrappedValue(str)))) // Verbatim string

          // Case num if it exists, insert stuff before and stuff after is another possibility
          str.indexOf(caseNum) match {
            case -1 =>
            case i =>
              nodes += ignoreInitState(___(C::Nil,
                ___(B::Nil, __(Append), ___(B::Nil, ___(Nil, __(Append), __(WrappedValue(str.substring(0, i)))), __(CaseNum))),
                __(WrappedValue(str.substring(i+caseNum.size, str.size)))))
          }

          // Try to find an occurrence of the string in the initial state
          // So we need to find a function that returns the left position and one that returns the right position
          if (textOpts.accessToStartState) {
            val startStates = states.map(_.head)
            val leftPositions = (startStates zip strs).map { pair =>
              val (startState, str) = pair
              startState.contents.indexOf(str, startState.pos)
            }
            if (leftPositions.forall(_ != -1)) {
              val rightPositions = (leftPositions zip strs).map { pair =>
                val (leftPos, str) = pair
                leftPos + str.size
              }
              val leftNode = createSimplePosGetter(startStates, leftPositions)
              val rightNode = createSimplePosGetter(startStates, rightPositions)
              nodes += ignoreCurrState(___(S::Nil, ___(S::Nil, ___(B::Nil, __(Substring), leftNode), rightNode), __(Id(StateType))))
            }
          }

          nodes.toList
        }
      }.defaultCandidate
    }

    def createInitOpNode(states:List[List[State]], ops:List[Op]) : MNode = {
      val currStates = states.map(_.last)
      if (worldOpts.verbose >= 3) Utils.track_printAll("createInitOpNode") {
        (currStates zip ops).foreach { pair =>
          val (state,op) = pair
          logs("%s : %s => %s", op, state, op.perform(state))
        }
      }
      require (ops.forall(_.getClass == ops.head.getClass), ops)
      def create(c:Combinator, node:MNode) = {
        if (textOpts.accessToStartState)
          ___(C::S::Nil, ___(B::B::Nil, __(c), node), __(Id(StateType))) // startState -> currState -> state
        else
          ___(S::Nil, ___(B::Nil, __(c), node), __(Id(StateType))) // currState -> state
      }
      ops.head match {
        case _:InsOp =>
          val insOps = ops.map(_.asInstanceOf[InsOp])
          val node = createStringGetter(states, insOps.map(_.str))
          create(Insert, node)
        case _:DelOp =>
          val delOps = ops.map(_.asInstanceOf[DelOp])
          val node = createPosGetter(states, mapList2(currStates, delOps, {(state:State,op:DelOp) => state.pos + op.count}))
          create(Delete, node)
        case _:MoveOp =>
          val moveOps = ops.map(_.asInstanceOf[MoveOp])
          val node = createPosGetter(states, mapList2(currStates, moveOps, {(state:State,op:MoveOp) => state.pos + op.count}))
          create(Move, node)
      }
    }

    def createInitRootNodes = {
      val opsList = {
        val opsList = track("Aligning") { trainTestCases.map(align) } // testcase -> [Op]
        // Assume we have the same number of operations (if not, then throw out test cases; ASSUME this isn't too bad)
        val enabled = mapList(opsList.size, { i:Int => opsList(i).size == opsList.head.size })
        foreach(trainTestCases.size, { i:Int => trainTestCases(i).enabled = enabled(i) })
        val newOpsList = (opsList zip enabled).filter(_._2).map(_._1)
        if (opsList.size != newOpsList.size) {
          warnings("Filtered out %s/%s test cases which didn't have exactly %s operations", opsList.size-newOpsList.size, opsList.size, opsList.head.size)
        }
        newOpsList
      }
      val numOps = opsList.head.size
      val C = opsList.size

      // For each operation, create a sequence of intermediate states (for each test case)
      val startStates = enabledTrainTestCases.map(_.startState)

      // Print out the intended behavior for all test cases
      if (worldOpts.verbose >= 4) Utils.track_printAll("Cases") {
        Utils.track_foreach("Case", C, { i:Int =>
          var state = startStates(i)
          opsList(i).foreach { op =>
            val newState = op.perform(state)
            logs("%s : %s => %s", op, state, newState)
            state = newState
          }
          require (state == enabledTrainTestCases(i).endState, "Doesn't match: "+state+" AND "+enabledTrainTestCases(i).endState)
          true
        })
      }

      // Try to construct a program that achieves the intended behavior
      var currStates = startStates
      var root : MNode = __(Id(StateType))
      foreach(numOps, { i:Int => // For each operation...
        if (worldOpts.verbose >= 2) begin_track("Op %s/%s", i, numOps)
        if (worldOpts.verbose >= 2) logs("%s ...", opsList.head(i))
        val opNode = createInitOpNode((startStates zip currStates).map { pair => pair._1 :: pair._2 :: Nil }, opsList.map(_(i)))

        val newStates : List[State] = (startStates zip currStates).map { pair =>
          val (startState, state) = pair
          val inputs = {
            if (textOpts.accessToStartState) WrappedValue(startState)::WrappedValue(state)::Nil
            else WrappedValue(state)::Nil
          }
          opNode.call(inputs, true) match {
            case WrappedValue(value:State) => value
            case x => throw fails("Returned %s, wanted a value", x)
          }
        }

        //if (worldOpts.verbose >= 4)
          //logs("Found op %s : %s", opNode.valueString, mapList2(currStates, newStates, { (state:State, newState:State) => state+"=>"+newState }).mkString(" ; "))
        //if (i == numOps-1)
          //(newStates zip enabledTrainTestCases.map(_.endState)).foreach { pair => require (pair._1 == pair._2, opNode + " doesn't work: " + pair) }
        currStates = newStates
        root = {
          if (textOpts.accessToStartState) ___(S::Nil, opNode, root)
          else ___(B::Nil, opNode, root)
        }
        if (worldOpts.verbose >= 2) end_track
      })

      roots += root
    }
  }
}

class TextOptions {
  import fig.basic.Option
  @Option(gloss="Path to yaml files") var inPath = ""
  @Option(gloss="Path to a file with yaml file paths") var inListPath = ""
  @Option(gloss="Whether operations have access to start state (allows for copy/paste)") var accessToStartState = false
}

class TextEditTestRunner extends Runnable {
  import fig.basic.OptionSet
  @OptionSet(name="world") val worldOpts = new WorldOptions
  @OptionSet(name="text") val textOpts = new TextOptions

  def run = {
    val errors = new fig.basic.FullStatFig
    Utils.track_record_foreach("Trial", worldOpts.numTrials, { trial:Int =>
      val world = new TextEditWorld(worldOpts, textOpts)
      world.initialize
      world.learn
      val error = world.getTestError
      logs("Trial error = %s", error)
      errors.add(error)
      Execution.putOutput("minTrialError", errors.getMin)
      Execution.putOutput("meanTrialError", errors.mean)
      Execution.putOutput("maxTrialError", errors.getMax)
      true
    })
    logs("Error over all trials: %s", errors)
  }
}

object TextEditTest {
  def main(args:Array[String]) = Execution.run(args, "main", new TextEditTestRunner)
}
