package programinduction

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,fails}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}
import scala.reflect.Manifest

import StandardLibrary._
import Router.{B,C,S}
import programinduction.{MPrimitiveNode => __}
import programinduction.{MCompositeNode => ___}
import MyUtils._

// We'd like to specify transformations in a simple declarative way.
// The issue is that with a simple transformation, the involved could contain
// a prefix of various routers which apply before the transformation and also
// a suffix of routers.
// We compile the logic needed to deal with matching transformations (augmented with other routers).
// A router is an element of {B,C,S}, denoted r.
// A routing vector is denoted rv or routers : List[Router].
// A configuration is a list of routing vectors, one for each QCompositeNode, denoted, rvs.
abstract class TransformTemplate(worldOpts:WorldOptions) {
  type RouterVec = List[Router]
  type Config = List[RouterVec]

  // A router expression specifies a list of routers.
  trait RouterExpr
  case class LitRouterExpr(router:Router) extends RouterExpr {
    override def toString = router.toString
  }
  case class VarRouterExpr(name:String) extends RouterExpr { // Wildcard
    override def toString = name
  }
  implicit def toRouterExpr(router:Router) = LitRouterExpr(router)

  trait QNode {
    def numCompositeNodes : Int
  }
  case class QPrimitiveNode(name:String) extends QNode {
    val numCompositeNodes = 0
    override def toString = name
  }
  case class QCompositeNode(routers:List[RouterExpr], func:QNode, arg:QNode) extends QNode {
    val numCompositeNodes = 1 + func.numCompositeNodes + arg.numCompositeNodes
    override def toString = "("+(if (routers==Nil) "" else routers.mkString(",")+" ")+func+" "+arg+")"
  }

  // Use these handy functions to define source and target
  val vx = QPrimitiveNode("x")
  val vy = QPrimitiveNode("y")
  val vz = QPrimitiveNode("z")
  val vI = QPrimitiveNode("I") // Special node which cannot receive any arguments
  val vr = VarRouterExpr("r")
  def Q(routers:List[RouterExpr], func:QNode, arg:QNode) = QCompositeNode(routers, func, arg)

  def source : QNode
  def target : QNode

  private val transformMap = new HashMap[Config, ArrayBuffer[Config]]
  private val untransformMap = new HashMap[Config, ArrayBuffer[Config]]
  def transform(rvs:Config) = sample(transformMap, untransformMap, rvs)
  def untransform(rvs:Config) = sample(untransformMap, transformMap, rvs)
  // Return a sample choiced from map and the bias [p(rvs -> new_rvs)/p(new_rvs -> rvs)]
  def sample(map:HashMap[Config,ArrayBuffer[Config]], rev_map:HashMap[Config,ArrayBuffer[Config]], rvs:Config) = {
    map.get(rvs) match {
      case None => None
      case Some(choices) =>
        val new_rvs = choices(worldOpts.random.nextInt(choices.size))
        val rev_choices = rev_map(new_rvs)
        Some((new_rvs, 1.0*rev_choices.size/choices.size))
    }
  }

  //// Build the transform maps: this is very brute force and inefficient.
  private def construct = {
    val all_rvs = {
      val all_rvs = new ArrayBuffer[RouterVec]
      def recurse(rv:RouterVec) : Unit = {
        all_rvs += rv
        if (rv.size < worldOpts.transformMaxLen) {
          recurse(B::rv)
          recurse(C::rv)
          recurse(S::rv)
        }
      }
      recurse(Nil)
      all_rvs
    }
    //dbgs("%s", all_rvs.mkString(" "))

    // Strategy: we try prefixing all QCompositeNode with routers.
    // This induces some behavior (specifies which arguments get routed to which leaf nodes).
    type Behavior = HashMap[QPrimitiveNode, List[Int]]
    class BadBehavior extends Exception

    // return a subset of _args_ which contains args(i) if guide.contains(rv(i))
    def filterArgs(rv:RouterVec, guide:List[Router], args:List[Int]) : List[Int] = {
      if (rv.size != args.size) throw new BadBehavior
      rv match {
        case Nil => Nil
        case r :: rest_rv =>
          if (guide.contains(r)) args.head :: filterArgs(rest_rv, guide, args.tail)
          else                                filterArgs(rest_rv, guide, args.tail)
      }
    }

    // If behavior is bad (e.g., routes different stuff to two leaf nodes with the same name), return None
    // rvs: for each composite node in _node_ (in DFS order), rvs specifies the routers to use
    def getBehavior(node:QNode, rvs:List[RouterVec]) : Behavior = {
      val behavior = new Behavior // Record what arguments end up in each leaf node
      // Return the router vectors not used up by subtree rooted at _node_
      def recurse(node:QNode, rvs:List[RouterVec], args:List[Int]) : List[RouterVec] = {
        //dbgs("%s %s %s", node, rvs, args)
        //require (node.numCompositeNodes == rvs.size, node.numCompositeNodes + " " + rvs)
        node match {
          case node:QPrimitiveNode =>
            behavior.get(node) match {
              case None =>
                if (node == vI) {
                  // Don't record this
                  if (args != Nil) throw new BadBehavior // Identity node cannot receive variables (see CI-moves)
                }
                else
                  behavior(node) = args // Record what args this QPrimitiveNode got
              case Some(existingArgs) =>
                if (args != existingArgs) throw new BadBehavior // If another QPrimitiveNode got different args, bad
            }
            rvs
          case node:QCompositeNode =>
            recurse(node.arg,
              recurse(node.func, rvs.tail, filterArgs(rvs.head, C::S::Nil, args)), 
              filterArgs(rvs.head, B::S::Nil, args))
        }
      }
      recurse(node, rvs, (1 to rvs.head.size).toList)
      behavior
    }

    // For each node, prefix it with a router vector, and see what the behavior is.
    def buildMap(node:QNode) = {
      val map = new HashMap[Behavior, List[RouterVec]]
      val K = node.numCompositeNodes
      
      // Try prefix all possible router vectors for all possible composite nodes.
      def recurse(rvs:List[RouterVec]) : Unit = {
        if (rvs.size == K) {
          try {
            val behavior = getBehavior(node, rvs)
            if (map.contains(behavior))
              throw fails("For %s, two ways of getting same behavior (%s): %s AND %s", node, behavior, map(behavior), rvs)
            map(behavior) = rvs
          } catch {
            case e:BadBehavior => // Do nothing
          }
        }
        else {
          all_rvs.foreach { rv => recurse(rv :: rvs) }
        }
      }
      recurse(Nil)
      map
    }

    // Get behaviors for source and target
    val sourceMap = buildMap(source)
    val targetMap = buildMap(target)
    require (sourceMap.size == targetMap.size)

    // Get a list of all possible variable routers (vrs).
    val vr_set = new HashSet[VarRouterExpr]
    def findVarRouters(node:QNode) : Unit = {
      node match {
        case node:QCompositeNode =>
          node.routers.foreach { vr =>
            vr match {
              case vr:VarRouterExpr => vr_set += vr
              case _ =>
            }
          }
          findVarRouters(node.func)
          findVarRouters(node.arg)
        case _ =>
      }
    }
    findVarRouters(source)
    findVarRouters(target)
    val vrs = vr_set.toList

    // For each variable router, need to map it to some router
    val vr_map = new HashMap[VarRouterExpr, RouterVec]
    def addTransforms = {
      def finalConstruct(node:QNode, rvs:List[RouterVec]) = {
        val final_rvs = new ArrayBuffer[RouterVec]
        def recurse(node:QNode, rvs:List[RouterVec]) : List[RouterVec] = { // Return remaining rvs
          node match {
            case node:QPrimitiveNode => rvs
            case node:QCompositeNode =>
              final_rvs += rvs.head ++ node.routers.flatMap { r:RouterExpr =>
                r match {
                  case r:LitRouterExpr => List(r.router)
                  case r:VarRouterExpr => vr_map(r)
                }
              }
              recurse(node.arg, recurse(node.func, rvs.tail))
          }
        }
        recurse(node, rvs)
        require (final_rvs.size == rvs.size)
        final_rvs.toList
      }
      sourceMap.keysIterator.foreach { behavior =>
        val s = finalConstruct(source, sourceMap(behavior))
        val t = finalConstruct(target, targetMap(behavior))
        if (worldOpts.verbose >= 6) dbgs("%s %s", s, t)
        append(transformMap, s, t)
        append(untransformMap, t, s)
      }
    }
    // Try all possible router vectors for the variable routers
    def recurse(vrs:List[VarRouterExpr]) : Unit = {
      vrs match {
        case Nil =>
          addTransforms  
        case vr :: rest_vrs =>
          all_rvs.foreach { rv =>
            vr_map(vr) = rv
            recurse(rest_vrs)
          }
      }
    }
    recurse(vrs)
  }
  
  construct
  logs("%s (%s) => %s (%s)", source, transformMap.size, target, untransformMap.size)
}
