package textedit

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,fails}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}
import MyUtils._

////////////////////////////////////////////////////////////
// Routers
case class Router(goesFunc:Boolean, goesArg:Boolean) {
  override def toString = {
    if (goesFunc && goesArg) "S"
    else if (goesFunc) "C"
    else if (goesArg) "B"
    else throw impossible
  }
}
object Router {
  val B = Router(false, true)
  val C = Router(true, false)
  val S = Router(true, true)

  def parse(c:Char) = Router(c == 'S' || c == 'C', c == 'S' || c == 'B')

  // Example: s = "BBCS"
  def parse(s:String) : List[Router] = {
    if (s.size == 0) Nil
    else parse(s.charAt(0)) :: parse(s.substring(1))
  }
}

object Identity { override def toString = "I" }

// Represents a random variable that takes on values which are needed by the subtree.
// Also keep track of how many holes there are 
class ExtractInfo(val numHoles:Int, val values:List[Any]) {
  require (numHoles >= numValues)
  def any = numHoles != 0
  def numValues = values.size
  def add(that:ExtractInfo) = {
    new ExtractInfo(this.numHoles + that.numHoles, (this.values ++ that.values).distinct)
  }
  override def toString = fmts("%s{%s}", numHoles, values.map(escape).mkString(","))
}

////////////////////////////////////////////////////////////
// Key nodes
trait KNode {
}
case class KPrimitiveNode(value:Any) extends KNode {
  override def toString = escape(value)
}
case class KCompositeNode(routers:List[Router], func:KNode, arg:KNode) extends KNode {
  override def toString = {
    "(" +
      (if (routers == Nil) "" else routers.mkString("") + " ") +
      func + " " + arg +
    ")"
  }
}

////////////////////////////////////////////////////////////

trait Func {
  def name : String
  def requiresState : Boolean // Whether this function is mutable; if so, first argument is the state
  def call(state:Any, args:Array[Any]) : Any
  override def toString = name
}

trait Node {
  def requiresState : Boolean
  var parent : Node = null // These are set once by the parent and not changed

  var extracted : ExtractInfo = null // What values were extracted out
  var key : KNode = null // Hash for the key

  // Only compute if invalid
  def computeExtracted : Unit
  def computeKey : Unit

  def invalidateUp : Unit = {
    extracted = null
    key = null
    if (parent != null) parent.invalidateUp
  }

  def call(state:Any) : Any
  def initialize(initRandom:Random, WO:WorldOptions) : Unit
  def computeCardinality : Double
  def foreachNode(includeUnchosen:Boolean, f:Node=>Any) : Unit
}

// Hidden variable: which candidate we've chosen
// Acts like the chosen node
case class NodeSet(candidates:Array[Node]) extends Node {
  require (candidates.size > 0)
  candidates.foreach(_.parent = this)

  var choice = 0 // Which candidate we've selected (filled in later)

  def initialize(initRandom:Random, WO:WorldOptions) = {
    choice = initRandom.nextInt(candidates.size)
    candidates.foreach(_.initialize(initRandom, WO))
  }

  def theNode = candidates(choice)

  // Act like the chosen node
  val requiresState = theNode.requiresState
  def computeExtracted = if (extracted == null) { theNode.computeExtracted; extracted = theNode.extracted }
  def computeKey = if (key == null) { candidates(choice).computeKey; key = theNode.key }
  def call(state:Any) = candidates(choice).call(state)

  override def toString = candidates(choice).toString

  def computeCardinality = {
    (0.0 /: candidates)(_ + _.computeCardinality)
  }

  def foreachNode(includeUnchosen:Boolean, f:Node=>Any) = {
    f(this)
    if (includeUnchosen) candidates.foreach(_.foreachNode(includeUnchosen, f)) // Follow all paths
    else candidates(choice).foreachNode(includeUnchosen, f) // Only follow chosen one
  }
}

// Hidden variable: whether this value is extrated or not
case class ValueNode(value:Any) extends Node { // Leaf node
  var canExtract = false
  var hole = false
  val requiresState = false
  def computeExtracted = { extracted = if (hole) new ExtractInfo(1, value :: Nil) else new ExtractInfo(0, Nil) }
  def computeKey = if (key == null) { key = KPrimitiveNode(if (hole) Identity else value) }
  def call(state:Any) = value
  def initialize(initRandom:Random, WO:WorldOptions) = {
    canExtract = !WO.onlyExtractStrings || value.isInstanceOf[String]
    hole = canExtract && initRandom.nextDouble < WO.initFracExtracted
  }
  def computeCardinality : Double = 2 // Either extract or not extract
  def foreachNode(includeUnchosen:Boolean, f:Node=>Any) = f(this)

  override def toString = {
    val s = escape(value)
    if (hole) "["+s+"]" else s
  }
}
case class ExprNode(func:Func, args:Array[NodeSet]) extends Node {
  args.foreach(_.parent = this)

  val prefixRequiresState = {
    val values = new Array[Boolean](args.size+1)
    values(0) = func.requiresState
    foreach(args.size, { i:Int =>
      values(i+1) = args(i).requiresState || values(i)
    })
    values
  }
  val requiresState = prefixRequiresState.last

  val prefixExtracted = new Array[ExtractInfo](args.size+1)
  def computeExtracted = if (extracted == null) {
    args.foreach(_.computeExtracted)
    prefixExtracted(0) = new ExtractInfo(0, Nil)
    foreach(args.size, { i:Int =>
      prefixExtracted(i+1) = prefixExtracted(i) add args(i).extracted
    })
    extracted = prefixExtracted.last
  }

  def computeKey = if (key == null) {
    args.foreach(_.computeKey) 
    key = createKey(args.size)
  }
  def createKey(n:Int) : KNode = {
    if (n == 0) KPrimitiveNode(func)
    else {
      var routers : List[Router] = Nil
      if (prefixRequiresState(n)) {
        routers = {
          val goesLeft = prefixRequiresState(n-1)
          val goesRight = args(n-1).requiresState
          Router(goesLeft, goesRight)
        } :: routers
      }
      if (prefixExtracted(n).any) { // If we need an argument
        routers = {
          val goesLeft = prefixExtracted(n-1).any
          val goesRight = args(n-1).extracted.any
          Router(goesLeft, goesRight)
        } :: routers
      }
      KCompositeNode(routers, createKey(n-1), args(n-1).key)
    }
  }

  def call(state:Any) = func.call(state, args.map(_.call(state)))
  def initialize(initRandom:Random, WO:WorldOptions) = args.foreach(_.initialize(initRandom, WO))
  def computeCardinality = {
    (1.0 /: args)(_ * _.computeCardinality)
  }
  def foreachNode(includeUnchosen:Boolean, f:Node=>Any) = {
    f(this)
    args.foreach(_.foreachNode(includeUnchosen, f))
  }

  override def toString = {
    if (args.size == 0) "("+func+")"
    else fmts("(%s %s)", func, args.mkString(" "))
  }
}

trait GNode {
  def k : Int
}
case class GPrimitiveNode(k:Int) extends GNode
case class GCompositeNode(k:Int, left:GNode, right:GNode) extends GNode

// The semantics of the top node is defined by function composition of the statements
// Hidden variable: the hierarchical grouping
case class ProgramNode(statements:Array[NodeSet]) extends Node { // Composition of statements
  statements.foreach(_.parent = this)

  val requiresState = true
  require (statements.forall(_.requiresState))

  var grouping : GNode = null // Random variable

  def initialize(initRandom:Random, WO:WorldOptions) = {
    def initGrouping(i:Int, j:Int) : GNode = {
      if (i+1 == j) GPrimitiveNode(i)
      else {
        // Choose a random split
        val k = i+1+initRandom.nextInt(j-i-1)
        GCompositeNode(k, initGrouping(i, k), initGrouping(k, j))
      }
    }
    grouping = initGrouping(0, statements.size)
    statements.foreach(_.initialize(initRandom, WO))
  }

  val prefixExtracted = new Array[ExtractInfo](statements.size+1)
  val suffixExtracted = new Array[ExtractInfo](statements.size+1)
  def computeExtracted = if (extracted == null) {
    statements.foreach(_.computeExtracted)
    prefixExtracted(0) = new ExtractInfo(0, Nil)
    foreach(statements.size, { i:Int =>
      prefixExtracted(i+1) = prefixExtracted(i) add statements(i).extracted
    })
    suffixExtracted(statements.size) = new ExtractInfo(0, Nil)
    Utils.foreachReverse(statements.size, { i:Int =>
      suffixExtracted(i) = statements(i).extracted add suffixExtracted(i+1)
    })
    extracted = prefixExtracted.last
  }

  def computeKey = if (key == null) {
    statements.foreach(_.computeKey)
    key = createKey(grouping)
  }
  def createKey(g:GNode) : KNode = {
    g match {
      case GPrimitiveNode(k) => statements(k).key
      case GCompositeNode(k, left, right) =>
        // See if we need a router for inputs
        // Note that left and right are reversed when using combinators due to composition
        val goesLeft = suffixExtracted(k).any
        val goesRight = prefixExtracted(k).any
        val routers = if (goesLeft || goesRight) Router(goesLeft, goesRight)::Router.B::Nil else Router.B::Nil
        KCompositeNode(routers, createKey(right), createKey(left))
    }
  }

  def computeCardinality : Double = {
    (catalan(statements.size-1).toDouble /: statements)(_ * _.computeCardinality)
  }
  def foreachNode(includeUnchosen:Boolean, f:Node=>Any) = {
    f(this)
    statements.foreach(_.foreachNode(includeUnchosen, f))
  }

  def call(state:Any) = {
    var currState = state
    foreach(statements.size, { i:Int =>
      currState = statements(i).call(currState)
    })
    currState
  }

  def computeAll = { computeExtracted; computeKey }

  def log = output(logs("%s", _))
  def output(puts:String=>Any) = {
    //statements.foreach { s => puts(s.toString) }
    def recurse(ind:String, g:GNode) : Unit = {
      g match {
        case GPrimitiveNode(k) => puts(ind+statements(k))
        case GCompositeNode(k, left, right) =>
          puts(ind+"-")
          recurse(ind+" ", left)
          recurse(ind+" ", right)
      }
    }
    puts("Extracted: "+extracted)
    recurse("", grouping)
  }
  override def toString = fmts("(PROG %s)", statements.mkString(" "))
}
