package textedit

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,fails,unknownCase}
import tea.Utils.{track,begin_track,end_track,logs,dbgs,warnings}

import java.util.{Set,ArrayList,Random}
import MyUtils._

import org.yaml.snakeyaml.Yaml

/**
Text editing.
*/

trait Action {
  def actionStr : String
  def inState : State
  def outState : State
}
case class MoveAction(actionStr:String, inState:State, outState:State, pos:Int) extends Action
case class SelectAction(actionStr:String, inState:State, outState:State, selection:(Int,Int)) extends Action
case class PasteAction(actionStr:String, inState:State, outState:State) extends Action
case class CutAction(actionStr:String, inState:State, outState:State) extends Action
case class CopyAction(actionStr:String, inState:State, outState:State) extends Action
case class DeleteAction(actionStr:String, inState:State, outState:State, pos:Int) extends Action
case class InsertAction(actionStr:String, inState:State, outState:State, str:String) extends Action
case class DeleteSelectionAction(actionStr:String, inState:State, outState:State) extends Action

// caseNum (starts at 1)
case class State(caseNum:Int, pos:Int, selection:Option[(Int,Int)], clipboard:String, contents:String) {
  require (validPos(pos))
  def log = Utils.track_printAll("State") {
    logs("Clipboard: %s", escape(clipboard))
    var markers = (pos, "__") :: Nil
    selection match {
      case Some((i,j)) => markers = (i, "{{") :: (j, "}}") :: markers
      case _ =>
    }
    markers = markers.sortWith(_._1<_._1)
    val output = new StringBuilder
    var last = 0
    markers.foreach { pair =>
      output ++= contents.substring(last, pair._1)
      output ++= pair._2
      last = pair._1
    }
    output ++= contents.substring(last, contents.size)
    logs("%s", output)
  }

  lazy val coarseContents = State.coarsify(contents)

  def validPos(i:Int) = i >= 0 && i <= contents.size

  def getAction(actionStr:String) : Action = {
    val tokens = actionStr.split(" ", 2)
    tokens(0) match {
      case "move" =>
        val pos = tokens(1).toInt
        MoveAction(actionStr, this, this.move(pos), pos)
      case "select" =>
        val (i, j) = parseSelection(tokens(1))
        SelectAction(actionStr, this, this.select(i, j), (i, j))
      case "paste" => PasteAction(actionStr, this, this.paste)
      case "cut" => CutAction(actionStr, this, this.cut)
      case "copy" => CopyAction(actionStr, this, this.copy)
      case "delete" =>
        val pos = tokens(1).toInt
        DeleteAction(actionStr, this, this.delete(pos), pos)
      case "insert" => InsertAction(actionStr, this, this.insert(tokens(1)), tokens(1))
      case "deleteselection" => DeleteSelectionAction(actionStr, this, this.deleteselection)
      case _ => throw fails("Unknown command: %s", actionStr)
    }
  }

  def move(newPos:Int) = if (validPos(newPos)) State(caseNum, newPos, None, clipboard, contents) else this
  def select(i:Int, j:Int) = if (validPos(i) && validPos(j) && i < j) State(caseNum, pos, Some((i,j)), clipboard, contents) else this
  def paste = insert(clipboard)
  def cut = selection match {
    case Some(selection) => State(caseNum, selection._1, None, contents.substring(selection._1, selection._2), contents.substring(0, selection._1)+contents.substring(selection._2, contents.size))
    case None => this
  }
  def copy = selection match {
    case Some(selection) => State(caseNum, pos, Some(selection), contents.substring(selection._1, selection._2), contents)
    case None => this
  }
  def delete(newPos:Int) = if (validPos(newPos)) State(caseNum, pos min newPos, selection, clipboard, contents.substring(0, pos min newPos)+contents.substring(pos max newPos, contents.size)) else this
  def insert(str:String) = State(caseNum, pos+str.size, None, clipboard, contents.substring(0, pos)+str+contents.substring(pos, contents.size))
  def deleteselection = selection match {
    case Some(selection) => State(caseNum, selection._1, None, clipboard, contents.substring(0, selection._1)+contents.substring(selection._2, contents.size))
    case None => this
  }

  def find(excludeSelf:Boolean, f:Char=>Boolean) : Int = {
    var i = pos + (if (excludeSelf) 1 else 0)
    while (i < contents.size) {
      if (f(contents(i))) return i
      i += 1
    }
    -1
  }
  def beginOfWord = {
    val i = find(false, Character.isWhitespace(_))
    if (i == -1) -1
    else {
      val j = find(false, !Character.isWhitespace(_))
      j
    }
  }
  def findWhitespace = find(false, Character.isWhitespace(_))

  def endOfFile = contents.size

  // contents: string to search in
  // endsWith: whether to return the location where the matching string ends
  // reverse: whether to search backwards rather than forwards
  // excludeSelf: whether to allow a match at the current location
  // skip: number of matches to jump ahead
  // str: string to search for
  // pos: where we start searching
  def find(contents:String, endsWith:Boolean, reverse:Boolean, excludeSelf:Boolean, skip:Int, str:String, pos:Int) : Int = {
    val startPos = {
      (if (endsWith && reverse) (pos-str.size) max 0 else pos) +
      (if (excludeSelf) (if (reverse) -1 else 1) else 0)
    }
    val foundPos = {
      if (!reverse) contents.indexOf(str, startPos)
      else contents.lastIndexOf(str, startPos)
    }
    val retPos = if (endsWith && foundPos != -1) foundPos+str.size else foundPos
    if (skip == 0 || retPos == -1) retPos
    else find(contents, endsWith, reverse, true, skip-1, str, retPos) // Skip ahead
  }
  def find(endsWith:Boolean, reverse:Boolean, excludeSelf:Boolean, skip:Int, str:String) : Int = find(contents, endsWith, reverse, excludeSelf, skip, str, pos)
  def coarseFind(endsWith:Boolean, reverse:Boolean, excludeSelf:Boolean, skip:Int, coarseStr:String) : Int = find(coarseContents, endsWith, reverse, excludeSelf, skip, coarseStr, pos)

  def substring(i:Int, j:Int) = { // Return none if out of bounds
    if (!validPos(i) || !validPos(j) || i > j) None
    else Some(contents.substring(i, j))
  }

  val window = 15
  override def toString = {
    "\""+sanitize(contents.substring(0 max (pos-window), pos))+"["+pos+"]"+sanitize(contents.substring(pos, contents.size min (pos+window)))+"\""
  }
}
object State {
  def coarsify(s:String) = {
    val buf = new StringBuilder(s.size)
    foreach(s.size, { i:Int =>
      val c = s(i)
      buf += {
        if (Character.isWhitespace(c)) s(i)
        else if (Character.isLetter(c)) 'a'
        //else if (Character.isUpperCase(c)) 'A'
        //else if (Character.isLowerCase(c)) 'a'
        else if (Character.isDigit(c)) '0'
        else '!'
      }
    })
    buf.toString
  }
}

case class TextEditWorld(WO:WorldOptions, TO:TextOptions) extends AbstractWorld {
  case class Example(startState:State, endState:State, actions:Array[String]) extends AbstractExample {
    val input = startState
    val output = endState
    override def getLoss(predOutput:Any) = {
      val predState = predOutput.asInstanceOf[State]
      if (predState.contents == endState.contents) {
        if (predState.pos != endState.pos) 0.01 else 0.0 // Tiny error for wrong position
      }
      else 1.0
    }
  }

  def allowCombination_?(func:KNode, arg:KNode) : Boolean = true

  def wrap[A,B,O](n:String, f:(A,B)=>O) : Func = new Func {
    def name = n
    def requiresState = false
    def call(state:Any, args:Array[Any]) = {
      require (args.size == 2)
      val x = args(0).asInstanceOf[A]
      val y = args(1).asInstanceOf[B]
      f(x, y)
    }
  }
  def wrapState[O](n:String, f:State=>O) : Func = new Func {
    def name = n
    def requiresState = true
    def call(state:Any, args:Array[Any]) = {
      require (args.size == 0)
      f(state.asInstanceOf[State])
    }
  }
  def wrapState[A,O](n:String, f:(State,A)=>O) : Func = new Func {
    def name = n
    def requiresState = true
    def call(state:Any, args:Array[Any]) = {
      require (args.size == 1)
      val a = args(0).asInstanceOf[A]
      f(state.asInstanceOf[State], a)
    }
  }
  def wrapState[A,B,O](n:String, f:(State,A,B)=>O) : Func = new Func {
    def name = n
    def requiresState = true
    def call(state:Any, args:Array[Any]) = {
      require (args.size == 2)
      val a = args(0).asInstanceOf[A]
      val b = args(1).asInstanceOf[B]
      f(state.asInstanceOf[State], a, b)
    }
  }
  def wrapState[A,B,C,O](n:String, f:(State,A,B,C)=>O) : Func = new Func {
    def name = n
    def requiresState = true
    def call(state:Any, args:Array[Any]) = {
      require (args.size == 3)
      val a = args(0).asInstanceOf[A]
      val b = args(1).asInstanceOf[B]
      val c = args(2).asInstanceOf[C]
      f(state.asInstanceOf[State], a, b, c)
    }
  }
  def wrapState[A,B,C,D,O](n:String, f:(State,A,B,C,D)=>O) : Func = new Func {
    def name = n
    def requiresState = true
    def call(state:Any, args:Array[Any]) = {
      require (args.size == 4)
      val a = args(0).asInstanceOf[A]
      val b = args(1).asInstanceOf[B]
      val c = args(2).asInstanceOf[C]
      val d = args(3).asInstanceOf[D]
      f(state.asInstanceOf[State], a, b, c, d)
    }
  }
  def wrapState[A,B,C,D,E,O](n:String, f:(State,A,B,C,D,E)=>O) : Func = new Func {
    def name = n
    def requiresState = true
    def call(state:Any, args:Array[Any]) = {
      require (args.size == 5)
      val a = args(0).asInstanceOf[A]
      val b = args(1).asInstanceOf[B]
      val c = args(2).asInstanceOf[C]
      val d = args(3).asInstanceOf[D]
      val e = args(4).asInstanceOf[E]
      f(state.asInstanceOf[State], a, b, c, d, e)
    }
  }

  // Very basic stuff
  val AddFunc = wrap("add", {(x:Int,y:Int) => x+y})
  val SubFunc = wrap("add", {(x:Int,y:Int) => x-y})
  val AppendFunc = wrap("append", {(str1:String,str2:String) => str1+str2})

  // Fetching state information
  val PosFunc = wrapState("pos", {state:State => state.pos})
  val CaseNumFunc = wrapState("caseNum", {state:State => state.caseNum.toString})

  // Find strings
  val FindFunc = wrapState("find", {(state:State,endsWith:Boolean,reverse:Boolean,excludeSelf:Boolean,skip:Int,str:String) => state.find(endsWith, reverse, excludeSelf, skip, str)})
  val CoarseFindFunc = wrapState("coarse-find", {(state:State,endsWith:Boolean,reverse:Boolean,excludeSelf:Boolean,skip:Int,coarseStr:String) => state.coarseFind(endsWith, reverse, excludeSelf, skip, coarseStr)})

  val BeginOfWord = wrapState("bow", {state:State => state.beginOfWord})
  val FindWhitespace = wrapState("find-whitespace", {state:State => state.findWhitespace})
  val EndOfFile = wrapState("eof", {state:State => state.endOfFile})

  // Actions
  val MoveFunc = wrapState("move", {(state:State,pos:Int) => state.move(pos)})
  val SelectFunc = wrapState("select", {(state:State,i:Int,j:Int) => state.select(i, j)})
  val PasteFunc = wrapState("paste", {state:State => state.paste})
  val CutFunc = wrapState("cut", {state:State => state.cut})
  val CopyFunc = wrapState("copy", {state:State => state.copy})
  val DeleteFunc = wrapState("delete", {(state:State,pos:Int) => state.delete(pos)})
  val InsertFunc = wrapState("insert", {(state:State,str:String) => state.insert(str)})
  val DeleteSelectionFunc = wrapState("deleteselection", {state:State => state.deleteselection})

  def populateScenarios = {
    import java.io._
    Utils.track_printAll("Loading scenarios from %s", TO.inPath) {
      val yaml = new Yaml
      var offset = 0
      val paths : Seq[String] = {
        if (new File(TO.inListPath).exists)
          Utils.readLines(TO.inListPath).flatMap { x =>
            var y = x.replaceAll("""\s*#.*$""", "")
            if (y.size == 0) None
            else Some(new File(TO.inListPath).getParent+"/"+y)
          }
        else
          new File(TO.inPath).list.toList.sortWith(_ < _).map(TO.inPath+"/"+_)
      }

      paths.foreach { path =>
        val name = new File(path).getName
        if (name.endsWith(".yaml")) {
          if (offset >= WO.scenarioOffset && scenarios.size < WO.maxScenarios) {
            track("%s", name) {
              val reader = fig.basic.IOUtils.openInHard(path)
              extractScenario(name, yaml.load(reader)) match {
                case Some(scenario) => scenarios += scenario
                case None =>
              }
              reader.close
            }
          }
          offset += 1
        }
      }
      Execution.putOutput("numScenarios", scenarios.size)
    }
  }

  def extractScenario(name:String, scenarioObj:AnyRef) : Option[Scenario] = {
    import java.util.{LinkedHashMap,ArrayList}
    def extractState(caseNum:Int, obj:AnyRef) = {
      val map = obj.asInstanceOf[LinkedHashMap[Object,Object]]
      val pos = map.get("pos").asInstanceOf[Int]
      val selection = map.get("selection").asInstanceOf[String] match {
        case null => None
        case s => Some(parseSelection(s))
      }
      val clipboard = map.get("clipboard").asInstanceOf[String]
      val contents = map.get("contents").asInstanceOf[String]
      State(caseNum, pos, selection, clipboard, contents)
    }

    val map = scenarioObj.asInstanceOf[LinkedHashMap[Object,Object]]
    var caseNum = 0
    val examples = map.get("examples").asInstanceOf[ArrayList[Object]].toArray(new Array[Object](0)).map { exObj =>
      val map = exObj.asInstanceOf[LinkedHashMap[Object,Object]]
      caseNum += 1
      val actions = map.get("actions").asInstanceOf[ArrayList[String]].toArray(new Array[String](0))
      Example(extractState(caseNum, map.get("before")), extractState(caseNum, map.get("after")), actions)
    }
    if (WO.verbose >= 4) {
      foreach(examples.size, { i:Int => val ex = examples(i)
        track("Example") {
          if (i == 0 && WO.verbose >= 5) ex.startState.log
          logs("Actions: %s", sanitize(ex.actions.mkString(" | ")))
          if (WO.verbose >= 5) ex.endState.log
        }
      })
    }

    // All examples must have the same sequence of actions for us to set up the empirical denotation properly
    def f(a:String) = a.split(" ", 2)(0)
    val allSameActions = examples.forall { ex => ex.actions.size == examples(0).actions.size } &&
      (0 to examples(0).actions.size-1).forall { i:Int =>
      examples.forall { ex =>
        //dbgs("%s %s %s", i, f(ex.actions(i)), f(examples(0).actions(i)))
        f(ex.actions(i)) == f(examples(0).actions(i))
      }
    }
    if (allSameActions)
      Some(Scenario(name, examples.toList))
    else {
      logs("Rejected because can't align the actions across all examples")
      None
    }
  }

  case class Scenario(name:String, examples:List[Example]) extends AbstractScenario {
    def createProgram = track("Scenario.createProgram(%s)", name) {
      //// Setup the program (candidate lists)

      // Break up each example into smaller chunks based on the actions
      val exActions = trainExamples.map { ex =>
        var state = ex.startState
        val actions = new ArrayBuffer[Action]
        ex.actions.foreach { actionStr =>
          val action = state.getAction(actionStr)
          actions += action
          state = action.outState
          //Utils.track_printAll("Result after %s", escape(actionStr)) { state.log }
        }

        // Make sure these actions actually produce the right thing
        if (state != ex.endState) track("Incorrect") {
          Utils.track_printAll("Given") { ex.endState.log }
          Utils.track_printAll("Simulated") { state.log }
          throw impossible
        }

        actions.slice(0, TO.maxActionsPerExample) // Slight hack
      }
      
      // Create a statement for each action
      val numActions = exActions.head.size
      val statements = (0 to numActions-1).map { i:Int =>
        NodeSet(Array(createStatement(exActions.map(_(i)))))
      }.toArray

      // Create program
      program = new ProgramNode(statements)
    }

    // Quick way to build expressions
    def __(func:Func, args:Any*) = {
      ExprNode(func, args.map { arg =>
        arg match {
          case arg:NodeSet => arg
          case arg:Node => NodeSet(Array(arg))
          case _ => NodeSet(Array(ValueNode(arg)))
        }
      }.toArray)
    }

    private def createStatement(actions:Seq[Action]) : ExprNode = {
      // Return all ways to get to a particular position
      def pos(getOutput:Action=>Int) = {
        val action = actions.head
        val state = action.inState
        val pos = getOutput(action)
        val candidates = new ArrayBuffer[Node]

        candidates += ValueNode(pos) // Absolute position
        candidates += __(AddFunc, __(PosFunc), pos-state.pos) // Relative position (character offset)
        candidates += __(BeginOfWord) // Beginning of the next word
        candidates += __(FindWhitespace)
        candidates += __(EndOfFile) // End of file

        // Search for a string
        Array((-1, 0), (0, +1)).foreach { dirPair => // -1: ends-with: +1: starts-with
          val (d1, d2) = dirPair
          val endsWith = d1 == -1
          // Search for a string of a certain length in the given direction
          Array(false, true).foreach { reverse => // Which direction to search
            Array(false, true).foreach { excludeSelf => // Whether to consider option of staying put
              foreach(TO.maxSearchSkip+1, { skip:Int => // How many occurrences to skip
                foreach(1, TO.maxSearchLen+1, { len:Int => // For each length
                  state.substring(pos+d1*len, pos+d2*len) match {
                    case Some(str) =>
                      candidates += __(FindFunc, endsWith, reverse, excludeSelf, skip, str) // Search verbatim 
                      candidates += __(CoarseFindFunc, endsWith, reverse, excludeSelf, skip, State.coarsify(str)) // Search coarse
                    case None =>
                  }
                })
              })
            }
          }
        }

        if (WO.verbose >= 7) {
          candidates.foreach { candidate =>
            Utils.track_printAll("%s", candidate) {
              actions.foreach { action =>
                dbgs("%s %s %s", action.inState, getOutput(action), candidate.call(action.inState))
              }
            }
          }
        }

        // Keep only the ones which are good
        val validCandidates = candidates.filter { candidate =>
          actions.forall { action =>
            candidate.call(action.inState).asInstanceOf[Int] == getOutput(action)
          }
        }.toArray
        if (WO.verbose >= 5) Utils.track_printAll("pos validCandidates") {
          validCandidates.foreach(logs("%s", _))
        }
        if (validCandidates.size == 0) Utils.track_printAll("BAD") {
          actions.foreach { action =>
            Utils.track_printAll("IN") { action.inState.log }
            logs("ACT: %s", action.actionStr)
            Utils.track_printAll("OUT") { action.outState.log }
          } 
        }
        NodeSet(validCandidates)
      }

      // Return possible ways to insert a string
      def str(getOutput:Action=>String) = {
        val action = actions.head
        val state = action.inState
        val str = getOutput(action)
        val candidates = new ArrayBuffer[Node]

        candidates += ValueNode(str) // Verbatim

        // Insert string caseNum string
        val caseNum = state.caseNum.toString
        str.indexOf(caseNum) match {
          case -1 =>
          case i =>
            candidates += __(AppendFunc, str.substring(0, i),
                                         __(AppendFunc, __(CaseNumFunc), str.substring(i+caseNum.size, str.size)))
        }

        if (WO.verbose >= 7) {
          candidates.foreach { candidate =>
            Utils.track_printAll("%s", candidate) {
              actions.foreach { action =>
                dbgs("%s %s %s", action.inState,
                  escape(getOutput(action).asInstanceOf[String]),
                  escape(candidate.call(action.inState).asInstanceOf[String]))
              }
            }
          }
        }

        // Keep only the ones which are good
        val validCandidates = candidates.filter { candidate =>
          actions.forall { action =>
            candidate.call(action.inState).asInstanceOf[String] == getOutput(action)
          }
        }.toArray
        if (WO.verbose >= 5) Utils.track_printAll("str validCandidates") {
          validCandidates.foreach(logs("%s", _))
        }
        if (validCandidates.size == 0) Utils.track_printAll("BAD") {
          actions.foreach { action =>
            Utils.track_printAll("IN") { action.inState.log }
            logs("ACT: %s", action.actionStr)
            Utils.track_printAll("OUT") { action.outState.log }
          } 
        }
        NodeSet(validCandidates)
      }

      actions.head match {
        case _:MoveAction => ExprNode(MoveFunc, Array(pos(_.asInstanceOf[MoveAction].pos)))
        case _:SelectAction => ExprNode(SelectFunc, Array(pos(_.asInstanceOf[SelectAction].selection._1),
                                                          pos(_.asInstanceOf[SelectAction].selection._2)))
        case _:PasteAction => ExprNode(PasteFunc, Array())
        case _:CutAction => ExprNode(CutFunc, Array())
        case _:CopyAction => ExprNode(CopyFunc, Array())
        case _:DeleteAction => ExprNode(DeleteFunc, Array(pos(_.asInstanceOf[DeleteAction].pos)))
        case _:InsertAction => ExprNode(InsertFunc, Array(str(_.asInstanceOf[InsertAction].str)))
        case _:DeleteSelectionAction => ExprNode(DeleteSelectionFunc, Array())
        case _ => throw unknownCase(actions.head)
      }
    }

    def log = track("Scenario %s", name) {
      logs("Extracted: %s", program.extracted)
      program.log
    }
  }
}

class TextEditTestRunner(WO:WorldOptions, TO:TextOptions) extends Runnable {
  def run = {
    val errors = new fig.basic.FullStatFig
    Utils.track_record_foreach("Trial", WO.numTrials, { trial:Int =>
      val world = new TextEditWorld(WO, TO)
      world.initialize
      world.learn
      val error = world.getBayesAvgTestError
      logs("Trial error = %s", error)
      errors.add(error)
      Execution.putOutput("minTrialError", errors.getMin)
      Execution.putOutput("meanTrialError", errors.mean)
      Execution.putOutput("maxTrialError", errors.getMax)
      true
    })
    logs("Error over all trials: %s", errors)
  }
}

object TextEditTest {
  def main(args:Array[String]) = {
    val WO = new WorldOptions
    val TO = new TextOptions
    Execution.run(args, new TextEditTestRunner(WO, TO), "world", WO, "text", TO)
  }
}
