package textedit

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import fig.record.Record
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,fails}
import tea.Utils.{track,begin_track,end_track,logs,dbgs,warnings}

import java.util.{Set,ArrayList,Random}

trait AbstractWorld {
  def WO : WorldOptions // ABSTRACT

  val primitives = new HashSet[Any] // SET IN initialize()
  val scenarios = new ArrayBuffer[Scenario] // SET IN initialize()
  val params = new Params(WO) // SET in initialize()

  val fullSamples = new CounterMap[List[Any]]
  val fullSampleLogLikelihoods = new HashMap[List[Any], Double]

  case class Event(description:String, acceptRatio:Double, accept:Boolean)
  val samplingStats = new HashMap[String, CounterMap[String]]

  val eventsOut = fig.basic.IOUtils.openOutHard(Execution.getFile("events"))
  def recordEvent(event:Event) = {
    val s = if (event.accept) "accept" else "reject"
    samplingStats.getOrElseUpdate(event.description, new CounterMap[String]).incr(s, 1)
    if (WO.outputEvents) {
      if (event.accept)
        eventsOut.println("STATE\t"+scenarios.map(_.program).mkString("\t"))
      eventsOut.println(event.description+"\t"+s+"\t"+event.acceptRatio)
    }
  }

  var iter : Int = -1
  def burnedIn_? = iter >= WO.burnInFrac*WO.numIters
  def temperature = {
    if (WO.numIters == 1 || burnedIn_?)
      WO.finalTemperature
    else
      WO.initTemperature + (WO.finalTemperature-WO.initTemperature) * iter / (WO.burnInFrac*WO.numIters - 1)
  }

  ////////////////////////////////////////////////////////////

  type Example <: AbstractExample
  trait AbstractExample {
    // When doing prediction averaging, accumulate possible outputs
    val predOutputs = new CounterMap[Any]

    def input : Any
    def output : Any

    def passes(program:ProgramNode) = program.call(input) == output
    def getLoss(program:ProgramNode) : Double = getLoss(program.call(input)) // Override if want to evaluate on something else
    def getLoss(predOutput:Any) = if (predOutput == output) 0.0 else 1.0 // Override if want to evaluate on something else

    override def toString = input+"=>"+output
  }

  type Scenario <: AbstractScenario
  trait AbstractScenario {
    def name : String
    def examples : List[Example]
    var program : ProgramNode = null

    def trainExamples = examples.slice(0, WO.numTrainExamples)
    def testExamples = examples.slice(WO.numTrainExamples, examples.size)
    def testError = {
      if (testExamples.size == 0) 0.0
      else ((0.0 /: testExamples) { (sum:Double,ex:Example) => sum + ex.getLoss(program) }) / testExamples.size
    }

    def createProgram : Unit // OVERRIDE
    def log : Unit // OVERRIDE
    def repCheck = { }

    def sample = {
      if (WO.sampleCandidates) sampleCandidates
      if (WO.sampleGrouping) sampleGrouping
      if (WO.sampleExtraction) sampleExtraction
    }

    // Try to make a change, check MH ratio, and then accept/reject
    def proposeChange(description:String, node:Node, forwChange : =>Any, backChange : =>Any) = {
      val oldKey = program.key
      val oldExtracted = program.extracted
      forwChange
      node.invalidateUp
      program.computeAll
      val newKey = program.key
      val newExtracted = program.extracted

      var modelRatio = 1.0
      modelRatio *= params.replaceCustomer(oldKey, newKey)
      modelRatio *= params.replaceCustomers(oldExtracted.values, newExtracted.values)
      modelRatio *= genProb(newExtracted) / genProb(oldExtracted)
      val acceptRatio = math.pow(modelRatio, 1.0/temperature)

      val accept = WO.random.nextDouble < acceptRatio
      if (!accept) {
        params.replaceCustomer(newKey, oldKey)
        params.replaceCustomers(newExtracted.values, oldExtracted.values)
        backChange
        node.invalidateUp
        program.computeAll
      }
      val event = Event(description, acceptRatio, accept)
      if (WO.verbose >= 5) Utils.track_printAll("proposeChange") {
        logs("%s", event)
        logs("old: %s | %s", oldKey, oldExtracted)
        logs("new: %s | %s", newKey, newExtracted)
      }
      recordEvent(event)
    }

    // Might have to worry about proposal probability
    def sampleCandidates = {
      program.foreachNode(false, { node:Node =>
        node match {
          case node:NodeSet =>
            val oldChoice = node.choice
            val newChoice = WO.random.nextInt(node.candidates.size)
            if (oldChoice != newChoice) {
              proposeChange("candidate", node, { node.choice = newChoice },
                                               { node.choice = oldChoice })
            }
          case _ =>
        }
      })
    }

    def sampleGrouping : Unit = {
      if (program.statements.size < 2) return // No grouping possible

      // Choose a non-root composite GNode at random as the pivot and rotate around that pivot
      def rotateRandom(k:Int, node:GNode) : GNode = {
        node match {
          case node:GCompositeNode =>
            if (k < node.k) {
              val left = node.left.asInstanceOf[GCompositeNode]
              if (left.k == k)
                GCompositeNode(left.k, left.left, GCompositeNode(node.k, left.right, node.right)) // Rotate right
              else
                GCompositeNode(node.k, rotateRandom(k, node.left), node.right)
            }
            else if (k > node.k) {
              val right = node.right.asInstanceOf[GCompositeNode]
              if (right.k == k)
                GCompositeNode(right.k, GCompositeNode(node.k, node.left, right.left), right.right) // Rotate left
              else
                GCompositeNode(node.k, node.left, rotateRandom(k, node.right))
            }
            else
              throw impossible
          case _ => node
        }
      }

      foreach(program.statements.size-1, { _:Int => // Repeat to hit every node once on average
        // Pick a random internal node (note that labels of internal nodes are 1..n-1)
        val k = 1+WO.random.nextInt(program.statements.size-1)
        if (program.grouping.k != k) { // Don't select root
          val oldGrouping = program.grouping
          val newGrouping = rotateRandom(k, program.grouping)
          if (oldGrouping != newGrouping) {
            proposeChange("group", program, { program.grouping = newGrouping },
                                            { program.grouping = oldGrouping })
          }
        }
      })
    }

    def sampleExtraction = {
      program.foreachNode(false, { node:Node =>
        node match {
          case node:ValueNode =>
            if (node.canExtract) {
              // Propose extracting or not extracting
              val oldHole = node.hole
              val newHole = WO.random.nextInt(2) == 1
              if (oldHole != newHole) {
                proposeChange("extract", node, { node.hole = newHole },
                                               { node.hole = oldHole })
              }
            }
          case _ =>
        }
      })
    }

    def savePredictions = {
      testExamples.foreach { ex =>
        ex.predOutputs.incr(program.call(ex.input), 1)
      }
    }

    // My params is to get a single-task baseline
    lazy val myParams = {
      val params = new Params(WO)
      params.initialize(primitives)
      params
    }
    def params = {
      if (WO.multitaskSharing) AbstractWorld.this.params
      else myParams
    }

    def initializeState = {
      program.initialize(WO.random, WO)
      program.computeAll
      logs("# programs: %s", program.computeCardinality)
      params.initialize(primitives)

      //dbgs("PROGRAM: %s", program)
      //dbgs("KEY: %s", program.key)
      params.addCustomer(program.key)
      params.addCustomers(program.extracted.values)
    }
  }

  ////////////////////////////////////////////////////////////

  def populateScenarios : Unit // ABSTRACT

  def initialize : Unit = track("AbstractWorld.initialize") {
    populateScenarios
    Utils.track_printAll("Create programs") {
      scenarios.foreach(_.createProgram)
    }

    // Get primitives
    scenarios.foreach { scenario =>
      val program = scenario.program
      program.foreachNode(true, { node =>
        node match {
          case ExprNode(func, _) => primitives += func
          case ValueNode(value) => primitives += value
          case _ =>
        }
      })
    }
    logs("# primitives: %s", primitives.size)

    // Initialize state
    Utils.track_printAll("Initialize state") {
      scenarios.foreach(_.initializeState)
    }

    // Print statistics
    val numTrain = new fig.basic.FullStatFig
    val numTest = new fig.basic.FullStatFig
    scenarios.foreach { scenario =>
      numTrain.add(scenario.trainExamples.size)
      numTest.add(scenario.testExamples.size)
    }
    logs("%s scenarios; #train: %s; #test: %s", scenarios.size, numTrain, numTest)

    if (WO.verbose >= 3) log
    repCheck
  }

  def outputAndRecordState = {
    if (iter == WO.numIters || iter % WO.recordStateInterval == 0) {
      Record.begin("iteration", iter)
      Execution.putLogRec("logLikelihood", logLikelihood)
      Utils.track_printAll("Evaluation") {
        if (WO.verbose >= 2) {
          scenarios.foreach { scenario =>
            Utils.track_printAll("%s", scenario.name) {
              scenario.program.log
              scenario.examples.foreach { ex =>
                val trueOutput = ex.output
                val predOutput = scenario.program.call(ex.input)
                logs("%s => TRUE: %s\tPRED:%s\t%s", ex.input, trueOutput, predOutput,
                  if (trueOutput == predOutput) "[correct]" else "[WRONG]")
              }
            }
          }
        }
        Execution.putLogRec("avgTestError", avgTestError)
      }
      Utils.track_printAll("Stats") {
        samplingStats.keysIterator.foreach { description =>
          Utils.track_printAll("%s", description) {
            samplingStats(description).write(Integer.MAX_VALUE, true, "", logs("%s", _))
          }
        }
      }
      Record.end
    }

    if (iter == WO.numIters || iter % WO.outputStateInterval == 0) {
      Utils.writeLines(Execution.getFile("state."+iter), { puts:(String=>Any) =>
        puts(fmts("log-likelihood: %s", logLikelihood))
        scenarios.foreach { scenario =>
          puts(fmts("%s (%s programs)", scenario.name, scenario.program.computeCardinality))
          scenario.program.output { s => puts("  "+s) }
        }
        //puts(fmts("avgTestError = %s", avgTestError))
      })

      params.write(Execution.getFile("params."+iter))

      // Count the occurrences of each combinators
      // Note that this is different from params's counts in dishes, because there's no collapsing due to the HDP.
      Utils.writeLines(Execution.getFile("combinators."+iter), { puts:(String=>Any) =>
        val map = new CounterMap[String]
        def recurse(node:KNode) : Unit = {
          map.incr(node.toString, 1)
          node match {
            case node:KCompositeNode =>
              recurse(node.func)
              recurse(node.arg)
            case _ =>
          }
        }
        scenarios.foreach { scenario => recurse(scenario.program.key) }
        map.write(Integer.MAX_VALUE, true, "", puts)
      })
    }

    // Bayesian averaging: record predictions
    if (burnedIn_?)
      scenarios.foreach(_.savePredictions)
  }

  def learn = {
    if (WO.online) onlineLearn
    else batchLearn
  }

  def onlineLearn = {
    // TODO 
  }

  def batchLearn = Utils.track_printAll("batchLearn") {
    track("Sampling") {
      Utils.track_foreach("Iteration", WO.numIters, { iter:Int =>
        this.iter = iter
        outputAndRecordState

        if (WO.rememberFullSamples) {
          val sample = scenarios.map { scenario => (scenario.program.key, scenario.program.extracted.numHoles, scenario.program.extracted.values.toSet) }.toList
          fullSamples.incr(sample, 1)
          fullSampleLogLikelihoods.getOrElseUpdate(sample, logLikelihood)
        }
        scenarios.foreach { scenario =>
          if (WO.verbose >= 5) track("Before") { log }
          scenario.sample
          if (WO.verbose >= 5) track("After") { log }
          repCheck
        }

        !Execution.shouldBail
      })

      this.iter = WO.numIters
      outputAndRecordState
    }

    track("Final samples") {
      scenarios.foreach { scenario => track("%s", scenario.name) { scenario.program.log } }
    }

    if (WO.rememberFullSamples) Utils.track_printAll("Full samples") {
      // Allow one to compute the true distribution over the explored space
      var logZ = Double.NegativeInfinity
      fullSampleLogLikelihoods.valuesIterator.foreach { ll => logZ = fig.basic.NumUtils.logAdd(logZ, ll) }

      fullSamples.foreachSorted(Integer.MAX_VALUE, { (sample,count) =>
        val empProb = fullSamples.getProb(sample)
        val trueLL = fullSampleLogLikelihoods(sample)
        val trueProb = math.exp(trueLL-logZ)
        logs("%s %s %s %s\t%s", count, empProb, trueProb, trueLL, sample.mkString("\t"))
      })
    }
  }

  def logLikelihood = {
    var sum = 0.0
    sum += params.logLikelihood // CRP
    scenarios.foreach { scenario => sum += math.log(genProb(scenario.program.extracted)) } // Extraction
    sum
  }

  // Return the probability of generating this extraction profile.
  // Don't count the generation of the actual values, because that's
  // taken care by the CRP.
  def genProb(extracted:ExtractInfo) = {
    var prob = 1.0
    prob *= math.pow(1-WO.genExtractValueStopProb, extracted.numValues) * WO.genExtractValueStopProb // Generate this many values (geometric distribution)
    prob *= math.pow(1.0/extracted.numValues, extracted.numHoles) // For each hole, generate a value
    prob
  }

  def repCheck = if (WO.repCheck) {
    params.repCheck
    scenarios.foreach(_.repCheck)

    // Check that params is consistent with the scenarios
    // Have everyone enter
    val checkDishes = new HashMap[KNode,Int]
    def recurse(node:KNode) : Unit = {
      checkDishes(node) = checkDishes.getOrElse(node, 0) + 1
      if (checkDishes(node) == 1) {
        node match {
          case node:KCompositeNode =>
            recurse(node.func)
            recurse(node.arg)
          case node:KPrimitiveNode =>
        }
      }
    }
    scenarios.foreach { scenario =>
      recurse(scenario.program.key)
    }

    checkDishes.keysIterator.foreach { node =>
      require (checkDishes(node) == params.counts.getCount(node), node)
    }
  }

  def log = Utils.track_printAll("World") {
    Utils.track_foreach("Scenario", scenarios.size, { i:Int =>
      scenarios(i).log
      true
    })
    Utils.track_printAll("Params") {
      params.log
    }
  }

  // Count each example once
  def avgTestError = {
    (0.0 /: scenarios) { (sum:Double,scenario:AbstractScenario) => sum + scenario.testError * scenario.testExamples.size } /
    (0 /: scenarios)(_ + _.testExamples.size)
  }

  // Use Bayesian averaging to make predictions
  def getBayesAvgTestError = track("getBayesAvgTestError") {
    val errors = new fig.basic.StatFig
    Record.begin("bayesAvgTestError")
    scenarios.foreach { scenario => track("Scenario %s", scenario.name) {
      scenario.testExamples.foreach { ex =>
        errors.add(ex.predOutputs.argmaxAndMax match {
          case (Some(predOutput), _) =>
            val loss = ex.getLoss(predOutput)
            if (WO.verbose >= 1) {
              Utils.track_printAll("Test case") {
                logs("Input: %s", ex.input)
                logs("True: %s", ex.output)
                logs("Pred: %s", predOutput)
                logs("Loss: %s", loss)
              }
            }
            Record.add(scenario.name, ""+loss)
            loss
          case _ => throw fails("Unable to get best prediction")
        })
      }
    } }
    Record.end
    errors.mean
  }
}
