package textedit

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,fails}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}
import MyUtils._

class Params(WO:WorldOptions) {
  val counts = new CounterMap[KNode] // c -> how many times c is used

  // Filld in later
  var alpha = Double.NaN
  var discount = Double.NaN
  var primitives = new HashSet[Any]

  def write(path:String) = {
    Utils.writeLines(path, { puts:(String=>Any) =>
      counts.write(Integer.MAX_VALUE, true, "", puts)
    })
  }

  // Local potential psi 
  def localPotential(node:KNode) = {
    if (WO.useLocalPotential) {
      node match {  
        case node:KPrimitiveNode => WO.genPrimitiveProb * { if (WO.numPrimitives == 0) 1.0/primitives.size else 1.0/WO.numPrimitives }
        case node:KCompositeNode => (1-WO.genPrimitiveProb) * 1.0/WO.numRouters
      }
    }
    else
      1.0
  }

  def logLikelihood : Double = {
    import fig.basic.NumUtils.logGamma
    var sum = 0.0
    val K = counts.numKeys
    if (K == 0) sum // Empty restaurant
    else {
      require (discount < 1)
      counts.keys.foreach { node =>
        sum += math.log((alpha + discount*(K-1)) * localPotential(node)) + logGamma(counts.getCount(node) - discount) - logGamma(1-discount)
      }
      sum += logGamma(alpha) - logGamma(alpha+counts.totalCount)
      sum
    }
  }

  def initialize(primitives:HashSet[Any]) = {
    this.alpha = WO.alpha
    this.discount = WO.discount
    this.primitives = primitives
  }

  def log = {
    logs("alpha = %s, N = %s", alpha, counts.totalCount)
    Utils.track_printAll("Dishes") {
      counts.keys.foreach { node =>
        logs("%s: n=%s, potential=%s", node, counts.getCount(node), localPotential(node))
      }
    }
  }

  // Return probability of adding
  def addCustomer(node:KNode) : Double = {
    val n = counts.incr(node, 1)
    val N = counts.totalCount
    val K = counts.numKeys
    var w = 1.0
    if (n == 1) { // First one of this dish
      w *= (alpha+discount*(K-1)) * localPotential(node)
      node match {
        case KCompositeNode(routers, func, arg) =>
          w *= addCustomer(func)
          w *= addCustomer(arg)
        case _ =>
      }
    }
    else { // Old
      w *= n-1-discount
    }
    //dbgs("ADD n=%s w=%s N=%s => %s", n, w, N, w / (alpha + N - 1))
    w / (alpha + N - 1)
  }

  def removeCustomer(node:KNode) : Double = {
    val N = counts.totalCount
    val K = counts.numKeys
    val n = counts.incr(node, -1) + 1
    var w = 1.0
    if (n == 1) { // Last one of this dish
      w /= (alpha+discount*(K-1)) * localPotential(node)
      node match {
        case KCompositeNode(routers, func, arg) =>
          w *= removeCustomer(func)
          w *= removeCustomer(arg)
        case _ =>
      }
    }
    else { // Old
      w /= n-1-discount
    }
    //dbgs("REM n=%s w=%s N=%s => %s", n, w, N, w * (alpha + N - 1))
    w * (alpha+N-1)
  }

  def replaceCustomer(oldNode:KNode, newNode:KNode) : Double = {
    if (WO.verbose >= 7) Utils.track_printAll("replaceCustomer") {
      logs("old: %s", oldNode)
      logs("new: %s", newNode)
    }
    var prob = 1.0
    prob *= removeCustomer(oldNode) 
    prob *= addCustomer(newNode)
    //dbgs("REPL %s", prob)
    prob
    // FUTURE: more efficient version if oldNode and newNode have common parts
  }

  def removeCustomers(values:List[Any]) : Double = {
    var prob = 1.0
    values.foreach { v => prob *= removeCustomer(KPrimitiveNode(v)) }
    prob
  }
  def addCustomers(values:List[Any]) : Double = {
    var prob = 1.0
    values.foreach { v => prob *= addCustomer(KPrimitiveNode(v)) }
    prob
  }

  def replaceCustomers(oldValues:List[Any], newValues:List[Any]) : Double = {
    if (WO.verbose >= 7) Utils.track_printAll("replaceCustomers") {
      logs("old: %s", oldValues.map(escape).mkString(" "))
      logs("new: %s", newValues.map(escape).mkString(" "))
    }
    var prob = 1.0
    oldValues.foreach { v => prob *= removeCustomer(KPrimitiveNode(v)) }
    newValues.foreach { v => prob *= addCustomer(KPrimitiveNode(v)) }
    prob
    // FUTURE: more efficient version if oldNode and newNode have common parts
  }

  def repCheck = { }
}
